// Copyright (c) 2021 The Foundry Visionmongers Ltd. All rights reserved.

// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are met:
//
//     * Redistributions of source code must retain the above copyright notice,
//       this list of conditions and the following disclaimer.
//     * Redistributions in binary form must reproduce the above copyright notice,
//       this list of conditions and the following disclaimer in the documentation
//       and/or other materials provided with the distribution.
//     * Neither the name of Foundry nor the names of its
//       contributors may be used to endorse or promote products derived from this
//       software without specific prior written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
// ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
// WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
// DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDERS OR CONTRIBUTORS BE LIABLE FOR
// ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
// (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
// LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
// ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
// SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

// This work is a C++ implementation compatible with version 1.2.8 of the Python-based Nuke gizmo
// authored at Psyop by Jonah Friedman and Andy Jones (see https://github.com/Psyop/Cryptomatte).

#include "CryptomatteLayerDialog.moc.h"

#include "EncryptomatteUtils.h"

#include "DDImage/LayerI.h"

#include <QtWidgets/QDialogButtonBox>
#include <QtWidgets/QGridLayout>
#include <QtWidgets/QLabel>
#include <QtWidgets/QLineEdit>
#include <QtWidgets/QMessageBox>

#include <cctype>
#include <iomanip>
#include <sstream>

namespace Foundry {
  namespace Encryptomatte {

    using namespace DD::Image;

    CryptomatteLayerDialog::CryptomatteLayerDialog(std::string cryptomatteLayerName, QWidget* parent)
    : QDialog(parent)
    , _cryptomatteLayerName(std::move(cryptomatteLayerName))
    {
      // Note: currently, we default to 3 crypto sublayers for a crypto layer.
      // TODO: When we offer configuration for the end users, make it dynamic.
      _layers.resize(kDefaultCryptoLayerDepth);

      setWindowTitle(_cryptomatteLayerName.empty() ? "New Cryptomatte Layer" : "Edit Cryptomatte Layer");

      // Create the dialog layout with a label and input field for the layer
      // name in the first row
      auto layout = new QGridLayout();
      setLayout(layout);
      _lineEditName = new QLineEdit(this);
      layout->addWidget(new QLabel("Name:"), 0, 0, 1, 1, Qt::AlignRight);
      layout->addWidget(_lineEditName, 0, 1);

      // Add some default crypto layer name and set the focus on this field as
      // this will be the primary input for the end user.
      _lineEditName->setText(_cryptomatteLayerName.c_str());
      _lineEditName->setFocus();
      _lineEditName->selectAll();

      // Add the standard OK and Cancel buttons in a row, and also the slots to
      // handle those clicks.
      _okCancelButtonBox = new QDialogButtonBox(QDialogButtonBox::Ok | QDialogButtonBox::Cancel);
      layout->addWidget(_okCancelButtonBox, 2, 0, 1, 2);
      connect(_okCancelButtonBox, &QDialogButtonBox::accepted, this, &CryptomatteLayerDialog::accept);
      connect(_okCancelButtonBox, &QDialogButtonBox::rejected, this, &CryptomatteLayerDialog::reject);
    }

    void CryptomatteLayerDialog::displayErrorMessage(const std::string& errorMessage) {
      auto errorMessageBox = new QMessageBox(QMessageBox::Warning, "Nuke", QString::fromStdString(errorMessage), QMessageBox::Ok, this, Qt::Dialog);
      errorMessageBox->setDefaultButton(QMessageBox::Ok);
      errorMessageBox->setEscapeButton(QMessageBox::Ok);     // needed so close box works
      errorMessageBox->exec();
    }

    void CryptomatteLayerDialog::accept()
    {
      _cryptomatteLayerName = _lineEditName->text().toStdString(); 
      const std::string errorMessage = ValidateLayerName(_cryptomatteLayerName);
      if (!errorMessage.empty()) {
        displayErrorMessage(errorMessage);
        return;
      }

      // Create the new layers with the standard RGBA channels, or
      // alternatively, select an existing one.
      for (size_t i = 0; i < _layers.size(); ++i) {
        std::stringstream cryptoStream;
        cryptoStream << std::setw(2) << std::setfill('0') << i;
        const std::string layerName = _cryptomatteLayerName + cryptoStream.str();
        _layers[i] = GetLayer(layerName);
        if (_layers[i] != nullptr) {
          continue;
        }
        for (auto const& channelName : kChannelNames) {
          const std::string layerChannelName = layerName + "." + channelName;
          getChannel(layerChannelName.c_str());
          _layers[i] = GetLayer(layerName);
        }
      }

      QDialog::accept();
    }
  }
}
