// Copyright (c) 2021 The Foundry Visionmongers Ltd. All Rights Reserved.

#include "CryptomatteLayerKnob.h"

// Include the Python header before including the Qt headers. This is necessary
// because the Python headers have "PyType_Slot *slots;". Qt headers #define
// 'slots' to an empty string, so "public slots:" is preprocessed into "public
// :".
#include "Python.h"

#include "CryptomatteLayerWidget.moc.h"
#include "EncryptomatteUtils.h"

#include "DDImage/Hash.h"
#include "DDImage/Iop.h"
#include "DDImage/WidgetContext.h"

#include <iomanip>

namespace Foundry {
  using namespace Cryptomatte;
  namespace Encryptomatte {

    using namespace DD::Image;

    struct CryptomatteLayerKnobObject {
      PyObject_HEAD
      CryptomatteLayer_Knob* _knob;
    };

    static PyObject* CryptomatteLayerKnobNew(PyTypeObject* type, PyObject* args, PyObject* kwargs)
    {
      Py_RETURN_NONE;
    }

    static void CryptomatteLayerKnobDealloc(PyObject* o)
    {
    }

    PyDoc_STRVAR(CryptomatteLayerKnobSetValue__doc__,
    "self.setValue(cryptomatteLayerName=None) -> None\n\n"
    "Set the cryptomatte layer name of the cryptomatte layer knob.\n"
    "@param cryptomatteLayerName: Optional parameter for specifying the new cryptomatte layer name. If omitted, the value will be set to None.\n"
    "@return: None");

    static PyObject* CryptomatteLayerKnobSetValue(CryptomatteLayerKnobObject* self, PyObject* args, PyObject* kwds)
    {
      const char* cryptomatteLayerName = nullptr;
      if (!PyArg_ParseTuple(args, "|z:setValue", &cryptomatteLayerName)) {
        return nullptr;
      }

      const bool isCryptomatteLayerNameNullPointer = cryptomatteLayerName == nullptr;
      const std::string errorMessage = isCryptomatteLayerNameNullPointer ? std::string() : ValidateLayerName(cryptomatteLayerName);
      if (!errorMessage.empty()) {
        PyErr_SetString(PyExc_ValueError, errorMessage.c_str());
        return nullptr;
      }

      CryptomatteLayer_Knob* cryptomatteLayerKnob = self->_knob;
      if (cryptomatteLayerKnob != nullptr) {
        cryptomatteLayerKnob->setSelectedCryptomatteLayerName(isCryptomatteLayerNameNullPointer ? std::string() : cryptomatteLayerName);
      }
      Py_RETURN_NONE;
    }

   PyDoc_STRVAR(CryptomatteLayerKnobGetValue__doc__,
   "self.getValue() -> str\n\n"
   "Get the cryptomatte layer name value of this knob as a string.\n"
   "@return: String value.\n"
   );

    static PyObject* CryptomatteLayerKnobGetValue(CryptomatteLayerKnobObject* self, PyObject* args, PyObject* kwds)
    {
      const CryptomatteLayer_Knob* cryptomatteLayerKnob = self->_knob;
      std::string selectedCryptomatteLayerName;
      if (cryptomatteLayerKnob != nullptr) {
        selectedCryptomatteLayerName = cryptomatteLayerKnob->getSelectedItemString();
      }

      return PyUnicode_FromString(selectedCryptomatteLayerName.empty() ? "none" : selectedCryptomatteLayerName.c_str());
    }

    static PyMethodDef CryptomatteLayerKnobMethods[] = {
      {"setValue", reinterpret_cast<PyCFunction>(CryptomatteLayerKnobSetValue), METH_VARARGS, CryptomatteLayerKnobSetValue__doc__},
      {"getValue", reinterpret_cast<PyCFunction>(CryptomatteLayerKnobGetValue), METH_NOARGS, CryptomatteLayerKnobGetValue__doc__},
      {"value",    reinterpret_cast<PyCFunction>(CryptomatteLayerKnobGetValue), METH_NOARGS, CryptomatteLayerKnobGetValue__doc__},
      {nullptr, nullptr, 0, nullptr}
    };

    // This is the python type that actually gets registered underneath. This
    // type connects the generic method callback entry point to the specific
    // implementations.
    PyTypeObject CryptomatteLayerKnobPythonType = {
      PyObject_HEAD_INIT(&PyType_Type)
      "CryptomatteLayer_Knob",
      sizeof(CryptomatteLayerKnobObject),
      0,
      CryptomatteLayerKnobDealloc,/* tp_dealloc */
      nullptr,                    /* tp_print */
      nullptr,                    /* tp_getattr */
      nullptr,                    /* tp_setattr */
      nullptr,                    /* tp_compare */
      nullptr,                    /* tp_repr */
      nullptr,                    /* tp_as_number */
      nullptr,                    /* tp_as_sequence */
      nullptr,                    /* tp_as_mapping */
      nullptr,                    /* tp_hash */
      nullptr,                    /* tp_call */
      nullptr,                    /* tp_str */
      PyObject_GenericGetAttr,    /* tp_getattro */
      PyObject_GenericSetAttr,    /* tp_setattro */
      nullptr,                    /* tp_as_buffer */
      Py_TPFLAGS_DEFAULT,         /* tp_flags */
      nullptr,                    /* tp_doc */
      nullptr,                    /* tp_traverse */
      nullptr,                    /* tp_clear */
      nullptr,                    /* tp_richcompare */
      0,                          /* tp_weaklistoffset */
      nullptr,                    /* tp_iter */
      nullptr,                    /* tp_iternext */
      CryptomatteLayerKnobMethods,/* tp_methods */
      nullptr,                    /* tp_members */
      nullptr,                    /* tp_getset */
      nullptr,                    /* tp_base */
      nullptr,                    /* tp_dict */
      nullptr,                    /* tp_descr_get */
      nullptr,                    /* tp_descr_set */
      0,                          /* tp_dictoffset */
      nullptr,                    /* tp_init */
      PyType_GenericAlloc,        /* tp_alloc */
      CryptomatteLayerKnobNew,    /* tp_new */
      PyObject_Del,               /* tp_free */
    };

    CryptomatteLayer_Knob::CryptomatteLayer_Knob(Knob_Closure* knobClosure, std::string* selectedCryptomatteLayerName, const char* name, const char* label)
      : Knob(knobClosure, name, label)
    {
      mFnAssert(selectedCryptomatteLayerName != nullptr);
      _selectedCryptomatteLayerName = *selectedCryptomatteLayerName;
      setPythonType(&CryptomatteLayerKnobPythonType);
    }

    bool CryptomatteLayer_Knob::not_default() const
    {
      return !_selectedCryptomatteLayerName.empty();
    }

    void CryptomatteLayer_Knob::store(StoreType storeType, void* data, Hash& hash, const OutputContext& context)
    {
      // Do not set the data back if the caller passed in a nullptr in the
      // knobs() method.
      if (data != nullptr) {
        // Store the current value back to the member variable of the Op
        // subclass passed in the knobs() method.
        auto dataString = static_cast<std::string*>(data);
        *dataString = _selectedCryptomatteLayerName;
      }

      hash.append(_selectedCryptomatteLayerName);
    }

    PluginPython_KnobI* CryptomatteLayer_Knob::pluginPythonKnob()
    {
      return this;
    }

    void CryptomatteLayer_Knob::updateUI(const OutputContext& context)
    {
      // Add existing cryptomatte layers from the main input.
      const auto cryptomatteObjects = GetAvailableCryptomatteObjects(op());
      _cryptomatteLayerNames.clear();
      for (const auto& cryptomatteObject : cryptomatteObjects) {
        _cryptomatteLayerNames.push_back(cryptomatteObject.name);
      }

      // Add the currently selected layer to drop-down if it is a new layer
      // not coming from the main input (i.e. created by this instance)
      if (!_selectedCryptomatteLayerName.empty()) {
        if (std::find_if(cryptomatteObjects.begin(), cryptomatteObjects.end(), [this](CryptomatteObject cryptomatteObject){
          return cryptomatteObject.name == _selectedCryptomatteLayerName; }) == cryptomatteObjects.end()) {
          _cryptomatteLayerNames.push_back(_selectedCryptomatteLayerName);
        }
      }

      updateWidgets();
    }

    void CryptomatteLayer_Knob::menu(const std::vector<std::string>& cryptomatteLayerNames)
    {
      _cryptomatteLayerNames = cryptomatteLayerNames;
    }

    const std::vector<std::string>& CryptomatteLayer_Knob::menu() const
    {
      return _cryptomatteLayerNames;
    }

    bool CryptomatteLayer_Knob::from_script(const char* value)
    {
      // Note: if value is an empty string, the selected cryptomatte layer name
      // would not be changed by stringstream.
      _selectedCryptomatteLayerName.clear();
      std::stringstream ss(value);
      ss >> _selectedCryptomatteLayerName;
      changed();
      return true;
    }

    void CryptomatteLayer_Knob::to_script(std::ostream& os, const OutputContext* context, bool quote) const
    {
      if (quote) {
        std::stringstream ss;
        ss << _selectedCryptomatteLayerName;
        os << cstring(ss.str().c_str());
      }
      else {
        os << _selectedCryptomatteLayerName;
      }
    }

    WidgetPointer CryptomatteLayer_Knob::make_widget(const WidgetContext& context)
    {
      CryptomatteLayerWidget* widget = new CryptomatteLayerWidget(this);
      widget->buildGroup();
      return widget;
    }

    void CryptomatteLayer_Knob::setSelectedCryptomatteLayerName(const std::string& selectedCryptomatteLayerName)
    {
      if (_selectedCryptomatteLayerName != selectedCryptomatteLayerName) {
        new_undo(this);
        _selectedCryptomatteLayerName = selectedCryptomatteLayerName;
        changed();
      }
    }

    int CryptomatteLayer_Knob::getSelectedItemIndex() const
    {
      const auto selectedCryptomatteLayerNameIterator = std::find(_cryptomatteLayerNames.cbegin(), _cryptomatteLayerNames.cend(), _selectedCryptomatteLayerName);
      return std::distance(_cryptomatteLayerNames.cbegin(), selectedCryptomatteLayerNameIterator);
    }

    std::string CryptomatteLayer_Knob::getSelectedItemString() const
    {
      return _selectedCryptomatteLayerName;
    }

    std::string CryptomatteLayer_Knob::getItemValueString(int itemIndex) const
    {
      const std::string cryptomatteLayerName = _cryptomatteLayerNames.at(itemIndex);
      std::stringstream valueStream;
      for (size_t sublayerIndex = 0; sublayerIndex < kDefaultCryptoLayerDepth; ++sublayerIndex) {
        if (sublayerIndex != 0) {
          valueStream << " ";
        }
        valueStream << cryptomatteLayerName << std::setw(2) << std::setfill('0') << sublayerIndex;
      }
      return valueStream.str();
    }
  }
}
