// Copyright (c) 2021 The Foundry Visionmongers Ltd. All Rights Reserved.

#ifndef CRYPTOMATTELAYER_KNOB_H
#define CRYPTOMATTELAYER_KNOB_H

#include "CryptomatteUtils.h"

#include "DDImage/Enumeration_KnobI.h"
#include "DDImage/Knob.h"
#include "DDImage/Knobs.h"

namespace Foundry {
  namespace Encryptomatte {

    // Cryptomatte layer selection knob. This knob lists all the cryptomatte
    // layers available in the current stream. It also offers a way to create
    // new cryptomatte layers. This class is the control part of the
    // view-controller architecture of knobs.
    class CryptomatteLayer_Knob : public DD::Image::Knob
                                , public DD::Image::Enumeration_KnobI
                                , public DD::Image::PluginPython_KnobI
    {
    public:
      CryptomatteLayer_Knob(DD::Image::Knob_Closure* knobClosure, std::string* cryptomatteLayerName, const char* name, const char* label = nullptr);
      CryptomatteLayer_Knob() = delete;

      const char* Class() const override { return "CryptomatteLayer_Knob"; }
      bool not_default() const override;

      // Serialising the currently selected cryptomatte layer name when saving
      // the script.
      void to_script(std::ostream& os, const DD::Image::OutputContext* context, bool quote) const override;

      // Deserialising the cryptomatte layer name that was selected before
      // saving the script.
      bool from_script(const char* value) override;

      // This method makes sure that the rendering path is rerun when the
      // cryptomatte layer selection changes. It adds the currently selected
      // cryptomatte layer name to the hash to store the state uniquely.
      void store(DD::Image::StoreType storeType, void* dest, DD::Image::Hash& hash, const DD::Image::OutputContext& context) override;

      // Makes the cryptomatte layer plugin knob accessible through Python.
      DD::Image::PluginPython_KnobI* pluginPythonKnob() override;

      // Creates the GUI element for this knob.
      WidgetPointer make_widget(const DD::Image::WidgetContext& context) override;

      // The available cryptomatte layer names can dynamically change in the
      // script, so therefore the available options are not static on creation.
      // This is a helper method to set up the drop-down entries
      // dynamically at any point in time.
      void updateUI(const DD::Image::OutputContext& context) override;

      // Sets all the available cryptomatte layer options in the drop-down list.
      void menu(const std::vector<std::string>& cryptomatteLayerNames) override;

      // Return all the available cryptomatte layer options in the drop-down.
      // list.
      const std::vector<std::string>& menu() const override;

      // This is called by the widget to set the current cryptomatte layer name
      // selection.
      void setSelectedCryptomatteLayerName(const std::string& selectedCryptomatteLayerName = std::string());

      // Return the index of the currently selected cryptomatte layer name.
      int getSelectedItemIndex() const override;

      // Return the currently selected cryptomatte layer name.
      std::string getSelectedItemString() const override;

      // Return the serialised list of Nuke layers for the given item.
      std::string getItemValueString(int itemIndex) const override;

    private:
      // Currently selected cryptomatte layer from the drop-down list.
      std::string _selectedCryptomatteLayerName;
      // The list of cryptomatte layer name options available in the drop-down
      // list of the knob.
      std::vector<std::string> _cryptomatteLayerNames;
    };
  }
}

#endif
