// Copyright (c) 2021 The Foundry Visionmongers Ltd. All Rights Reserved.

#include "CryptomatteLayerWidget.moc.h"

#include "CryptomatteLayerDialog.moc.h"
#include "CryptomatteLayerKnob.h"

#include "Build/fnAssert.h"

#include <QtWidgets/QHBoxLayout>
#include <QtWidgets/QMenu>
#include <QtWidgets/QTabWidget>
#include <QtWidgets/QToolButton>

namespace Foundry {
  namespace Encryptomatte {

    using namespace DD::Image;

    const QString kNoCryptomatteLayerOption = "none";
    const QString kNewCryptomatteLayerOption = "new";

    CryptomatteLayerWidget::CryptomatteLayerWidget(CryptomatteLayer_Knob* knob)
      : _knob(knob)
    {
      _knob->addCallback(WidgetCallback, this);
    }

    CryptomatteLayerWidget::~CryptomatteLayerWidget()
    {
      if (_knob != nullptr) {
        _knob->removeCallback(WidgetCallback, this);
      }
    }

    int CryptomatteLayerWidget::WidgetCallback(void* closure, Knob::CallbackReason reason)
    {
      // TODO: Bail out if not on the main GUI thread as it is better not to do
      // anything to GUI widgets in a non-GUI working thread. As an internal
      // plugin: if (!NukeApplication::nukeInMainThread()) { return 0; }
      // The issue is that NukeApplication is not part of the public API for an
      // example plugin.
      CryptomatteLayerWidget* widget = reinterpret_cast<CryptomatteLayerWidget*>(closure);
      mFnAssert(widget);
      switch (reason) {
        case Knob::kIsVisible:
          // We check for visibility up to the containing tab widget.
          // This means that a widget is still considered visible when its
          // NodePanel is hidden due to being in hidden tab in a dock.

          for (QWidget* w = widget->parentWidget(); w != nullptr; w = w->parentWidget()) {
            if (qobject_cast<QTabWidget*>(w)) {
              return static_cast<int>(widget->isVisibleTo(w));
            }
          }
          return static_cast<int>(widget->isVisible());

        case Knob::kUpdateWidgets:
          widget->update();
          return 0;

        case Knob::kDestroying:
          widget->destroy();
          return 0;

        default:
          return 0;
      }
    }

    void CryptomatteLayerWidget::adjustWidthToContents()
    {
      const QFontMetrics fontMetrics(_layerChoice->font());
      const int oneCharacterWidth = fontMetrics.horizontalAdvance(QLatin1Char('x'));
      // Some reasonable minimum width in case the list is empty or contains
      // short item names.
      int fixedWidth = 20 * oneCharacterWidth;
      if (_layerMenu != nullptr) {
        const QList<QAction*> actionList = _layerMenu->actions();
        const auto longestNameAction = std::max_element(actionList.cbegin(), actionList.cend(), [fontMetrics] (QAction* leftHandSideAction, QAction* rightHandSideAction) { return fontMetrics.horizontalAdvance(leftHandSideAction->text()) < fontMetrics.horizontalAdvance(rightHandSideAction->text());});
        // Add some extra space around the string for margin, etc, as the width
        // seems to be a bit short otherwise.
        const int longestNameWidth = fontMetrics.horizontalAdvance((*longestNameAction)->text()) + 4 * oneCharacterWidth;
        fixedWidth = std::max(fixedWidth, longestNameWidth);
      }

      _layerChoice->setFixedWidth(fixedWidth);
    }

    void CryptomatteLayerWidget::populateMenu()
    {
      mFnAssert(_knob != nullptr);
      if (_knob == nullptr) {
        return;
      }

      _layerMenu->clear();

      // Add the "none" option to the drop-down list and select it by default.
      QAction* action = _layerMenu->addAction(kNoCryptomatteLayerOption, [this](){this->_knob->setSelectedCryptomatteLayerName();});
      QFont layerFont = _layerMenu->font();
      layerFont.setItalic(true);
      action->setFont(layerFont);
      const std::string selectedCryptomatteLayerName = _knob->getSelectedItemString();
      if (selectedCryptomatteLayerName.empty()) {
        action->setCheckable(true);
        action->setChecked(true);
      }

      // Add all the cryptomatte layer names to the drop-down list.
      for (const auto& cryptomatteLayerName : _knob->menu()) {
        QAction* action = _layerMenu->addAction(QString::fromStdString(cryptomatteLayerName), [this, cryptomatteLayerName](){this->_knob->setSelectedCryptomatteLayerName(cryptomatteLayerName);});
        if (!selectedCryptomatteLayerName.empty() && selectedCryptomatteLayerName == cryptomatteLayerName) {
          action->setCheckable(true);
          action->setChecked(true);
        }
      }

      // Add the new layer creation option to the end of the drop-down list.
      _layerMenu->addAction(kNewCryptomatteLayerOption, this, &CryptomatteLayerWidget::showNewCryptomatteLayerDialog)->setFont(layerFont);

      adjustWidthToContents();
    }

    void CryptomatteLayerWidget::showNewCryptomatteLayerDialog()
    {
      CryptomatteLayerDialog dialog;
      const QDialog::DialogCode dialogCode = static_cast<QDialog::DialogCode>(dialog.exec());
      if (dialogCode == QDialog::Accepted) {
        const std::string cryptomatteLayerName = dialog.cryptomatteLayerName();
        if (!cryptomatteLayerName.empty()) {
          _knob->setSelectedCryptomatteLayerName(cryptomatteLayerName);
        }
        else {
          _knob->updateWidgets(); // gets the word "new" out of the widget
        }
      }
    }

    void CryptomatteLayerWidget::buildGroup()
    {
      _layout = new QHBoxLayout();
      _layout->setMargin(0);
      setLayout(_layout);

      _layerMenu = new QMenu();
      connect(_layerMenu, &QMenu::aboutToShow, this, &CryptomatteLayerWidget::populateMenu);
      _layerChoice = new QToolButton();
      _layerChoice->setPopupMode(QToolButton::InstantPopup);
      _layerChoice->setProperty("foundry.nuke.button.text.alignment", {Qt::AlignLeft | Qt::AlignVCenter});
      _layerChoice->update();

      _layerChoice->setFocusPolicy(Qt::NoFocus);
      _layerChoice->setMenu(_layerMenu);
      _layout->addWidget(_layerChoice, 0, Qt::AlignLeft);

      connect(_layerChoice, &QToolButton::destroyed, this, &CryptomatteLayerWidget::childDestroyed);
    }

    void CryptomatteLayerWidget::childDestroyed(QObject* destroyedObject)
    {
      if (destroyedObject == _layerChoice)  {
        _layerChoice = nullptr;
      }
    }

    void CryptomatteLayerWidget::update()
    {
      QString text = QString::fromStdString(_knob->getSelectedItemString());

      if (text.isEmpty()) {
        text = kNoCryptomatteLayerOption;
      }

      if (_layerChoice != nullptr) {
        const QString previousText = _layerChoice->text();
        if (text != previousText) {
          _layerChoice->setText(text);
        }
        else {
          populateMenu();
        }
      }
    }
  }
}
