// Copyright (c) 2021 The Foundry Visionmongers Ltd. All rights reserved.

// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are met:
//
//     * Redistributions of source code must retain the above copyright notice,
//       this list of conditions and the following disclaimer.
//     * Redistributions in binary form must reproduce the above copyright notice,
//       this list of conditions and the following disclaimer in the documentation
//       and/or other materials provided with the distribution.
//     * Neither the name of Foundry nor the names of its
//       contributors may be used to endorse or promote products derived from this
//       software without specific prior written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
// ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
// WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
// DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDERS OR CONTRIBUTORS BE LIABLE FOR
// ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
// (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
// LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
// ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
// SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

// This work is a C++ implementation compatible with version 1.2.8 of the Python-based Nuke gizmo
// authored at Psyop by Jonah Friedman and Andy Jones (see https://github.com/Psyop/Cryptomatte).

#include "CryptomatteManifest.h"
#include "CryptomatteUtils.h"
#include <Build/fnAssert.h>
#include <QJsonDocument>
#include <QJsonObject>
#include <QStringList>
#include <string>
#include <assert.h>

namespace Foundry {
  namespace Cryptomatte {

    Manifest::Manifest(const std::string& strJson)
    {
      // Normally this shouldn't happen on any of officially supported platforms,
      // however let's make sure that the size of int equals the size of float.
      if (sizeof(unsigned int) != sizeof(float)) {
        mFnAssertMsg(false, "The Manifest feature of the Cryptomatte plugin isn't "
          "available for platforms where the size of unsigned integer and the size "
          "of single-precision floating-point don't match.");
        return;
      }

      insert(strJson);
    }

    bool Manifest::insert(const std::string& strJson)
    {
      if (strJson.empty()) {
        return true;
      }

      // Parse JSON
      const QByteArray byteArray(strJson.c_str(), static_cast<int>(strJson.size()));
      QJsonParseError parseErrorHandler{};
      const QJsonDocument jsonDocument = QJsonDocument::fromJson(byteArray, &parseErrorHandler);

      // If parsing wasn't successful, return with an empty Manifest.
      if (parseErrorHandler.error != QJsonParseError::NoError) {
        return false;
      }

      // Loop through nested objects of JSON's root object,
      // and populate manifest items created from them.
      QJsonObject jsonRootObject = jsonDocument.object();
      for (const auto& key : jsonRootObject.keys()) {
        // Get the name of Cryptomatte Object
        const std::string objectName(key.toUtf8().constData());
        auto jsonObjectValue = jsonRootObject[key];
        if (!jsonObjectValue.isString()) {
          continue;
        }
        // Get the ID of Cryptomatte Object, presented in HEX format
        bool conversionSucceeded;
        unsigned int objectIDHex = jsonObjectValue.toString().toUInt(&conversionSucceeded, 16);
        if (!conversionSucceeded) {
          continue;
        }
        // Reinterpret HEX format of Cryptomatte Object's ID to the Floating-point format
        const float objectIDFloat = *(reinterpret_cast<float*>(&objectIDHex));
        insert(objectName, objectIDFloat);
      }
      return true;
    }

    bool Manifest::insert(const std::string& objectName, float objectID)
    {
      // Make sure the ID of Cryptomatte Object matches the hashed value of its name;
      // otherwise skip it, as this is not a valid ID-hash pair.
      if (GetMurmurHashValueFloat(objectName) == objectID) {
        // Insert items in both maps of the bimap, name-to-id and id-to-name
        _objectNameToIDMap.insert({objectName, objectID});
        _objectIDToNameMap.insert({objectID, objectName});
        return true;
      }
      return false;
    }

    float Manifest::find(const std::string& objectName) const
    {
      const auto itrNameAndID = _objectNameToIDMap.find(objectName);
      if (itrNameAndID == _objectNameToIDMap.end()) {
        return kFillingObjectId;
      }
      return itrNameAndID->second;
    }

    std::string Manifest::find(float objectID) const
    {
      const auto itrIDAndName = _objectIDToNameMap.find(objectID);
      if (itrIDAndName == _objectIDToNameMap.end()) {
        return std::string();
      }
      return itrIDAndName->second;
    }

    std::vector<std::string> Manifest::objectNames() const
    {
      std::vector<std::string> names;
      names.reserve(_objectIDToNameMap.size());
      for (const auto& pair : _objectIDToNameMap) {
        names.push_back(pair.second);
      }
      return names;
    }

    bool Manifest::empty() const
    {
      return _objectIDToNameMap.empty();
    }

    std::string Manifest::toJson() const
    {
      QJsonObject jsonObject;
      for (auto const& element : _objectNameToIDMap) {
        const std::string& matteName = element.first;
        jsonObject[QString::fromStdString(matteName)] = QString::fromStdString(GetMurmurHashValueFloatToHex(element.second));
      }
      return QJsonDocument(jsonObject).toJson(QJsonDocument::Compact).toStdString();
    }
  }
}
