// Copyright (c) 2021 The Foundry Visionmongers Ltd. All rights reserved.

// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are met:
//
//     * Redistributions of source code must retain the above copyright notice,
//       this list of conditions and the following disclaimer.
//     * Redistributions in binary form must reproduce the above copyright notice,
//       this list of conditions and the following disclaimer in the documentation
//       and/or other materials provided with the distribution.
//     * Neither the name of Foundry nor the names of its
//       contributors may be used to endorse or promote products derived from this
//       software without specific prior written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
// ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
// WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
// DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDERS OR CONTRIBUTORS BE LIABLE FOR
// ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
// (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
// LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
// ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
// SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

// This work is a C++ implementation compatible with version 1.2.8 of the Python-based Nuke gizmo
// authored at Psyop by Jonah Friedman and Andy Jones (see https://github.com/Psyop/Cryptomatte).

#ifndef CRYPTOMATTE_MANIFEST_H
#define CRYPTOMATTE_MANIFEST_H

#include <map>
#include <string>
#include <vector>

namespace Foundry {
  namespace Cryptomatte {

    // The manifest contains both the names and the IDs of all of the objects in the
    // Cryptomatte type. It may be included in full in the header, or it may be a separate file.
    // The format of the manifest is a JSON dictionary, where the names are the keys and the
    // IDs are the values.
    class Manifest
    {
    public:
      // Creates an empty Manifest instance.
      Manifest() = default;

      // Creates a Manifest instance and populates it with items from a given text in JSON format.
      explicit Manifest(const std::string& strJson);

      // Insert the items from a given text in JSON format.
      bool insert(const std::string& strJson);

      // Inserts a pair of object name and ID into the manifest, if the pair has valid values.
      // For any valid pair, a hashed value of object name should match the value of object ID.
      // Returns true, if the pair passed the validation and was inserted; otherwise false.
      bool insert(const std::string& objectName, float objectID);

      // Finds an associated object ID by the given name. If the given name is found, method
      // returns the associated object ID; otherwise, kInvalidObjectID is returned.
      float find(const std::string& objectName) const;

      // Finds an object name by the given associated ID. If the given name is found, method
      // returns the object name; otherwise, empty string is returned.
      std::string find(float objectID) const;

      // Returns a list of Matte names in this Manifest object.
      std::vector<std::string> objectNames() const;
      
      // Returns whether the Manifest is empty.
      bool empty() const;

      // Returns the manifest in json string format
      std::string toJson() const;

    private:
      // The manifest is implemented as a bimap for fast reverse lookup.
      // Below there is a pair of maps, name-to-id and id-to-name, that form the bimap.
      std::map<std::string, float> _objectNameToIDMap;
      std::map<float, std::string> _objectIDToNameMap;
    };

  }
}

#endif
