// Copyright (c) 2021 The Foundry Visionmongers Ltd. All rights reserved.

// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are met:
//
//     * Redistributions of source code must retain the above copyright notice,
//       this list of conditions and the following disclaimer.
//     * Redistributions in binary form must reproduce the above copyright notice,
//       this list of conditions and the following disclaimer in the documentation
//       and/or other materials provided with the distribution.
//     * Neither the name of Foundry nor the names of its
//       contributors may be used to endorse or promote products derived from this
//       software without specific prior written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
// ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
// WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
// DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDERS OR CONTRIBUTORS BE LIABLE FOR
// ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
// (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
// LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
// ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
// SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

// This work is a C++ implementation compatible with version 1.2.8 of the Python-based Nuke gizmo
// authored at Psyop by Jonah Friedman and Andy Jones (see https://github.com/Psyop/Cryptomatte).

#include "CryptomatteMatteExpression.h"
#include "CryptomatteManifest.h"
#include "CryptomatteUtils.h"
#include "Build/fnAssert.h"
#include <sstream>
#include <iomanip>
#include <assert.h>

using namespace Foundry::Cryptomatte;

namespace {

  constexpr char kFloatingPointIdRegex[] = "<(-?[0-9]+\\.[0-9]+(e[-+][0-9]+)?)>";

  // Quote 'matteName' if it begins or ends with space character, or contains comma
  std::string Quote(const std::string& matteName)
  {
    if (!matteName.empty() &&
        (matteName.front() == kSpace
          || matteName.back() == kSpace
          || matteName.find(',') != std::string::npos)) {
      return R"(")" + matteName + R"(")";
    }
    return matteName;
  }

  std::string KnobEscape(const std::string& matteName)
  {
    if (matteName.empty()) {
      return std::string();
    }
    std::string escaped;
    for (size_t i = 0; i < matteName.size(); ++i) {
      const char c = matteName[i];
      switch (c) {
        // Prepend a backslash to \* and \ in for GUI representation.
      case kEscape:
      case kOpenSquareBracket:
      case kCloseSquareBracket:
        escaped += kEscape;
        break;
        // Prepend double backslahes to a double quote for GUI representation.
      case '"':
        escaped += R"(\\)";
        break;
      default:
        break;
      }
      escaped += c;
    }
    return escaped;
  }
}

namespace Foundry {
  namespace Cryptomatte {

    MatteExpression::MatteExpression(const std::string& expression)
    : _format(IsFloatId(expression) ? eFloatingPointObjectID : eObjectName)
    {
      for (size_t i = 0; i < expression.size(); ++i) {
        if (EscapedChar(expression, i)) {
          _value += kEscape;
          ++i;
        }
        else if (expression[i] == '*') {
          _format = eWildcardExpression;
        }
        _value += expression[i];
      }
      _wildcardRegex = std::regex(CreateRegex(_value));
    }

    MatteExpression::MatteExpression(float objectID, const Manifest& manifest)
    : _format(eObjectName)
    {
      // Check if the manifest contains certain id. If the id was found, get its associated name.
      const auto objectName = manifest.find(objectID);
      if (!objectName.empty()) {
        _value = Escape(objectName);
        _wildcardRegex = std::regex(CreateRegex(_value));
      }
      else {
        // Otherwise, generate a name by the floating point value of the id,
        // wrapped with angle brackets.
        // Note: the precision value below was found empirically and works
        // well in terms of backward compatibility with the gizmo.
        std::ostringstream sstream;
        sstream << '<' << std::scientific << std::setprecision(11) << objectID << '>';
        _value = sstream.str();
        _format = eFloatingPointObjectID;
      }
      mFnAssert(_format == eObjectName || _format == eFloatingPointObjectID);
    }

    bool MatteExpression::match(const std::string& objectName) const
    {
      return std::regex_match(Unescape(objectName), _wildcardRegex);
    }

    std::vector<std::string> MatteExpression::matches(const Manifest& manifest) const
    {
      std::vector<std::string> selectedMattes;
      switch (_format) {
        case eObjectName:
        case eWildcardExpression: {
          for (const auto& name : manifest.objectNames()) {
            if (!name.empty() && match(name)) {
              selectedMattes.push_back(name);
            }
          }
          break;
        }
        case eFloatingPointObjectID: {
          break;
        }
      }
      return selectedMattes;
    }

    std::set<float> MatteExpression::resolveIDs(const Manifest& manifest) const
    {
      std::set<float> selectedIDs;
      switch (_format) {
      case eObjectName: {
        // If object name format, populate a floating point reinterpretation of 32-bit Murmur3 hash
        // of the object name.
        selectedIDs.insert(GetMurmurHashValueFloat(Unescape(_value)));
        break;
      }
      case eFloatingPointObjectID: {
        // If object ID format, extract the floating point value from the expression.
        // Populate the set with the object ID, if successfully extracted.
        selectedIDs.insert(getObjectIDFromFloatingPointExpression());
        break;
      }
      case eWildcardExpression:
        for (const auto& objectName : matches(manifest)) {
          selectedIDs.insert(GetMurmurHashValueFloat(objectName));
        }
      }
      // Make sure selected IDs don't contain the invalid object ID
      selectedIDs.erase(kFillingObjectId);
      return selectedIDs;
    }

    float MatteExpression::getObjectIDFromFloatingPointExpression() const
    {
      static const auto pattern = std::regex(kFloatingPointIdRegex);
      std::smatch match;
      if (std::regex_match(_value, match, pattern)) {
        return std::stof(match[1].str());
      }
      return kFillingObjectId;
    }

    bool MatteExpression::IsFloatId(const std::string& floatExpression)
    {
      static const auto kFloatRegex = std::regex(kFloatingPointIdRegex);
      return std::regex_match(floatExpression, kFloatRegex);
    }

    AddMatteExpression::AddMatteExpression(const std::string& expression)
      : MatteExpression(expression)
    {
    }

    AddMatteExpression::AddMatteExpression(float objectID, const Manifest& manifest)
      : MatteExpression(objectID, manifest)
    {      
    }

    void AddMatteExpression::apply(std::set<float>& ids, const Manifest& manifest) const
    {
      std::set<float> selectedIDs = resolveIDs(manifest);
      ids.insert(selectedIDs.begin(), selectedIDs.end());
    }

    std::string AddMatteExpression::str(bool firstInSection) const
    {
      return Quote(KnobEscape(_value));
    }

    RemoveMatteExpression::RemoveMatteExpression(const std::string& expression)
      : MatteExpression(expression)
    {
    }

    RemoveMatteExpression::RemoveMatteExpression(float objectID, const Manifest& manifest)
      : MatteExpression(objectID, manifest)
    {
    }

    void RemoveMatteExpression::apply(std::set<float>& ids, const Manifest& manifest) const
    {
      std::set<float> selectedIDs = resolveIDs(manifest);
      for (float id : selectedIDs) {
        ids.erase(id);
      }
    }

    std::string RemoveMatteExpression::str(bool firstInSection) const
    {
      std::string prefix;
      if (!firstInSection) {
        prefix += kSpace;
      }
      prefix += kMinusSign;
      if (!firstInSection) {
        prefix += kSpace;
      }      
      return prefix + Quote(KnobEscape(_value));
    }

  }
}
