// Copyright (c) 2021 The Foundry Visionmongers Ltd. All rights reserved.

// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are met:
//
//     * Redistributions of source code must retain the above copyright notice,
//       this list of conditions and the following disclaimer.
//     * Redistributions in binary form must reproduce the above copyright notice,
//       this list of conditions and the following disclaimer in the documentation
//       and/or other materials provided with the distribution.
//     * Neither the name of Foundry nor the names of its
//       contributors may be used to endorse or promote products derived from this
//       software without specific prior written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
// ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
// WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
// DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDERS OR CONTRIBUTORS BE LIABLE FOR
// ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
// (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
// LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
// ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
// SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

// This work is a C++ implementation compatible with version 1.2.8 of the Python-based Nuke gizmo
// authored at Psyop by Jonah Friedman and Andy Jones (see https://github.com/Psyop/Cryptomatte).

#ifndef CRYPTOMATTE_MATTEEXPRESSION_H
#define CRYPTOMATTE_MATTEEXPRESSION_H

#include <string>
#include <set>
#include <regex>
#include "CryptomatteManifest.h"

namespace Foundry {
  namespace Cryptomatte {

    // Matte expression is a single string expression line of a multiline Matte List.
    class MatteExpression
    {
    public:
      // Enumerates possible formats of the matte expression.
      enum Format {
        // The object name, that may be one of the manifest keys.
        eObjectName,
        // The floating-point value of object ID, wrapped with angle brackets (e.g. <2.07262543558e+26>).
        eFloatingPointObjectID,
        // The wildcard expression contains normal characters in combination with one or many appearances of
        // the special character `*`. Normal characters means any latter and digit character or `_` symbol.
        // The expression may be resolved into zero or many object IDs, depending on the pattern it represents
        // and supplied manifest.
        eWildcardExpression
      };

      // Constructs a matte expression instance from the string expression.
      MatteExpression(const std::string& expression);

      // Creates an instance of the matte expression based on given object ID
      // and the manifest that may contain a matte name associated with the ID.
      MatteExpression(float objectID, const Manifest& manifest);

      virtual ~MatteExpression() = default;

      // Returns true if objectName matches the current expression.
      bool match(const std::string& objectName) const;

      // Returns all matte names in 'manifest' that match the current expression.
      std::vector<std::string> matches(const Manifest& manifest) const;

      // Depending on the type, the expression either adds the resolved matte IDs to
      // or removes them from the passed set. The type and the format of expression define
      // how IDs are resolved.
      virtual void apply(std::set<float>& ids, const Manifest& manifest = {}) const = 0;

      // Serializes the expression to string.
      virtual std::string str(bool firstInSection) const = 0;

      // Gets the string value of the expression.
      inline const std::string& getValue() const { return _value; }

      // Gets the format of the expression.
      inline Format getFormat() const { return _format; }

      // Determines whether the given expression is a floating-point object ID expression.
      static bool IsFloatId(const std::string& expression);

  protected:

      // Returns a set of IDs, resolved by the expression in regards to a given manifest.
      std::set<float> resolveIDs(const Manifest& manifest) const;

      // If the format of expression expression is object ID, extracts and returns
      // a floating-point value from it. Otherwise returns InvalidObjectID.
      float getObjectIDFromFloatingPointExpression() const;

      // The value of the expression.
      std::string _value;
      // The format of the expression.
      Format _format = eObjectName;
      // Regular expression of wildcard matte expression, if appropriate format;
      // otherwise empty regular expression.
      std::regex _wildcardRegex;
    };

    inline bool operator==(const MatteExpression& lhs, const MatteExpression& rhs) { return lhs.getValue() == rhs.getValue(); }
    inline bool operator!=(const MatteExpression& lhs, const MatteExpression& rhs) { return !operator==(lhs, rhs); }
    inline bool operator< (const MatteExpression& lhs, const MatteExpression& rhs) { return lhs.getValue() < rhs.getValue(); }
    inline bool operator> (const MatteExpression& lhs, const MatteExpression& rhs) { return  operator<(rhs, lhs); }
    inline bool operator<=(const MatteExpression& lhs, const MatteExpression& rhs) { return !operator>(lhs, rhs); }
    inline bool operator>=(const MatteExpression& lhs, const MatteExpression& rhs) { return !operator<(lhs, rhs); }

    // The expression that adds resolved IDs to a set of selected IDs.
    class AddMatteExpression final : public MatteExpression
    {
    public:
      AddMatteExpression(const std::string& expression);
      AddMatteExpression(float objectID, const Manifest& manifest);

      void apply(std::set<float>& ids, const Manifest& manifest = {}) const override;
      std::string str(bool firstInSection) const override;
    };

    // The expression that removes resolved IDs from a set of selected IDs.
    // This expression is designated by a '-' special character in the matte list.
    class RemoveMatteExpression final : public MatteExpression
    {
    public:
      RemoveMatteExpression(const std::string& expression);
      RemoveMatteExpression(float objectID, const Manifest& manifest);

      void apply(std::set<float>& ids, const Manifest& manifest = {}) const override;
      std::string str(bool firstInSection) const override;
    };

  }
}

#endif
