// Copyright (c) 2021 The Foundry Visionmongers Ltd. All rights reserved.

// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are met:
//
//     * Redistributions of source code must retain the above copyright notice,
//       this list of conditions and the following disclaimer.
//     * Redistributions in binary form must reproduce the above copyright notice,
//       this list of conditions and the following disclaimer in the documentation
//       and/or other materials provided with the distribution.
//     * Neither the name of Foundry nor the names of its
//       contributors may be used to endorse or promote products derived from this
//       software without specific prior written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
// ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
// WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
// DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDERS OR CONTRIBUTORS BE LIABLE FOR
// ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
// (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
// LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
// ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
// SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

// This work is a C++ implementation compatible with version 1.2.8 of the Python-based Nuke gizmo
// authored at Psyop by Jonah Friedman and Andy Jones (see https://github.com/Psyop/Cryptomatte).

#include "CryptomatteMatteList.h"
#include "CryptomatteMatteExpression.h"
#include "CryptomatteUtils.h"
#include "Build/fnAssert.h"
#include <algorithm>
#include <cassert>
#include <cctype>

namespace {
  using namespace Foundry::Cryptomatte;

  bool IsSpace(const std::string& value)
  {
    return std::all_of(value.begin(), value.end(), [](char c) { return std::isspace(c); });
  }

  class MatteListTokenizer
  {
    // Enumerates types the matte expression.
    enum class MatteExpressionType
    {
      eAdd,
      eRemove
    };

    std::string _token;
    // Not to remove spaces and tabs between words outside quotes
    bool _inWord = false;
    // Type of next created matte expression
    MatteExpressionType _expressionType = MatteExpressionType::eAdd;
    // Determines if new matte section should be opened for the next expression
    bool _openNewSection = true;
    // To indicate whether we are inside a quoted sequence to take characters
    // literally, except special characters ('\', '*', and '"')
    bool _quoteSequence = false;
    // Counting trailing spaces and tabs outside quotes for popping, otherwise
    // it would be considered _inWord
    size_t _trailingSpaceTabCounter = 0;

    void resetState(bool openNewSection = true, MatteExpressionType expressionType = MatteExpressionType::eAdd)
    {
      _token.clear();
      _inWord = false;
      _expressionType = expressionType;
      _openNewSection = openNewSection;
      _quoteSequence = false;
      _trailingSpaceTabCounter = 0;
    }

    // Concatenates trailing expressions in the last section of the given `row` with
    // `trailingExpressions` and checks if this could match the floating-point id
    // expression format.
    // If matches, trailing expressions are extracted from the last section of the row,
    // a new floating-point expression is added, and true is returned.
    // Otherwise, false is returned.
    // To handle the case with two minus signs in the floating-point expression (one
    // for the sign of value and one for the sign of mantissa), the method is
    // implemented in the recursive manner.
    bool processTrailingFloatId(MatteRow& row, const std::string& trailingExpressions,
      size_t expressionBackIndex = 0, size_t maxExpressions = 2)
    {
      // Make sure the trailing expressions end with the closing angle bracket
      if (trailingExpressions.back() != kClosingAngleBracket) {
        return false;
      }
      // Make sure the row isn't empty
      if (row.empty()) {
        return false;
      }
      // Make sure the current index is in range of expressions of the last section
      auto& lastSection = row.back();
      const size_t numberExpressions = lastSection.size();
      if (numberExpressions <= expressionBackIndex) {
        return false;
      }
      // Make sure the current expression is of the object name format
      const auto& expression = lastSection[numberExpressions - expressionBackIndex - 1];
      if (expression->getFormat() != MatteExpression::Format::eObjectName) {
        return false;
      }
      // Determine whether the current expression is an add expression
      const bool isAddExpression = dynamic_cast<AddMatteExpression*>(expression.get()) != nullptr;
      // Merge the current expression with the trailing expressions and
      // check if it matches the floating-point ID expression format
      const std::string expressionString = expression->getValue();
      std::string mergedExpression = expressionString + trailingExpressions;
      if (MatteExpression::IsFloatId(mergedExpression)) {
        std::shared_ptr<MatteExpression> expression;
        if (isAddExpression) {
          expression = std::make_shared<AddMatteExpression>(mergedExpression);
        }
        else {
          expression = std::make_shared<RemoveMatteExpression>(mergedExpression);
        }
        lastSection.erase(lastSection.end() - expressionBackIndex - 1, lastSection.end());
        lastSection.push_back(expression);
        return true;
      }
      // If not floating-point ID, check if the next level of recursion is available
      if (++expressionBackIndex < maxExpressions) {
        if (!isAddExpression) {
          mergedExpression = kMinusSign + mergedExpression;
        }
        return processTrailingFloatId(row, mergedExpression, expressionBackIndex, maxExpressions);
      }
      // The max recursion level has been exceeded, but floating-point ID hasn't been detected
      return false;
    }

    bool processDelimiter(MatteRow& row)
    {
      if (_token.empty() ||  IsSpace(_token)) {
        return false;
      }
      // Remove any trailing whitespaces when reaching the token delimiter
      if (_trailingSpaceTabCounter > 0) {
        _token.erase(_token.size() - _trailingSpaceTabCounter);
      }
      // Process a floating-point id expression, if concatenation of the token with
      // trailing expressions matches the floating-point expression format.
      const std::string floatIdTrailingExpressions = (_expressionType == MatteExpressionType::eAdd ? _token : kMinusSign + _token);
      if (processTrailingFloatId(row, floatIdTrailingExpressions)) {
        return true;
      }
      // Add the token to the line of tokens
      std::shared_ptr<MatteExpression> expression;
      if (_expressionType == MatteExpressionType::eAdd) {
        expression = std::make_shared<AddMatteExpression>(_token);
      }
      else if (_expressionType == MatteExpressionType::eRemove) {
        expression = std::make_shared<RemoveMatteExpression>(_token);
      }
      else {
        mFnAssertMsg(false, "Expression type isn't supported.");
        return false;
      }
      if (_openNewSection) {
        row.push_back({});
      }
      row.back().emplace_back(expression);
      return true;
    }

  public:

    void tokenize(const std::string& source, std::vector<MatteRow>& matteExpressions)
    {
      if (source.empty()) {
        return;
      }
      resetState();
      MatteRow row;
      for (char c : source) {
        if (c != kSpace && c != kTab) {
          _inWord = true;
          if (c != kComma && c != kLineBreak && c != kMinusSign) {
            _trailingSpaceTabCounter = 0;
          }
        }
        switch (c) {
        case kDoubleQuote:
          // Remove the escaping backslash if followed by a double quote
          if (!_token.empty() && _token.back() == kEscape) {
            _token.pop_back();
            _token += c;
          }
          else {
            // This quote has a special meaning, so do not treat it as literal.
            // Do not add it. Mark the state that we are inside quotes and
            // words now.
            _quoteSequence = !_quoteSequence;
            _inWord = _quoteSequence;
          }
          break;
        case kComma:
          // Comma is literal, not delimiter inside quotes
          if (_quoteSequence) {
            _token += c;
          }
          else {
            processDelimiter(row);
            resetState();
          }
          break;
        case kLineBreak:
          processDelimiter(row);
          resetState();
          if (!row.empty()) {
            // Add the line of tokens to the overall token container
            matteExpressions.push_back(row);
          }
          // Reset the state for the first token of the next line
          row.clear();
          break;
        case kSpace:
        case kTab:
          // Take these literally inside quotes
          if (_quoteSequence) {
            _token += c;
          }
          // Take these literally after a word, while also counting the spaces
          // and tabs for later removal when no more word characters.
          else if (_inWord) {
            _token += c;
            ++_trailingSpaceTabCounter;
          }
          break;
        case kMinusSign:
          // If dash is in quotes or escaped, treat it as a regular character
          if (_quoteSequence || (!_token.empty() && _token.back() == kEscape)) {
            _token += c;
          }
          else {
            const bool expressionAdded = processDelimiter(row);
            resetState(!expressionAdded, MatteExpressionType::eRemove);
          }
          break;
        default:
          // Other characters are not special, so just add them literally.
          _token += c;
          break;
        }
      }

      // If no matching quote pairs, just invalidate the token.
      if (_quoteSequence) {
        _token.clear();
      }

      // If no comma or line break found for the last token,
      // add the token to the list of tokens
      processDelimiter(row);

      // If no comma or line break found for the last line, add the list of
      // tokens to the tokens container if not empty 
      if (!row.empty()) {
        matteExpressions.push_back(row);
      }
    }
  };
}

namespace Foundry {
  namespace Cryptomatte {

    MatteList::MatteList(const std::string& strMatteList)
    {
      MatteListTokenizer tokenizer;
      tokenizer.tokenize(strMatteList, _matteExpressions);
    }

    std::set<float> MatteList::resolveIDs(const Manifest& manifest) const
    {
      // Go through each matte expression and populate the out set
      // with underlying object IDs.
      std::set<float> resolvedIDs;
      for (const auto& matteRow : _matteExpressions) {
        for (const auto& matteSection : matteRow) {
          for (const auto& matteExpression : matteSection) {
            matteExpression->apply(resolvedIDs, manifest);
          }
        }
      }
      return resolvedIDs;
    }

    void MatteList::selectByID(float objectID, const Manifest& manifest)
    {
      const auto expression = std::make_shared<AddMatteExpression>(objectID, manifest);
      // Remove exact matches with the value of new expression in the matte list
      removeExactMatches(expression->getValue());
      // Add the new expression to the list.
      _matteExpressions.push_back({{expression}});
    }

    void MatteList::unselectByID(float objectID, const Manifest& manifest)
    {
      const auto expression = std::make_shared<RemoveMatteExpression>(objectID, manifest);
      // Remove exact matches with the value of new expression in the matte list
      const bool anyWildcardMatch = removeExactMatches(expression->getValue());
      if (anyWildcardMatch) {
        // Add the new expression to the list.
        _matteExpressions.push_back({ {expression} });
      }
    }

    // Converts the internal matte list representation to the user visible text
    // in the corresponding knob. It will escape and quote the special
    // characters in the expression as it is needed. It will also add the
    // necessary delimiters automatically - commas in the same line, and
    // line break for new lines.
    std::string MatteList::str() const
    {
      std::string matteListString;
      for (const auto& matteRow : _matteExpressions) {
        if (!matteListString.empty()) {
          matteListString += kLineBreak;
        }
        bool firstSectionInRow = true;
        for (const auto& matteSection : matteRow) {
          bool firstExpressionInSection = true;
          if (!firstSectionInRow) {
            matteListString += kComma;
          }
          for (const auto& matteExpression : matteSection) {
            matteListString += matteExpression->str(firstExpressionInSection);
            firstExpressionInSection = false;
          }
          firstSectionInRow = false;
        }
      }
      return matteListString;
    }

    bool MatteList::removeExactMatches(const std::string& matteName)
    {
      bool anyWildcardMatch = false;
      // Remove exact matches.
      for (auto& matteRow : _matteExpressions) {
        matteRow.erase(std::remove_if(matteRow.begin(), matteRow.end(),
          [&matteName, &anyWildcardMatch](const MatteSection& section) {
            if (section.front()->getFormat() == MatteExpression::eWildcardExpression) {
              // If any wildcard expression of the current section matches the matte name,
              // set the corresponding flag.
              anyWildcardMatch |= std::any_of(section.begin(), section.end(),
                [&matteName](const std::shared_ptr<MatteExpression>& expression) {
                  return expression->match(matteName);
                });
              // Since we got to this point, just return false.
              // The return expression below will always be negative.
              return false;
            }
            // Return whether the current section consists of a single matching expression that
            // the Picker Add or the Picker Remove could have added to the matte list.
            return section.size() == 1 && section.front()->match(matteName);
          }), matteRow.end());
      }
      // Remove empty rows
      _matteExpressions.erase(std::remove_if(_matteExpressions.begin(), _matteExpressions.end(),
        [](const MatteRow& row) { return row.empty(); }), _matteExpressions.end());
      return anyWildcardMatch;
    }
  }
}
