// Copyright (c) 2021 The Foundry Visionmongers Ltd. All rights reserved.

// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are met:
//
//     * Redistributions of source code must retain the above copyright notice,
//       this list of conditions and the following disclaimer.
//     * Redistributions in binary form must reproduce the above copyright notice,
//       this list of conditions and the following disclaimer in the documentation
//       and/or other materials provided with the distribution.
//     * Neither the name of Foundry nor the names of its
//       contributors may be used to endorse or promote products derived from this
//       software without specific prior written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
// ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
// WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
// DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDERS OR CONTRIBUTORS BE LIABLE FOR
// ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
// (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
// LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
// ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
// SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

// This work is a C++ implementation compatible with version 1.2.8 of the Python-based Nuke gizmo
// authored at Psyop by Jonah Friedman and Andy Jones (see https://github.com/Psyop/Cryptomatte).

#ifndef CRYPTOMATTE_MATTELIST_H
#define CRYPTOMATTE_MATTELIST_H

#include <string>
#include <vector>
#include <set>

#include "CryptomatteMatteExpression.h"

namespace Foundry {
  namespace Cryptomatte {

    // The section of matte names in the matte list, separated by a comma.
    typedef std::vector<std::shared_ptr<MatteExpression>> MatteSection;

    // The row of matte names that the matte list consists of.
    typedef std::vector<MatteSection> MatteRow;

    // Container of matte expressions, that apart from the items manipulation functionality
    // provides string serialization/deserialization and object IDs resolution services.
    class MatteList
    {
    public:
      // Default constructor.
      MatteList() = default;

      // Uses `strMatteList` as input string and deserializes a matte list from
      // it. When `strMatteList` contains wildcard expressions,
      // `manifestObjectNames` is used to select all matte names that match the
      // expressions.
      MatteList(const std::string& strMatteList);

      // Resolves contained expressions into the set of IDs.
      std::set<float> resolveIDs(const Manifest& manifest = {}) const;

      // Selects matte by the given ID, where a new add expression is appended to the end of list.
      // Also, existing expressions with exact match to the selected matte are removed.
      // The manifest is used for determining the object name from object ID.
      void selectByID(float objectID, const Manifest& manifest);

      // Selects mattes by the given range of object IDs ([first,last)), where new add expressions
      // are appended to the end of list. Also, existing expressions that match exactly to the selected
      // mattes are removed. The manifest is used for determining the object name from object ID.
      template <class InputIterator>
      void selectByID(InputIterator first, InputIterator last, const Manifest& manifest)
      {
        while (first != last) {
          const float objectID = *first;
          selectByID(objectID, manifest);
          ++first;
        }
      }

      // Unselects matte by the given ID, where a new remove expression is appended to the end of list.
      // Also, existing expressions with exact match to the selected matte are removed.
      // The manifest is used for determining the object name from object ID.
      void unselectByID(float objectID, const Manifest& manifest);

      // Unselects mattes by the given range of object IDs ([first,last)), where new remove expressions
      // are appended to the end of list. Also, existing expressions with exact match to the selected
      // mattes are removed. The manifest is used for determining the object name from object ID.
      template <class InputIterator>
      void unselectByID(InputIterator first, InputIterator last, const Manifest& manifest)
      {
        while (first != last) {
          const float objectID = *first;
          unselectByID(objectID, manifest);
          ++first;
        }
      }

      // Return current selection text. The original text may have had added or
      // removed lines.
      std::string str() const;

    private:
      // Removes Object Name expressions that match with the given matte name.
      // Matching expressions of other types (e.g. Wildcard or Floating-point expressions)
      // remain in the matte list.
      // Returns true, if any wildcard expression matches the given matte name;
      // otherwise false is returned.
      bool removeExactMatches(const std::string& matteName);

      // Matte expressions that define the matte list.
      std::vector<MatteRow> _matteExpressions;
    };

  }
}

#endif
