#include "CryptomattePickingManager.h"
#include "CryptomatteMatteList.h"
#include "DDImage/Iop.h"
#include <algorithm>
#include <cctype>
#include <string>

namespace 
{
  constexpr int kMantissaScaleRed = 1;
  constexpr int kMantissaScaleGreen = 16;
  constexpr int kMantissaScaleBlue = 64;

  constexpr float Unpremultiply(float coverage, float alpha) {
    return coverage / (alpha > 0.0f ? alpha : 1.0f);
  }

  // ID-Coverage pair encodes multiple ID mattes together using two channels, ID and coverage.
  // The ID channel encodes the ID of one object per pixel.
  // The Coverage channel encodes how much of the pixel's value is contributed by this object.
  struct IDCoveragePair
  {
    float id = 0.0f;
    float coverage = 0.0f;
  };
}

namespace Foundry
{
  namespace Cryptomatte
  {
    using namespace DD::Image;

    const char* PickingManager::kPreviewEnabled = "previewEnabled";  // Name from gizmo
    const char* PickingManager::kPickerAddKnob = "pickerAdd";        // Name from gizmo
    const char* PickingManager::kPickerAddText = "pickerAddText";
    const char* PickingManager::kPickerRemoveKnob = "pickerRemove";  // Name from gizmo
    const char* PickingManager::kPickerRemoveText = "pickerRemoveText";
    const char* PickingManager::kMatteListKnob = "matteList";        // Name from gizmo
    const char* PickingManager::kClearKnob = "clear";                // Name from gizmo

    PickingManager::PickingManager(DD::Image::Iop* parent)
    : _parent(parent)
    {}

    void PickingManager::setLabelNaming(const std::string& camelCaseNaming)
    {
      _label = camelCaseNaming;
      _lowerCaseLabel = camelCaseNaming;

      // Make all letters lower case (the label might include multiple words
      // so we need to do it for each letter and not just the first one).
      std::transform(_lowerCaseLabel.begin(), _lowerCaseLabel.end(), _lowerCaseLabel.begin(),
        [](unsigned char c) { return std::tolower(c); });
    }

    void PickingManager::setDefaultPreviewEnabled(bool previewEnabled)
    {
      _previewEnabled = previewEnabled;
    }

    void PickingManager::setDefaultPickerActivated(DefaultPickerActivated pickerActivated)
    {
      _pickerDefaultActivated = pickerActivated;
    }

    void PickingManager::addKnobs(Knob_Callback callback)
    {
      const std::string listLabel = _label + " List";

      const std::string previewEnabledLabel = "Preview " + _label + "s";
      Bool_knob(callback, &_previewEnabled, kPreviewEnabled, previewEnabledLabel.c_str());
      const std::string previewEnabledTooltip = "Preview available and selected " + _lowerCaseLabel + "s. Turn off for RGB pass-through.";
      Tooltip(callback, previewEnabledTooltip.c_str());

      Divider(callback);

      Eyedropper_knob(callback, _pickerAddValue, kPickerAddKnob, "Picker Add");
      const std::string pickerAddTooltip = "Sample objects in a viewer to add to the " + listLabel + ".";
      Tooltip(callback, pickerAddTooltip.c_str());
      if (_pickerDefaultActivated != DefaultPickerActivated::ePickerAdd) {
        ClearFlags(callback, Knob::CHECKED);
      }
      Named_Text_knob(callback, kPickerAddText, "", "Picker Add");

      Eyedropper_knob(callback, _pickerRemoveValue, kPickerRemoveKnob, "Picker Remove");
      if (_pickerDefaultActivated != DefaultPickerActivated::ePickerRemove) {
        ClearFlags(callback, Knob::CHECKED);
      }
      const std::string pickerRemoveTooltip = "Sample objects in a viewer to remove from the " + listLabel + ".";
      Tooltip(callback, pickerRemoveTooltip.c_str());
      Named_Text_knob(callback, kPickerRemoveText, "", "Picker Remove");

      Multiline_String_knob(callback, &_matteList, kMatteListKnob, listLabel.c_str());
      const std::string matteListTooltip = "The list of expressions for selected object names."
        "<p>Type in object names directly to add their corresponding ID " + _lowerCaseLabel + "s to the output. Use commas "
        "to create a list of object names to pick in a single line.</p>"
        "<p>Use the wilcard character '*' to select all of the objects matching a given wildcard expression "
        "throughout the sequence (e.g. flower* can be used to select objects named flower1, flower2 and flower3).</p>"
        "<p>Escape special characters with '\\\\' if they appear in an object name (e.g. flower\\\\* would allow an object "
        "named 'flower*' to be selected exactly instead of being resolved as a wildcard expression).</p>";
      Tooltip(callback, matteListTooltip.c_str());
      Newline(callback);

      Button(callback, kClearKnob, "Clear");
      const std::string clearTooltip = "Clears all the expressions in the " + listLabel + ".";
      Tooltip(callback, clearTooltip.c_str());
    }

    int PickingManager::knob_changed(Knob* k)
    {
      if (k->is(kPickerAddKnob) || k->is(kPickerRemoveKnob)) {
        // Get a bbox of the selected area
        const Box selectionArea(
          k->get_value(4),
          k->get_value(5),
          k->get_value(6),
          k->get_value(7)
        );
        // Determine whether Add or Remove
        const auto selectionMode = k->is(kPickerAddKnob) ? eSelectByTopCoverage : eUnselectByTopCoverage;
        // Select IDs by the bbox
        updateMatteListBySelectionArea(selectionArea, selectionMode);
        return 1;
      }
      if (k->is(kClearKnob)) {
        resetMatteList();
        return 1;
      }
      return 0;
    }

    void PickingManager::enableKnobs(bool enable)
    {
      _parent->knob(kPreviewEnabled)->enable(enable);
      _parent->knob(kPickerAddKnob)->enable(enable);
      _parent->knob(kPickerAddText)->enable(enable);
      _parent->knob(kPickerRemoveKnob)->enable(enable);
      _parent->knob(kPickerRemoveText)->enable(enable);
      _parent->knob(kMatteListKnob)->enable(enable);
      _parent->knob(kClearKnob)->enable(enable);
    }

    bool PickingManager::isPickingKnob(DD::Image::Knob* k) const
    {
      return (k->is(kPreviewEnabled)|| k->is(kPickerAddKnob)
        || k->is(kPickerRemoveKnob) || k->is(kMatteListKnob)
        || k->is(kClearKnob));
    }

    bool PickingManager::previewIsEnabled() const
    {
      return _previewEnabled;
    }

    std::set<float> PickingManager::getResolvedIDs(const Cryptomatte::Manifest& manifest) const
    {
      // Note, we don't use _matteList directly to avoid the issue, where
      // backslashes get removed for a baked value of Multiline_String_knob.
      const Knob* matteListKnob = _parent->knob(kMatteListKnob);
      if (matteListKnob != nullptr) {
        const std::string strMatteList(_matteList == nullptr ? "" : _matteList);
        // Resolve IDs from Matte List for further use in pixel_engine()
        const MatteList matteList(strMatteList);
        return matteList.resolveIDs(manifest);
      }
      return std::set<float>();
    }

    void PickingManager::resetMatteList()
    {
      auto matteListKnob = _parent->knob(kMatteListKnob);
      if (matteListKnob != nullptr) {
        matteListKnob->set_text("");
      }
    }

    void PickingManager::updateMatteListBySelectionArea(const Box& selectionArea, MatteSelectionMode selectionMode)
    {
      PickingInterface* pickingOp = dynamic_cast<PickingInterface*>(_parent);
      if (pickingOp == nullptr) {
        return;
      }

      auto matteListKnob = _parent->knob(kMatteListKnob);
      if (matteListKnob == nullptr) {
        return;
      }

      const auto cryptomatte = pickingOp->getSelectedCryptomatteObject();

      // Process pixels of the Cryptomatte Layer that lay inside the selection area.
      // For each pixel, read ID-Coverage pairs and find the pair with the highest coverage value.
      // The found pair would contain the ID that we are aiming to manipulate (select or unselect).
      
      // We only need to go through the area where the selection and the image bounds
      // are intersected.
      _parent->validate();
      Box bounds = selectionArea;
      bounds.intersect(_parent->info().box());

      const int areaY = bounds.y();
      const int areaT = bounds.t();
      const int areaX = bounds.x();
      const int areaR = bounds.r();

      const size_t width = static_cast<size_t>(areaR - areaX);
      const PickingInterface::PixelRowFetcher pixelRowFetcher = pickingOp->getCryptomattePixelRowFetcher();

      ChannelSet cryptomatteLayers;
      for (const auto& layer : cryptomatte.layers) {
        cryptomatteLayers += layer;
      }

      std::set<float> selectedIDs;
      for (int y = areaY; y < areaT; ++y) {
        // Get pointers for the current row
        Row row(areaX, areaR);
        pixelRowFetcher(y, areaX, areaR, cryptomatteLayers, row);
        std::array<std::vector<Row::ReadablePtr>, 4> rgbaCoverage =
          retrieveRowPointersPerChannelIndex(row, areaX, cryptomatte.layers);

        for (size_t pixel = 0; pixel < width; ++pixel) {
          IDCoveragePair topCoverage = { 0.0f, 0.0f };
          for (auto colour : { Chan_Red - 1, Chan_Blue - 1 }) {
            for (size_t i = 0; i < rgbaCoverage[colour].size(); ++i) {
              const float coverage = *(rgbaCoverage[colour + 1][i] + pixel);
              if (coverage > topCoverage.coverage) {
                topCoverage.id = *(rgbaCoverage[colour][i] + pixel);
                topCoverage.coverage = coverage;
              }
            }
          }
          selectedIDs.insert(topCoverage.id);
        }
      }

      // Skip zero value, as this doesn't represent any object ID.
      // Zero value means that the ID-Coverage pair has no associated object.
      selectedIDs.erase(0.0f);

      // Load the manifest from the selected source.
      const Manifest manifest = pickingOp->getManifestFromSelectedSource(cryptomatte);

      // Deserialize a MatteList object from a string value of the matte list knob.
      const char* matteListText = matteListKnob->get_text(&_parent->uiContext());
      const std::string strOldMatteList = matteListText == nullptr ? "" : matteListText;
      MatteList matteList(strOldMatteList);

      // According to the selection mode, selected IDs are either added to
      // or removed from the MatteList object.
      if (selectionMode == eSelectByTopCoverage) {
        matteList.selectByID(selectedIDs.begin(), selectedIDs.end(), manifest);
      }
      else {
        matteList.unselectByID(selectedIDs.begin(), selectedIDs.end(), manifest);
      }

      // Serialize modified MatteList object into a string, and
      // set the string as a new value to the matte list knob.
      const std::string strNewMatteList = matteList.str();
      matteListKnob->set_text(strNewMatteList.c_str());
    }

    std::array<std::vector<Row::ReadablePtr>, 4> PickingManager::retrieveRowPointersPerChannelIndex(const Row& row, int x, const std::vector<ChannelSet>& layers) const
    {
      std::array<std::vector<Row::ReadablePtr>, 4> rgbaCoverage;

      for (const auto& layer : layers) {
        for (const Channel& channel : layer) {
          rgbaCoverage[colourIndex(channel)].push_back(row[channel] + x);
        }
      }

      return rgbaCoverage;
    }

    bool PickingManager::renderPreview(Iop* op, const Row& in, int y, int x, int r, 
      const std::vector<ChannelSet>& cryptomatteLayers, const std::set<float>& resolvedIDs,
      Channel* matteOutChannel, bool unpremultiplyEnabled,  Row& out) const
    {
      const size_t width = static_cast<size_t>(r - x);

      // Stores input channel pointers in a convenient way for Cryptomatte
      // data processing.
      std::array<std::vector<Row::ReadablePtr>, 4> rgbaCoverage = 
        retrieveRowPointersPerChannelIndex(in, x, cryptomatteLayers);

      const float* inAlpha = in[Chan_Alpha] + x;

      float* outRed;
      float* outGreen;
      float* outBlue;
      float* outMatte;

      if (_previewEnabled) {
        outRed = out.writable(Chan_Red) + x;
        outGreen = out.writable(Chan_Green) + x;
        outBlue = out.writable(Chan_Blue) + x;
      }

      if (matteOutChannel != nullptr) {
        outMatte = out.writable(*matteOutChannel) + x;
      }

      std::vector<float> totalCoverage(width);
      for (size_t pixel = 0; pixel < width; ++pixel) {
        if (op->aborted() || op->cancelled()) {
          return false;
        }

        float totalCoverage = 0.0f;

        // Evaluate total coverage.
        for (auto colour : { Chan_Red - 1, Chan_Blue - 1 }) {
          for (size_t i = 0; i < rgbaCoverage[colour].size(); ++i) {
            const float id = *(rgbaCoverage[colour][i] + pixel);
            if (resolvedIDs.find(id) != resolvedIDs.end()) {
              totalCoverage += *(rgbaCoverage[colour + 1][i] + pixel);
            }
          }
        }

        if (unpremultiplyEnabled) {
          totalCoverage = Unpremultiply(totalCoverage, inAlpha[pixel]);
        }

        if (matteOutChannel != nullptr) {
          outMatte[pixel] = totalCoverage;
        }

        // Convert obtained pairs into their visual representation (RGB Preview), and output it to RGB channels.
        // Furthermore, by adding the total coverage to R and G channels the selected mattes get
        // a bright yellow overlay, so that the user can see what is currently selected.
        if (_previewEnabled) {
          outRed[pixel] = totalCoverage;
          outGreen[pixel] = totalCoverage;
          outBlue[pixel] = 0.0f;
          for (auto colour : { Chan_Red - 1, Chan_Blue - 1 }) {
            for (size_t i = 0; i < rgbaCoverage[colour].size(); ++i) {
              const float id = *(rgbaCoverage[colour][i] + pixel);
              const float value = *(rgbaCoverage[colour + 1][i] + pixel);
              outRed[pixel] += ChannelPreview(id, value, kMantissaScaleRed);
              outGreen[pixel] += ChannelPreview(id, value, kMantissaScaleGreen);
              outBlue[pixel] += ChannelPreview(id, value, kMantissaScaleBlue);
            }
          }
        }
      }

      return true;
    }
  }
}