#ifndef CRYPTOMATTE_PICKING_MANAGER_H
#define CRYPTOMATTE_PICKING_MANAGER_H

#include "CryptomatteManifest.h"
#include "DDImage/Knobs.h"
#include "CryptomatteUtils.h"
#include "DDImage/Row.h"
#include <array>
#include <set>
#include <vector>
#include <functional>

namespace DD
{
  namespace Image
  {
    class Iop;
    class Box;
    class Row;
  }
}

namespace Foundry
{
  namespace Cryptomatte
  {
    // An interface for all ops that want to use the picking functionality
    class PickingInterface
    {
    public:
      typedef std::function<void(int, int, int, DD::Image::ChannelSet, DD::Image::Row& row)> PixelRowFetcher;

      // Retrieves the selected cryptomatte object
      virtual CryptomatteObject getSelectedCryptomatteObject() const = 0;

      // Loads the manifest from the selected source.
      virtual Manifest getManifestFromSelectedSource(const CryptomatteObject& cryptomatte) = 0;

      // Returns a function that takes (y, x, r, ChannelSet, &Row) as input and 
      // returns the row with those parameters by reference.
      virtual PixelRowFetcher getCryptomattePixelRowFetcher() const = 0;
    };

    class PickingManager
    {
    public:

      enum DefaultPickerActivated
      {
        ePickerAdd = 0,
        ePickerRemove,
        eNone
      };

      PickingManager(DD::Image::Iop* parent);

      void setLabelNaming(const std::string& camelCaseNaming);

      void setDefaultPreviewEnabled(bool previewEnabled);
      void setDefaultPickerActivated(DefaultPickerActivated pickerActivated);

      void addKnobs(DD::Image::Knob_Callback callback);
      int knob_changed(DD::Image::Knob* k);
      void enableKnobs(bool enable);
      bool isPickingKnob(DD::Image::Knob* k) const;

      bool previewIsEnabled() const;

      std::set<float> getResolvedIDs(const Manifest& manifest) const;

      void resetMatteList();

      // Renders the cryptomatte preview into the outRow's rgb if preview has been enabled
      // and the matte into the matteOutChannel of outRow if it has been specified.
      bool renderPreview(DD::Image::Iop* op, const DD::Image::Row& in, int y, int x, int r,
        const std::vector<DD::Image::ChannelSet>& cryptomatteLayers,
        const std::set<float>& resolvedIDs,
        DD::Image::Channel* matteOutChannel, bool unpremultiplyEnabled,
        DD::Image::Row& out) const;

      // Knob names
      static const char* kPreviewEnabled;
      static const char* kPickerAddKnob;
      static const char* kPickerAddText;
      static const char* kPickerRemoveKnob;
      static const char* kPickerRemoveText;
      static const char* kMatteListKnob;
      static const char* kClearKnob;

    private:
      // Enumerates possible modes of Matte List selection by the user.
      enum MatteSelectionMode
      {
        eSelectByTopCoverage = 0,
        eUnselectByTopCoverage
      };

      // Updates the Matte List with mattes, picked by associated IDs in the area of interest (user-selected bbox).
      void updateMatteListBySelectionArea(const DD::Image::Box& selectionArea, MatteSelectionMode selectionMode);

      // Stores pointers to the underlying channel data from the input Row at position
      // x for each channel of the cryptomatte layers. The pointers are organized such 
      // that each channel is added to their respective channel index in the array,
      // eg Layer.red = 0, Layer.green = 1, Layer.blue = 2, Layer.alpha = 3. This 
      // makes it easier and faster to fetch all the ids (red and blue channels) 
      // and the coverage values (green and alpha channels).
      std::array<std::vector<DD::Image::Row::ReadablePtr>, 4> retrieveRowPointersPerChannelIndex(
        const DD::Image::Row& row, int x,
        const std::vector<DD::Image::ChannelSet>& layers) const;

      DD::Image::Iop* _parent;

      // Knob values
      bool _previewEnabled = true;
      float _pickerAddValue[8]{};
      float _pickerRemoveValue[8]{};

      // User-selected Matte List string
      const char* _matteList = nullptr;

      // If the picking should be available from start
      // or needs to be enabled by the user
      DefaultPickerActivated _pickerDefaultActivated = DefaultPickerActivated::ePickerAdd;

      // What the naming of the labels should be.
      std::string _label = "Matte";
      std::string _lowerCaseLabel = "matte";
    };
  }
}

#endif