// Copyright (c) 2021 The Foundry Visionmongers Ltd. All rights reserved.

// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are met:
//
//     * Redistributions of source code must retain the above copyright notice,
//       this list of conditions and the following disclaimer.
//     * Redistributions in binary form must reproduce the above copyright notice,
//       this list of conditions and the following disclaimer in the documentation
//       and/or other materials provided with the distribution.
//     * Neither the name of Foundry nor the names of its
//       contributors may be used to endorse or promote products derived from this
//       software without specific prior written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
// ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
// WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
// DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDERS OR CONTRIBUTORS BE LIABLE FOR
// ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
// (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
// LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
// ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
// SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

// This work is a C++ implementation compatible with version 1.2.8 of the Python-based Nuke gizmo
// authored at Psyop by Jonah Friedman and Andy Jones (see https://github.com/Psyop/Cryptomatte).

#include "CryptomatteUtils.h"
#include "CryptomatteManifest.h"
#include "DDImage/Iop.h"
#include "DDImage/Row.h"
#include "DDImage/MetaData.h"
#include "murmur3.h"
#include <QDir>
#include <QFileInfo>
#include <QString>
#include <cstdlib>
#include <cmath>
#include <fstream>
#include <functional>
#include <iomanip>
#include <regex>

namespace {

  using namespace DD::Image;
  using namespace Foundry;
  using namespace Cryptomatte;

  // Defines the signature of a function used to check for a metadata key.
  typedef std::function<bool(const std::string&)> MetaDataKeyChecker;

  // Returns a value of metadata property for the given key.
  // If the metadata property isn't found, an empty string is returned.
  std::string FetchMetaDataByKey(Iop* input, const MetaDataKeyChecker& checker);

  // Reads a sidecar file that contains the manifest, defined in the text JSON format. If the file
  // is found and successfully read, the function returns a string that contains the content of file;
  // otherwise, an empty string is returned.
  std::string ReadSidecarManifestFile(const std::string& sidecarFileName);

  constexpr char kInputFilenameKey[] = "input/filename";
  constexpr char kMetadataCryptomatte[] = "cryptomatte";
  constexpr char kMetadataConversion[] = "conversion";
  constexpr char kMetadataHash[] = "hash";
  constexpr char kMetadataName[] = "name";
  constexpr char kMetadataManifestFile[] = "manif_file";
  constexpr char kMetadataManifest[] = "manifest";
  constexpr char kMetadataDelimiter = '/';

  const size_t kMetadataCryptomatteLength = std::strlen(kMetadataCryptomatte);
}

namespace Foundry {
  namespace Cryptomatte {

    using namespace DD::Image;

    void ReplaceSpecialCharacters(std::string& name)
    {
      std::replace_if(name.begin(), name.end(), [](char c){ return c == ' ' || c == '.'; }, '_');
    }

    std::vector<CryptomatteObject> GetAvailableCryptomatteObjects(const Op* op)
    {
      const auto& outputContext = op->uiContext();
      auto input = op_cast<Iop*>(op->node_input(0, Op::EXECUTABLE_INPUT, &outputContext));
      if (input == nullptr) {
        return std::vector<CryptomatteObject>();
      }

      // Forces the input to populate channels
      input->validate();

      return GetAvailableCryptomatteObjects(input->fetchMetaData(nullptr), input->channels());
    }

    std::vector<CryptomatteObject> GetAvailableCryptomatteObjects(
      const MetaData::Bundle& metadata, const ChannelSet& channels)
    {
      std::vector<CryptomatteObject> cryptomatteObjects;

      // Iterate through all the metadata properties of the given input
      // and populate a vector of cryptomatte objects, created from
      // metadata properties whose key matches the certain pattern,
      // that complies with the Cryptomatte Specification (e.g. "cryptomatte/adb0a84/name").
      std::map<std::string, size_t> nameToObjectIndexMap;
      std::map<std::string, CryptomatteObject> keyToObjectMap;
      for (const auto& metadataItem : metadata) {
        const std::string& metadataKey = metadataItem.first;
        size_t cryptomattePosition = metadataKey.find(kMetadataCryptomatte);
        if (cryptomattePosition == std::string::npos) {
          continue;
        }
        const size_t keyPosition = cryptomattePosition + kMetadataCryptomatteLength + 1;
        const size_t metadataPosition = metadataKey.find(kMetadataDelimiter, keyPosition) + 1;
        const std::string cryptomatteMetadataKey = metadataKey.substr(keyPosition, metadataPosition - keyPosition - 1);
        if (!keyToObjectMap.count(cryptomatteMetadataKey)) {
          CryptomatteObject cryptomatteObject;
          cryptomatteObject.metadataKey = cryptomatteMetadataKey;
          keyToObjectMap.insert({cryptomatteMetadataKey, cryptomatteObject});
        }
        CryptomatteObject& cryptomatteObject = keyToObjectMap[cryptomatteMetadataKey];
        const std::string cryptomatteMetadata = metadataKey.substr(metadataPosition);
        const std::string cryptomatteMetadataValue = MetaData::getPropertyString(metadataItem.second, 0);
        if (cryptomatteMetadata == kMetadataHash) {
          cryptomatteObject.hash = cryptomatteMetadataValue;
        }
        else if (cryptomatteMetadata == kMetadataConversion) {
          cryptomatteObject.conversion = cryptomatteMetadataValue;
        }
        else if (cryptomatteMetadata == kMetadataName) {
          cryptomatteObject.name = cryptomatteMetadataValue;
          // For populating the GUI layer selection knob, replace the special
          // characters with underscores.
          ReplaceSpecialCharacters(cryptomatteObject.name);
        }
        else if (cryptomatteMetadata == kMetadataManifestFile) {
          cryptomatteObject.manifestFile = cryptomatteMetadataValue;
        }
      }
      for (const auto& keyToObject : keyToObjectMap) {
        CryptomatteObject cryptomatteObject = keyToObject.second;
        // Populate new cryptomatte objects
        cryptomatteObjects.push_back(cryptomatteObject);
        nameToObjectIndexMap.insert({cryptomatteObject.name, cryptomatteObjects.size() - 1});
      }

      // Iterate through each channel of the given input (except for
      // the beauty channels) and select those, whose name is associated
      // with a cryptomatte name from the metadata. Once the channel is
      // selected, populate corresponding cryptomatte object with
      // the layer that the channel belongs to.
      ChannelSet inputChannels = channels - Mask_RGBA;
      while (!inputChannels.empty()) {
        // Get a name of the channel
        const Channel channel = inputChannels.first();
        // Get a layer that the current channel belongs to
        ChannelSet layer = Mask_None;
        layer.addBrothers(channel, 4);
        // Reduce the set of channels to process by the current layer
        inputChannels -= layer;
        // Skip layers which name doesn't comply with the Cryptomatte Specification,
        // where the value should have trailing digits (e.g. 00, 01, 02, ...)
        const std::string layerName(getLayerName(channel));
        static const std::regex regexLayerNameWithTrailingDigits("^(.*[^0-9])\\d+$");
        std::smatch matchLayerNameWithTrailingDigits;
        if (!std::regex_match(layerName, matchLayerNameWithTrailingDigits, regexLayerNameWithTrailingDigits)) {
          continue;
        }
        // Skip the layer, if its name doesn't mach any cryptomatte object's name
        const std::string& cryptomatteName = matchLayerNameWithTrailingDigits[1];
        auto itrNameAndObjectIndex = nameToObjectIndexMap.find(cryptomatteName);
        if (itrNameAndObjectIndex == nameToObjectIndexMap.end()) {
          continue;
        }
        // Skip the layer, if it doesn't consist of 4 channels. In regards to the
        // Cryptomatte Specification, each cryptomatte layer should consists of
        // two pairs of ID-Coverage values (4 channels).
        if (layer.size() != 4) {
          continue;
        }
        // Add the layer to corresponding cryptomatte object
        cryptomatteObjects[itrNameAndObjectIndex->second].layers.push_back(layer);
      }
      // Remove cryptomatte objects that have zero layers
      cryptomatteObjects.erase(std::remove_if(cryptomatteObjects.begin(), cryptomatteObjects.end(),
        [](const CryptomatteObject& cryptomatteObject) {
          return cryptomatteObject.layers.empty();
        }), cryptomatteObjects.end());
      return cryptomatteObjects;
    }

    std::pair<std::string, std::string> GetLayerHashAndAttribute(const std::string& metadataKey)
    {
      // Get the position in the key where the cryptomatte word occurs.
      const size_t cryptomattePosition = metadataKey.find(kMetadataCryptomatte);

      // Return a pair of empty strings if it is not a cryptomatte entry.
      if (cryptomattePosition == std::string::npos) {
        return std::pair<std::string, std::string>();
      }

      // Get the position in the key where the layer hash starts.
      const size_t layerHashPosition = cryptomattePosition + kMetadataCryptomatteLength + 1;

      // Get the position in the key where the cryptomatte attribute starts.
      // This is the last part of the key after the last delimiter.
      const size_t attributePosition = metadataKey.find(kMetadataDelimiter, layerHashPosition) + 1;

      // Get the actual layer hash string.
      std::string layerHash = metadataKey.substr(layerHashPosition, attributePosition - layerHashPosition - 1);

      // Get the actual attribute string.
      std::string attribute = metadataKey.substr(attributePosition);

      return std::make_pair(layerHash, attribute);
    }

    std::vector<std::string> FindLayerHashes(const MetaData::Bundle& metaDataBundle, const std::string& layerName)
    {
      std::vector<std::string> nameLayerHashes;

      // Loop through all the key/value pairs of the metadata
      for (const auto& metaDataBundleItem : metaDataBundle) {
        std::string layerHash;
        std::string attribute;
        std::tie(layerHash, attribute) = GetLayerHashAndAttribute(metaDataBundleItem.first);

        // Get all the layer hashes that belong to a particular layer name.
        // Check if the current key is for the layer name. If so, add it to the
        // list, otherwise get the next key-value pair from the metadata.
        if (!layerHash.empty() && attribute == kMetadataName) {
          std::string layerPropertyString = MetaData::getPropertyString(metaDataBundleItem.second, 0);
          // Since the metadata contains the layer names with special
          // characters, we need to replace those with underscores to match the
          // selected layer name on the GUI layer selection knob.
          ReplaceSpecialCharacters(layerPropertyString);
          if (layerPropertyString == layerName) {
            nameLayerHashes.push_back(layerHash);
          }
        }
      }

      return nameLayerHashes;
    }

    Manifest FindManifest(const MetaData::Bundle& metaDataBundle, const std::vector<std::string>& nameLayerHashes)
    {
      Manifest manifest;

      // Loop through all the key/value pairs of the metadata
      for (const auto& metaDataBundleItem : metaDataBundle) {
        std::string layerHash;
        std::string attribute;
        std::tie(layerHash, attribute) = GetLayerHashAndAttribute(metaDataBundleItem.first);

        // The layer hashes for a particular layer name are collected at this
        // point. If the current hash read belongs to the layer that we are
        // interested in, then get the manifest data. Otherwise, get the next
        // key-value pair from them metadata.
        if (!layerHash.empty() && attribute == kMetadataManifest && std::find(nameLayerHashes.begin(), nameLayerHashes.end(), layerHash) != nameLayerHashes.end()) {
          manifest.insert(MetaData::getPropertyString(metaDataBundleItem.second, 0));
        }
      }

      return manifest;
    }

    Manifest GetManifestEmbedded(Iop* input, const CryptomatteObject& object)
    {
      Manifest manifest;
      if (input == nullptr) {
        return manifest;
      }

      const auto& metaDataBundle = input->fetchMetaData(nullptr);

      return GetManifestFromMetadata(metaDataBundle, object);
    }

    Manifest GetManifestFromMetadata(
      const DD::Image::MetaData::Bundle& metaDataBundle,
      const CryptomatteObject& object)
    {
      // Find the hashes in the metadata for the given layer name.
      // Unfortunately, different software programs generate different hashes
      // interpreting the cryptomatte specification differently. In order to
      // work with their Cryptomatte images, we need to be flexible to
      // recognise the manifest for them. This means we cannot rely on the
      // layer hash, but the layer name.
      std::vector<std::string> nameLayerHashes = FindLayerHashes(metaDataBundle, object.name);

      // Go through the metadata again to collect the manifest for each entry
      // that has the correct layer name associated with it. Unfortunately, we
      // need to iterate twice because usually, the manifest comes in the
      // metadata earlier than the name.
      return FindManifest(metaDataBundle, nameLayerHashes);
    }

    Manifest GetManifestSidecar(const std::string& fileName)
    {
      // Read the sidecar file and parse the manifest from its content.
      const std::string sidecarManifestJson = ReadSidecarManifestFile(fileName);
      return Manifest(sidecarManifestJson);
    }

    uint32_t GetMurmurHashValueUInt(const std::string& objectName)
    {
      uint32_t hash;
      MurmurHash3_x86_32(objectName.c_str(), static_cast<int>(objectName.size()), 0, &hash);
      // According to Cryptomatte specification, in order to avoid NaN floats
      // we should clamp the exponent to a valid value
      return hash;
    }

    float GetMurmurHashValueFloat(const std::string& objectName)
    {
      auto hash = GetMurmurHashValueUInt(objectName);
      const uint32_t exponent = hash >> 23 & 255;
      if (exponent == 0 || exponent == 255) {
        hash ^= 1 << 23;
      }
      float f;
      memcpy(&f, &hash, 4);
      return f;
    }

    std::string GetMurmurHashValueFloatToHex(float value)
    {
      uint32_t floatBits;
      memcpy(&floatBits, &value, 4);
      std::stringstream hexStream;
      hexStream << std::hex << std::setw(8) << std::setfill('0') << floatBits;
      return hexStream.str();
    }

    std::string GetMurmurHashValueHex(const std::string& name)
    {
      return GetMurmurHashValueFloatToHex(GetMurmurHashValueFloat(name));
    }

    std::string GetAbsoluteFilePathToSidecarManifest(Iop* input, const std::string& sidecarFileName)
    {
      // Try to open a sidecar file by the path, created from the directory of Cryptomatte's EXR file,
      // combined with the relative path to sidecar file name.
      const std::string inputFileName = FetchMetaDataByKey(input, [](const std::string& metaDataKey) {
        return kInputFilenameKey == metaDataKey;
      });
      const QFileInfo inputFileInfo(QString::fromStdString(inputFileName));
      const QString absolutePathQString = QDir::cleanPath(inputFileInfo.absolutePath()
        + QDir::separator() + QString::fromStdString(sidecarFileName));
      return absolutePathQString.toStdString();
    }

    // Escape cryptomatte especial characters.
    std::string Escape(const std::string& matteName)
    {
      if (matteName.empty()) {
        return std::string();
      }
      std::string escaped(matteName);
      size_t pos = 0;
      static const std::string kSpecialChars{kAsterisk, kMinusSign};
      while ((pos = escaped.find_first_of(kSpecialChars, pos)) != std::string::npos) {
        escaped.insert(pos, 1, kEscape);
        pos += 2;
      }
      return escaped;
    }

    // Remove wildcard escape sequence '\\'.
    std::string Unescape(const std::string& matteExpressionStr)
    {
      std::string matteName(matteExpressionStr);
      size_t pos = 0;
      while ((pos = matteName.find(kEscape, pos))!= std::string::npos)
      {
        matteName.erase(pos, 1);
      }
      return matteName;
    }

    std::string CreateRegex(const std::string& expression)
    {
      const std::string regexSpecialChars(kRegexCharacters);
      std::string pattern;
      for (size_t i = 0; i < expression.size(); ++i) {
        switch (expression[i]) {
        case kAsterisk:
          pattern += ".";
          break;
        case kEscape: {
          static const std::string kCharsToEscape{kAsterisk, kDoubleQuote, kMinusSign};
          char escapedChar = EscapedChar(expression, i);
          if (kCharsToEscape.find(escapedChar) != std::string::npos) {
            pattern += kEscape;
            ++i;
          }
        }
        default:
          if (regexSpecialChars.find(expression[i]) != std::string::npos) {
            pattern += kEscape;
          }
          break;
        }
        pattern += expression[i];
      }
      return pattern;
    }

    float ChannelPreview(float id, float coverage, int mantissaScale)
    {
      int exponent = 0;
      // Note, frexp function that is used below returns a mantissa value of
      // passed matte ID. 'exponent' is unused.
      return std::fmod(std::frexp(std::abs(id), &exponent) * static_cast<float>(mantissaScale), 0.25f) * coverage;
    }

    void HandleCryptomatteObjectErrors(const CryptomatteObject& cryptomatteObject, Iop* iop)
    {
      if (iop == nullptr) {
        // Iop has to be valid to be able to display errors
        return;
      }

      std::string errorString = cryptomatteObject.hasSupportedHash();
      if (!errorString.empty()) {
        iop->error(errorString.c_str());
      }
      errorString = cryptomatteObject.hasSupportedConversion();
      if (!errorString.empty()) {
        iop->error(errorString.c_str());
      }
    }

  }
}

namespace {

  std::string FetchMetaDataByKey(Iop* input, const MetaDataKeyChecker& checker)
  {
    // Op::fetchMetaData() doesn't support fetching data by key.
    // Therefore we do brute-force keyname comparison.
    const auto& metadata = input->fetchMetaData(nullptr);
    for (const auto& metadataItem : metadata) {
      // Check if current metadata item is the input file name (Cryptomatte's EXR file).
      if (checker(metadataItem.first)) {
        return MetaData::getPropertyString(metadataItem.second, 0);
      }
    }

    return std::string();
  }

  std::string ReadSidecarManifestFile(const std::string& sidecarFileName)
  {
    const std::ifstream sidecarFile = std::ifstream(sidecarFileName);
    // Check if the path to file is valid and file can be read.
    if (sidecarFile.good()) {
      std::stringstream buffer;
      buffer << sidecarFile.rdbuf();
      return buffer.str();
    }
    return std::string();
  }
}
