// Copyright (c) 2021 The Foundry Visionmongers Ltd. All rights reserved.

// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are met:
//
//     * Redistributions of source code must retain the above copyright notice,
//       this list of conditions and the following disclaimer.
//     * Redistributions in binary form must reproduce the above copyright notice,
//       this list of conditions and the following disclaimer in the documentation
//       and/or other materials provided with the distribution.
//     * Neither the name of Foundry nor the names of its
//       contributors may be used to endorse or promote products derived from this
//       software without specific prior written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
// ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
// WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
// DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDERS OR CONTRIBUTORS BE LIABLE FOR
// ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
// (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
// LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
// ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
// SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

// This work is a C++ implementation compatible with version 1.2.8 of the Python-based Nuke gizmo
// authored at Psyop by Jonah Friedman and Andy Jones (see https://github.com/Psyop/Cryptomatte).

#ifndef CRYPTOMATTE_UTILS_H
#define CRYPTOMATTE_UTILS_H

#include <vector>
#include <string>

namespace DD {
  namespace Image {
    class ChannelSet;
    class Iop;
    class Op;
    class Row;
    namespace MetaData {
      class Bundle;
    }
  }
}

namespace Foundry {
  namespace Cryptomatte {

    class Manifest;

    constexpr char kEscape = '\\';
    constexpr char kSpace = ' ';
    constexpr char kTab = '\t';
    constexpr char kComma = ',';
    constexpr char kLineBreak = '\n';
    constexpr char kDoubleQuote = '"';
    constexpr char kAsterisk = '*';
    constexpr char kMinusSign = '-';
    constexpr char kClosingAngleBracket = '>';
    constexpr char kOpenSquareBracket = '[';
    constexpr char kCloseSquareBracket = ']';

    // List of special characters for wildcard expressions.
    constexpr char kRegexCharacters[] = R"(.+?^$()[]{}|\)";

    static const std::string kSupportedMetadataHash = "MurmurHash3_32";
    static const std::string kSupportedMetadataConversion = "uint32_to_float32";

    // Represents an invalid object ID value, which in regards to the Cryptomatte Specification
    // means no object is selected. This value cannot appear in any manifest.
    constexpr float kFillingObjectId = 0.0f;

    constexpr float kCoverageEmpty = 0.0f;
    constexpr float kCoverageFull = 1.0f;

    // Represents a cryptomatte object, created from metadata items
    // in correspondence to availability of associated EXR's AOV layers.
    //
    // Example of metadata properties and corresponding layer names that define
    // a Cryptomatte object.
    // Metadata properties:
    //    /exr/cryptomatte/adb0a84/name       : uCryptoAsset
    //    /exr/cryptomatte/adb0a84/hash       : MurmurHash3_32
    //    /exr/cryptomatte/adb0a84/conversion : uint32_to_float32
    //    /exr/cryptomatte/adb0a84/manif_file : bunny_CryptoObject_manifest.json
    //  Corresponding layer names:
    //    uCryptoAsset00, uCryptoAsset01, uCryptoAsset02, uCryptoAsset03
    //  Where the struct would represent it as follows:
    //    {
    //      name          : uCryptoAsset
    //      metadataKey   : adb0a84
    //      layers        : { uCryptoAsset00, uCryptoAsset01, uCryptoAsset02, uCryptoAsset03 }
    //      manifestFile  : bunny_CryptoObject_manifest.json
    //    }
    struct CryptomatteObject
    {
      std::string name;
      std::string hash;
      std::string conversion;
      std::string metadataKey;
      std::string manifestFile;
      std::vector<DD::Image::ChannelSet> layers;

      // If the hash is unsupported, it returns an error string, otherwise an
      // empty string.
      std::string hasSupportedHash() const
      {
        if (hash != kSupportedMetadataHash) {
          return R"(Invalid hash value: ")" + hash
                 + R"(", for key: ")" + metadataKey + R"(")";
        }
        return std::string();
      }

      // If the conversion is unsupported, it returns an error string,
      // otherwise an empty string.
      std::string hasSupportedConversion() const
      {
        if (conversion != kSupportedMetadataConversion) {
          return R"(Invalid conversion value: ")" + conversion
                 + R"(", for key: ")" + metadataKey + R"(")";
        }
        return std::string();
      }
    };

    // Checks that the properties of the Cryptomatte object are supported by
    // the current implementation. Displays an error otherwise.
    void HandleCryptomatteObjectErrors(const CryptomatteObject& cryptomatteObject, DD::Image::Iop* iop);

    // Replaces special characters (e.g. spaces or dots) in strings, like
    // imported cryptomatte layer names, with underscores. E.g. what is
    // potentially coming from Blender, as these are not supported characters
    // in standard Nuke layer names.
    void ReplaceSpecialCharacters(std::string& name);

    // Returns a vector of cryptomatte objects, available on the input 0 of the given op.
    // A wrapper function around the GetAvailableCryptomatteObjects(metadata, channels)
    // below. Grabs all data from the input of the given op.
    std::vector<CryptomatteObject> GetAvailableCryptomatteObjects(const DD::Image::Op* op);

    // Returns a vector of cryptomatte objects, available in the given
    // metadata and channels.
    // To determine available cryptomatte objects the function fetches
    // cryptomatte information from the metadata. Then it looks into
    // available layers and finds matches between cryptomatte name
    // from metadata and a layer name. Finally, it creates and returns
    // cryptomatte objects, based on found matches.
    std::vector<CryptomatteObject> GetAvailableCryptomatteObjects(
      const DD::Image::MetaData::Bundle& metadata, 
      const DD::Image::ChannelSet& channels);

    // Gets the layer hash and attribute sections from the key of the
    // cryptomatte metadata entry.
    std::pair<std::string, std::string> GetLayerHashAndAttribute(const std::string& metadataKey);

    // Finds the layer hashes for a particular layer name
    std::vector<std::string> FindLayerHashes(const DD::Image::MetaData::Bundle& metaDataBundle, const std::string& layerName);

    // Finds the manifest for particular layer hashes
    Manifest FindManifest(const DD::Image::MetaData::Bundle& metaDataBundle, const std::vector<std::string>& nameLayerHashes);

    // Gets a manifest for given cryptomatte object from metadata of the input.
    Manifest GetManifestEmbedded(DD::Image::Iop* input, const CryptomatteObject& object);

    // Gets a manifest for given cryptomatte object from the given metadata.
    Manifest GetManifestFromMetadata(const DD::Image::MetaData::Bundle& metaDataBundle, const CryptomatteObject& object);

    // Gets a manifest, loaded from the file by given path.
    // The `input` parameter is used to determine the absolute path to `fileName`.
    Manifest GetManifestSidecar(const std::string& fileName);

    // Convert ID-Coverage pairs to intermediate either Red, Green, or Blue value
    // using the original algorithm from gizmo.
    float ChannelPreview(const float id, const float coverage, int mantissaScale);

    // Returns the unsigned integer ID value for a given objectName using MurmurHash.
    uint32_t GetMurmurHashValueUInt(const std::string& objectName);

    // Returns the floating-point ID value for a given objectName using MurmurHash.
    float GetMurmurHashValueFloat(const std::string& objectName);

    // Returns the floating-point id value in hex format from the floating-point ID value
    std::string GetMurmurHashValueFloatToHex(float value);

    // Returns the floating-point ID value in hex format for a given string using MurmurHash
    std::string GetMurmurHashValueHex(const std::string& name);

    // Returns the absolute path to sidecar manifest file, if given cryptomatte object contains any;
    // otherwise returns an empty string.
    // The `input` parameter is used to determine the absolute path to `fileName`.
    // `fileName` is relative to the folder where the input file (EXR) is located.
    std::string GetAbsoluteFilePathToSidecarManifest(DD::Image::Iop* input, const std::string& sidecarFileName);

    // If a escape sequence starts at 'pos' in the 'source' string, returns the
    // escaped character, otherwise returns 0 (zero).
    inline char EscapedChar(const std::string& source, const size_t pos)
    {
      // Check if inside range.
      if (pos < 0 || (pos + 1) >= source.size()) {
        return 0;
      }
      if (source[pos] != kEscape)
        return 0;
      return source[pos + 1];
    }

    // Escape cryptomatte especial characters.
    std::string Escape(const std::string& matteName);

    // Remove wildcard escape sequence '\\'.
    std::string Unescape(const std::string& matteExpressionStr);

    std::string CreateRegex(const std::string& expression);
  }
}

#endif
