// Copyright (c) 2021 The Foundry Visionmongers Ltd. All rights reserved.

// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are met:
//
//     * Redistributions of source code must retain the above copyright notice,
//       this list of conditions and the following disclaimer.
//     * Redistributions in binary form must reproduce the above copyright notice,
//       this list of conditions and the following disclaimer in the documentation
//       and/or other materials provided with the distribution.
//     * Neither the name of Foundry nor the names of its
//       contributors may be used to endorse or promote products derived from this
//       software without specific prior written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
// ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
// WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
// DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDERS OR CONTRIBUTORS BE LIABLE FOR
// ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
// (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
// LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
// ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
// SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

// This work is a C++ implementation compatible with version 1.2.8 of the Python-based Nuke gizmo
// authored at Psyop by Jonah Friedman and Andy Jones (see https://github.com/Psyop/Cryptomatte).

#include "EncryptomatteUtils.h"

#include "DDImage/LayerI.h"

#include <algorithm>
#include <cctype>

namespace Foundry {
  namespace Encryptomatte {

    using namespace DD::Image;
    using namespace Cryptomatte;

    CoverageIdsMap InitialisePixel(float matteId, float matteAlpha, size_t rowIndex)
    {
      // New Layer or a refurbished new layer. When we remove mattes from a layer, it seems to become blank. If it could remain 0.0/1.0f, we would not need to redo this part.
      CoverageIdsMap coverageIdPairs;
      float matteAlphaInvert = 1 - matteAlpha;
      for (size_t i = 0; i < 4; ++i) {
        coverageIdPairs[kFillingObjectId].push_front(kCoverageEmpty);
      };
      coverageIdPairs[matteAlphaInvert].push_front(kFillingObjectId);
      coverageIdPairs[matteAlpha].push_front(matteId);
      return coverageIdPairs;
    }

    CoverageIdsMap CopyPixelOutsideMatte(const std::vector<const float*>& inputRowData, size_t rowIndex, float matteAlpha)
    {
      CoverageIdsMap coverageIdPairs;
      // If no matte alpha for coverage, copy the input verbatim into the
      // output and then go to the next pixel. Maybe, experiment with
      // out.copy(in, channels, x, r); if it could be more performant
      if (matteAlpha <= 0.0f) {
        for (int channelIndex = inputRowData.size() - 1; channelIndex > 0;) {
          const float coverage = inputRowData[channelIndex--][rowIndex];
          const float id = inputRowData[channelIndex--][rowIndex];
          coverageIdPairs[coverage].push_front(id);
        }
      }
      return coverageIdPairs;
    }

    CoverageIdsMap CreateCoverageIdPairs(const std::vector<const float*>& inputRowData, size_t rowIndex, float matteId, float matteAlpha, bool isOver)
    {
      CoverageIdsMap coverageIdPairs;
      float totalCoverage = 0.0f;
      const float matteAlphaInvert = 1 - matteAlpha;
      for (size_t inChannelIndex = 0; inChannelIndex < inputRowData.size();) {
        const float id = inputRowData[inChannelIndex++][rowIndex];
        const float coverage = inputRowData[inChannelIndex++][rowIndex];
        if (isOver) {
          coverageIdPairs[coverage * matteAlphaInvert].push_front(id);
        }
        else {
          if (id == kFillingObjectId) {
            coverageIdPairs[coverage * matteAlphaInvert].push_front(id);
            continue;
          }
          totalCoverage += coverage;
          coverageIdPairs[coverage].push_front(id);
        }
      }
      // Insert matte ID
      coverageIdPairs[matteAlpha * (1 - totalCoverage)].push_front(matteId);
      return coverageIdPairs;
    }

    void WriteOutputRowData(const CoverageIdsMap& coverageIdPairs, const std::vector<float*>& outputRowData, size_t rowIndex)
    {
      size_t numberOfChannels = outputRowData.size();
      size_t outputIndex = 0;
      for (auto it = coverageIdPairs.rbegin(); it != coverageIdPairs.rend() && outputIndex < numberOfChannels; ++it) {
        auto ids = it->second;
        for (int idIndex = ids.size() - 1; idIndex >= 0 && outputIndex < numberOfChannels; --idIndex) {
          outputRowData[outputIndex++][rowIndex] = ids[idIndex];
          outputRowData[outputIndex++][rowIndex] = it->first;
        }
      }
    }

    std::string ValidateLayerName(const std::string& layerName)
    {
      // Quote the cryptomatte layer name so that the error message becomes
      // easier to read due to the delimiters, especially when e.g. the input
      // cryptomatte layer name is an empty string (the user has not entered
      // anything), etc.
      const std::string quotedLayerName = R"(")" + layerName + R"(")";

      // Do not allow empty cryptomatte layer names.
      if (layerName.empty()) {
        return quotedLayerName + " is not allowed as the name of a layer, it must start with a letter or underscore";
      }

      // The first character of a standard Nuke, and therefore cryptomatte,
      // layer name must be either a letter or an underscore.
      // 0x80 is inherited from the ChannelKnob creation rules. 
      char const& layerNameFirstCharacter = layerName.front();
      if (!isalpha(layerNameFirstCharacter) && layerNameFirstCharacter != '_'
             && !(layerNameFirstCharacter & 0x80)) {
        return quotedLayerName + " is not allowed as the name of a layer, it must start with a letter or underscore";
      }

      // While the specification does not explicitly mention what "typename"
      // is, therefore up to the implementation what valid characters are, do
      // not allow trailing digits in the cryptomatte layer name for simplicity.
      if (std::isdigit(layerName.back())) {
        return quotedLayerName + " cannot be used as the name of a layer, it has trailing digit";
      }

      // Do not allow dots in the layername as this would conflict with the
      // layerName.channelName schema.
      if (layerName.find('.') != std::string::npos) {
        return quotedLayerName + " cannot be used as the name of a layer, it contains periods";
      }

      // Reserved word for selecting none of the cryptomatte layers.
      if (layerName == "none") {
        return quotedLayerName + " cannot be used as the name of a layer";
      }

      // From the second character of a standard Nuke, and therefore
      // cryptomatte, layer, the character must be a letter, digit or
      // underscore.
      // 0x80 is inherited from the ChannelKnob creation rules. 
      const bool validLayerName = std::all_of(layerName.cbegin() + 1, layerName.cend(), [](const char c) {
        return (isalnum(c) || c == '_' || (c & 0x80));
      });

      if (!validLayerName) {
        return quotedLayerName + " is not allowed as the name of a layer, it must contain only letters, digits, or underscore";
      }

      const LayerI* layer = GetLayer(layerName);
      if (layer && layer->builtIn()) {
        return "Can't change built-in layers";
      }

      return std::string();
    }
  }
}
