// Copyright (c) 2021 The Foundry Visionmongers Ltd. All rights reserved.

// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are met:
//
//     * Redistributions of source code must retain the above copyright notice,
//       this list of conditions and the following disclaimer.
//     * Redistributions in binary form must reproduce the above copyright notice,
//       this list of conditions and the following disclaimer in the documentation
//       and/or other materials provided with the distribution.
//     * Neither the name of Foundry nor the names of its
//       contributors may be used to endorse or promote products derived from this
//       software without specific prior written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
// ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
// WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
// DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDERS OR CONTRIBUTORS BE LIABLE FOR
// ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
// (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
// LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
// ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
// SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

// This work is a C++ implementation compatible with version 1.2.8 of the Python-based Nuke gizmo
// authored at Psyop by Jonah Friedman and Andy Jones (see https://github.com/Psyop/Cryptomatte).

#ifndef ENCRYPTOMATTE_UTILS_H
#define ENCRYPTOMATTE_UTILS_H

#include "CryptomatteUtils.h"

#include <deque>
#include <map>

namespace Foundry {
  namespace Encryptomatte {

    // Note: in general, it is not recommended to have a float key for a map as
    // it may cause equality issues. But it is acceptable here because it is
    // unlikely that both the coverage values would be very precise float
    // numbers and there would be two object ids with their coverage so close
    // to each other.
    typedef std::map<float, std::deque<float>> CoverageIdsMap;

    constexpr size_t kDefaultCryptoLayerDepth = 3;
    const std::vector<std::string> kChannelNames = { "red", "green", "blue", "alpha" };

    // Initialises the pixels to explicit filling for now after layer creation
    // or if the last matte is removed from the layer and then a matte is
    // placed on it again.
    CoverageIdsMap InitialisePixel(float matteId, float matteAlpha, size_t rowIndex);

    // Takes the pixel values from the main input when the alpha is 0
    CoverageIdsMap CopyPixelOutsideMatte(const std::vector<const float*>& inputRowData, size_t rowIndex, float matteAlpha);

    // Create an intermediate map to utilise its insertion. Maybe slower and
    // one extra step compared to direct insertion into the output row data,
    // but less code.
    CoverageIdsMap CreateCoverageIdPairs(const std::vector<const float*>& inputRowData, size_t rowIndex, float matteId, float matteAlpha, bool isOver);

    // Write out the intermediate map to the output row data
    void WriteOutputRowData(const CoverageIdsMap& coverageIdPairs, const std::vector<float*>& outputRowData, size_t rowIndex);

    // Validates the layer name whether it is a Nuke and Cryptomatte
    // compatible layer name. Returns an error string if the layer name is
    // invalid, otherwise an empty string.
    std::string ValidateLayerName(const std::string& layerName);
  }
}

#endif
