// codecWhitelist.cpp
// Copyright (c) 2014 The Foundry Visionmongers Ltd.  All Rights Reserved.
//
// It's easier to include those formats and codecs we DO want
//
// Note, we ignore ProRes as this is handled seperately.
// Note, keep formats and codecs synchronized with hiero/core/FnExporterBase.py
// TODO use a common source for formats and codecs

#include "codecWhitelist.h"
#include "DDImage/Application.h"
#include <stddef.h>
#include <stdlib.h>
#include <string>
#include <string.h>
#include <map>

namespace Foundry
{
namespace Nuke
{
namespace
{
  struct FilterEntry
  {
      const char* name;
      bool        enableReader;
      bool        enableWriter;
  };

  // Bug 11027 - Nuke write: ffmpeg codec fails has details on individual codecs

  // For a full list of formats, define FN_FFMPEGWRITER_PRINT_CODECS in ffmpegWriter.cpp
  const FilterEntry kFormatWhitelist[] =
  {
    { "avi",            true,  true },
    { "h264",           true,  false}, // Change enableWriter to true if building to add h264
    { "mov",            true,  true },
    { "mp4",            true,  true },
    { "mpeg",           true,  true },
    { nullptr,          false, false }
  };

  // For a full list of formats, define FN_FFMPEGWRITER_PRINT_CODECS in ffmpegWriter.cpp
  // A range of codecs are omitted for licensing reasons, or because they support obselete/unnecessary
  // formats that confuse the interface. 
  
  const FilterEntry kCodecWhitelist[] =
  {
    // Video codecs.
    { "avrp",           true,  false }, // Avid 1:1 10-bit RGB Packer - write not supported as not official qt readable with relevant 3rd party codec.
    { "avui",           true,  false }, // Avid Meridien Uncompressed - write not supported as this is an SD only codec.
    { "ayuv",           true,  false }, // Uncompressed packed MS 4:4:4:4 - write not supported as not official qt readable.
    { "h264",           true,  true}, // Change enableWriter to true if building to add h264
    { "jpeg2000",       true,  false }, // JPEG 2000 - write not supported as it looks terrible.
    { "jpegls",         true,  false }, // JPEG-LS - write not supported as can't be read in in official qt.
    { "ljpeg",          true,  false }, // Lossless JPEG - write not supported as can't be read in in official qt.
    { "mjpeg",          true,  true}, // MJPEG (Motion JPEG) - this looks to be MJPEG-A. MJPEG-B encoding is not supported by FFmpeg so is not included here. To avoid confusion over the MJPEG-A and MJPEG-B variants, this codec is displayed as 'Photo JPEG'. This is done to i) avoid the confusion of the naming, ii) be consistent with Apple QuickTime, iii) the term 'Photo JPEG' is recommend for progressive frames which is appropriate to Nuke/NukeStudio as it does not have interlaced support.
    { "mpeg1video",     true,  true }, // MPEG-1 video
    { "mpeg4",          true,  true }, // MPEG-4 part 2
    { "msmpeg4v2",      true,  false }, // MPEG-4 part 2 Microsoft variant version 2 - write not supported as doesn't read in official qt.
    { "msmpeg4",        true,  false }, // MPEG-4 part 2 Microsoft variant version 3 - write not supported as doesn't read in official qt.
    { "png",            true,  true }, // PNG (Portable Network Graphics) image
    { "qtrle",          true,  true }, // QuickTime Animation (RLE) video
    { "r10k",           true,  false }, // AJA Kono 10-bit RGB - write not supported as not official qt readable with relevant 3rd party codec.
    { "r210",           true,  false }, // Uncompressed RGB 10-bit - write not supported as not official qt readable with relevant 3rd party codec without colourshifts.
    { "rawvideo",       true,  false }, // Uncompressed 4:2:2 8-bit - write not supported as not official qt readable.
    { "tiff",           true,  false }, // TIFF Image - write not supported as not official qt readable.
    { "v210",           true,  true }, // Uncompressed 4:2:2 10-bit
    { "v308",           true,  false }, // Uncompressed packed 4:4:4 - write not supported as not official qt readable and 8-bit only.
    { "v408",           true,  false }, // Uncompressed packed QT 4:4:4:4 - write not supported as official qt can't write, so bad round trip choice and 8-bit only.
    { "v410",           true,  false }, // Uncompressed 4:4:4 10-bit - write not supported as not official qt readable with standard codecs.
    { "prores",          true,  true }, // Apple ProRes
    { "dnxhd",           true,  true }, // Avid DNxHD
    { "mjpegb",          true,  true}, //MJPEG-B
    { "mjpega",          true,  true}, //MJPEG-A

    // Audio codecs.
    { "pcm_alaw",       true,  true }, // PCM A-law / G.711 A-law
    { "pcm_f32be",      true,  true }, // PCM 32-bit floating point big-endian
    { "pcm_f32le",      true,  true }, // PCM 32-bit floating point little-endian
    { "pcm_f64be",      true,  true }, // PCM 64-bit floating point big-endian
    { "pcm_f64le",      true,  true }, // PCM 64-bit floating point little-endian
    { "pcm_mulaw",      true,  true }, // PCM mu-law / G.711 mu-law
    { "pcm_s16be",      true,  true }, // PCM signed 16-bit big-endian
    { "pcm_s16le",      true,  true }, // PCM signed 16-bit little-endian
    { "pcm_s24be",      true,  true }, // PCM signed 24-bit big-endian
    { "pcm_s24le",      true,  true }, // PCM signed 24-bit little-endian
    { "pcm_s32be",      true,  true }, // PCM signed 32-bit big-endian
    { "pcm_s32le",      true,  true }, // PCM signed 32-bit little-endian
    { "pcm_s8",         true,  true }, // PCM signed 8-bit
    { "pcm_u16be",      true,  true }, // PCM unsigned 16-bit big-endian
    { "pcm_u16le",      true,  true }, // PCM unsigned 16-bit little-endian
    { "pcm_u24be",      true,  true }, // PCM unsigned 24-bit big-endian
    { "pcm_u24le",      true,  true }, // PCM unsigned 24-bit little-endian
    { "pcm_u32be",      true,  true }, // PCM unsigned 32-bit big-endian
    { "pcm_u32le",      true,  true }, // PCM unsigned 32-bit little-endian
    { "pcm_u8",         true,  true }, // PCM unsigned 8-bit
    { nullptr,          false, false }
  };

  // codecs to disable in NC mode only
  const FilterEntry kPLECodecBlacklist[] =
  {
    { "h264",           false, false }, // h264 is disabled for NC due to licensing issues when we cannot count the number of seats
    { "mpeg4",          false, false }, // mpeg4 is disabled for NC due to licensing issues when we cannot count the number of seats
    { "msmpeg4v2",      false, false }, // MPEG-4 part 2 Microsoft variant version 2 is disabled for NC due to licensing issues when we cannot count the number of seats
    { "msmpeg4",        false, false }, // MPEG-4 part 2 Microsoft variant version 3 is disabled for NC due to licensing issues when we cannot count the number of seats
    { nullptr,          false, false }
  };

  typedef std::map<std::string, std::string> CodecMap;
  CodecMap CreateCodecOverridenLongNamesMap()
  {
    CodecMap m;
    m["mjpeg"]         = "Photo - JPEG";
    m["mpeg1video"]    = "MPEG-1 Video";
    m["mpeg2video"]    = "MPEG-2 Video";
    m["mpeg4"]         = "MPEG-4";
    m["png"]           = "PNG";
    m["qtrle"]         = "Animation";
    m["v210"]          = "Uncompressed";
    m["mjpegb"]        = "Motion JPEG B";
    m["mjpega"]        = "Motion JPEG A";
    return m;
  }
  
  static const CodecMap kCodecOverridenLongNames = CreateCodecOverridenLongNamesMap();
  
  CodecMap CreateCodecKnobLabelsMap()
  {
    CodecMap m;
    
    // Video codecs.
    m["mjpeg"]         = "jpeg\tPhoto - JPEG";
    m["mpeg1video"]    = "mp1v\t\a";
    m["mpeg2video"]    = "mp2v\tMPEG-2 Video";
    m["mpeg4"]         = "mp4v\tMPEG-4";
    m["png"]           = "png \tPNG";
    m["qtrle"]         = "rle \tAnimation";
    m["v210"]          = "v210\tUncompressed";
    m["mjpegb"]        = "mjpb\tMotion JPEG B";
    m["mjpega"]        = "mjpa\tMotion JPEG A";
    return m;
  }
  
  static const CodecMap kCodecKnobLabels = CreateCodecKnobLabelsMap();
  
  #define CODEC_TRANSLATION_SCRIPT(knob_name)                             \
  "#puts \"Loading codec label '$value'\" \n"                             \
  " \n"                                                                   \
  "# Translate from old enumeration values to new ones \n"                \
  " \n"                                                                   \
  "# A list of lists of old labels against new labels \n"                 \
  "set label_table [list "                                                \
  "  {{Apple ProRes 4444}         {ap4h\tApple ProRes 4444}} "            \
  "  {{Apple ProRes 422 HQ}       {apch\tApple ProRes 422 HQ}} "          \
  "  {{Apple ProRes 422}          {apcn\tApple ProRes 422}} "             \
  "  {{Apple ProRes 422 LT}       {apcs\tApple ProRes 422 LT}} "          \
  "  {{Apple ProRes 422 Proxy}    {apco\tApple ProRes 422 Proxy}} "       \
  "  {{Photo - JPEG}              {jpeg\tPhoto - JPEG}} "                 \
  "  {{MPEG-1 Video}              {mp1v\tMPEG-1 Video}} "                 \
  "  {{MPEG-4 Video}              {mp4v\tMPEG-4}} "                       \
  "  {{PNG}                       {png \tPNG}} "                          \
  "  {{Animation}                 {rle \tAnimation}} "                    \
  "  {{Uncompressed 10-bit 4:2:2} {v210\tUncompressed}} "                 \
  ""                                                                      \
  "  {{MJPEG (Motion JPEG)}       {jpeg\tPhoto - JPEG}} "                 \
  "  {{MPEG-1 video}              {mp1v\tMPEG-1 Video}} "                 \
  "  {{MPEG-4 part 2}             {mp4v\tMPEG-4}} "                       \
  "  {{PNG image}                 {png \tPNG}} "                          \
  "  {{QuickTime Animation (RLE) video} {rle \tAnimation}} "              \
  "  {{Uncompressed 10-bit 4:2:2} {v210\tUncompressed}}] \n"              \
  " \n"                                                                   \
  "set old_index 0 \n"                                                    \
  "set new_index 1 \n"                                                    \
  " \n"                                                                   \
  "set found_label false \n"                                              \
  " \n"                                                                   \
  "# Look up our value in the list \n"                                    \
  "foreach pair $label_table { \n"                                        \
  "  if {$value eq [lindex $pair $old_index]} { \n"                       \
  "    knob "#knob_name" [lindex $pair $new_index] \n"                    \
  "    set found_label true \n"                                           \
  "    #puts -nonewline \"Replacing '$value' \" \n"                       \
  "    #puts \"with '[lindex $pair $new_index]'\" \n"                     \
  "    break \n"                                                          \
  "  } \n"                                                                \
  "} \n"                                                                  \
  "# If not found then set directly \n"                                   \
  "if {!$found_label} { \n"                                               \
  "  knob "#knob_name" $value \n"                                         \
  "  #puts \"Passing through '$value'\" \n"                               \
  "} \n";
  
  // It is necessary to translate knobs for the mov64Writer and the wrapper movWriter
  // so we generate a translation script for each of these.
  //
  // The mov64 script is used when loading old scripts with ffmpeg writers on linux.
  // The wrapper script is used when loading old scripts with ffmpeg writers on windows and mac.
  // Linux should be updated to use the wrapper movWriter at which point the mov64 translation
  // script may be removed.
  const char * const kMov64CodecTranslationScript = CODEC_TRANSLATION_SCRIPT(mov64_codec);
  const char * const kMetaCodecTranslationScript = CODEC_TRANSLATION_SCRIPT(meta_codec);
}

namespace
{
  const FilterEntry* getEntry(const char* name, const FilterEntry* whitelist, std::string& outReason, const FilterEntry* blacklist = nullptr)
  {
    const FilterEntry* iterWhitelist = whitelist;

    const size_t nameLength = strlen(name);
    // check for normal mode
    while (iterWhitelist->name != nullptr) {
      size_t iteNameLength = strlen(iterWhitelist->name);
      size_t maxLength = (nameLength > iteNameLength) ? nameLength : iteNameLength;
      if (strncmp(name, iterWhitelist->name, maxLength) == 0) {
        // Found in whitelist, now check blacklist
        if (blacklist) {
          const FilterEntry* iterBlacklist = blacklist;

          while (iterBlacklist->name != nullptr) {
            iteNameLength = strlen(iterBlacklist->name);
            maxLength = (nameLength > iteNameLength) ? nameLength : iteNameLength;
            if (strncmp(name, iterBlacklist->name, maxLength) == 0) {
              //if the blacklist where we found the codec was the NC one
              //set the appropiate message
              if ( kPLECodecBlacklist == blacklist ) {
                outReason = std::string(name)+
                    " codecs are disabled in non-commercial "+
                    DD::Image::Application::GetProductName()+
                    ".  \nPlease see the User Guide for details.  ";
              }
              // Found in codec whitelist but blacklisted too
              return nullptr;
            }

            ++iterBlacklist;
          }
        }
        // Found in whitelist and not in blacklist
        return iterWhitelist;
      }

      ++iterWhitelist;
    }
    return nullptr;
  }

  const FilterEntry* getEntry(const char* name, const FilterEntry* whitelist, const FilterEntry* blacklist = nullptr)
  {
    std::string dummyReason;
    return getEntry(name, whitelist, dummyReason, blacklist);
  }

  const FilterEntry* getBlacklist() {
    if (DD::Image::Application::IsNukePLE()) {
      return kPLECodecBlacklist;
    }
    return nullptr;
  }
}


bool isFormatWhitelistedForReading(const char* formatName)
{
  const FilterEntry* whitelistEntry = getEntry(formatName, kFormatWhitelist);
  return (whitelistEntry && whitelistEntry->enableReader);
}

bool isFormatWhitelistedForWriting(const char* formatName)
{
  const FilterEntry* whitelistEntry = getEntry(formatName, kFormatWhitelist);
  return (whitelistEntry && whitelistEntry->enableWriter);
}

bool isCodecWhitelistedForReading(const char* codecName, std::string& outReason)
{
  const FilterEntry* whitelistEntry = getEntry(codecName, kCodecWhitelist, outReason, getBlacklist() );
  return (whitelistEntry && whitelistEntry->enableReader);
}

bool isCodecWhitelistedForWriting(const char* codecName, std::string& outReason)
{
  const FilterEntry* whitelistEntry = getEntry(codecName, kCodecWhitelist, outReason, getBlacklist());
  return (whitelistEntry && whitelistEntry->enableWriter);
}

const char* getCodecOverridenLongName(const char* codecShortName)
{
  CodecMap::const_iterator it = kCodecOverridenLongNames.find(std::string(codecShortName));
  if (it != kCodecOverridenLongNames.end()) {
    return it->second.c_str();
  } else {
    return "";
  }
}
  
const char* getCodecKnobLabel(const char* codecShortName)
{
  CodecMap::const_iterator it = kCodecKnobLabels.find(std::string(codecShortName));
  if (it != kCodecKnobLabels.end()) {
    return it->second.c_str();
  } else {
    return "";
  }
}
  
const char* getMov64CodecTranslationScript()
{
  return kMov64CodecTranslationScript;
}
  
const char* getMetaCodecTranslationScript()
{
  return kMetaCodecTranslationScript;
}
  
}
}
