////////////////////////////////////////////////////////////////////////////////
// Copyright (c) 2013 The Foundry Visionmongers Ltd.  All Rights Reserved.
////////////////////////////////////////////////////////////////////////////////

// This file contains code which is common to movReader, movWriter, mov64Reader
// and mov64Writer.

#ifndef NUKE_MOVCOMMON_H
#define NUKE_MOVCOMMON_H

////////////////////////////////////////////////////////////////////////////////
// Includes
////////////////////////////////////////////////////////////////////////////////

// DDImage headers
#include "DDImage/Enumeration_KnobI.h"
#include "DDImage/LUT.h"

////////////////////////////////////////////////////////////////////////////////
// Static variables
////////////////////////////////////////////////////////////////////////////////
static const char kDefaultLabel_Prefix[] = "default (";
static const char kDefaultLabel_Suffix[] = ")";

static const size_t kDefaultLabel_PrefixSize = (mFnSizeofArray(kDefaultLabel_Prefix) - 1);
static const size_t kDefaultLabel_SuffixSize = (mFnSizeofArray(kDefaultLabel_Suffix) - 1);
static const size_t kDefaultLabel_AppendSize = (kDefaultLabel_PrefixSize + kDefaultLabel_SuffixSize);

// The fourccs for MJPEGA, MJPEGB, and Photo JPEG
static const char*  kJpegCodecs[] = { "jpeg", "mjpa", "mjpb" };
static const int    kNumJpegCodecs = mFnSizeofArray(kJpegCodecs);

static const int kUseInternalKnobValue = -1;

//These are currently also defined in Core/Quicktime which we don't want to drag
//into the mov64 reader.
static const char kMetaKeyApplication[]         = "uk.co.thefoundry.Application";
static const char kMetaKeyApplicationVersion[]  = "uk.co.thefoundry.ApplicationVersion";
static const char kMetaKeyYCbCrMatrix[]         = "uk.co.thefoundry.YCbCrMatrix";
static const char kMetaKeyPixelFormat[]         = "uk.co.thefoundry.PixelFormat";
static const char kMetaKeyColorspace[]          = "uk.co.thefoundry.Colorspace";
static const char kMetaKeyWriter[]              = "uk.co.thefoundry.Writer";
static const char kMetaValueWriter32[]          = "mov32";
static const char kMetaValueWriter64[]          = "mov64";

////////////////////////////////////////////////////////////////////////////////
// Static functions
////////////////////////////////////////////////////////////////////////////////

// Safety cut out in case of bad data
#define MAX_LUTS 100

// Removes 'default (<<lut_name>>)' string from the given lut name
//
// param string lutName:  a string with the lut name
// returns: a string with the stripped name
static std::string stripDefaultLutName(const std::string& lutName)
{
  std::string strippedLutName = lutName;

  const bool startsWithDefaultPrefix = lutName.find(kDefaultLabel_Prefix) == 0;
  const bool endsWithDefaultSuffix = lutName.rfind(kDefaultLabel_Suffix) == lutName.length() - kDefaultLabel_SuffixSize;

  if (startsWithDefaultPrefix && endsWithDefaultSuffix) {
    strippedLutName = lutName.substr(kDefaultLabel_PrefixSize, lutName.length() - kDefaultLabel_AppendSize);
  }

  return strippedLutName;
}

// Find the current LUT name from the built in list, otherwise return null.
//
// param pLut:    LUT pointer to get the name for.
static const char* const GetLutName(const DD::Image::LUT* const pLut)
{
  const char* lutName = nullptr;
  if ( pLut ) {
    for ( int i = 1; DD::Image::LUT::builtin_names[i] && i <= MAX_LUTS; i++ ) {
      if ( pLut == DD::Image::LUT::GetBuiltinLUT(i) ) {
        lutName = DD::Image::LUT::builtin_names[i];
        break;
      }
    }
  }  
  return lutName;
}

// Set the knob value from a label.
//
// param knob:      Input knob
// param menu:      Menu options
// param label:     Target label
static void SetKnobValue(DD::Image::Knob* knob, const StringArray& menu, const std::string& label)
{
  StringArray::const_iterator it = std::find(menu.begin(), menu.end(), label);
  const bool foundLabel = (it != menu.end());
  
  if (!foundLabel) {
    // Decide whether to change the value.
    // This will only occur if the value has been initialised and is not default.
    const bool changeValue = (knob->get_value() != 0.0);
    if (changeValue) {
      // Issue a warning if changing an initialised non-default value
      DD::Image::Knob::warning(knob, "%s value not supported: %s", knob->name().c_str(), label.c_str());
      knob->set_value(0.0);
    }
  }
  else {
    knob->set_value(static_cast<double>(it - menu.begin()));
  }
  
  knob->updateWidgets();
}

// Get the current label from a knob.
// We need range checking for these labels to handle loaded files.
//
// param knob:      Input knob
// param menu:      Menu options
// param knobValue: Current knob value (default to knob->get_value())
// returns:         Current label (empty if not found)
static std::string GetKnobLabel(DD::Image::Knob* knob, const StringArray& menu, int knobValue = kUseInternalKnobValue)
{
  // Use the knob for the default value
  if (knobValue == kUseInternalKnobValue) {
    knobValue = int(knob->get_value());
  }
  
  // Get the label if the value is valid for this list
  // For some reason enumeration knobs are loaded in two phases; by storing the
  // knob value and then copying the menu options into the knob.
  // The LUT is updated during this process. We need to defend against the knob
  // value being out of range at this stage since a new value will be used with
  // an old set of options.
  // We don't need to worry about the incorrect evaluation here as this will be
  // reevaluated on loading the menu options.
  std::string label;
  const bool isValidSelection = (knobValue >= 0 && knobValue < int(menu.size()));
  if (isValidSelection) {
    label = menu[knobValue];
#ifdef DEBUG_PRINT_KNOB_UPDATES
    std::cerr << "GetKnobLabel: knob->get_value()=" << knobValue << ", " << label << endl;
#endif
  }
  
  // Empty if not found
  return label;
}

// Get the label from an enumeration knob.
//
// param outCurrentLabel:     Label for the current value
// param knob:                Input knob
// param knobValue:           Current knob value (default to knob->get_value())
// param outEnumerationKnob:  Enumeration knob if available (ignored if NULL)
// returns:                   Whether a label was found
static bool GetEnumerationKnobLabel( std::string& outCurrentLabel,
                              DD::Image::Knob* const knob,
                              const int knobValue = kUseInternalKnobValue,
                              DD::Image::Enumeration_KnobI** outEnumerationKnob = nullptr )
{
  // Knob may not be initialised
  if(knob) {
    DD::Image::Enumeration_KnobI* enumerationKnob = knob->enumerationKnob();
    mFnAssertMsg(enumerationKnob, "GetEnumerationKnobLabel: Must be an enumeration knob");
    if (enumerationKnob) {
      const StringArray& options = enumerationKnob->menu();
      if (outEnumerationKnob) {
        *outEnumerationKnob = enumerationKnob;
      }
    
      // Options may not be initialised
      if (!options.empty()) {
        outCurrentLabel = GetKnobLabel(knob, options, knobValue);
        return true;
      }
    }
  }
  
  return false;
}

static bool IsJpeg(DD::Image::Knob* codecKnob, int codecValue)
{
  std::string strCodec;
  GetEnumerationKnobLabel(strCodec, codecKnob, codecValue);

  // Check the fourcc for a JPEG codec
  for (int i = 0; i < kNumJpegCodecs; ++i) {
    const char* codec = kJpegCodecs[i];
    // All labels start with the fourcc.  This should never change as it would
    // break existing Nuke scripts.
    const bool startsWithCodec = (strCodec.find(codec) == 0);
    if (startsWithCodec) {
      return true;
    }
  }
  
  return false;
}

#if !defined NUKE_MOV64WRITER_CPP

static std::string GetPixelTypeString(DD::Image::Knob* pixelTypeKnob, int pixelTypeValue) 
{
  std::string strPixelType;
  GetEnumerationKnobLabel(strPixelType, pixelTypeKnob, pixelTypeValue);

  // Handle default
  if (strPixelType.find(kDefaultLabel_Prefix) == 0 && strPixelType.rfind(kDefaultLabel_Suffix) == strPixelType.length()-kDefaultLabel_SuffixSize) {
    strPixelType = strPixelType.substr(kDefaultLabel_PrefixSize, strPixelType.length()-kDefaultLabel_AppendSize);
  }
  
  return strPixelType;
}

#endif // !defined NUKE_MOV64WRITER_CPP
#endif // NUKE_MOVCOMMON_H
