// Copyright (c) 2014 The Foundry Visionmongers Ltd.  All Rights Reserved.
#ifdef _WIN32
#define _CRT_NONSTDC_NO_DEPRECATE // Warnings complaining about the use of POSIX complient but non-standard C methods.
#define _CRT_SECURE_NO_WARNINGS   // Microsoft specific CRT warnings most of which have no real cross-platform solution.
#pragma warning(disable: 4251)    // Supress warnings from DD::Image - class needs to have dll-interface to be used by clients of class 'DD::Image::...'
#pragma warning(disable: 4996)    // We cannot control the Windows OS version checking in this plugin.
#endif

#include "movReader.h"
#include "movReaderFormat.h"
#include "DDImage/TimelineRead.h"
#include "DDImage/MetaData.h"
#include "DDImage/TextKnobI.h"
#include <sstream>
#include <string>
#include <list>
#include <map>

#ifdef FN_LICENSED_PRORES_DNXHD_CODEC
  #include "AvidDNxCodec.h"
#endif

#if PLUGIN_UPDATE
# include "version.h"
#endif

#ifndef _WIN32
  #include <unistd.h>
#endif

using namespace DD::Image;

#define Mov64ReaderErrPrefix "Mov64Reader: "

static const char* const kDefaultColorMatrixLabel = "default (ycbcr_matrix)";
static const char* const kDefaultDecodeVideoLevels = "default (decode_video_levels)";

static const int kVideoRange = 1;
static const int kColorConverterRowPadMultiple = 16;

#define CHECK(x) \
{\
  int error = x;\
  if (error<0) {\
    setInternalError(error);\
    return;\
  }\
}\

#ifdef FN_LICENSED_PRORES_DNXHD_CODEC
static const std::string kUnknownCodecTypeName = NukeCodecs::CodecFactory::codecNameFromType(NukeCodecs::CodecFactory::eUnknownCodec);
#else
static const std::string kUnknownCodecTypeName = "Unknown";
#endif

// Get the display string for a COLR/NCLC atom's primaries, based on index.
//
// param index:     NCLC primaries index
static const std::string GetNCLCPrimariesLabelFromIndex(const int index)
{
  if (0 > index || index >= kNCLCPrimariesLabel.size()) {
    // out of range
    return kNCLCUnknownLabel;
  }

  return kNCLCPrimariesLabel[index];
}

// Get the display string for a COLR/NCLC atom's transfer func, based on index.
//
// param index:     NCLC transfer index
static const std::string GetNCLCTransferLabelFromIndex(const int index)
{
  if (0 > index || index >= kNCLCTransferLabel.size()) {
    // out of range
    return kNCLCUnknownLabel;
  }

  return kNCLCTransferLabel[index];
}

// Get the display string for a COLR/NCLC atom's YCbCr matrix, based on index.
//
// param index:     NCLC matrix index
static const std::string GetNCLCMatrixLabelFromIndex(const int index)
{
  if (0 > index || index >= kNCLCMatrixLabel.size()) {
    // out of range
    return kNCLCUnknownLabel;
  }

  return kNCLCMatrixLabel[index];
}

typedef std::vector<std::string> MetaPrefixStripList;
static const MetaPrefixStripList kMetaPrefixStripList = {
    "com.",
    "uk.",
    "co.", //Note order to allow uk.co. to be fully stripped.
    "de.",
    "org.",
    "net.",
    "gov.",
};


//Reformat meta key parses the incoming string to optionally strip inverse
//DNS notation, as well as replace subdomain .'s with /s to match Nuke
//standard metadata formatting.
static std::string ReformatMetaKey(const std::string& origKey, const bool doNotAttachPrefix,
                                   const bool matchMetaFormat)
{
  std::string key = origKey;
  if(matchMetaFormat) {
    //Strip leading reverse dns for those we can identify, otherwise leave.
    for (const auto& prefix : kMetaPrefixStripList) {
      if((prefix.size() < key.size()) && (key.substr(0, prefix.size()) == prefix))
        key = key.substr(prefix.size(), key.size());
    }
    
    std::replace(key.begin(), key.end(), '.', '/');
  }
  if(doNotAttachPrefix) return key;
  return std::string(MetaData::QuickTime::QUICKTIME_PREFIX).append(key);
}

typedef std::map<std::string, std::string> MetaKeyMap;
MetaKeyMap CreateMetaKeyMap()
{
  MetaKeyMap m;
  m["author"]     = MetaData::CREATOR;
  m["comment"]    = MetaData::COMMENT;
  m["album"]      = MetaData::PROJECT;
  m["copyright"]  = MetaData::COPYRIGHT;
  return m;
}
static const MetaKeyMap kMetaKeyMap = CreateMetaKeyMap();


//Remap deals with special cased keys going into shared meta namespaces
static std::string RemapMetaKey(const std::string& origKey, const bool doNotAttachPrefix,
                                const bool matchMetaFormat)
{
  MetaKeyMap::const_iterator it = kMetaKeyMap.find(origKey);
  if (it != kMetaKeyMap.end())
    return ReformatMetaKey(it->second, doNotAttachPrefix, matchMetaFormat);
  //It's not in the direct mappings list, but we still need the prefix.
  std::string newKey = ReformatMetaKey(origKey, doNotAttachPrefix, matchMetaFormat);
  return newKey;
}

//Set metadata item is the entry point for special case metadata items that don't need
//to be checked against an exclude list.
static void SetMetaDataItem(MetaData::Bundle& metadata, const std::string& key,
                            const std::string& value, const bool doNotAttachPrefix,
                            const bool matchMetaFormat)
{
  std::string newKey = RemapMetaKey(key, doNotAttachPrefix, matchMetaFormat);
  metadata.setData(newKey, value);
}

//The MetaKeyIgnoreList deals with stripping metadata inserted by ffmpeg, but which
//we wish to suppress from the meta list inserted in the stream, either because it's
//inconsistent with the standard mov metadata, or because it's values used to convey
//internal TF data that's interpreted separately (eg the NCLC atom keys).
typedef std::vector<std::string> MetaKeyList;
MetaKeyList CreateMetaKeyIgnoreList()
{
  MetaKeyList l;
  l.push_back("creation_time");
  l.push_back("language");
  l.push_back("handler_name");
  l.push_back("compatible_brands");
  l.push_back("major_brand");
  l.push_back("minor_version");
  l.push_back("timecode");  //This is specifically searched out and push_backed separately.
  l.push_back("com.arri.camera.LookFileXml"); //This formats terribly, since its a long xml string.
  l.push_back(kNCLCPrimariesKey);
  l.push_back(kNCLCMatrixKey);
  l.push_back(kNCLCTransferKey);
  std::sort(l.begin(), l.end());
  return l;
}
static const MetaKeyList kMetaKeyIgnoreList = CreateMetaKeyIgnoreList();

static bool IsMetaKeyIgnored(const std::string& key)
{
  return (std::find(kMetaKeyIgnoreList.begin(), kMetaKeyIgnoreList.end(),
                    key) != kMetaKeyIgnoreList.end());
}

//Validate and set metadata item is the generic entry point for adding general metadata items
//that need to be checked against an exclude list.
static void ValidateAndSetMetaDataItem(MetaData::Bundle& metadata, const std::string& key,
                                       const std::string& value, const bool doNotAttachPrefix,
                                       const bool matchMetaFormat)
{
  if(!IsMetaKeyIgnored(key))
    SetMetaDataItem(metadata, key, value, doNotAttachPrefix, matchMetaFormat);
}

static std::string GetCodecName(FFmpegFile& reader)
{
  // Return early if the reader is in an invalid state.
  if (reader.invalid()) {
    return kUnknownCodecTypeName;
  }

  // Get the name from the metadata.
  DD::Image::MetaData::Bundle metaData;
  reader.metadata(metaData);
  const std::string codecName = metaData.getString(MetaData::QuickTime::CODEC_NAME);
  return codecName.empty() ? kUnknownCodecTypeName : codecName;
}

NukeCodecs::IDecoder * FFmpegFile::Stream::getColorConvertor(AVPixelFormat srcPixelFormat, int width, int height, int srcColorRange)
{
  bool reset_ctx = _resetConvertCtx;
  if (_resetConvertCtx) {
    _resetConvertCtx = false;
    _color_convertor.reset();
  }

  if (!_color_convertor) {
    _color_convertor.reset(NukeCodecs::CodecFactory::makeDecoder(NukeCodecs::CodecFactory::eColorSpaceConvertor));
    _csc_params.colorMatrix = isRec709Format() ? NukeCodecs::DecoderParams::kRec709 : NukeCodecs::DecoderParams::kRec601;
    _csc_params.width = width;
    _csc_params.height = height;
    _csc_params.out_planes = 1;
    // Optional color space override
    if (_colorMatrixTypeOverride > 0) {
      if (_colorMatrixTypeOverride == 1) {
        _csc_params.colorMatrix = NukeCodecs::DecoderParams::kRec709;
      }
      else {
        _csc_params.colorMatrix = NukeCodecs::DecoderParams::kRec601;
      }
    }
    
    switch (srcColorRange) {
    case AVCOL_RANGE_MPEG:
      _csc_params.srcDataRange = NukeCodecs::DecoderParams::kVideoRange;
      break;
    case AVCOL_RANGE_JPEG:
      _csc_params.srcDataRange = NukeCodecs::DecoderParams::kFullRange;
      break;
    case AVCOL_RANGE_UNSPECIFIED:
    default:
      // If the colour range wasn't specified, set the flag according to
      // whether the data is YUV or not.
      _csc_params.srcDataRange = isYUV() ? NukeCodecs::DecoderParams::kVideoRange : NukeCodecs::DecoderParams::kFullRange;
      break;
    }
      
    switch (srcPixelFormat) {
      case AV_PIX_FMT_NV12:
        _csc_params.srcSubsampling = NukeCodecs::DecoderParams::k420;
        _csc_params.bitDepth = 8;
        _csc_params.in_planes = 2;
        break;
      case AV_PIX_FMT_YUV420P:
      case AV_PIX_FMT_YUVJ420P:
        _csc_params.srcSubsampling = NukeCodecs::DecoderParams::k420;
        _csc_params.bitDepth = 8;
        _csc_params.in_planes = 3;
        break;
      case AV_PIX_FMT_YUV422P:
      case AV_PIX_FMT_YUVJ422P:
        _csc_params.srcSubsampling = NukeCodecs::DecoderParams::k422;
        _csc_params.bitDepth = 8;
        _csc_params.in_planes = 3;
        break;
      case AV_PIX_FMT_YUV444P:
      case AV_PIX_FMT_YUVJ444P:
        _csc_params.srcSubsampling = NukeCodecs::DecoderParams::k444;
        _csc_params.bitDepth = 8;
        _csc_params.in_planes = 3;
        break;
      case AV_PIX_FMT_YUV420P16LE:
        _csc_params.srcSubsampling = NukeCodecs::DecoderParams::k420;
        _csc_params.bitDepth = 16;
        _csc_params.in_planes = 3;
        break;
      case AV_PIX_FMT_YUV422P16LE:
        _csc_params.srcSubsampling = NukeCodecs::DecoderParams::k422;
        _csc_params.bitDepth = 16;
        _csc_params.in_planes = 3;
        break;
      case AV_PIX_FMT_YUV444P16LE:
        _csc_params.srcSubsampling = NukeCodecs::DecoderParams::k444;
        _csc_params.bitDepth = 16;
        _csc_params.in_planes = 3;
        break;
      case AV_PIX_FMT_YUV420P12LE:
        _csc_params.srcSubsampling = NukeCodecs::DecoderParams::k420;
        _csc_params.bitDepth = 12;
        _csc_params.in_planes = 3;
        break;
      case AV_PIX_FMT_YUV422P12LE:
        _csc_params.srcSubsampling = NukeCodecs::DecoderParams::k422;
        _csc_params.bitDepth = 12;
        _csc_params.in_planes = 3;
        break;
      case AV_PIX_FMT_YUV444P12LE:
        _csc_params.srcSubsampling = NukeCodecs::DecoderParams::k444;
        _csc_params.bitDepth = 12;
        _csc_params.in_planes = 3;
        break;
      case AV_PIX_FMT_YUV420P10LE:
        _csc_params.srcSubsampling = NukeCodecs::DecoderParams::k420;
        _csc_params.bitDepth = 10;
        _csc_params.in_planes = 3;
        break;
      case AV_PIX_FMT_YUV422P10LE:
        _csc_params.srcSubsampling = NukeCodecs::DecoderParams::k422;
        _csc_params.bitDepth = 10;
        _csc_params.in_planes = 3;
        break;
      case AV_PIX_FMT_YUV444P10LE:
        _csc_params.srcSubsampling = NukeCodecs::DecoderParams::k444;
        _csc_params.bitDepth = 10;
        _csc_params.in_planes = 3;
        break;
      default:
        _resetConvertCtx = reset_ctx;
        _color_convertor.reset();
        break;
    }
  }
    
  return _color_convertor.get();
}

SwsContext* FFmpegFile::Stream::getConvertCtx(AVPixelFormat srcPixelFormat, int srcWidth, int srcHeight, int srcColorRange, AVPixelFormat dstPixelFormat, int dstWidth, int dstHeight)
{
  // Reset is flagged when the UI colour matrix selection is
  // modified. This causes a new convert context to be created
  // that reflects the UI selection.
  if (_resetConvertCtx) {
    _resetConvertCtx = false;
    if (_convertCtx) {
      sws_freeContext(_convertCtx);
      _convertCtx = nullptr;
    }
  }

  if (!_convertCtx) {

    //Preventing deprecated pixel format used error messages, see:
    //https://libav.org/doxygen/master/pixfmt_8h.html#a9a8e335cf3be472042bc9f0cf80cd4c5
    //This manually sets them to the new versions of equivalent types.
    switch (srcPixelFormat) {
      case AV_PIX_FMT_YUVJ420P :
        srcPixelFormat = AV_PIX_FMT_YUV420P;
        break;
      case AV_PIX_FMT_YUVJ422P  :
        srcPixelFormat = AV_PIX_FMT_YUV422P;
        break;
      case AV_PIX_FMT_YUVJ444P   :
        srcPixelFormat = AV_PIX_FMT_YUV444P;
        break;
      case AV_PIX_FMT_YUVJ440P :
        srcPixelFormat = AV_PIX_FMT_YUV440P;
      default:
        break;
    }

    _convertCtx = sws_getContext(srcWidth, srcHeight, srcPixelFormat, // src format
                                 dstWidth, dstHeight, dstPixelFormat,        // dest format
                                 SWS_BICUBIC, nullptr, nullptr, nullptr);

    // Set up the SoftWareScaler to convert colorspaces correctly.
    // Colorspace conversion makes no sense for RGB->RGB conversions
    if (!isYUV()) {
      return _convertCtx;
    }

    int colorspace = isRec709Format() ? SWS_CS_ITU709 : SWS_CS_ITU601;
    // Optional color space override
    if (_colorMatrixTypeOverride > 0) {
      if (_colorMatrixTypeOverride == 1) {
        colorspace = SWS_CS_ITU709;
      }
      else {
        colorspace = SWS_CS_ITU601;
      }
    }

    // sws_setColorspaceDetails takes a flag indicating the white-black range of the input:
    //     0  -  mpeg, 16..235
    //     1  -  jpeg,  0..255 
    int srcRange;
    // Set this flag according to the color_range reported by the codec context.
    switch (srcColorRange) {
    case AVCOL_RANGE_MPEG:
      srcRange = 0;
      break;
    case AVCOL_RANGE_JPEG:
      srcRange = 1;
      break;
    case AVCOL_RANGE_UNSPECIFIED:
    default:
      // If the colour range wasn't specified, set the flag according to 
      // whether the data is YUV or not.
      srcRange = isYUV() ? 0 : 1;  
      break;
    }
    
    int result = sws_setColorspaceDetails(_convertCtx,
                                          sws_getCoefficients(colorspace), // inv_table
                                          srcRange, // srcRange -flag indicating the white-black range of the input (1=jpeg / 0=mpeg) 0 = 16..235, 1 = 0..255
                                          sws_getCoefficients(SWS_CS_DEFAULT), // table
                                          1, // dstRange - 0 = 16..235, 1 = 0..255
                                          0, // brightness fixed point, with 0 meaning no change,
                                          1 << 16, // contrast   fixed point, with 1<<16 meaning no change,
                                          1 << 16); // saturation fixed point, with 1<<16 meaning no change);

    assert(result != -1);
  }

  return _convertCtx;
}

/*static*/ double FFmpegFile::Stream::GetStreamAspectRatio(Stream* stream)
{
  if (stream->_avstream->sample_aspect_ratio.num) {
    return av_q2d(stream->_avstream->sample_aspect_ratio);
  }
  else if (stream->_codecContext->sample_aspect_ratio.num) {
    return av_q2d(stream->_codecContext->sample_aspect_ratio);
  }
  return stream->_aspect;
}

// get stream start time
int64_t FFmpegFile::getStreamStartTime(Stream& stream)
{
  FoundryAVPacket avPacket;

  // Read from stream. If the value read isn't valid, get it from the first frame in the stream that provides such a
  // value.
  int64_t startPTS = stream._avstream->start_time;
  int64_t startDTS = stream._avstream->start_time;

  if (startPTS < 0) {
    // Seek 1st key-frame in video stream.
    avcodec_flush_buffers(stream._codecContext);

    // Here, avPacket needs to be local as we don't need to keep any information outside this function context.
    // Do not replace this with _avPacket, which is global, because _avPacket is associated with the playback process
    // and that may produces inconsistencies. _avPacket is now used only in the |decode| method and we need to ensure
    // that it remains valid even after we get out of the |decode| method context, because the information stored by
    // _avPacket may be displayed on the screen for multiple frames. Please have a look at TP 162892 for more information
    // https://foundry.tpondemand.com/entity/162892
    if (av_seek_frame(_context, stream._idx, startPTS, AVSEEK_FLAG_BACKWARD) >= 0) {
      // Read frames until we get one for the video stream that contains a valid PTS.
      while (av_read_frame(_context, &avPacket) >= 0) {
        if (avPacket.stream_index != stream._idx) {
          continue;
        }
        // Packet read for video stream. Get its PTS.
        startPTS = avPacket.pts;
        startDTS = avPacket.dts;

        // Loop will continue if the current packet doesn't end after 0
        if (startPTS + avPacket.duration > 0) {
          break;
        }
      }
    }
  }

  // If we still don't have a valid initial PTS, assume 0. (This really shouldn't happen for any real media file, as
  // it would make meaningful playback presentation timing and seeking impossible.); 
  // TP 162519 - We discard the samples with a negative timestamp to make mov64 match the Quick Time Player;
  // Video streams are usually inter-coded (most frames rely on other frames to be decoded). So, when such a stream
  // is trimmed those reference frames, even if not within the trimmed time range, have to be included in the output
  // in order for the video to be playable. These frames are assigned to a negative time-stamp;
  // Based on the experimental work we concluded that for the streams that have included frames which are assigned 
  // to negative timestamps it doesn't make sense to pick a value grater than 0 for the first frame timestamp. The first
  // frame timestamp is going to match the packet which starts just before 0 and ends after 0 if that exists. Otherwise
  // it will be 0.
  // For more information please have a look at TP 162519
  const bool isStartPTSValid = (startPTS + avPacket.duration > 0);
  if (!isStartPTSValid) {
    startPTS = 0;
    startDTS = 0;
  }

  stream._startDTS = startDTS;

  return startPTS;
}

// Get the video stream duration in frames...
int64_t FFmpegFile::getStreamFrames(Stream& stream)
{
  // JHyde: previous versions of this code sometimes reported an incorrect frame count which could cause
  // Nuke to try to decode non existent frames (TP 466816), so we want to make sure the frame count is
  // as accurate as possible.
  // To summarise, the issue here is that there are several ways to measure the frame count, we were previously
  // preferring AVFormatContext::Duration but this would sometimes be inaccurate, because it may use the duration
  // of the longest stream in the file, which could sometimes be an audio stream that's longer than the video 
  // stream we're interested in. 
  // The rule we should follow is to prefer AVStream::duration if it is > 0 and less than AVContext::duration,
  // otherwise use AVContext::duration if it exists, or the alternative methods after. The justification is that
  // the AVContext duration will either come from the file's edit info, and it would make no sense to specify a 
  // duration longer than the length of the video stream, so should only be considered when it is shorter, or it
  // will be be calculated based on the longest stream in the file, such as the audio stream, but for the purposes
  // of video decoding we are only interested in the length of the video stream, which AVStream::Duration is a more
  // direct measure of.
  int64_t frames = 0;
  int64_t movieDurationFrames = 0;
  // calculate from stream's duration, fps and timebase.
  const auto streamDurationNum = static_cast<int64_t>(stream._avstream->duration) * stream._avstream->time_base.num  * stream._fpsNum;
  const auto streamDurationDen = static_cast<int64_t>(stream._avstream->time_base.den) * stream._fpsDen;
  const int64_t streamDurationframes = static_cast<int64_t>(std::round(streamDurationNum / static_cast<long double>(streamDurationDen)));

  /// Obtain from movie duration if specified (JHyde: and less than stream duration). This is preferred since mov/mp4 
  /// formats allow the media in tracks (=streams) to be remapped in time to the final movie presentation without
  /// needing to recode the underlying tracks content; the movie duration thus correctly describes the final presentation.
  if (_context->duration != 0) {
    // Annoyingly, FFmpeg exposes the movie duration converted (with round-to-nearest semantics) to units of
    // AV_TIME_BASE (microseconds in practice) and does not expose the original rational number duration
    // from a mov/mp4 file's "mvhd" atom/box. Accuracy may be lost in this conversion; a duration that was
    // an exact number of frames as a rational may end up as a duration slightly over or under that number
    // of frames in units of AV_TIME_BASE.
    // Conversion to whole frames rounds up the resulting number of frames because a partial frame is still
    // a frame. However, in an attempt to compensate for AVFormatContext's inaccurate representation of
    // duration, with unknown rounding direction, the conversion to frames subtracts 1 unit (microsecond)
    // from that duration. The rationale for this is thus:
    // * If the stored duration exactly represents an exact number of frames, then that duration minus 1
    //   will result in that same number of frames once rounded up.
    // * If the stored duration is for an exact number of frames that was rounded down, then that duration
    //   minus 1 will result in that number of frames once rounded up.
    // * If the stored duration is for an exact number of frames that was rounded up, then that duration
    //   minus 1 will result in that number of frames once rounded up, while that duration unchanged would
    //   result in 1 more frame being counted after rounding up.
    // * If the original duration in the file was not for an exact number of frames, then the movie timebase
    //   would have to be >= 10^6 for there to be any chance of this calculation resulting in the wrong
    //   number of frames. This isn't a case that I've seen. Even if that were to be the case, the original
    //   duration would have to be <= 1 microsecond greater than an exact number of frames in order to
    //   result in the wrong number of frames, which is highly improbable.
    int64_t divisor = int64_t(AV_TIME_BASE) * stream._fpsDen;
    movieDurationFrames = ((_context->duration - 1) * stream._fpsNum + divisor - 1) / divisor;
  }

  if ((movieDurationFrames > 0) && ((movieDurationFrames < streamDurationframes) || (streamDurationframes <= 0))) {
    frames = movieDurationFrames;
  }
  else if (streamDurationframes > 0) {
    frames = streamDurationframes;
  }

  // If the number of frames is still unknown, attempt to measure it from the last frame PTS for the stream in the
  // file relative to first (which we know from earlier).
  if (!frames) {
    int64_t maxPts = stream._startPTS;

    // Seek last key-frame.
    avcodec_flush_buffers(stream._codecContext);
    av_seek_frame(_context, stream._idx, stream.frameToPts(1<<29), AVSEEK_FLAG_BACKWARD);

    // Read up to last frame, extending max PTS for every valid PTS value found for the video stream.
    FoundryAVPacket avPacket;
    // Here, avPacket needs to be local as we don't need to keep any information outside this function context.
    // Do not replace this with _avPacket, which is global, because _avPacket is associated with the playback process
    // and that may produces inconsistencies. _avPacket is now used only in the |decode| method and we need to ensure
    // that it remains valid even after we get out of the |decode| method context, because the information stored by
    // _avPacket may be displayed on the screen for multiple frames. Please have a look at TP 162892 for more information
    // https://foundry.tpondemand.com/entity/162892

    while (av_read_frame(_context, &avPacket) >= 0) {
      if (avPacket.stream_index == stream._idx && avPacket.pts != int64_t(AV_NOPTS_VALUE) && avPacket.pts > maxPts)
        maxPts = avPacket.pts;
    }

    // Compute frame range from min to max PTS. Need to add 1 as both min and max are at starts of frames, so stream
    // extends for 1 frame beyond this.
    frames = 1 + stream.ptsToFrame(maxPts);
  }

  if (!frames) {
    // As a last resort use the number of avFrames, note this may not match the number of 'nuke frames'
    // if the stream has a variable frame length
    frames = stream._avstream->nb_frames;
  }

  return frames;
}

// Returns true if the properties of the two streams are considered to match in terms of
// codec, resolution, frame rate, time base, etc. The motivation for this is that streams
// that match in this way are more likely to contain multiple views rather then unrelated
// content. This is somewhat arbitrary but seems like a reasonable starting point until
// users tell us otherwise.
bool CheckStreamPropertiesMatch(const AVStream* streamA, const AVStream* streamB)
{
  const AVCodecParameters* codecA = streamA->codecpar;
  const AVCodecParameters* codecB = streamB->codecpar;

  // Sanity check //
  if (codecA == nullptr || codecB == nullptr) {
    return false;
  }

  // Not matching even if both reads failed
  if (codecA->format == AV_PIX_FMT_NONE || codecB->format == AV_PIX_FMT_NONE) {
    return false;
  }

  const AVPixFmtDescriptor* pixFmtDescA = av_pix_fmt_desc_get((AVPixelFormat)codecA->format);
  const AVPixFmtDescriptor* pixFmtDescB = av_pix_fmt_desc_get((AVPixelFormat)codecB->format);
  
  return
    (codecA->codec_id                 == codecB->codec_id) &&
    (codecA->bits_per_raw_sample      == codecB->bits_per_raw_sample) &&
    (codecA->width                    == codecB->width) &&
    (codecA->height                   == codecB->height) &&
    (codecA->sample_aspect_ratio.num  == codecB->sample_aspect_ratio.num) &&
    (codecA->sample_aspect_ratio.den  == codecB->sample_aspect_ratio.den) &&
    (pixFmtDescA->nb_components       == pixFmtDescB->nb_components) &&
    (streamA->sample_aspect_ratio.num == streamB->sample_aspect_ratio.num) &&
    (streamA->sample_aspect_ratio.den == streamB->sample_aspect_ratio.den) &&
    (streamA->time_base.num           == streamB->time_base.num) &&
    (streamA->time_base.den           == streamB->time_base.den) &&
    (streamA->start_time              == streamB->start_time) &&
    (streamA->duration                == streamB->duration) &&
    (streamA->nb_frames               == streamB->nb_frames) &&
    (streamA->r_frame_rate.num        == streamB->r_frame_rate.num) &&
    (streamA->r_frame_rate.den        == streamB->r_frame_rate.den);
}

// constructor
FFmpegFile::FFmpegFile(const char* filename, int viewIndex)
: _context(nullptr)
, _format(nullptr)
, _streams()
, _selectedStream(nullptr)
, _errorMsg()
, _invalidState(false)
, _lock()
{
  const char* strippedFilename = FileOp::StripPrefix(filename);

  AVDictionary* demuxerOptions = nullptr;
  // enabling drefs to allow reading from external tracks
  // this enables quicktime reference files demuxing
  av_dict_set(&demuxerOptions, "enable_drefs", "1", 0);
  CHECK( avformat_open_input(&_context, strippedFilename, _format, &demuxerOptions) );
  if (demuxerOptions != nullptr) {
    // demuxerOptions is destroyed and replaced, on avformat_open_input return,
    // with a dict containing the options that were not found
    mFnAssertMsg(false, "invalid options passed to demuxer");
    av_dict_free(&demuxerOptions);
  }
  mFnAssert(_context);
  // Bug 51016 - probesize is the maximum amount of data, in bytes, which is read to determine
  // frame information in avformat_find_stream_info. It's important that when reading
  // stereo quicktimes that the probe size be large enough to read data from both video tracks.
  // Otherwise the streams could be reported as having different properties, when really they're
  // the same but due to an insufficient probesize the second stream didn't have all the relevent data
  // loaded. This defaults to 5meg. 100meg should be enough for large stereo quicktimes.
  _context->probesize = 100000000;

  CHECK( avformat_find_stream_info(_context, nullptr) );

  // fill the array with all available video streams
  bool unsuported_codec = false;

  // find all streams that the library is able to decode
  for (unsigned i = 0; i < _context->nb_streams; ++i) {
    AVStream* avstream = _context->streams[i];
    AVCodecContext* codecCtx = avcodec_alloc_context3(nullptr);
    avcodec_parameters_to_context(codecCtx, avstream->codecpar);

    // be sure to have a valid stream
    if (!avstream || !codecCtx) {
      continue;
    }

    // considering only video streams, skipping audio
    if (codecCtx->codec_type != AVMEDIA_TYPE_VIDEO) {
      continue;
    }

    // find the codec
    AVCodec* videoCodec = avcodec_find_decoder(codecCtx->codec_id);
    if (videoCodec == nullptr) {
      continue;
    }

    // skip codecs not in the white list
    std::string reason;
    if (!Foundry::Nuke::isCodecWhitelistedForReading(videoCodec->name,reason)) {
      if (!reason.empty()) {
        setInternalError(-1, reason.c_str() );
        return;
      }
      unsuported_codec = true;
      continue;
    }

    // Some codecs support multi-threaded decoding (eg mpeg). Its fast but causes problems when opening many readers
    // simultaneously since each opens as many threads as you have cores. This leads to resource starvation and failed reads.
    // For now, revert to the previous ffmpeg behaviour (single-threaded decode) unless overridden by env var.
    char* envVarVal = std::getenv("NUKE_MULTI_THREADED_CODEC_SUPPORT");
    std::string multiThreadedString = envVarVal ? envVarVal: "";
    if (multiThreadedString.empty() || multiThreadedString != "1") {
     // Default behaviour
      codecCtx->thread_count = 1;
      codecCtx->thread_type = FF_THREAD_FRAME;
    }

    // skip if the codec can't be open
    if (avcodec_open2(codecCtx, videoCodec, nullptr) < 0) {
      continue;
    }

    avcodec_parameters_from_context(avstream->codecpar, codecCtx);
    
    if (!_streams.empty()) {
      // Assume that if this stream's properties don't match those of the zeroth stream then it doesn't
      // correspond to an alternative view.
      // This may turn out to be either too stringent or pointlessly lax, we'll have to see what users
      // make of it. The way to handle this properly is to provide a knob allowing the user to map views
      // to streams.
      if (!CheckStreamPropertiesMatch(_streams[0]->_avstream, avstream)) {
        continue;
      }
    }

    Stream* stream = new Stream();
    stream->_idx = i;
    stream->_avstream = avstream;
    stream->_codecContext = codecCtx;
    stream->_videoCodec = videoCodec;
    stream->_avFrame = av_frame_alloc();
    stream->_avIntermediateFrame = av_frame_alloc();

#ifdef FN_LICENSED_PRORES_DNXHD_CODEC
    if (AV_CODEC_ID_PRORES ==stream->_codecContext->codec_id) {
      // Always assume there is alpha.
      // This is done so that if alpha is present in the encoded
      // bitstream, then it will not be discarded. If there is
      // no alpha in the encoded bitstream, then the decoder
      // returns opaque values for alpha.
      stream->_bitDepth = 16;
      stream->_numberOfComponents = 4;

      // Use the RGBA component re-order swizzle. This avoids having
      // to use another buffer to convert from BGRA to RGBA.
      stream->_componentPosition[0] = 2; // Red position.
      stream->_componentPosition[1] = 1; // Green position.
      stream->_componentPosition[2] = 0; // Blue position.
      stream->_componentPosition[3] = 3; // Alpha position.
    }
    else if (AV_CODEC_ID_DNXHD == stream->_codecContext->codec_id) {
      // Avid DNxHD specific.
      stream->_bitDepth = stream->_codecContext->bits_per_raw_sample;
      stream->_numberOfComponents = 4; // Always assume there is alpha.

      // Use the RGBA component re-order swizzle. This avoids having
      // to use another buffer to convert from BGRA to RGBA.
      stream->_componentPosition[0] = 2; // Red position.
      stream->_componentPosition[1] = 1; // Green position.
      stream->_componentPosition[2] = 0; // Blue position.
      stream->_componentPosition[3] = 3; // Alpha position.

      // In the QuickTime movie (in the sample description atom
      // 'stsd') there is an 'ACLR' atom that contains a parameter
      // which indicates if the data range of the component data
      // is full range (0-255) or legal (16-240). The FFmpeg
      // library does provide that value in avCodecContext->color_range;
      // 'ACLR' defines a range field as 1=video range, 2=full range.
      // That is mapped here to AVCOL_RANGE_JPEG=full range, AVCOL_RANGE_MPEG=legal range.
      if (kVideoRange == stream->_codecContext->color_range) {
        stream->_defaultDecodeVideoLevels = eDecodeVideoLevelsVideoRange;
      }
      else {
        stream->_defaultDecodeVideoLevels = eDecodeVideoLevelsFullRange;
      }

      AVDictionaryEntry* t = av_dict_get(stream->_avstream->metadata, "fn_aalp_alpha_coding_type", nullptr, AV_DICT_IGNORE_SUFFIX);
      if (t) {
        // 'AALP' atom defines the alpha channel coding type:
        // 'AVdn' for DNx-encoded (uses same compression as encoded video)
        // 'AUA8' for 8 bit raw alpha
        // 'AUAX' for 10 bit packet raw alpha
        //
        const auto aalpAlphaCodingType = strtoul(t->value, nullptr, 10);
        if (aalpAlphaCodingType == 'AVdn') {
          stream->_alphaChannelType = NukeCodecs::DecoderParams::kDNxHDEncodedAlpha;
        }
        else if (aalpAlphaCodingType == 'AUA8') {
          stream->_alphaChannelType = NukeCodecs::DecoderParams::k8BitRawAlpha;
        }
        else if (aalpAlphaCodingType == 'AUAX') {
          stream->_alphaChannelType = NukeCodecs::DecoderParams::k10BitRawAlpha;
        }
        else {
          stream->_alphaChannelType = NukeCodecs::DecoderParams::kNoAlpha;
        }
      }
    } else
#endif
    {
      // In |engine| the output bit depth was hard coded to 16-bits.
      // Now it will use the bit depth reported by the decoder so
      // that if a decoder outputs 10-bits then |engine| will convert
      // this correctly. This means that the following change is
      // required for FFmpeg decoders. Currently |_bitDepth| is used
      // internally so this change has no side effects.
      //stream->_bitDepth = avstream->codec->bits_per_raw_sample;
      stream->_bitDepth = 16;

      const AVPixFmtDescriptor* avPixFmtDescriptor = av_pix_fmt_desc_get(stream->_codecContext->pix_fmt);
      // Sanity check the number of components.
      // Only 3 or 4 components are supported by |engine|, that is
      // Nuke/NukeStudio will only accept 3 or 4 component data.
      // For a monochrome image (single channel) promote to 3
      // channels. This is in keeping with all the assumptions
      // throughout the code that if it is not 4 channels data
      // then it must be three channel data. This ensures that
      // all the buffer size calculations are correct.
      stream->_numberOfComponents = avPixFmtDescriptor->nb_components;
      if (3 > stream->_numberOfComponents)
        stream->_numberOfComponents = 3;
      // AVCodecContext::bits_pre_raw_sample may not be set, if
      // it's not set, try with the following utility function.
      if (0 == stream->_bitDepth)
        stream->_bitDepth = av_get_bits_per_pixel(avPixFmtDescriptor) / stream->_numberOfComponents;
    }

    if (stream->_bitDepth > 8) {
      // TODO: This will sometimes be sub-optimal for the timeline, which can deal directly with 3 x 10 bit pixel formats.
      stream->_outputPixelFormat = (4 == stream->_numberOfComponents) ? AV_PIX_FMT_RGBA64LE : AV_PIX_FMT_RGB48LE; // 16-bit.
    }
    else {
      stream->_outputPixelFormat = (4 == stream->_numberOfComponents) ? AV_PIX_FMT_RGBA : AV_PIX_FMT_RGB24; // 8-bit
    }

    // If FPS is specified, record it.
    // Otherwise assume 1 fps (default value).
    if ( avstream->r_frame_rate.num != 0 &&  avstream->r_frame_rate.den != 0 ) {
      stream->_fpsNum = avstream->r_frame_rate.num;
      stream->_fpsDen = avstream->r_frame_rate.den;
    }

    stream->_width  = stream->_codecContext->width;
    stream->_height = stream->_codecContext->height;

#ifdef FN_LICENSED_PRORES_DNXHD_CODEC
    // Adjust the output Width for Thin-Raster DNxHD encodes
    if (stream->isCodecTypeDNxHD()) {
      if (stream->_codecContext->width == 1440) {
        stream->_width = 1920;
      }
      else if (stream->_codecContext->width == 960) {
        stream->_width = 1280;
      }
    }
#endif

    // set aspect ratio
    stream->_aspect = Stream::GetStreamAspectRatio(stream);
    
    // set stream start time and numbers of frames
    stream->_startPTS = getStreamStartTime(*stream);
    stream->_frames   = getStreamFrames(*stream);

    // save the stream
    _streams.push_back(stream);
  }

  if (_streams.empty()) {
    setError( unsuported_codec ? "unsupported codec..." : "unable to find video stream" );
    _selectedStream = nullptr;
  }
  else {
    mFnAssertMsg(viewIndex >= 0, "Negative view index specified.");
    
    if (static_cast<size_t>(viewIndex) < _streams.size()) {
      _selectedStream = _streams[viewIndex];
    }
    else {
      _selectedStream = _streams[0];
    }
  }
}

// destructor
FFmpegFile::~FFmpegFile()
{
  std::for_each(_streams.begin(), _streams.end(), Stream::destroy);

  if (_context)
    avformat_close_input(&_context);
}

void FFmpegFile::setSelectedStream(int streamIndex)
{
  if ((streamIndex >= 0) && (streamIndex < static_cast<int>(_streams.size()))) {
    _selectedStream = _streams[streamIndex];
  }
  else {
    mFnAssertMsg(false,"setSelectedStream: Invalid streamIndex");
    _selectedStream = !_streams.empty() ? _streams[0] : nullptr;
  }
}

std::string FFmpegFile::getTimecodeForFrame(int frameIndex) const
{
#ifdef FN_EXAMPLE_PLUGIN
  // FIXME: replace with shippable time code handling
  return std::string();
#else
  if (_streams.empty()) {
    return std::string();
  }

  mFnAssertMsg(_selectedStream, "getTimecodeForFrame: null _selectedStream");

  Foundry::Base::TimeBase timeBase(_selectedStream->_fpsNum, _selectedStream->_fpsDen);
  Foundry::Base::Timecode::DisplayType displayType = Foundry::Base::Timecode::displayTypeFromString(_selectedStream->_timecode);
  Foundry::Base::Time time = Foundry::Base::Timecode::stringToTime(_selectedStream->_timecode, timeBase, displayType);
  time += frameIndex;
  return Foundry::Base::Timecode::timeToString(time, timeBase, displayType);
#endif
}

const char* FFmpegFile::getColorspace() const {
  //The preferred colorspace is figured out from a number of sources - initially we look for a number
  //of different metadata sources that may be present in the file. If these fail we then fall back
  //to using the codec's underlying storage mechanism - if RGB we default to gamma 1.8, if YCbCr we
  //default to gamma 2.2 (note prores special case). Note we also ignore the NCLC atom for reading
  //purposes, as in practise it tends to be incorrect.
  
  //First look for the meta keys that (recent) Nukes would've written, or special cases in Arri meta.
  //Doubles up searching for lower case keys as the ffmpeg searches are case sensitive, and the keys
  //have been seen to be lower cased (particularly in old Arri movs).
  if(_context && _context->metadata) {
    AVDictionaryEntry* t;
    
    t = av_dict_get(_context->metadata, "uk.co.thefoundry.Colorspace", nullptr, AV_DICT_IGNORE_SUFFIX);
    if (!t)
        av_dict_get(_context->metadata, "uk.co.thefoundry.colorspace", nullptr, AV_DICT_IGNORE_SUFFIX);
    if (t) {
      //Validate t->value against root list, to make sure it's been written with a LUT
      //we have a matching conversion for.
      bool found = false;
      int  i     = 0;
      while (!found && LUT::builtin_names[i] != nullptr) {
        found = !strcasecmp(t->value, LUT::builtin_names[i++]);
      }
      if (found) {
        return t->value;
      }
    }
    
    t = av_dict_get(_context->metadata, "com.arri.camera.ColorGammaSxS", nullptr, AV_DICT_IGNORE_SUFFIX);
    if (!t)
        av_dict_get(_context->metadata, "com.arri.camera.colorgammasxs", nullptr, AV_DICT_IGNORE_SUFFIX);
    if (t && !strncasecmp(t->value, "LOG-C", 5)) {
      return "AlexaV3LogC";
    }
    if (t && !strncasecmp(t->value, "REC-709", 7)) {
      return "rec709";
    }
  }

  // get the colorspace from ffmpeg transfer characteristics
  mFnAssertMsg(_streams.empty() || _selectedStream, "_streams not empty but null _selectedStream");
  if (!_streams.empty() && _selectedStream->_codecContext) {
    if (_selectedStream->_codecContext->color_trc == AVCOL_TRC_BT709 ||
        _selectedStream->_codecContext->color_trc == AVCOL_TRC_SMPTE240M) {
      return "rec709";
    }
    else if (_selectedStream->_codecContext->color_trc == AVCOL_TRC_GAMMA22) {
      return "Gamma2.2";
    }
    else if (_selectedStream->_codecContext->color_trc == AVCOL_TRC_SMPTE170M) {
      return "rec601";
    }
    else if (_selectedStream->_codecContext->color_trc == AVCOL_TRC_SMPTE2084) {
      return "st2084";
    }
    else if (_selectedStream->_codecContext->color_trc == AVCOL_TRC_IEC61966_2_1) {
      return "sRGB";
    }
    else if (_selectedStream->_codecContext->color_trc == AVCOL_TRC_LINEAR) {
      return "linear";
    }
  }
  return (isYUV() ? "Gamma2.2" : "Gamma1.8");
}

#ifdef FN_LICENSED_PRORES_DNXHD_CODEC
NukeCodecs::CodecFactory::CodecType FFmpegFile::Stream::getDNxCodecType() const
{
  if (isCodecTypeDNxHR()) {
    return getDNxHRCompressionId();
  }

  // The four character code 'AVdn' aliases all DNxHD
  // compression IDs. The 'ARES' atom contains the actual
  // compression ID. The 'ACLR' atom contains a parameter
  // that indicates if the data range of the component
  // data is full range (0-255) or legal (16-240). The
  // FFmpeg library was modified to provide the 'ARES' and
  // 'ACLR' atoms in metadata for DNxHD, so attempt to
  // locate and use that information here.
  AVDictionaryEntry* aresCompressionID = av_dict_get(_avstream->metadata, kARESCompressionID, nullptr, AV_DICT_IGNORE_SUFFIX);
  if (!aresCompressionID) {
    return NukeCodecs::CodecFactory::eUnknownCodec;
  }
  return static_cast<NukeCodecs::CodecFactory::CodecType>(strtoul(aresCompressionID->value, nullptr, 10));
}
#endif

// decode a single frame into the buffer. Thread safe
bool FFmpegFile::decode(unsigned char* buffer, unsigned frame)
{
  Guard guard(_lock);

  if (_streams.empty()) {
    return false;
  }

  // Check if the requested frame index is in range
  if ((static_cast<int64_t>(frame) >= _selectedStream->_frames) || 
      (static_cast<int64_t>(frame) < _selectedStream->ptsToFrame(_selectedStream->_avstream->start_time))) {
    return false;
  }

  bool hasPicture = false;
  AVFrame* avFrameOut = _selectedStream->_avFrame;

  // Old behaviour of expecting the next frame out of the decoder to be current frame + 1 is wrong.
  // For some codecs i.e H.264 and HEVC the decoder can output frames in decode order.
  // This is here to prevent any performace regression where we assume the decoder will always
  // output the frames in presentation order (true for the SW H.264 and MPEG4 decoders in libavcodec).
  // This may come back to haunt us one day.

  // Only seek and reset for non-sequential frames as this can be very costly.
  if (_selectedStream->ptsToFrame(avFrameOut->pts) + 1 != frame) {
    seekToFrame(_selectedStream->frameToPts(frame), AVSEEK_FLAG_BACKWARD);
  }

  // Setup the output frame struct with the buffer passed in
  avFrameOut->width   = _selectedStream->_width;
  avFrameOut->height  = _selectedStream->_height;
  avFrameOut->format  = _selectedStream->_outputPixelFormat;

  int res = 0;

  if ((res = av_image_fill_linesizes(avFrameOut->linesize, _selectedStream->_outputPixelFormat, _selectedStream->_width)) < 0) {
    setInternalError(res, Mov64ReaderErrPrefix "Failed to fill image linesizes: ");
    return false;
  }

  res = av_image_fill_pointers(
    avFrameOut->data,
    _selectedStream->_outputPixelFormat,
    _selectedStream->_height,
    buffer,
    avFrameOut->linesize
  );

  if (res < 0) {
    setInternalError(res, Mov64ReaderErrPrefix "Failed to fill image pointers: ");
    return false;
  }

  hasPicture = demuxAndDecode(avFrameOut, frame);

  if (!hasPicture) {
    // If we've failed to get the picture try seeking to the frame again. For an 'inter-frame' codec
    // we should seek behind the frame by the 'group of pictures' size to ensure the frame gets decoded.
    // Gop size will be 0 for 'intra-only' codecs. More info on this here: https://en.wikipedia.org/wiki/Group_of_pictures
    // TODO: Consider evaluating the ffmpeg error code and not doing this if it's an end of file error.
    const auto gopSize = _selectedStream->_codecContext->gop_size;
    const auto targetFrame = std::max(static_cast<int>(frame) - gopSize, 0);
    seekToFrame(_selectedStream->frameToPts(targetFrame), AVSEEK_FLAG_FRAME | AVSEEK_FLAG_BACKWARD);
    hasPicture = demuxAndDecode(avFrameOut, frame);
  }

  return hasPicture;
}


bool FFmpegFile::seekToFrame(int64_t frame, int seekFlags) 
{
  avcodec_flush_buffers(_selectedStream->_codecContext);
  int res = 0;
  
  if ((res = av_seek_frame(_context, _selectedStream->_idx, frame, seekFlags)) < 0) {
    setInternalError(res, Mov64ReaderErrPrefix "Failed to seek frame: ");
    return false;
  }

  return true;
}


bool FFmpegFile::demuxAndDecode(AVFrame* avFrameOut, int64_t frame)
{
  bool isCodecFFmpeg = true;
  if ((_selectedStream->_codecContext->codec_id == AV_CODEC_ID_PRORES) ||
      (_selectedStream->_codecContext->codec_id == AV_CODEC_ID_DNXHD)) {
      isCodecFFmpeg = false;
  }
  
  av_frame_unref(_selectedStream->_avIntermediateFrame);

  int res                   = 0;
  bool hasPicture           = false;
  int frameDecoded          = 0;
  AVFrame* avFrameDecodeDst = isCodecFFmpeg ? _selectedStream->_avIntermediateFrame: avFrameOut;

  Stream* stream = _selectedStream;
  AVFormatContext* formatCtx = _context;

  enum FoundFrameResult
  {
    eFrameFound,
    eFrameBehind,
    eFrameAhead
  };

  auto foundCorrectFrame = [stream, formatCtx](AVFrame* decodedFrame, int64_t targetFrameIdx) -> FoundFrameResult {
    bool found = false;

    // Fallback to using the DTS from the AVPacket that triggered returning this frame if no PTS is found
    if (decodedFrame->pts == AV_NOPTS_VALUE) {
      decodedFrame->pts = decodedFrame->pkt_dts;
    }

    int64_t decodedFrameIdx = stream->ptsToFrame(decodedFrame->pts);

    if (decodedFrameIdx == targetFrameIdx) {
      found = true;
    }
    // If the current frame needs to be displayed longer than current_pts - prev_pts
    else if ((decodedFrameIdx < targetFrameIdx) && (stream->ptsToFrame(decodedFrame->pts + decodedFrame->pkt_duration) > targetFrameIdx)) {
      found = true;
    }

    if (found) {
      return eFrameFound;
    }

    return decodedFrameIdx < targetFrameIdx ? eFrameBehind : eFrameAhead;
  };

  auto processFrameInternal = [stream](const FoundryAVPacket& pkt, AVFrame* decodedFrame, int& packetNew, int& frameDecoded) -> int {
    int ret = 0;
    int gotFrame = 0;

    if (packetNew) {
      ret = avcodec_send_packet(stream->_codecContext, &pkt);
      if (ret == AVERROR(EAGAIN)) {
        ret = 0;
      }
      else if (ret >= 0 || ret == AVERROR_EOF) {
        ret = 0;
        packetNew = 0;
      }
    }
    if (ret >= 0) {
      AVFrame* frameOut = av_frame_alloc();
      ret = avcodec_receive_frame(stream->_codecContext, frameOut);
      if (ret >= 0) {
        gotFrame = 1;
        frameDecoded = 1;
        av_frame_unref(decodedFrame);
        av_frame_move_ref(decodedFrame, frameOut);
      }
      else {
        av_frame_free(&frameOut);
        if (ret == AVERROR(EAGAIN) || ret == AVERROR_EOF) {
          ret = 0;
        }
      }
    }

    if (ret < 0) {
      return ret;
    }
    return gotFrame || packetNew;
  };

  // Begin reading from the newly seeked position
  FoundryAVPacket avPacket;
  while (!(res = av_read_frame(_context, &avPacket))) {

    if (avPacket.stream_index == _selectedStream->_idx) {

      if (isCodecFFmpeg) {
        int packetNew = 1;
        while (processFrameInternal(avPacket, avFrameDecodeDst, packetNew, frameDecoded) > 0) {
          if (frameDecoded) {
            auto result = foundCorrectFrame(avFrameDecodeDst, frame);
            if (result == FoundFrameResult::eFrameFound) {
              hasPicture = isCodecFFmpeg ? imageConvert(avFrameDecodeDst, avFrameOut) : true;
              return hasPicture;
            }
            else if (result == FoundFrameResult::eFrameAhead) {
              // Return early if we've gone past the frame
              return false;
            }
          }
        }
      }
      else {
        if (decodeLicensed(avFrameDecodeDst, frameDecoded, avPacket) > 0) {
          if (frameDecoded) {
            auto result = foundCorrectFrame(avFrameDecodeDst, frame);
            if (result == FoundFrameResult::eFrameFound) {
              hasPicture = isCodecFFmpeg ? imageConvert(avFrameDecodeDst, avFrameOut) : true;
              return hasPicture;
            }
            else if (result == FoundFrameResult::eFrameAhead) {
              // Return early if we've gone past the frame
              return false;
            }
          }
        }
        else {
          return false;
        }
      }
    }
  }


  // Check for any read failures
  if (res < 0 && res != AVERROR_EOF) {
    setInternalError(res, Mov64ReaderErrPrefix "Failed to read frame: ");
    return false;
  }


  // Flush the decoder of remaining frames
  if (res == AVERROR_EOF) {
    if (isCodecFFmpeg) {
      int packetNew = 1;
      while (processFrameInternal(avPacket, avFrameDecodeDst, packetNew, frameDecoded) > 0) {
        if (frameDecoded) {
          if (foundCorrectFrame(avFrameDecodeDst, frame) == FoundFrameResult::eFrameFound) {
            return imageConvert(avFrameDecodeDst, avFrameOut);
          }
        }
      }
    }
  }

  return hasPicture;
}


int FFmpegFile::decodeProRes(AVFrame* avFrameOut, int& frameDecoded, AVPacket& avPacketIn)
{
#ifdef FN_LICENSED_PRORES_DNXHD_CODEC
  frameDecoded = 0;

  if (!avFrameOut) {
    setError("Invalid output frame", Mov64ReaderErrPrefix);
    return 0;
  }

  if (_selectedStream->_codecContext->codec_id != AV_CODEC_ID_PRORES) {
    setError("Invalid Codec ID, Failed to create ProRes decoder", Mov64ReaderErrPrefix);
    return 0;
  }

  // Create an instance of a decoder if one does not already exist.
  const auto codecType = static_cast<NukeCodecs::CodecFactory::CodecType>(SWAP32(_selectedStream->_codecContext->codec_tag));

  if (!_decoder.get()) {
    _decoder.reset(NukeCodecs::CodecFactory::makeDecoder(codecType));
    if (!_decoder.get()) {
      setError("Failed to create ProRes decoder", Mov64ReaderErrPrefix);
      return 0;
    }
  }

  NukeCodecs::DecoderParams params;
  params.in     = avPacketIn.data;
  params.inSize = avPacketIn.size;

  _codecInfo._codecType   = codecType;
  _decoder->GetDecoderInfo(_codecInfo._codecType, _selectedStream->_codecContext->width, _selectedStream->_codecContext->height, _codecInfo, &params);

  // ProRes is an intra only codec so every frame is a key-frame
  avFrameOut->key_frame    = 1;
  avFrameOut->pts          = avPacketIn.pts == AV_NOPTS_VALUE ? avPacketIn.dts: avPacketIn.pts;
  avFrameOut->pkt_duration = avPacketIn.duration;
  avFrameOut->format       = _codecInfo._bitsPerComponent > 8 ? AV_PIX_FMT_BGRA64LE : AV_PIX_FMT_BGRA;
  avFrameOut->color_range  = _codecInfo._bitsPerComponent > 8 ? AVCOL_RANGE_JPEG : AVCOL_RANGE_MPEG;

  _selectedStream->_bitDepth = _codecInfo._bitsPerComponent;

  params.width        = _codecInfo._width;
  params.height       = _codecInfo._height;
  params.out          = avFrameOut->data[0];
  params.outRowBytes  = _codecInfo._rowBytes;
  params.codecType    = _codecInfo._codecType;
  params.alphaType    = _selectedStream->_alphaChannelType;

  int error = _decoder->Decode(params);

  frameDecoded = (error <= 0) ? 0 : 1;

  if (!frameDecoded) {
    switch (error)
    {
      case -1: setError("Failed to decode frame, ProRes decoder creation failed", Mov64ReaderErrPrefix);              break;
      case -2: setError("Failed to decode frame, ProRes colour space conversion failed", Mov64ReaderErrPrefix);       break;
      case -3: setError("Failed to decode frame, ProRes decoder does not support this codec", Mov64ReaderErrPrefix);  break;
      case -4: setError("Failed to decode frame, the output buffer is not 16-bit aligned", Mov64ReaderErrPrefix);       break;
      case -5: setError("Failed to decode frame, ProRes decoder setup is invalid", Mov64ReaderErrPrefix);             break;
      default: setError("Failed to decode frame, unknown error", Mov64ReaderErrPrefix);                               break;
    }
  }

  return error;
#else
  frameDecoded = 0;
  return 0;
#endif
}

int FFmpegFile::decodeDNx(AVFrame* avFrameOut, int& frameDecoded, AVPacket& avPacketIn)
{
#ifdef FN_LICENSED_PRORES_DNXHD_CODEC
  frameDecoded = 0;

  if (!avFrameOut) {
    setError("Invalid output frame", Mov64ReaderErrPrefix);
    return 0;
  }

  AVPixelFormat srcPixelFormat;
  
  if (!_selectedStream->isCodecTypeDNx()) {
    setError("Invalid Codec ID: Failed to create DNx decoder", Mov64ReaderErrPrefix);
    return 0;
  }

  const auto codecType = _selectedStream->getDNxCodecType();

  // The 'ACLR' atom defines a range field as 1=video range,
  // 2=full range. This is stored in Stream::_decodeVideoLevels
  // member variable which can be modified by the UI.
  AVColorRange srcColourRange = AVCOL_RANGE_JPEG;
  if (eDecodeVideoLevelsVideoRange == _selectedStream->_decodeVideoLevels) {
    srcColourRange = AVCOL_RANGE_MPEG;
  }

  // Create an instance of a decoder if one does not already exist.
  if (!_decoder.get())
    _decoder.reset(NukeCodecs::CodecFactory::makeDecoder(codecType));

  if (_decoder.get()) {
    _codecInfo._codecType = codecType;

    if ((NukeCodecs::CodecFactory::eDNxHDCodec1235 == codecType) ||
        (NukeCodecs::CodecFactory::eDNxHDCodec1241 == codecType) ||
        (NukeCodecs::CodecFactory::eDNxHDCodec1250 == codecType) ||
        (NukeCodecs::CodecFactory::eDNxHDCodec1256 == codecType)) {
      srcPixelFormat = AV_PIX_FMT_BGRA64LE; // 10-bit DNxHD.
    }
    else {
      srcPixelFormat = AV_PIX_FMT_RGBA; // 8-bit DNxHD.
    }
  }
  else {
    setError("Failed to create DNx decoder", Mov64ReaderErrPrefix);
    return 0;
  }

  NukeCodecs::DecoderParams params;
  params.in     = avPacketIn.data;
  params.inSize = avPacketIn.size;

  _decoder->GetDecoderInfo(_codecInfo._codecType, _selectedStream->_codecContext->width, _selectedStream->_codecContext->height, _codecInfo, &params);

  bool isThinRaster = false;

  if ((NukeCodecs::CodecFactory::eDNxHDCodec1244 == _codecInfo._codecType) ||
      (NukeCodecs::CodecFactory::eDNxHDCodec1258 == _codecInfo._codecType) ||
      (NukeCodecs::CodecFactory::eDNxHDCodec1259 == _codecInfo._codecType) ||
      (NukeCodecs::CodecFactory::eDNxHDCodec1260 == _codecInfo._codecType)) {
    isThinRaster = true;
  }

  params.width  = _codecInfo._width;
  params.height = _codecInfo._height;

  // The following DNxHD codec profiles are 'thin raster',
  // either 1440x1080 or 960x720. These require a scaling
  // operation and so will use an intermediate buffer.
  AVFrame* intermediateFrame = nullptr;
  if (isThinRaster) {
    intermediateFrame = _selectedStream->_avIntermediateFrame;

    intermediateFrame->width  = _codecInfo._width;
    intermediateFrame->height = _codecInfo._height;
    intermediateFrame->format = srcPixelFormat;
    
    int res = av_image_alloc(
      intermediateFrame->data, 
      intermediateFrame->linesize,
      intermediateFrame->width,
      intermediateFrame->height,
      srcPixelFormat, 32
    );

    if (res < 0) {
      setInternalError(res, Mov64ReaderErrPrefix "Failed to allocate intermediate frame: ");
      av_frame_unref(intermediateFrame);
      return res;
    }

    params.out = intermediateFrame->data[0];
  }
  else {
    params.out = avFrameOut->data[0];
  }

  params.outRowBytes  = _codecInfo._rowBytes;
  params.codecType    = _codecInfo._codecType;
  params.alphaType    = _selectedStream->_alphaChannelType;
  params.srcDataRange = (srcColourRange == AVCOL_RANGE_JPEG) ? NukeCodecs::DecoderParams::kFullRange : NukeCodecs::DecoderParams::kVideoRange;

  int error = _decoder->Decode(params);

  // DNxHD and DNxHR are intra only codecs so every frame is a key-frame
  avFrameOut->key_frame    = 1;
  avFrameOut->pts          = avPacketIn.pts == AV_NOPTS_VALUE ? avPacketIn.dts: avPacketIn.pts;
  avFrameOut->pkt_duration = avPacketIn.duration;
  avFrameOut->color_range  = srcColourRange;
  avFrameOut->format       = srcPixelFormat;

  if(error > 0 && isThinRaster) {
    imageConvert(intermediateFrame, avFrameOut);
  }

  if (intermediateFrame) {
    av_frame_unref(intermediateFrame);
  }

  frameDecoded = (error <= 0) ? 0 : 1; // Mimic the behaviour of |avcodec_decode_video2| on |frameDecoded|.

  if (!frameDecoded) {
    switch (error)
    {
      case NukeCodecs::CodecFactory::kError_FailedToCreateDecoder:
        setError("Failed to decode frame, failed to create DNx decoder", Mov64ReaderErrPrefix);             break;
      case NukeCodecs::CodecFactory::kError_FailedToConfigureDecoder:
        setError("Failed to decode frame, failed to configure DNx decoder", Mov64ReaderErrPrefix);          break;
      case NukeCodecs::CodecFactory::kError_FailedToApplyAlpha:
        setError("Failed to decode frame, DNx decodeer failed to apply alpha", Mov64ReaderErrPrefix);       break;
      case NukeCodecs::CodecFactory::kError_CodecNotSupported:
        setError("Failed to decode frame, DNx decoder does not support this codec", Mov64ReaderErrPrefix);  break;
      case NukeCodecs::CodecFactory::kError_IncorrectMemAlignment:
        setError("Failed to decode frame, the output buffer is not byte aligned", Mov64ReaderErrPrefix);    break;
      case NukeCodecs::CodecFactory::kError_ColorspaceConversion:
        setError("Failed to decode frame, DNx colour space conversion failed", Mov64ReaderErrPrefix);       break;
      default:
        setError("Failed to decode frame, unknown error", Mov64ReaderErrPrefix); break;
    }
  }

  return error;
#else
  frameDecoded = 0;
  return 0;
#endif
}

// @return >0 If successful, this is the size in bytes of the decoded frame.
// @return <=0 For any failure to decode.
int FFmpegFile::decodeLicensed(AVFrame* avFrameOut, int& frameDecoded, AVPacket& avPacketIn)
{
  frameDecoded = 0;

  if (!avFrameOut) {
    setError("Invalid output frame", Mov64ReaderErrPrefix);
    return 0;
  }

  switch (_selectedStream->_codecContext->codec_id)
  {
#ifdef FN_LICENSED_PRORES_DNXHD_CODEC
    case AV_CODEC_ID_PRORES:
      return decodeProRes(avFrameOut, frameDecoded, avPacketIn);

    case AV_CODEC_ID_DNXHD:
      return decodeDNx(avFrameOut, frameDecoded, avPacketIn);
#endif
    default:
      setError("Could not find a vaild decoder for this stream", Mov64ReaderErrPrefix);
      break;
  }

  return 0;
}


bool FFmpegFile::imageConvert(AVFrame* avFrameIn, AVFrame* avFrameOut)
{
  if (!avFrameIn || !avFrameOut) {
    setError("mov64Reader no input or output frame provided for conversion");
    return false;
  }

  AVPixelFormat dstPixFmt = (AVPixelFormat)avFrameOut->format;
  AVPixelFormat srcPixFmt = (AVPixelFormat)avFrameIn->format;

  avFrameOut->pts          = avFrameIn->pts;
  avFrameOut->pkt_dts      = avFrameIn->pkt_dts;
  avFrameOut->pkt_duration = avFrameIn->pkt_duration;

  if (!avFrameOut->data[0]) {
    int res = av_image_alloc(avFrameOut->data, avFrameOut->linesize, avFrameOut->width, avFrameOut->height, dstPixFmt, 32);
    if (res <= 0) {
      setInternalError(res, Mov64ReaderErrPrefix "Output frame allocation failed during format conversion");
      return false;
    }
  }

  NukeCodecs::IDecoder * color_convertor = _selectedStream->getColorConvertor(
    srcPixFmt, 
    avFrameIn->width, 
    avFrameIn->height, 
    avFrameIn->color_range
  );
  SwsContext* context = _selectedStream->getConvertCtx(
    srcPixFmt, 
    avFrameIn->width, 
    avFrameIn->height,
    avFrameIn->color_range,
    dstPixFmt, 
    avFrameOut->width, 
    avFrameOut->height
  );

  // Scale if any of the decoding path has provided a convert
  // context. Otherwise, no scaling/conversion is required after
  // decoding the frame.
  if (color_convertor) {
    NukeCodecs::DecoderParams & p = _selectedStream->getColorConvertorParams();
    p.in_planar   = reinterpret_cast<const void **>(const_cast<const uint8_t **>(avFrameIn->data));
    p.in_linesize = avFrameIn->linesize;
    p.out         = avFrameOut->data[0];
    p.outRowBytes = getRowSize();

    color_convertor->Decode(p);
  }
  else if (context) {
    sws_scale(
      context,
      avFrameIn->data,
      avFrameIn->linesize,
      0,
      avFrameIn->height,
      avFrameOut->data,
      avFrameOut->linesize
    );
  }
  else {
    return false;
  }
  
  return true;
}

// Fill the metada information for the selected stream.
bool FFmpegFile::metadata(MetaData::Bundle& metadata)
{
  Guard guard(_lock);

  if (_streams.empty())
    return false;

  int bitsPerComponent = 8; // 8 bits per channel (Fix for Bug 29986. data_ is unsigned char)
  std::string codecName;

#ifdef FN_LICENSED_PRORES_DNXHD_CODEC
  if (AV_CODEC_ID_PRORES == _selectedStream->_codecContext->codec_id) {
    // |bitsPerComponent| is just for display purposes so
    // report the data from the movie.
    bitsPerComponent = _selectedStream->_codecContext->bits_per_raw_sample;

    // Set the full (and correct) ProRes codec profile name.
    // The |codec_tag| member variable is the four character
    // code (FCC) of the codec profile. Use this to identify
    // the ProRes encoding profile and set the appropriate
    // codec profile name.
    switch (SWAP32(_selectedStream->_codecContext->codec_tag)) {
    case NukeCodecs::CodecFactory::eProResCodec4444XQ:
      codecName = APPLE_PRORES_4444XQ_NAME;
      break;

    case NukeCodecs::CodecFactory::eProResCodec4444:
      codecName = APPLE_PRORES_4444_NAME;
      break;

    case NukeCodecs::CodecFactory::eProResCodec422HQ:
      codecName = APPLE_PRORES_422HQ_NAME;
      break;

    case NukeCodecs::CodecFactory::eProResCodec422:
      codecName = APPLE_PRORES_422_NAME;
      break;

    case NukeCodecs::CodecFactory::eProResCodec422LT:
      codecName = APPLE_PRORES_422LT_NAME;
      break;

    case NukeCodecs::CodecFactory::eProResCodec422Proxy:
      codecName = APPLE_PRORES_422PROXY_NAME;
      break;

    default:
      break;
    }
  }
  else if (_selectedStream->isCodecTypeDNx()) {
      codecName = NukeCodecs::CodecFactory::codecNameFromType(_selectedStream->getDNxCodecType());
      // |bitsPerComponent| is just for display purposes so
      // report the data from the movie. Note that for DNxHD
      // the actual bits per sample is forced to 10 for
      // reasons described above. The following will ensure
      // that the correct bit depth is displayed irrespective
      // of the internal shennigans of this reader.
      bitsPerComponent = _selectedStream->_codecContext->bits_per_raw_sample;
  } else
#endif
  {
    //Add metadata sourced from non-meta sources in ffmpeg.
    metadata.setData(MetaData::FFmpegStream::PIXEL_FORMAT, av_get_pix_fmt_name(_selectedStream->_codecContext->pix_fmt));

    codecName = std::string(Foundry::Nuke::getCodecOverridenLongName(_selectedStream->_videoCodec->name));
    if (codecName.empty()) {
      codecName = _selectedStream->_videoCodec->name;
    }
  }
  if (!codecName.empty())
    metadata.setData(MetaData::QuickTime::CODEC_NAME, codecName);

  if (_selectedStream->_codecContext->codec_tag) {
    char buff[AV_FOURCC_MAX_STRING_SIZE];
    const char* codecID = av_fourcc_make_string(buff, _selectedStream->_codecContext->codec_tag);
    if (0 != strcmp(codecID, "")) {
      metadata.setData(MetaData::QuickTime::CODEC_ID, codecID);
    }
  }
  
  metadata.setData(MetaData::DEPTH, MetaData::DEPTH_FIXED(bitsPerComponent));

  metadata.setData(MetaData::FRAME_RATE, double(_selectedStream->_fpsNum) / double(_selectedStream->_fpsDen));
  
  //trim to a couple of degrees of precision.
  float ar = ((int)(Stream::GetStreamAspectRatio(_selectedStream) * 100 + .5) / 100.0f);
  metadata.setData(MetaData::PIXEL_ASPECT, ar);

  //Add special cased metadata that we can't add as part of the subsequent main loop.

  // The predefined 'MetaData::TIMECODE' tag will be picked up by
  // Nuke/NukeStudio as the timecode for the clip.
  // There are a number of methods to retrieve the timecode using
  // the FFmpeg libraries and are listed here in case the current
  // method does not cover all cases.
  // * Use av_dict_get() on AVFormatContext::metadata (current/preferred method).
  // * Use av_dict_get() on video stream AVStream::metadata
  // * Use av_dict_get() on data stream AVStream::metadata
  // * Read the timecode sample from the data stream using
  //   av_read_frame, check the AVPacket::stream_index matches
  //   the data stream and decode/translate the value to a
  //   time code value.
  //
  // First attempt to extract timecode from the format context. If
  // this does not yield any timecode information, try the video
  // stream.
  //
  // Match the official reader by producing per frame TC. This is
  // achieved by caching the start timecode counter value and then
  // calculating the correct timecode using this value and the
  // index of the currently selected frame. This is performed in
  // |prefetchMetaData|.
  AVDictionaryEntry* avDictionaryEntry = av_dict_get(_context->metadata, "timecode", nullptr, 0);
  if (!avDictionaryEntry)
    avDictionaryEntry = av_dict_get(_selectedStream->_avstream->metadata, "timecode", nullptr, 0);
  if (avDictionaryEntry) {
    _selectedStream->_timecode = avDictionaryEntry->value;
    metadata.setData(MetaData::TIMECODE, avDictionaryEntry->value);
  }
  
  // The following has been removed as |timecode_frame_start| is unreliable.
  // This has been replaced by search for a 'timecode' key in the format
  // metadata, r.e. SetMetaDataItem() with MeteData::TIMECODE above.
  //metadata.setData("mov64/codec/timecodeFrameStart", (unsigned int)_selectedStream->_codecContext->timecode_frame_start);
  // Similarly startTime has been removed as mismatches with official quicktime.
  //metadata.setData("mov64/codec/startTime", (unsigned int)_context->start_time);
  //In practise official quicktime may add a com.apple.quicktime.starttime meta key.
  //We'd need to convert the TC source above with the fps to give us the equivalent
  //as both timecode_frame_start and start_time are unreliable in testing.
  
  AVDictionaryEntry* t;
  t = av_dict_get(_selectedStream->_avstream->metadata, kNCLCPrimariesKey, nullptr, AV_DICT_IGNORE_SUFFIX);
  if (t) {
    metadata.setData(MetaData::QuickTime::NCLC_PRIMARIES,GetNCLCPrimariesLabelFromIndex(atoi(t->value)));
    t = nullptr;
  }
  t = av_dict_get(_selectedStream->_avstream->metadata, kNCLCTransferKey, nullptr, AV_DICT_IGNORE_SUFFIX);
  if (t) {
    metadata.setData(MetaData::QuickTime::NCLC_TRANSFER,GetNCLCTransferLabelFromIndex(atoi(t->value)));
    t = nullptr;
  }
  t = av_dict_get(_selectedStream->_avstream->metadata, kNCLCMatrixKey, nullptr, AV_DICT_IGNORE_SUFFIX);
  if (t) {
    metadata.setData(MetaData::QuickTime::NCLC_MATRIX,GetNCLCMatrixLabelFromIndex(atoi(t->value)));
    t = nullptr;
  }

#ifdef FN_LICENSED_PRORES_DNXHD_CODEC
  if (_selectedStream->isCodecTypeDNxHR()) {
    const auto codecType = _selectedStream->getDNxHRCompressionId();
    if (codecType != NukeCodecs::CodecFactory::eUnknownCodec) {
      std::stringstream codecTypeStream;
      codecTypeStream << codecType;
      ValidateAndSetMetaDataItem(
          metadata,
          kARESCompressionID,
          codecTypeStream.str(),
          _selectedStream->_doNotAttachPrefix,
          _selectedStream->_matchMetaFormat
      );
    }
  }
#endif

  //Add optional top level meta when found that'll also be duplicated to a general place in the subsequent loop.
  t = av_dict_get(_context->metadata, "com.arri.camera.ReelName", nullptr, AV_DICT_IGNORE_SUFFIX);
  if (t) {
    metadata.setData(MetaData::QuickTime::REEL, t->value);
    t = nullptr;
  }

  //Add in all arbitrary metadata from the stream context, renaming/formatting special cases.
  while ((t = av_dict_get(_selectedStream->_avstream->metadata, "", t, AV_DICT_IGNORE_SUFFIX)) != nullptr) {
    ValidateAndSetMetaDataItem(metadata, t->key, t->value,
                               _selectedStream->_doNotAttachPrefix, _selectedStream->_matchMetaFormat);
  }
  
  //Add in all arbitrary metadata from the format context, renaming/formatting special cases.
  while ((t = av_dict_get(_context->metadata, "", t, AV_DICT_IGNORE_SUFFIX)) != nullptr) {
    ValidateAndSetMetaDataItem(metadata, t->key, t->value,
                               _selectedStream->_doNotAttachPrefix, _selectedStream->_matchMetaFormat);
  }
  
  // If we've found more than one 'matching' video stream then assume they might be for multiple
  // views and fill in the VIEW_NAMES entry. We've only got the track IDs to use as distinguishing
  // names and since we don't even know for sure whether the tracks are for multiple views we
  // enclose them in [] so that client code can detect this uncertainty (e.g. see readviewscheck.py).
  const size_t numStreams = _streams.size();
  if (numStreams > 1) {
    
    std::stringstream viewNames;
    for (size_t streamCount = 0; streamCount < numStreams; ++streamCount) {
      viewNames << "[Track ID " << _streams[streamCount]->_avstream->id << "]";
      if (streamCount < (numStreams - 1)) {
        viewNames << "\n";
      }
    }
    
    metadata.setData(MetaData::VIEW_NAMES, viewNames.str());
  }
  
  return true;
}

// get stream information
bool FFmpegFile::info( int& width,
           int& height,
           double& aspect,
           int& frames)
{
  Guard guard(_lock);

  if (_streams.empty())
    return false;

  mFnAssertMsg(_selectedStream, "Null _selectedStream");
  width  = _selectedStream->_width;
  height = _selectedStream->_height;
  aspect = _selectedStream->_aspect;
  frames = static_cast<int>(_selectedStream->_frames);

  return true;
}

int FFmpegFile::getRowSize() const
{
  if (canUseColorConvertor()) {
    return mFnAlignSize(getNumberOfComponents() * getWidth() * sizeof(float), kColorConverterRowPadMultiple);
  }
  // returns 0 if no stream
  const int bitDepth = getBitDepth();

  if (bitDepth > 8) {
    return getNumberOfComponents() * getWidth() * sizeof(U16);
  }
  return getNumberOfComponents() * getWidth();
}


int FFmpegFile::getBufferSize() const
{
  return getRowSize() * getHeight();
}

bool FFmpegFile::canUseColorConvertor() const
{
  if (!_selectedStream) {
    return false;
  }

  //we dont do csc for ProRes and Dnx:
  //ProRes: csc done by Apple SDK
  //Dnx: csc done in NukeCodecs
  const auto usingProRes = _selectedStream->_codecContext->codec_id == AV_CODEC_ID_PRORES;
  const auto usingDNxHD = _selectedStream->_codecContext->codec_id == AV_CODEC_ID_DNXHD;
  if (usingProRes || usingDNxHD) {
    return false;
  }

  if (_selectedStream->_color_convertor) {
    return true;
  }

  switch (_selectedStream->_codecContext->pix_fmt) {
    case AV_PIX_FMT_NV12:
    case AV_PIX_FMT_YUV420P:
    case AV_PIX_FMT_YUVJ420P:
    case AV_PIX_FMT_YUV422P:
    case AV_PIX_FMT_YUVJ422P:
    case AV_PIX_FMT_YUV444P:
    case AV_PIX_FMT_YUVJ444P:
    case AV_PIX_FMT_YUV420P16LE:
    case AV_PIX_FMT_YUV422P16LE:
    case AV_PIX_FMT_YUV444P16LE:
    case AV_PIX_FMT_YUV420P12LE:
    case AV_PIX_FMT_YUV422P12LE:
    case AV_PIX_FMT_YUV444P12LE:
    case AV_PIX_FMT_YUV420P10LE:
    case AV_PIX_FMT_YUV422P10LE:
    case AV_PIX_FMT_YUV444P10LE:
      return true;

    default:
      return false;
  }
}

//// FFmpegFileManager
//// Keeps track of all FFmpegFile mapped against file name.

// A lock manager function for FFmpeg, enabling it to use mutexes managed by this reader. Pass to av_lockmgr_register().
int FFmpegFileManager::FFmpegLockManager(void** mutex, enum AVLockOp op)
{
  switch (op) {
  case AV_LOCK_CREATE: // Create a mutex.
    try {
      *mutex = static_cast< void* >(new Lock());
      return 0;
    }
    catch(...) {
      // Return error if mutex creation failed.
      return 1;
    }

  case AV_LOCK_OBTAIN: // Lock the specified mutex.
    try {
      static_cast< Lock* >(*mutex)->lock();
      return 0;
    }
    catch(...) {
      // Return error if mutex lock failed.
      return 1;
    }

  case AV_LOCK_RELEASE: // Unlock the specified mutex.
    // Mutex unlock can't fail.
    static_cast< Lock* >(*mutex)->unlock();
    return 0;

  case AV_LOCK_DESTROY: // Destroy the specified mutex.
    // Mutex destruction can't fail.
    delete static_cast< Lock* >(*mutex);
    *mutex = nullptr;
    return 0;

  default: // Unknown operation.
    mFnAssert(false);
    return 1;
  }
}

// singleton
static FFmpegFileManager s_readerManager;

// constructor
FFmpegFileManager::FFmpegFileManager()
{
  av_log_set_level(AV_LOG_ERROR);
}

FFmpegFile::Ptr FFmpegFileManager::get(const char* filename, int viewIndex, ShareMode shareMode)
{
  // If the caller has specified no sharing, just create and return a new FFMpegFile
  // instance, we don't have anything to do with the _readers map.
  if (shareMode == eDontAllowSharing) {
    FFmpegFile::Ptr retVal;
    retVal.allocate(filename, viewIndex);
    return retVal;
  }
  
  Guard guard(_lock);
  FFmpegFile::Ptr retVal;

  // TODO: This approach will create a separate entry for each requested view of a given file, even if that file
  // doesn't contains fewer views than we're asking for, e.g. reading from a mono mov in a stereo project. If this
  // is likely to be a problem we should modify this code to prevent that.
  const ReaderMapKey mapKey(std::string(filename), viewIndex);
  
  ReaderMap::iterator it = _readers.find(mapKey);
  if (it == _readers.end()) {
    retVal.allocate(filename, viewIndex);
    _readers.insert(std::make_pair(mapKey, retVal));
  }
  else {
    retVal = it->second;
  }
  return retVal;
}

// release a specific reader
void FFmpegFileManager::release(const char* filename, int viewIndex)
{
  Guard guard(_lock);
  ReaderMap::iterator it = _readers.find(ReaderMapKey(filename, viewIndex));
  // Safely cope if there's no FFMpegFile object in the _readers map for the specified name, we
  // just don't need to do anything.
  if (it != _readers.end()) {
    if (it->second.refcount() == 1) {
      _readers.erase(it);
    }
  }
}


FFmpegFileManager FFmpegFileManager::s_readerManager;


///////////////////////////////////////////////////////////////////////////
/// mov64Reader

void mov64Reader::memoryInfo(Memory::MemoryInfoArray& output, const void* restrict_to) const
{
  if (restrict_to && iop->node() != (const Node*)restrict_to)
    return;

  Memory::MemoryInfo memInfo = Memory::MemoryInfo(iop, _data.size());

  std::stringstream buf;
  buf << w() << "x" << h();
  memInfo.addUserData("ImgSize", buf);

  output.push_back(memInfo);
}


///////////////////////////////////////////////////////
/// mov64ReaderFormat
mov64ReaderFormat::mov64ReaderFormat(Read* iop)
  : _iop(iop)
  , _decoder(0)
  , _colorMatrixType(0)
  , _defaultColorMatrixType(0)
  , _doNotAttachPrefix(false)
  , _matchMetaFormat(true)
  , _isDNxSdkVersionVisible(false)
  , _decodeVideoLevels(eDecodeVideoLevelsFullRange)
  , _defaultDecodeVideoLevels(eDecodeVideoLevelsFullRange)
  , _firstTrackOnly(true)
{
  // Create labels for all the colour matrices, based on mov32Reader behaviour
  // This enables the default to change without affecting the knob value
  int i = 0;
  // Add a placeholder default
  _colourMatrixTypeLabels[i++] = kDefaultColorMatrixLabel;
  // Set the matrix labels from Foundry::Color
  for ( ; i < eNumYCbCrEncodings + kNumDefaultLabels; i++) {
    _colourMatrixTypeLabels[i] = YCbCrEncodingNames[i - kNumDefaultLabels];
  }
  // Add a list delimiter
  _colourMatrixTypeLabels[i++] = nullptr;
  mFnAssertMsg( i == mFnSizeofArray(_colourMatrixTypeLabels),
                "Color matrix labels not initialised correctly" );

  i = 0;
  // Add a placeholder default
  _decodeVideoLevelsTypeLabels[i++] = kDefaultDecodeVideoLevels;
  for ( ; i < eNumDecodeVideoLevels + kNumDefaultLabels; i++) {
    _decodeVideoLevelsTypeLabels[i] = DecodeVideoLevelsNames[i - kNumDefaultLabels];
  }
  // Add a list delimiter
  _decodeVideoLevelsTypeLabels[i++] = nullptr;
  mFnAssertMsg( i == mFnSizeofArray(_decodeVideoLevelsTypeLabels),
                "Decode video levels labels not initialised correctly" );

}

namespace {
#ifdef FN_LICENSED_PRORES_DNXHD_CODEC
  std::string DNxHR_SDK_Version()
  {
    std::stringstream versionNumber;
    versionNumber
        << "Avid DNxCodec "
        << DNX_VERSION_MAJOR << "."
        << DNX_VERSION_MINOR << "."
        << DNX_VERSION_PATCH
    ;
    return versionNumber.str();
  }
#endif

  std::string FFmpeg_Version()
  {
    std::stringstream versionNumber;
    versionNumber
        << "libavcodec "
        << LIBAVCODEC_VERSION_MAJOR << "."
        << LIBAVCODEC_VERSION_MINOR  << "."
        << LIBAVCODEC_VERSION_MICRO
    ;
    return versionNumber.str();
  }
}

void mov64ReaderFormat::knobs(Knob_Callback c)
{
  static const char* kDecoderLabels[] = {"mov64", "mov32", nullptr};
  Enumeration_knob(c, &_decoder, kDecoderLabels, "decoder", "decoder");
  SetFlags(c, Knob::INVISIBLE);
  Tooltip(c, "Select library used to decode");

  Enumeration_knob(c, &_colorMatrixType, _colourMatrixTypeLabels, "mov64_ycbcr_matrix", "YCbCr matrix");
  SetFlags(c, Knob::NO_ANIMATION);
  Tooltip(c, "Choose color matrix used to convert from YCbCr to RGB. "
             "Only enabled when working with a YCbCr based pixel type. ");

  Enumeration_knob(c, &_decodeVideoLevels, _decodeVideoLevelsTypeLabels, "mov64_decode_video_levels", "Data Range");
  Tooltip(c, "For Avid DNxHD use this option to adjust the minimum and maximum values of the "
          "decoded output. If the source was encoded with a reduced range (0-240 for 8-bit or "
          "0-960 for 10-bit) then select 'Video Range' to decode to the full range "
          "(0-255 for 8-bit or 0-1023 for 10-bit).");
  SetFlags(c, Knob::NO_ANIMATION);
  
  Bool_knob(c, &_firstTrackOnly, "mov64_first_track_only", "first track only");
  Tooltip(c, "Causes the reader to ignore all but the first video track it finds in the file. This should be "
          "selected in a multiview project if the file happens to contain multiple video tracks that don't "
          "correspond to different views.");
  SetFlags(c, Knob::STARTLINE | Knob::NO_ANIMATION);
    
  Named_Text_knob(c, "mov64_metadata", "Metadata", "");
  SetFlags(c, Knob::STARTLINE | Knob::NO_ANIMATION);

  Bool_knob(c, &_doNotAttachPrefix, "mov64_no_prefix", "Do not attach prefix");
  Tooltip(c, "By default the 'quicktime' prefix is attached to general metadata keys to make it "
               "distinct from other metadata in the tree.\n\nEnable this option to read the general "
               "metadata 'as is' without attaching the prefix. Note that quicktime container "
               "specific metadata will still be inserted with the prefix.");

  Bool_knob(c, &_matchMetaFormat, "mov64_match_meta_format", "Match key format");
  Tooltip(c, "Attempts to clean up Quicktime meta keys to match standard Nuke metadata formatting.\n\n"
            "This will strip reverse DNS key names where possible, and place sub domains into child "
            "metadata namespaces.");
  SetFlags(c, Knob::NO_ANIMATION);
  ClearFlags(c, Knob::STARTLINE);

  Divider(c);
  Named_Text_knob(c, "mov64_codec_name", "Video Codec:", kUnknownCodecTypeName.c_str());

#ifdef FN_LICENSED_PRORES_DNXHD_CODEC
  static const std::string DNxHRSDKVersion = DNxHR_SDK_Version();
  Named_Text_knob(c, "DNxHR_SDK_Version", "SDK Build:", DNxHRSDKVersion.c_str());
  SetFlags(c, Knob::STARTLINE | Knob::ALWAYS_SAVE | Knob::HIDDEN);
  Tooltip(c, "Avid Codec SDK version that this file was originally loaded with.");
#endif

#ifdef PLUGINS_UPDATE
  Named_Text_knob(c, "Nuke_FileIO_Version", "Plugins Build:", PluginsInstallerVersion);
  SetFlags(c, Knob::ALWAYS_SAVE);
  Tooltip(c, "Plugins package version that this file was originally loaded with.");
#endif

  // We no longer support mov32... All these knobs would error the script
  Obsolete_knob(c, "mov32_new_qt_behavior", nullptr);
  Obsolete_knob(c, "mov32_ycbcr_matrix", nullptr);
  Obsolete_knob(c, "mov32_codec", nullptr);
  Obsolete_knob(c, "mov32_pixel_format", nullptr);

  Obsolete_knob(c, "ycbcr_matrix", "knob mov64_ycbcr_matrix $value");
  Obsolete_knob(c, "decode_video_levels", "knob mov64_decode_video_levels $value");

  Obsolete_knob(c, "DNxHR_SDK_Divider", nullptr);
}

void mov64ReaderFormat::updateCodecNameTextKnob(std::string codecName)
{
  if (Knob* const knob = _iop->knob("mov64_codec_name")) {
    assert(dynamic_cast<Text_KnobI*>(knob));
    auto& textKnob = *dynamic_cast<Text_KnobI*>(knob);
    textKnob.text(codecName);
  }
}

void mov64ReaderFormat::setKnobsVisibility(bool isVisible)
{
  if (Knob* const knob = _iop->knob("mov64_ycbcr_matrix")) {
    knob->visible(isVisible);
  }
  if (Knob* const knob = _iop->knob("mov64_metadata")) {
    knob->visible(isVisible);
  }
  if (Knob* const knob = _iop->knob("mov64_no_prefix")) {
    knob->visible(isVisible);
  }
  if (Knob* const knob = _iop->knob("mov64_match_meta_format")) {
    knob->visible(isVisible);
  }
  if (Knob* const knob = _iop->knob("mov64_decode_video_levels")) {
    knob->visible(isVisible);
  }

  if (Knob* const knob = _iop->knob("DNxHR_SDK_Version")) {
    knob->visible(isVisible && _isDNxSdkVersionVisible);
  }
}

void mov64ReaderFormat::append(Hash& hash)
{
  hash.append(_colorMatrixType);
  hash.append(_doNotAttachPrefix);
  hash.append(_matchMetaFormat);
  hash.append(_decodeVideoLevels);
  hash.append(_firstTrackOnly);
  hash.append(_decoder);
}

void mov64ReaderFormat::initDynamicKnobs()
{
  Knob* knob = _iop->knob("mov64_ycbcr_matrix");
  if (knob)
    updateColorMatrixKnob(knob);

  knob = _iop->knob("mov64_decode_video_levels");
  if (knob)
    updateDecodeVideoLevelsKnob(knob);

}

void mov64ReaderFormat::setYCbCrMatrixEnabled(bool enabled)
{
  Knob* colorMatrixKnob = _iop->knob("mov64_ycbcr_matrix");
  if (colorMatrixKnob) {
    if (enabled) {
      colorMatrixKnob->enable();
    }
    else {
      colorMatrixKnob->disable();
    }
  }
}

void mov64ReaderFormat::updateColorMatrixKnob(Knob* colorMatrixKnob)
{
  // Get the enumeration knob
  mFnAssert(colorMatrixKnob);
  Enumeration_KnobI* colorMatrixEnumerationKnob = colorMatrixKnob->enumerationKnob();
  if (colorMatrixEnumerationKnob) {
    // Change the menu items //

    StringArray colorMatrices = colorMatrixEnumerationKnob->menu();

    // Get the label if the value is valid for this list
    std::string label = GetKnobLabel(colorMatrixKnob, colorMatrices);

    const size_t arraySize = colorMatrices.size();
    const size_t labelIndex = _defaultColorMatrixType; // 1 = Rec709, 2 = Rec601;
    mFnAssert(arraySize > labelIndex);

    const char* defaultColorMatrixName = YCbCrEncodingNames[labelIndex];
    colorMatrices[0] = std::string(kDefaultLabel_Prefix) + std::string(defaultColorMatrixName) + std::string(kDefaultLabel_Suffix);

    colorMatrixEnumerationKnob->menu(colorMatrices);

    // Set the knob value using the new list
    SetKnobValue(colorMatrixKnob, colorMatrices, label);
  }
}

void mov64ReaderFormat::updateDecodeVideoLevelsKnob(Knob* knob)
{
  // Get the enumeration knob
  mFnAssert(knob);
  Enumeration_KnobI* enumerationKnob = knob->enumerationKnob();
  if (enumerationKnob) {
    // Change the menu items //

    StringArray decodeVideoLevels = enumerationKnob->menu();

    // Get the label if the value is valid for this list
    std::string label = GetKnobLabel(knob, decodeVideoLevels);

    const size_t arraySize = decodeVideoLevels.size();
    const size_t labelIndex = _defaultDecodeVideoLevels; // 0 = Full range, 1 = Video range
    mFnAssert(arraySize > labelIndex);

    const char* defaultDecodeVideoLevelsName = DecodeVideoLevelsNames[labelIndex];
    decodeVideoLevels[0] = std::string(kDefaultLabel_Prefix) + std::string(defaultDecodeVideoLevelsName) + std::string(kDefaultLabel_Suffix);

    enumerationKnob->menu(decodeVideoLevels);

    // Set the knob value using the new list
    SetKnobValue(knob, decodeVideoLevels, label);
  }
}

////////////////////////////////////////////////////////////////////////////////
mov64Reader::mov64Reader(Read* iop, const char* optionalFilename)
  : Reader(iop)
  , IBufferFillHeap("mov64Reader")
  , _data(this)
  , _memNeeded(0)
  , _viewIndex(0)
  , _numFrames(0)
  , _currentFrame(0)
  , _isTimelineReader(false)
  , _optionalFilename(optionalFilename)
  , _planarReadBuffer(nullptr)
{
  // get the reader from the FFmpeg reader, using optional filename if supplied
  const char* filename = _optionalFilename ? _optionalFilename : iop->filename();
  
  // Find the ReaderFormat, if there is one, for access to the knobs.
  // Note that preview will not create a ReaderFormat so we have to tolerate this.
  FileHandler* fileHandler = iop->handler();
  mov64ReaderFormat* readerFormat = fileHandler ? dynamic_cast<mov64ReaderFormat*>(fileHandler) : nullptr;
  const bool isMov64PluginStandalone = readerFormat != nullptr;   // if false then the knobs visibility is controlled by the movReader plugin
#ifndef FN_OS_LINUX
  // If this is a component of the movReader, then filehandler will be a movReaderFormat instead
  if (!readerFormat) {
    movReaderFormat* readerFormatMeta = fileHandler ? dynamic_cast<movReaderFormat*>(fileHandler) : nullptr;
    if (readerFormatMeta) {
      // Get the internal reader format to use here
      readerFormat = readerFormatMeta->getMov64ReaderFormat();
    }
  }
#endif
  
  if (!readerFormat) {
    iop->internalError("This format is not supported by mov64 reader.");
    return;
  }

  // mov64 is the only supported encoder since mov32 was deprecated
  if (readerFormat->getSelectedDecoder() != 0) {
    iop->internalError("mov32 is no longer supported!");
    return;
  }
  
  _isTimelineReader = (dynamic_cast<TimelineRead*>(iop) != nullptr);
  
  // Get the requested view index from the owning op. If we're reading for the comp we subtract one to account for
  // the hidden defaul view, though not if that's actually been specified!
  _viewIndex = _isTimelineReader ? iop->view_for_reader() : std::max((iop->view_for_reader() - 1), 0);
  
  // If the 'first track only' knob has been set then force the requested index to 0 so we select the 0th video stream.
  // For now, don't do this on the timeline, which handles the override at a higher level and doesn't use the knobs,
  // so we're always getting the default knob value here.
  if (readerFormat && readerFormat->getFirstTrackOnly() && !_isTimelineReader) {
    _viewIndex = 0;
  }
  
  // For performance reason and for different use cases Hiero prefers to decode multiple time the same file.
  // That means allocating more resources by guaranteeing best performance when the application needs
  // to seek in consecutive way from different position in the same video sequence.
  const FFmpegFileManager::ShareMode shareMode  = _isTimelineReader
                                                ? FFmpegFileManager::eDontAllowSharing
                                                : FFmpegFileManager::eAllowSharing;
  
  _reader = FFmpegFileManager::s_readerManager.get(filename, _viewIndex, shareMode);
  
  if (_reader->invalid()) {
    const bool componentOfmovReader = (_optionalFilename != nullptr);
    if (componentOfmovReader) {
      // Originally we gave a warning here, but warnings seem unpopular in the nuke universe.
    }
    else {
      iop->internalError(_reader->error());
    }
    return;
  }
  
  if (readerFormat) {
    readerFormat->_isDNxSdkVersionVisible = _reader->isCodecTypeDNx();
    if (isMov64PluginStandalone) {
      readerFormat->setKnobsVisibility(true);
    }
  }
  
  // If we've only got one stream available disable the 'first track only' knob as it's redundant.
  // Note: Don't do this if we're reading for the timeline, we'll be on a background thread (and
  // there won't be any knobs anyway).
  if ((_reader->streams() == 1) && !_isTimelineReader) {
    if (Knob* firstTrackOnlyKnob = iop->knob("mov64_first_track_only")) {
      firstTrackOnlyKnob->disable();
    }
  }
  
  // set iop info size
  int width, height, frames;
  double aspect;

  // get metadata from file
  if (_reader->info(width, height, aspect, frames)) {
    const int numberOfComponents = _reader->getNumberOfComponents();

    info_.channels((numberOfComponents == 4) ? Mask_RGBA : Mask_RGB);
    set_info(width, height, numberOfComponents, aspect);
    info_.first_frame(1);
    info_.last_frame(frames);

    setRowPadMultiple(_reader->canUseColorConvertor() ? kColorConverterRowPadMultiple : 0);
    _numFrames = frames;
    _memNeeded = _reader->getBufferSize();
  }

  // Set the YCbCr colorspace on the ReaderFormat, if we have one (preview will not create one).

  if (readerFormat) {
    readerFormat->updateCodecNameTextKnob(GetCodecName(*_reader));

    // YCbCr Matrix knob only enabled for YUV pixel formats
    if (_reader->isYUV()) {
      readerFormat->setYCbCrMatrixEnabled(true);
      // YCbCrMatrix is based on image height
      const int defaultColorMatrixType = _reader->isRec709Format() ? eYCbCrEncodingRec709 : eYCbCrEncodingRec601;
      readerFormat->setDefaultColorMatrixType(defaultColorMatrixType);
      // Based on Enumeration_knob choice
      const int currentColorMatrixType = readerFormat->getColorMatrixType();
      _reader->setColorMatrixTypeOverride(currentColorMatrixType);
    }
    else {
      // Disable for RGB formats
      readerFormat->setYCbCrMatrixEnabled(false);
    }
    const bool doNotAttachPrefix = readerFormat->getDoNotAttachPrefix();
    _reader->setDoNotAttachPrefix(doNotAttachPrefix);
    const bool matchMetaFormat   = readerFormat->getMatchMetaFormat();
    _reader->setMatchMetaFormat(matchMetaFormat);

    // DNxHD specific.
    // The UI is configure so that index 0 is the 'default' value
    // which has to be set after the reader format is created.
    // Index 1 and 2 are video range and full range respectively.
    // This must be converted to 0-1 range for the decoder.
    readerFormat->setDefaultDecodeVideoLevels(_reader->getDefaultDecodeVideoLevels());
    int decodeVideoLevels = readerFormat->getDecodeVideoLevels(); // This is in the range 0-2.
    if (0 < decodeVideoLevels) {
      --decodeVideoLevels; // Convert to a zero based index (r.e. DecodeVideoLevelsEnum)
    }
    else {
      decodeVideoLevels = _reader->getDefaultDecodeVideoLevels(); // Use the 'default' value.
    }
    _reader->setDecodeVideoLevels(decodeVideoLevels); // This is in the range 0-1.
  }
  
  // get stream 0 metadata
  _reader->metadata(meta);

  // set buffer size
  _data.resize(_memNeeded);

  // Set colorspace knob based on pixel format and errata.
  this->setLUT(LUT::Builtin(_reader->getColorspace(), this));
}

mov64Reader::~mov64Reader()
{
  // If we're being used for the timeline then the refcount *must* be 1 as we shouldn't have allowed
  // sharing of FFMpegFile objects between readers or have it referenced by the FFmpegFileManager.
  mFnAssert(!_isTimelineReader || (_reader.refcount() == 1));
  
  // Remove a reference from the FFMpegFile object we've been using, possibly deleting it.
  // If _isTimelineReader is true this should definitely result in the object being deleted.
  _reader.clear();

  if (!_isTimelineReader) {
    // release the reader if is not used anymore
    const char* filename = _optionalFilename ? _optionalFilename : iop->filename();
    FFmpegFileManager::s_readerManager.release(filename, _viewIndex);
  }
  
  delete[] _planarReadBuffer;
}

std::string mov64Reader::getDefaultColorspace() const
{
  const std::string colorspaceMetaKeyValue = meta.getString("quicktime/thefoundry/Colorspace");
  return colorspaceMetaKeyValue;
}

bool mov64Reader::invalid()
{
  if (_reader) {
    return _reader->invalid();
  }
  return true;
}

bool mov64Reader::supports_stereo() const
{
  return true;
}

bool mov64Reader::fileStereo() const
{
  if (_isTimelineReader) {
    // See commit message for more details.
    mFnAssertMsg(false, "mov64Reader::fileStereo: needs full implementation for timeline");
    return false;
  }
  
  // Say we're a stereo/multiview file if there's more than one video stream and the project is
  // currently set up with more than one view. We can't know that multiple video streams really do
  // contain different views, they might have entirely unrelated data, so only report we're stereo
  // if the user is already working in a multiview project, and hope that if they're loading
  // multi-stream QuickTime files in that scenario that they are actually multiview.
  return _reader && (_reader->streams() > 1) && OutputContext::hasMultipleViews(iop);
}

bool mov64Reader::fillBuffer(void* buffer)
{
  if (buffer == nullptr) {
    iop->internalError("mov64Reader has run out of memory", _memNeeded);
    return false;
  }

  // Translate from the 1-based frames expected by Nuke to 0-based frame offsets used during decoding
  if (!_reader->decode( static_cast<unsigned char*>(buffer), frame() - 1 )) {
    iop->internalError(_reader->error());
    return false;
  }

  return true;
}

void mov64Reader::engine(int y, int x, int rx, ChannelMask channels, Row& out)
{
  MemoryBufferGuard guard(_data);
  if (!guard.buffer()) {
    // for some reason the buffer is not available and the Iop error
    // is already called inside fillBuffer method.
    return;
  }
  if (_reader->canUseColorConvertor()) {
    engineFloat(y, x, rx, channels, out, static_cast<float*>(guard.buffer()));
  }
  else if (_reader->getBitDepth() > 8) {
    engine16(y, x, rx, channels, out, static_cast<U16*>(guard.buffer()));
  }
  else {
    engine8(y, x, rx, channels, out, static_cast<unsigned char*>(guard.buffer()));
  }
}

void mov64Reader::engine8(int y, int x, int rx, ChannelMask channels, Row& out, unsigned char* data)
{
  int numberOfComponents = _reader->getNumberOfComponents();
  foreach ( z, channels ) {
    float* TO = out.writable(z) + x;
    int componentPosition = _reader->getComponentPosition(z); // In-place RGBA swizzle.
    unsigned char* FROM = data + componentPosition;
    FROM += (height() - y - 1) * width() * numberOfComponents;
    FROM += x * numberOfComponents;
    from_byte(z, TO, FROM, nullptr, rx - x, numberOfComponents);
  }
}

void mov64Reader::engine16(int y, int x, int rx, ChannelMask channels, Row& out, U16* data)
{
  int numberOfComponents = _reader->getNumberOfComponents();
  int bitsPerComponent = _reader->getBitDepth();
  foreach ( z, channels ) {
    float* TO = out.writable(z) + x;
    int componentPosition = _reader->getComponentPosition(z);
    U16* FROM = data + componentPosition; // In-place RGBA swizzle.
    FROM += (height() - y - 1) * width() * numberOfComponents;
    FROM += x * numberOfComponents;
    from_short(z, TO, FROM, nullptr, rx - x, bitsPerComponent, numberOfComponents);
  }
}

void mov64Reader::engineFloat(int y, int x, int rx, ChannelMask channels, Row& out, float* data)
{
  int numberOfComponents = _reader->getNumberOfComponents();
  foreach ( z, channels ) {
    float* TO = out.writable(z) + x;
    int componentPosition = _reader->getComponentPosition(z); // In-place RGBA swizzle.
    float* FROM = data + componentPosition;
    FROM += (height() - y - 1) * _reader->getRowSize() / sizeof(float);
    FROM += x * numberOfComponents;
    from_float(z, TO, FROM, nullptr, rx - x, numberOfComponents);
  }
}

void mov64Reader::prefetchMetaData()
{
  meta.setData(MetaData::TIMECODE, _reader->getTimecodeForFrame(frame() - 1));
  Reader::prefetchMetaData();
}

void mov64Reader::open()
{
  int f = frame();

  if (f < info_.first_frame())
    f = info_.first_frame();
  if (f > info_.last_frame())
    f = info_.last_frame();
  if (f != _currentFrame)
    _data.invalidate();

  _currentFrame = f;
}

////////////////////////////////////////////////////////////////////////////////
// Get information about the planar image, including the size of the buffer need
// to store the encoded image and the ideal decoded image format for
// the specified channel.
PlanarReadInfo mov64Reader::planarReadInfo(const ChannelSet &channels)
{
  // Map from the LUT to a ColorCurveEnum to report the correct colour curve to
  // Hiero.
  ColorCurveEnum colorCurveType = eColorCurveUnknown;
  if (lut_ == LUT::GetBuiltinLutByName("linear"))
    colorCurveType = eColorCurveLinear;
  else if (lut_ == LUT::GetBuiltinLutByName("sRGB"))
    colorCurveType = eColorCurveSRGB;
  else if (lut_ == LUT::GetBuiltinLutByName("rec709"))
    colorCurveType = eColorCurveRec709;
  else if (lut_ == LUT::GetBuiltinLutByName("Cineon"))
    colorCurveType = eColorCurveLog;
  else if (lut_ == LUT::GetBuiltinLutByName("Gamma1.8"))
    colorCurveType = eColorCurveGamma18;
  else if (lut_ == LUT::GetBuiltinLutByName("Gamma2.2"))
    colorCurveType = eColorCurveGamma22;
  else if (lut_ == LUT::GetBuiltinLutByName("Gamma2.4"))
    colorCurveType = eColorCurveGamma24;
  else if (lut_ == LUT::GetBuiltinLutByName("Gamma2.6"))
    colorCurveType = eColorCurveGamma26;
  else if (lut_ == LUT::GetBuiltinLutByName("Panalog"))
    colorCurveType = eColorCurvePanalog;
  else if (lut_ == LUT::GetBuiltinLutByName("REDLog"))
    colorCurveType = eColorCurveREDLog;
  else if (lut_ == LUT::GetBuiltinLutByName("ViperLog"))
    colorCurveType = eColorCurveViperLog;
  else if (lut_ == LUT::GetBuiltinLutByName("AlexaV3LogC"))
    colorCurveType = eColorCurveAlexaV3LogC;
  else if (lut_ == LUT::GetBuiltinLutByName("PLogLin"))
    colorCurveType = eColorCurvePLogLin;
  else if (lut_ == LUT::GetBuiltinLutByName("SLog"))
    colorCurveType = eColorCurveSLog;
  else if (lut_ == LUT::GetBuiltinLutByName("SLog1"))
    colorCurveType = eColorCurveSLog1;
  else if (lut_ == LUT::GetBuiltinLutByName("SLog2"))
    colorCurveType = eColorCurveSLog2;
  else if (lut_ == LUT::GetBuiltinLutByName("SLog3"))
    colorCurveType = eColorCurveSLog3;
  else if (lut_ == LUT::GetBuiltinLutByName("CLog"))
    colorCurveType = eColorCurveCLog;
  else if (lut_ == LUT::GetBuiltinLutByName("Log3G10"))
    colorCurveType = eColorCurveLog3G10;
  else if (lut_ == LUT::GetBuiltinLutByName("Log3G12"))
    colorCurveType = eColorCurveLog3G12;
  else if (lut_ == LUT::GetBuiltinLutByName("HybridLogGamma"))
    colorCurveType = eColorCurveHybridLogGamma;
  else if (lut_ == LUT::GetBuiltinLutByName("Protune"))
    colorCurveType = eColorCurveProtune;
  else if (lut_ == LUT::GetBuiltinLutByName("BT1886"))
    colorCurveType = eColorCurveBT1886;
  else if (lut_ == LUT::GetBuiltinLutByName("st2084"))
    colorCurveType = eColorCurveST2084;

  const int sourceBitDepth = _reader->getBitDepth();

  DataTypeEnum dataType = eDataTypeNone;
  const int minValue = 0;
  int maxValue = (1 << sourceBitDepth) - 1;
  const int whitePoint = 1;

  // TODO: For the timeline node context, add support for telling FFmpeg to decode to a 3 x 10 bit
  //  _output_ pixel format for 3 x 10 bit source data, then use eDataTypeUInt32_10bit here.
  //  Wihtout that we're still using the very suboptimal approach of planarReadPass() into a local
  //  buffer and planarDecodePass() on that, rather than planarReadPass() directly to the timeline's
  //  playback cache buffer.

  //TODO: do we need it in Studio? seems not
  //if (_reader->canUseColorConvertor()) {
  //  dataType = eDataTypeFloat32;
  //}
    
  if (_reader->canUseColorConvertor()) {
    dataType = eDataTypeFloat32;
    maxValue = 1;
  }
  else {
    switch (sourceBitDepth) {
    case 8:
      dataType = eDataTypeUInt8;
      break;
    case 10:  // TODO: It would be much better to use eDataTypeUInt32_10bit
    case 12:
    case 14:
    case 16:
      dataType = eDataTypeUInt16;
      break;
    default:
      mFnAssertMsg(false, "Unsupported bit depth:" << sourceBitDepth);
    }
  }

  const ChannelSet srcChannels = channels;
  const int numberOfComponents = _reader->getNumberOfComponents();

  const Box bounds(0, 0, w(), h());
  ImagePlaneDescriptor baseDesc(bounds, true, srcChannels, numberOfComponents);

  DataInfo dataInfo(dataType, true, minValue, maxValue, whitePoint);

  // planarDecodePass was hard-coded to flip vertically for correct display in the timeline,
  // which expects image data with y increasing with address.
  auto direction = ImageDirection::eDecreasingY;

  // #rick: Everything _may_ work fine for channel counts other than 3 or 4 but a lot of the mov64Reader code
  // assumes that's what we've got, i.e. permutations of RGB(A), and probably needs fixing-up for anything else.
  mFnAssert((numberOfComponents == 3) || (numberOfComponents == 4));

  // Report any swizzling clients need to use to restore RGBA ordering (we may decode the image
  // data into BGRA, ARGB, etc).
  std::vector<int> channelSwizzling;
  for (int count = 0; count < numberOfComponents; ++count) {
    Channel channel = static_cast<Channel>(Chan_Red + count);
    const auto channelIndex = _reader->getComponentPosition(channel);
    mFnAssert((channelIndex >= 0) && (channelIndex < numberOfComponents));
    channelSwizzling.push_back(channelIndex);
  }

  GenericImagePlaneDescriptor desc(baseDesc, dataInfo, colorCurveType, direction, channelSwizzling);

  // Currently passing 0 as readPassBufferSize.
  // In future it may get used by Core/Media to determine that it should allocate a buffer itself.
  return PlanarReadInfo(desc, /* readPassBufferSize = */ 0, /* isDecodeThreadable */ false, /* isValid */ true);
}

////////////////////////////////////////////////////////////////////////////////
// Planar read and decode the specified channels of the image in one go.
void mov64Reader::planarReadAndDecode(GenericImagePlane& image, const ChannelSet& channels, int priority)
{
  // Think we've got big problems if the client code requests channels that don't match those
  // specified in the passed GenericImagePlane, especially if the number of channels doesn't match.
  mFnAssert(image.desc().channels() == channels);

  bool canWriteDirect = false;

  const bool channelsMatch = (image.desc().channels().size() == _reader->getNumberOfComponents()) &&
                             (channels.size() == _reader->getNumberOfComponents());

  // TODO: The existing mov64Reader/FFmpeg implementation only supports decoded _output_ pixel formats
  //  of 8 and 16 bits (AV_PIX_FMT_RGB24, AV_PIX_FMT_RGBA, AV_PIX_FMT_RGB48LE or AV_PIX_FMT_RGBA64LE).
  //  Thus, the common 10 bit case gets decoded into a 16 bit buffer. To display this correctly we need
  //  to either:
  //    - Ideally, for 3 channel cases, update mov64Reader/FFmpeg to support decoding into a 10 pixel
  //      format and handle that directly as eDataTypeUInt32_10bit (as for 10 bit dpx files).
  //    - Update the timeline code to make use of the DataInfo's max value and perform the scaling itself,
  //      ideally on the GPU when we read from the texture.
  //    - Continue to use planarDecodePass() to scale up the data over the 16 bit range (otherwise
  //      the image will be darker than it's supposed to be).
  const int sourceBitDepth = _reader->getBitDepth();
  const bool noScalingRequired = (sourceBitDepth == 8) || (sourceBitDepth == 16);

  if (channelsMatch && noScalingRequired) {

    // TODO: Think this will exclude three channel 10 bit from direct write, though in principle
    //  we can actually do that if we set everything up correctly.
    const auto idealInfo = planarReadInfo(channels);

    const auto& idealDesc = idealInfo.genericImagePlaneDescriptor();

    canWriteDirect = (image.desc().imageDirection() == idealDesc.imageDirection()) &&
                     (image.desc().channelSwizzling() == idealDesc.channelSwizzling());
  }

  if (canWriteDirect) {
    switch (image.desc().dataInfo().dataType()) {
      case eDataTypeUInt8 : planarReadPass(&image.writableAt<uint8_t>(0, 0, 0), channels); break;
      case eDataTypeUInt16: planarReadPass(&image.writableAt<uint16_t>(0, 0, 0), channels); break;
      case eDataTypeFloat32: planarReadPass(&image.writableAt<float>(0, 0, 0), channels); break;
      default: mFnAssert(false);  // Need to add support, and possibly updating channelsMatch and bitDepthsMatch, e.g. for eDataTypeUInt32_10bit.
    }
  }
  else {
    if (!_planarReadBuffer) {
      _planarReadBuffer = new uint8_t[_reader->getBufferSize()];
    }
    planarReadPass(_planarReadBuffer, channels);
    planarDecodePass(_planarReadBuffer, image, channels, 0, 1, priority);
  }
}

////////////////////////////////////////////////////////////////////////////////
// Planar read the specified channels of the whole file into the buffer in one
// go, returning the number of bytes read.
int mov64Reader::planarReadPass(void* buffer, const ChannelSet& channels)
{
  fillBuffer(buffer);
  return _reader->getBufferSize();
}

////////////////////////////////////////////////////////////////////////////////
// Convert from an 8 bits per component source frame.
// This has two template parameters that flag the presence of an alpha channel
// in the source and destination buffers.
//
template<bool srcHasAlpha, bool dstHasAlpha>
void convert_8bit(GenericImagePlane& image,
                  int dstLine,
                  int dstLineInc,
                  const uint8_t* srcData,
                  long srcRowBytes,
                  int width,
                  int height,
                  int maxValue,
                  uint8_t* componentPosition)
{
  int redIndex = componentPosition[0];
  int greenIndex = componentPosition[1];
  int blueIndex = componentPosition[2];
  int alphaIndex = componentPosition[3];

  int line = 0;
  while (height > line) {

    const uint8_t* src = reinterpret_cast<const uint8_t*>(srcData);
    srcData += srcRowBytes;

    uint8_t* dst = &(image.writableAt<uint8_t>(0, dstLine, 0));

    int pixel = width;
    while (pixel--) {
      *dst++ = src[redIndex];
      *dst++ = src[greenIndex];
      *dst++ = src[blueIndex];

      uint8_t value;
      if (srcHasAlpha) {
        value = src[alphaIndex];
        src += 4;
      }
      else {
        value = maxValue;
        src += 3;
      }

      if (dstHasAlpha)
        *dst++ = value;
    }

    ++line;
    dstLine += dstLineInc;
  }
}

////////////////////////////////////////////////////////////////////////////////
// Convert from a source frame that has between 9 and 16 bits per component.
// This has two template parameters that flag the presence of an alpha channel
// in the source and destination buffers.
//
template<bool srcHasAlpha, bool dstHasAlpha>
void convert_9bit_to_16bit(GenericImagePlane& image,
                           int dstLine,
                           int dstLineInc,
                           const uint8_t* srcData,
                           long srcRowBytes,
                           int width,
                           int height,
                           int maxValue,
                           uint8_t* componentPosition,
                           int bitsPerComponent)
{
  int redIndex = componentPosition[0];
  int greenIndex = componentPosition[1];
  int blueIndex = componentPosition[2];
  int alphaIndex = componentPosition[3];

  int diffDynamicRange = 16 - bitsPerComponent;
  int dynamicRangeLeftShift = diffDynamicRange;
  int dynamicRangeRightShift = bitsPerComponent - diffDynamicRange;

  int line = 0;
  while (height > line) {

    const uint16_t* src = reinterpret_cast<const uint16_t*>(srcData);
    srcData += srcRowBytes;

    uint16_t* dst = &(image.writableAt<uint16_t>(0, dstLine, 0));

    int pixel = width;
    while (pixel--) {
      uint16_t value = src[redIndex];
      *dst++ = (value << dynamicRangeLeftShift) | (value >> dynamicRangeRightShift);
      value = src[greenIndex];
      *dst++ = (value << dynamicRangeLeftShift) | (value >> dynamicRangeRightShift);
      value = src[blueIndex];
      *dst++ = (value << dynamicRangeLeftShift) | (value >> dynamicRangeRightShift);

      if (srcHasAlpha) {
        value = src[alphaIndex];
        src += 4;
      }
      else {
        value = maxValue;
        src += 3;
      }
      value = (value << dynamicRangeLeftShift) | (value >> dynamicRangeRightShift);

      if (dstHasAlpha)
        *dst++ = value;
    }

    ++line;
    dstLine += dstLineInc;
  }
}

////////////////////////////////////////////////////////////////////////////////
// Do a planar decode of the image data in the specified source buffer into the
// GenericImagePlane, for the specified channels.
void mov64Reader::planarDecodePass(void* srcBuffer, GenericImagePlane& image, const ChannelSet& channels,
                                   int threadIndex, int nDecodeThreads, int priority)
{
  mFnAssert(threadIndex >= 0);
  mFnAssert(threadIndex < nDecodeThreads);
  mFnAssert(nDecodeThreads > 0);

  // Assert that the image plane we're filling in is configured to hold the
  // whole image, not a sub area of it. Although the dpx format supports
  // specifying the offset and size of the 'original' image (in the orientation
  // header) we don't bother with that, either in this new planar read/decode
  // code or in the original.
  mFnAssert(image.desc().bounds().x() == 0);
  mFnAssert(image.desc().bounds().y() == 0);
  mFnAssert(image.desc().bounds().w() == w());
  mFnAssert(image.desc().bounds().h() == h());

  int readerHeight = _reader->getHeight();
  int height = (readerHeight + (nDecodeThreads - 1)) / nDecodeThreads; // Number of lines for this thread to process.
  int width = _reader->getWidth();

  // Determine the offset of the first line for this thread to
  // convert. Simply use the thread index and buffer rowbytes
  // to calculate the byte offset into the source buffer.
  long srcRowBytes = _reader->getRowSize();
  const uint8_t* srcData = static_cast<const uint8_t*>(srcBuffer) + (threadIndex * height * srcRowBytes);

  // Determine the location in the destination buffer.
  // If flipping vertically, move in reverse from the
  // last line to the first line of the destination
  // buffer. If multiple threads are in use, skip the
  // rows that are processed by other threads.
  int dstLine, dstLineInc;
  const bool flipVertically = false;  // Our planarReadInfo has specified ImageDirection::eDecreasingY, so flipping should be handled by the client.
  if (flipVertically) {
    dstLine = (readerHeight - 1) - (threadIndex * height);
    dstLineInc = -1;
  }
  else {
    dstLine = threadIndex * height;
    dstLineInc = 1;
  }

  const int numberOfComponents = _reader->getNumberOfComponents();
  const bool srcHasAlpha = (4 == numberOfComponents) ? true : false;
  const bool dstHasAlpha = (image.desc().channels() == Mask_RGBA) ? true : false;
  uint8_t componentPositionLUT[4] = { 0, 1, 2, 3 }; // Channel swizzling now handled by the client, from the info provided by planarReadInfo().

  // The entire frame is divided amoung the worker threads. The frame
  // may not be an integer mulitple of the number of threads. This
  // means that one of the threads will process fewer lines than the
  // other threads. Test here for that condition and set the number
  // of lines appropriately.
  int linesRemaining = readerHeight - (height * threadIndex);
  if (height > linesRemaining)
    height = linesRemaining;

  int bitDepth = _reader->getBitDepth();
  int maxValue = (1 << bitDepth) - 1;
  if (8 == bitDepth) {
    if (srcHasAlpha && dstHasAlpha)
      convert_8bit<true, true>(image, dstLine, dstLineInc, srcData, srcRowBytes, width, height, maxValue, componentPositionLUT);
    else if (srcHasAlpha && !dstHasAlpha)
      convert_8bit<true, false>(image, dstLine, dstLineInc, srcData, srcRowBytes, width, height, maxValue, componentPositionLUT);
    else if (!srcHasAlpha && dstHasAlpha)
      convert_8bit<false, true>(image, dstLine, dstLineInc, srcData, srcRowBytes, width, height, maxValue, componentPositionLUT);
    else
      convert_8bit<false, false>(image, dstLine, dstLineInc, srcData, srcRowBytes, width, height, maxValue, componentPositionLUT);
  }
  else if (16 >= bitDepth) {
    if (srcHasAlpha && dstHasAlpha)
      convert_9bit_to_16bit<true, true>(image, dstLine, dstLineInc, srcData, srcRowBytes, width, height, maxValue, componentPositionLUT, bitDepth);
    else if (srcHasAlpha && !dstHasAlpha)
      convert_9bit_to_16bit<true, false>(image, dstLine, dstLineInc, srcData, srcRowBytes, width, height, maxValue, componentPositionLUT, bitDepth);
    else if (!srcHasAlpha && dstHasAlpha)
      convert_9bit_to_16bit<false, true>(image, dstLine, dstLineInc, srcData, srcRowBytes, width, height, maxValue, componentPositionLUT, bitDepth);
    else
      convert_9bit_to_16bit<false, false>(image, dstLine, dstLineInc, srcData, srcRowBytes, width, height, maxValue, componentPositionLUT, bitDepth);
  }
}

////////////////////////////////////////////////////////////////////

static Reader* Build(Read* iop, int fd, const unsigned char* b, int n)
{
  if (fd >= 0) {
    ::close(fd);
  }
  return new mov64Reader(iop);
}

static ReaderFormat* BuildFormat(Read* iop)
{
  return new mov64ReaderFormat(iop);
}

static bool Test(int fd, const unsigned char* block, int length)
{
  return true;
}

const Reader::Description mov64Reader::d("mov\0mov64\0m4v\0mp4\0ffmpeg\0", "mov64", Build, Test, BuildFormat);