// Copyright (c) 2014 The Foundry Visionmongers Ltd.  All Rights Reserved.
#ifndef NUKE_MOV64READER_H
#define NUKE_MOV64READER_H

#include "Build/fnBuild.h"
#ifndef FN_EXAMPLE_PLUGIN
#include "Base/fnTimecode.h" // FIXME: replace with shippable time code handling
#endif
#include "DDImage/DDString.h"
#include "DDImage/Reader.h"
#include "DDImage/Row.h"
#include "DDImage/MemoryBuffer.h"
#include "DDImage/RefCountedObject.h"
#include "DDImage/Thread.h"
#include "DDImage/LUT.h"

#include "codecWhitelist.h"

#include "mov64Common.h"
#define NUKE_MOV64READER_CPP
#include "movCommon.h"
#include "codecCommon.h"

#ifdef _WIN32
  #include <io.h>
#endif

#ifndef INT64_C
# define INT64_C(c) (c ## LL)
#endif
#ifndef UINT64_C
# define UINT64_C(c) (c ## ULL)
#endif

extern "C" {
  #ifdef _WIN32
    #pragma warning(push, 0) // We dont care about warnings from the libav headers
  #endif
  #include <errno.h>
  #include <libavformat/avformat.h>
  #include <libavutil/imgutils.h>
  #include <libavformat/avio.h>
  #include <libavcodec/avcodec.h>
  #include <libavutil/opt.h>
  #include <libswscale/swscale.h>
  #include <libavutil/avutil.h>
  #include <libavutil/error.h>
  #ifdef _WIN32
    #pragma warning(pop)
  #endif
}

#ifndef FN_EXAMPLE_PLUGIN
// Comment out this preprocessor definition if building this file WITHOUT a
// valid Apple ProRes license or valid Avid DNxHD license.
# define FN_LICENSED_PRORES_DNXHD_CODEC
#endif

#ifdef FN_LICENSED_PRORES_DNXHD_CODEC
# include "NukeCodecs/fnNukeCodecs.h" // propriatary Foundry codec management
#endif

// Subset of Foundry::Colour::YCbCrEncodingEnum
enum YCbCrEncodingEnum {
  eYCbCrEncodingRec709,
  eYCbCrEncodingRec601,
  eNumYCbCrEncodings
};

const char* const YCbCrEncodingNames[eNumYCbCrEncodings + 1] =
{
  "Rec 709",
  "Rec 601",
  nullptr // NULL terminator
};

enum DecodeVideoLevelsEnum {
  eDecodeVideoLevelsFullRange,
  eDecodeVideoLevelsVideoRange,
  eNumDecodeVideoLevels
};

const char* const DecodeVideoLevelsNames[eNumDecodeVideoLevels + 1] =
{
  "Full Range",
  "Video Range",
  nullptr // NULL terminator
};

class FFmpegFile : public DD::Image::RefCountedObject
{
  struct Stream
  {
    int _idx;                      // stream index
    AVStream* _avstream;           // video stream
    AVCodecContext* _codecContext; // video codec context
    AVCodec* _videoCodec;
    AVFrame* _avFrame;             // decoding frame
    AVFrame* _avIntermediateFrame; // decode into this if an image conversion is required
    SwsContext* _convertCtx;
    std::unique_ptr<NukeCodecs::IDecoder> _color_convertor; //Nuke float math YUV to rgb convertor
    NukeCodecs::DecoderParams _csc_params;
    bool _resetConvertCtx;

    int _fpsNum;
    int _fpsDen;

    int64_t _startPTS;     // PTS of the first frame in the stream
    int64_t _startDTS;     // DTS of the first frame in the stream
    int64_t _frames;       // video duration in frames

    bool _ptsSeen;                      // True if a read AVPacket has ever contained a valid PTS during this stream's decode,
    // indicating that this stream does contain PTSs.
    int64_t AVPacket::*_timestampField; // Pointer to member of AVPacket from which timestamps are to be retrieved. Enables
    // fallback to using DTSs for a stream if PTSs turn out not to be available.

    int                 _width;
    int                 _height;
    double              _aspect;
    int                 _bitDepth;
    int                 _numberOfComponents;
    enum AVPixelFormat  _outputPixelFormat;
    uint8_t             _componentPosition[4];
    int                 _colorMatrixTypeOverride;       // Option to override the default YCbCr color matrix. 0 means no override (default).
    bool                _doNotAttachPrefix;
    bool                _matchMetaFormat;
    int                 _defaultDecodeVideoLevels;
    int                 _decodeVideoLevels;

#ifdef FN_LICENSED_PRORES_DNXHD_CODEC
    NukeCodecs::DecoderParams::eAlphaType _alphaChannelType = NukeCodecs::DecoderParams::kNoAlpha; // Only for DNxHD codec
#endif

    int _decodeNextFrameIn; // The 0-based index of the next frame to be fed into decode. Negative before any
    // frames have been decoded or when we've just seeked but not yet found a relevant frame. Equal to
    // frames_ when all available frames have been fed into decode.

    int _decodeNextFrameOut; // The 0-based index of the next frame expected out of decode. Negative before
    // any frames have been decoded or when we've just seeked but not yet found a relevant frame. Equal to
    // frames_ when all available frames have been output from decode.

    int _accumDecodeLatency; // The number of frames that have been input without any frame being output so far in this stream
    // since the last seek. This is part of a guard mechanism to detect when decode appears to have
    // stalled and ensure that FFmpegFile::decode() does not loop indefinitely.

    std::string _timecode;

    Stream()
      : _idx(0)
      , _avstream(nullptr)
      , _codecContext(nullptr)
      , _videoCodec(nullptr)
      , _avFrame(nullptr)
      , _avIntermediateFrame(nullptr)
      , _convertCtx(nullptr)
      , _resetConvertCtx(true)
      , _fpsNum(1)
      , _fpsDen(1)
      , _startPTS(0)
      , _startDTS(0)
      , _frames(0)
      , _ptsSeen(false)
      , _timestampField(&AVPacket::pts)
      , _width(0)
      , _height(0)
      , _aspect(1.0)
      , _bitDepth(8)
      , _numberOfComponents(3)
      , _outputPixelFormat(AV_PIX_FMT_RGB24)
      , _colorMatrixTypeOverride(0)
      , _doNotAttachPrefix(true)
      , _matchMetaFormat(true)
      , _defaultDecodeVideoLevels(false)
      , _decodeVideoLevels(false)
      , _decodeNextFrameIn(-1)
      , _decodeNextFrameOut(-1)
      , _accumDecodeLatency(0)
      , _timecode("00:00:00:00")
    {
      // The purpose of this is to avoid an RGB->RGB conversion.
      // This saves memory and improves performance. For example
      // Nuke expects RGBA component order. So for any RGB pixel
      // format which does not have this ordering, the following
      // should be used to remap the components, e.g. ARGB->RGBA,
      // BGRA->RGBA, BGR->RGB, etc. The output of the DNxHD decoder
      // is an example of this in action.
      _componentPosition[0] = 0; // Red position.
      _componentPosition[1] = 1; // Green position.
      _componentPosition[2] = 2; // Blue position.
      _componentPosition[3] = 3; // Alpha position.
    }

    ~Stream()
    {

      if (_avFrame)
        av_frame_free(&_avFrame);

      if (_avIntermediateFrame)
        av_frame_free(&_avIntermediateFrame);

      if (_codecContext)
        avcodec_free_context(&_codecContext);

      if (_convertCtx)
        sws_freeContext(_convertCtx);
    }

    static void destroy(Stream* s)
    {
      delete(s);
    }

    int64_t frameToDts(int frame) const
    {
      int64_t numerator = int64_t(frame) * _fpsDen *  _avstream->time_base.den;
      int64_t denominator = int64_t(_fpsNum) * _avstream->time_base.num;
      return _startDTS + (numerator / denominator);
    }

    int64_t frameToPts(int frame) const
    {
      int64_t numerator = int64_t(frame) * _fpsDen *  _avstream->time_base.den;
      int64_t denominator = int64_t(_fpsNum) * _avstream->time_base.num;
      return _startPTS + (numerator / denominator);
    }

    int ptsToFrame(int64_t pts) const
    {
      int64_t numerator = int64_t(pts - _startPTS) * _avstream->time_base.num * _fpsNum;
      int64_t denominator = int64_t(_avstream->time_base.den) * _fpsDen;
      return static_cast<int>(numerator / denominator);
    }

    bool isRec709Format()
    {
      // First check for codecs which require special handling:
      //  * JPEG codecs always use Rec 601.
      AVCodecID codecID = _codecContext->codec_id;

      if (codecID == AV_CODEC_ID_MJPEG || codecID == AV_CODEC_ID_MJPEGB || codecID == AV_CODEC_ID_MJPEGA)  {
        return false;
      }

      // Using method described in step 5 of QuickTimeCodecReader::setPreferredMetadata
      return (_height >= 720);
    }

    bool isYUV() const
    {
      // from swscale_internal.h
      const AVPixFmtDescriptor *desc = av_pix_fmt_desc_get(_codecContext->pix_fmt);
      return !(desc->flags & AV_PIX_FMT_FLAG_RGB) && desc->nb_components >= 2;
    }

    bool isCodecTypeDNx() const
    {
      return (AV_CODEC_ID_DNXHD == _codecContext->codec_id);
    }

    bool isCodecTypeDNxHR() const
    {
      return isCodecTypeDNx() && (FF_PROFILE_DNXHD != _codecContext->profile);
    }

    bool isCodecTypeDNxHD() const
    {
      return isCodecTypeDNx() && (FF_PROFILE_DNXHD == _codecContext->profile);
    }

#ifdef FN_LICENSED_PRORES_DNXHD_CODEC
    NukeCodecs::CodecFactory::CodecType getDNxHRCompressionId() const
    {
      switch (_codecContext->profile) {
      case FF_PROFILE_DNXHR_444:
        return NukeCodecs::CodecFactory::eDNxHRCodec1270;
      case FF_PROFILE_DNXHR_HQX:
        return NukeCodecs::CodecFactory::eDNxHRCodec1271;
      case FF_PROFILE_DNXHR_HQ:
        return NukeCodecs::CodecFactory::eDNxHRCodec1272;
      case FF_PROFILE_DNXHR_SQ:
        return NukeCodecs::CodecFactory::eDNxHRCodec1273;
      case FF_PROFILE_DNXHR_LB:
        return NukeCodecs::CodecFactory::eDNxHRCodec1274;
      }
      return NukeCodecs::CodecFactory::eUnknownCodec;
    }

    NukeCodecs::CodecFactory::CodecType getDNxCodecType() const;
#endif

    static double GetStreamAspectRatio(Stream* stream);

    //returns Nuke float math based YUV to RGB convertor
    //or nullptr if unsupported pixel format or scaling needed
    // |reset| forces recalculation of cached context.
    NukeCodecs::IDecoder * getColorConvertor(AVPixelFormat srcPixelFormat, int width, int height, int srcColorRange);
    NukeCodecs::DecoderParams & getColorConvertorParams() { return _csc_params; }
    // Generate the conversion context used by SoftWareScaler if not already set.
    // |reset| forces recalculation of cached context.
    SwsContext* getConvertCtx(AVPixelFormat srcPixelFormat, int srcWidth, int srcHeight, int srcColorRange, AVPixelFormat dstPixelFormat, int dstWidth, int dstHeight);

    // Return the number of input frames needed by this stream's codec before it can produce output. We expect to have to
    // wait this many frames to receive output; any more and a decode stall is detected.
    // In FFmpeg 2.1.4, I found that some codecs now support multithreaded decode which appears as latency
    // I needed to add the thread_count onto the codec delay - pickles
    int getCodecDelay() const
    {
      return ((_videoCodec->capabilities & AV_CODEC_CAP_DELAY) ? _codecContext->delay : 0)
        + _codecContext->has_b_frames
        + _codecContext->thread_count;
    }
  };

  struct FoundryAVPacket : public AVPacket
  {
  public:
    FoundryAVPacket()
    {
      InitPacket();
    }
    ~FoundryAVPacket()
    {
      FreePacket();
    }
      
    AVPacket * operator&() {
      FreePacket();
      return this;
    }
  public:
    void InitPacket()
    {
      data = nullptr;
      size = 0;

      av_init_packet(this);
    }
    void FreePacket()
    {
      av_packet_unref(this);
    }
  };


  // AV structure
  AVFormatContext* _context;
  AVInputFormat*   _format;

  // store all video streams available in the file
  std::vector<Stream*> _streams;
  
  // The stream selected as the one from which data will be read. The constructor sets this to the
  // viewIndex^th stream if there is one, otherwise the first (i.e. zeroth) stream.
  Stream* _selectedStream;

  // reader error state
  std::string _errorMsg;  // internal decoding error string
  bool _invalidState;     // true if the reader is in an invalid state

  // internal lock for multithread access
  DD::Image::Lock _lock;

#ifdef FN_LICENSED_PRORES_DNXHD_CODEC
  // Apple ProRes and Avid DNxHD codec specific.
  std::unique_ptr<NukeCodecs::IDecoder> _decoder = nullptr;
  NukeCodecs::CodecInfo _codecInfo;
#endif

  // set reader error
  void setError(const char* msg, const char* prefix = nullptr)
  {
    if (prefix) {
      _errorMsg = prefix;
      _errorMsg += msg;
    }
    else {
      _errorMsg = msg;
    }
    _invalidState = true;
  }

  // set FFmpeg library error
  void setInternalError(const int error, const char* prefix = nullptr)
  {
    char errorBuf[1024];
    av_strerror(error, errorBuf, sizeof(errorBuf));
    setError(errorBuf, prefix);
  }

  // get stream start time
  int64_t getStreamStartTime(Stream& stream);

  // Get the video stream duration in frames...
  int64_t getStreamFrames(Stream& stream);

public:

  typedef DD::Image::RefCountedPtr<FFmpegFile> Ptr;

  // The viewIndex argument specifies which video stream, i.e. 'track' in QuickTime terminology,
  // to use as the selected stream for decoding. This is intended for multiview support, hence
  // only those streams that have the same properties is the zeroth stream are considered as candidates
  // for the selected stream. If there are not enough candidate streams for the specified index then
  // the zeroth stream is used.
  FFmpegFile(const char* filename, int viewIndex);

  // destructor
  ~FFmpegFile() override;

  // get the internal error string
  const char* error() const
  {
    return _errorMsg.c_str();
  }

  // return true if the reader can't decode the frame
  bool invalid() const
  {
    return _invalidState;
  }
    

  // return the numbers of streams supported by the reader
  size_t streams() const
  {
    return _streams.size();
  }
  
  void setSelectedStream(int streamIndex);
  
  void setColorMatrixTypeOverride(int colorMatrixType) const
  {
    if (_streams.empty()) {
      return;
    }

    mFnAssert(_selectedStream);
    _selectedStream->_colorMatrixTypeOverride = colorMatrixType;
    _selectedStream->_resetConvertCtx = true;
  }
  
  void setDoNotAttachPrefix(bool doNotAttachPrefix) const
  {
    if (_streams.empty()) {
      return;
    }
    
    mFnAssert(_selectedStream);
    _selectedStream->_doNotAttachPrefix = doNotAttachPrefix;
  }

  void setMatchMetaFormat(bool matchMetaFormat) const
  {
    if (_streams.empty()) {
      return;
    }
    
    _selectedStream->_matchMetaFormat = matchMetaFormat;
  }

  bool getDefaultDecodeVideoLevels() const
  {
    if (_streams.empty()) {
      return false;
    }
    
    mFnAssert(_selectedStream);
    return _selectedStream->_defaultDecodeVideoLevels;
  }

  void setDecodeVideoLevels(int decodeVideoLevels) const
  {
    if (_streams.empty()) {
      return;
    }
    
    mFnAssert(_selectedStream);
    _selectedStream->_decodeVideoLevels = decodeVideoLevels;
  }
  
  bool isRec709Format() const
  {
    if (_streams.empty()) {
      return false;
    }

    mFnAssert(_selectedStream);
    return _selectedStream->isRec709Format();
  }

  bool isYUV() const
  {
    if (_streams.empty()) {
      return false;
    }

    mFnAssert(_selectedStream);
    return _selectedStream->isYUV();
  }

  bool isCodecTypeDNx() const
  {
    if (_streams.empty()) {
      return false;
    }

    mFnAssert(_selectedStream);
    return _selectedStream->isCodecTypeDNx();
  }

  int getBitDepth() const
  {
    if (_streams.empty()) {
      return 0;
    }
    
    mFnAssert(_selectedStream);

    // Sometimes AVCodec reports a bitdepth of 0 (eg PNG codec).
    // In this case, assume 8 bit.
    return std::max(8, _selectedStream->_bitDepth);
  }

  AVPixelFormat getOutputPixelFormat() const
  {
    if (_streams.empty()) {
      return AV_PIX_FMT_NONE;
    }

    mFnAssert(_selectedStream);

    return _selectedStream->_outputPixelFormat;
  }

  int getNumberOfComponents() const
  {
    if (_streams.empty()) {
      return 0;
    }

    mFnAssert(_selectedStream);
    return _selectedStream->_numberOfComponents;
  }

  int getComponentPosition(DD::Image::Channel channel) const
  {
    if (_streams.empty()) {
      return 0;
    }

    int componentIndex;
    if (DD::Image::Chan_Red == channel)
      componentIndex = 0;
    else if (DD::Image::Chan_Green == channel)
      componentIndex = 1;
    else if (DD::Image::Chan_Blue == channel)
      componentIndex = 2;
    else // Assume alpha.
      componentIndex = 3;

    mFnAssert(_selectedStream);
    return _selectedStream->_componentPosition[componentIndex];
  }

  int getComponentPosition(DD::Image::ChannelSetInit mask) const
  {
    if (_streams.empty()) {
      return 0;
    }

    int componentIndex;
    if (DD::Image::Mask_Red == mask)
      componentIndex = 0;
    else if (DD::Image::Mask_Green == mask)
      componentIndex = 1;
    else if (DD::Image::Mask_Blue == mask)
      componentIndex = 2;
    else // Assume alpha.
      componentIndex = 3;

    mFnAssert(_selectedStream);
    return _selectedStream->_componentPosition[componentIndex];
  }

  int getWidth() const
  {
    if (_streams.empty()) {
      return 0;
    }

    mFnAssert(_selectedStream);
    return _selectedStream->_width;
  }

  int getHeight() const
  {
    if (_streams.empty()) {
      return 0;
    }

    mFnAssert(_selectedStream);
    return _selectedStream->_height;
  }

  std::string getTimecodeForFrame(int frameIndex) const;
  bool canUseColorConvertor() const;

  // Decode a single frame of the selected stream into the buffer. Thread safe
  bool decode(unsigned char* buffer, unsigned frame);



  // fill the metada information of the selected stream.
  bool metadata(DD::Image::MetaData::Bundle& metadata);

  // Get basic info for the selected stream.
  bool info( int& width,
             int& height,
             double& aspect,
             int& frames);

  const char* getColorspace() const;

  int getBufferSize() const;
  int getRowSize() const;

private:

  bool seekToFrame(int64_t frame, int seekFlags);

  bool demuxAndDecode(AVFrame* avFrameOut, int64_t frame);

  int decodeProRes(AVFrame* avFrameOut, int& frameDecoded, AVPacket& avPacketIn);

  int decodeDNx(AVFrame* avFrameOut, int& frameDecoded, AVPacket& avPacketIn);

  int decodeLicensed(AVFrame* avFrameOut, int& frameDecoded, AVPacket& avPacketIn);

  bool imageConvert(AVFrame* avFrameIn, AVFrame* avFrameOut);

};

//////////////////////////////////////////////////////////////////////////////
// Keeps track of all FFmpegFile mapped against file name.
class FFmpegFileManager
{
  typedef std::pair<std::string, int> ReaderMapKey;
  typedef std::map<ReaderMapKey, FFmpegFile::Ptr> ReaderMap;

  ReaderMap _readers;

  // internal lock
  DD::Image::Lock _lock;

  // A lock manager function for FFmpeg, enabling it to use mutexes managed by this reader. Pass to av_lockmgr_register().
  static int FFmpegLockManager(void** mutex, enum AVLockOp op);

public:

  // singleton
  static FFmpegFileManager s_readerManager;

  // constructor
  FFmpegFileManager();

  //! Argument to get() to specify whether to potentially return an existing FFMpegFile object
  //! for sharing between mov64Reader instances.
  enum ShareMode
  {
    eAllowSharing = 0,
    eDontAllowSharing
  };
  
  //! Get an FFMpegFile instance for the specified file.
  //! If shareMode is eAllowSharing and a previous call to get() has already created an FFMpegFile
  //! for the specified file that that instance will be returned for sharing between the mov64Reader objects
  //! (assuming release() hasn't yet been called between the get() calls, of course). If shareMode is
  //! eDontAllowSharing then a new FFMpegFile object will be returned by each call to get().
  FFmpegFile::Ptr get(const char* filename, int viewIndex, ShareMode shareMode);

  // release a specific reader
  void release(const char* filename, int viewIndex);
};

/////////////////////////////////////////////////////////////////////////////

class mov64Reader : public DD::Image::Reader, public DD::Image::IBufferFillHeap
{
public:
  // If supplied, optionalFilename will be used over the iops FileKnob.
  // This allows the movReader to perform an early check. See invalid().
  explicit mov64Reader(DD::Image::Read* iop, const char* optionalFilename = nullptr);
  ~mov64Reader() override;

  // Expose protected function for use by movReader
  void prefetchMetaData() override;
  const DD::Image::MetaData::Bundle& fetchMetaData(const char* key) override { return meta; }
  bool supports_stereo() const override;
  bool fileStereo() const override;
  bool videosequence() const override { return true; }
  void engine(int y, int x, int r, DD::Image::ChannelMask channels, DD::Image::Row& row) override;
  void open() override;

  // redifined from Reader::getDefaultColorspace()
  std::string getDefaultColorspace() const override;

  // Planar reader extensions
  DD::Image::PlanarReadInfo planarReadInfo(const DD::Image::ChannelSet& channels) override;
  void planarReadAndDecode(DD::Image::GenericImagePlane& image,
                                   const DD::Image::ChannelSet& channels, int priority) override;
  int planarReadPass(void* buffer,
                             const DD::Image::ChannelSet& channels) override;
  void planarDecodePass(void* srcBuffer,
                                DD::Image::GenericImagePlane& image,
                                const DD::Image::ChannelSet& channels,
                                int threadIndex,
                                int nDecodeThreads,
                                int priority) override;

  // For movReader, to query the ability of mov64 to read a particular codec
  bool invalid();

private:

  // redefined from DD::Image::IBufferFill
  bool fillBuffer(void* data) override;

  void memoryInfo(DD::Image::Memory::MemoryInfoArray& output, const void* restrict_to) const override;

  void engine8( int y, int x, int r, DD::Image::ChannelMask channels, DD::Image::Row& row, unsigned char* data);
  void engine16(int y, int x, int r, DD::Image::ChannelMask channels, DD::Image::Row& row, U16* data);
  void engineFloat(int y, int x, int r, DD::Image::ChannelMask channels, DD::Image::Row& row, float* data);

private:

  FFmpegFile::Ptr _reader;
  DD::Image::MemoryBuffer _data;    // decoding buffer
  size_t _memNeeded;                // memory needed for decoding a single frame
  
  // The view index for which we're fetching data, i.e. for which our _reader is configured.
  // This starts at 0 for the 0th user-visible view, i.e. it excludes the hidden default view
  // of the OutputContext's legacy global views list, hence it is not necessarily quite the same
  // as our owning Read op's view_for_reader_ member.
  int _viewIndex;

  int  _numFrames;                  // number of frames in the selected stream
  int  _currentFrame;               // current decoding frame
  
  // Whether this instance is being used for reading for the NukeStudio/Hiero timeline (or possibly
  // for timeline/bin thumbnails), rather than for Nuke's node graph viewer (or any other regular
  // Nuke usage).
  bool _isTimelineReader;

  DD::Image::MetaData::Bundle meta;
  static const Reader::Description d;

  const char* const _optionalFilename;

  uint8_t* _planarReadBuffer;
};

/////////////////////////////////////////////////////////
class mov64ReaderFormat : public DD::Image::ReaderFormat
{
  friend class mov64Reader;

public:
  mov64ReaderFormat(DD::Image::Read* iop);

  void knobs(DD::Image::Knob_Callback c) override;
  void append(DD::Image::Hash& hash) override;
  void initDynamicKnobs() override;

  int  getColorMatrixType()   const { return _colorMatrixType; }
  bool getDoNotAttachPrefix() const { return _doNotAttachPrefix; }
  bool getMatchMetaFormat()   const { return _matchMetaFormat; }
  int  getDecodeVideoLevels() const { return _decodeVideoLevels; }
  int  getSelectedDecoder()   const { return _decoder; }
  void setDefaultColorMatrixType(int colorMatrixType) { _defaultColorMatrixType = colorMatrixType; }
  void setYCbCrMatrixEnabled(bool enabled);
  void setDefaultDecodeVideoLevels(int decodeVideoLevels) { _defaultDecodeVideoLevels = decodeVideoLevels; }
  bool getFirstTrackOnly() const { return _firstTrackOnly; }
  void setKnobsVisibility(bool visible);
  void updateCodecNameTextKnob(std::string codecName);

private:
  void updateColorMatrixKnob(DD::Image::Knob* colorMatrixKnob);
  void updateDecodeVideoLevelsKnob(DD::Image::Knob* knob);

private:
  DD::Image::Read* _iop;
  int _decoder;
  int _colorMatrixType;
  int _defaultColorMatrixType;
  bool _doNotAttachPrefix;
  bool _matchMetaFormat;
  bool _isDNxSdkVersionVisible;

  // DNxHD specific.
  int _decodeVideoLevels;
  int _defaultDecodeVideoLevels;
  
  // When true, causes all but the first video track/stream to be ignored.
  bool _firstTrackOnly;

  // Based on Mov32Reader
  static const int kNumDefaultLabels = 1;
  static const int kNumListDelimiters = 1;
  static const int kNumColorMatrixLabels = kNumDefaultLabels + eNumYCbCrEncodings + kNumListDelimiters;
  const char* _colourMatrixTypeLabels[kNumColorMatrixLabels];

  static const int kNumDecodeVideoLevelsLabels = kNumDefaultLabels + eNumDecodeVideoLevels + kNumListDelimiters;
  const char* _decodeVideoLevelsTypeLabels[kNumDecodeVideoLevelsLabels];
};

#endif // NUKE_MOV64READER_H

