// Copyright (c) 2014 The Foundry Visionmongers Ltd.  All Rights Reserved.
#ifdef _WIN32
#define _CRT_NONSTDC_NO_DEPRECATE // Warnings complaining about the use of POSIX complient but non-standard C methods.
#define _CRT_SECURE_NO_WARNINGS   // Microsoft specific CRT warnings most of which have no real cross-platform solution.
#pragma warning(disable: 4251)    // Supress warnings from DD::Image - class needs to have dll-interface to be used by clients of class 'DD::Image::...'
#pragma warning(disable: 4996)    // We cannot control the Windows OS version checking in this plugin.
#endif

#include "Build/fnBuild.h"

#ifndef FN_EXAMPLE_PLUGIN
#include "Base/fnTimecode.h" // FIXME: replace with shippable time code handling
#endif // ndef FN_EXAMPLE_PLUGIN

#include "DDImage/Version.h"
#include "DDImage/ddImageVersion.h"
#include "DDImage/DDString.h"
#include "DDImage/Row.h"
#include "DDImage/Knobs.h"
#include "DDImage/LUT.h"
#include "DDImage/Enumeration_KnobI.h"

#include "movWriter.h"
#include "codecWhitelist.h"
#include "codecCommon.h"

#ifndef INT64_C
# define INT64_C(c) (c ## LL)
#endif
#ifndef UINT64_C
# define UINT64_C(c) (c ## ULL)
#endif

#include <assert.h>
#include <sstream>
#include <string>
#include <cstring>

#include <boost/format.hpp>

#include "AvidDNxCodec.h"
#if PLUGIN_UPDATE
# include "version.h"
#endif

#define FN_MOV64WRITER_PRINT_CODECS 0

/**
  * @deprecated there is no libavcodec-wide limit on the number of B-frames
  */
#define FF_MAX_B_FRAMES 16

using namespace DD::Image;
using namespace boost;

namespace {
  std::string DNxHR_SDK_Version()
  {
    std::stringstream versionNumber;
    versionNumber
      << "Avid DNxCodec "
      << DNX_VERSION_MAJOR << "."
      << DNX_VERSION_MINOR << "."
      << DNX_VERSION_PATCH
      ;
    return versionNumber.str();
  }

  constexpr const char* kPixelFormatKnob = "mov64_pixel_format";
}

////////////////////////////////////////////////////////////////////////////////

# define SIZEOF_ARRAY(array) (sizeof(array) / sizeof(array[0]))

////////////////////////////////////////////////////////////////////////////////

// Constants
static const char* const kVideoRangeStr = "1";
static const char* const kFullRangeStr = "2";

static const char* kEncodeVideoQualityLabels[] = {
  "Best", "High", "Medium", "Low", "Least", "Custom", nullptr
};

static const int kDefaultBitrate = 200000;
static const int kDefaultBitrateTolerance = 4000000;

static const std::map<std::string, std::map<QualityLevel, int>> kMinimumQuantizeThresholdPerCodec = {
  {"mpeg1video",  {{eQualityLevel_Least, 12}, {eQualityLevel_Low, 6}, {eQualityLevel_Medium, 3}, {eQualityLevel_High, 1}, {eQualityLevel_Best, 1}}},
  {"mpeg4",       {{eQualityLevel_Least, 12}, {eQualityLevel_Low, 6}, {eQualityLevel_Medium, 3}, {eQualityLevel_High, 1}, {eQualityLevel_Best, 1}}},
  {"mjpega",      {{eQualityLevel_Least, 12}, {eQualityLevel_Low, 6}, {eQualityLevel_Medium, 3}, {eQualityLevel_High, 1}, {eQualityLevel_Best, 1}}},
  {"mjpegb",      {{eQualityLevel_Least, 12}, {eQualityLevel_Low, 6}, {eQualityLevel_Medium, 3}, {eQualityLevel_High, 1}, {eQualityLevel_Best, 1}}},  
  {"mjpeg",       {{eQualityLevel_Least, 12}, {eQualityLevel_Low, 6}, {eQualityLevel_Medium, 3}, {eQualityLevel_High, 1}, {eQualityLevel_Best, 1}}},
  {"fn_h264",     {{eQualityLevel_Least, 12}, {eQualityLevel_Low, 6}, {eQualityLevel_Medium, 3}, {eQualityLevel_High, 1}, {eQualityLevel_Best, 1}}}
};

static const std::map<std::string, std::map<QualityLevel, int>> kMaximumQuantizeThresholdPerCodec = {
  {"mpeg1video",  {{eQualityLevel_Least, 23}, {eQualityLevel_Low, 15}, {eQualityLevel_Medium, 7}, {eQualityLevel_High, 3}, {eQualityLevel_Best, 1}}},
  {"mpeg4",       {{eQualityLevel_Least, 23}, {eQualityLevel_Low, 15}, {eQualityLevel_Medium, 7}, {eQualityLevel_High, 3}, {eQualityLevel_Best, 1}}},
  {"mjpega",      {{eQualityLevel_Least, 23}, {eQualityLevel_Low, 15}, {eQualityLevel_Medium, 7}, {eQualityLevel_High, 3}, {eQualityLevel_Best, 1}}},
  {"mjpegb",      {{eQualityLevel_Least, 23}, {eQualityLevel_Low, 15}, {eQualityLevel_Medium, 7}, {eQualityLevel_High, 3}, {eQualityLevel_Best, 1}}},  
  {"mjpeg",       {{eQualityLevel_Least, 23}, {eQualityLevel_Low, 15}, {eQualityLevel_Medium, 7}, {eQualityLevel_High, 3}, {eQualityLevel_Best, 1}}},
  {"fn_h264",     {{eQualityLevel_Least, 23}, {eQualityLevel_Low, 15}, {eQualityLevel_Medium, 7}, {eQualityLevel_High, 3}, {eQualityLevel_Best, 1}}},
};

static const std::map<QualityLevel, uint32_t> kH264FullHDBitrate = {
  {eQualityLevel_Least,  4000},
  {eQualityLevel_Low,    11000},
  {eQualityLevel_Medium, 18000},
  {eQualityLevel_High,   28000},
  {eQualityLevel_Best,   50000},
};
//

////////////////////////////////////////////////////////////////////////////////
// Apple ProRes and Avid DNxHD codec support.
// This is a brief guide to the changes for adding Apple ProRes and Avid DNxHD
// encoding to the mov64 writer.
// * Uses existing mov64 writer configuration code.
// * When ProRes or DNxHD are selected as the codec, these vendor provided
//   libraries will be used to encode the uncompressed source frame.
// * The ProRes or DNxHD encoded frame is then written to the selected file type
//   in the same way as any other codec.
//
// Encoding data involves converting from floating point RGB (Nuke) to the input
// pixel format of the encoding, then encoding and finally writing the encoded
// bit stream to the file.
//
////////////////////////////////////////////////////////////////////////////////
// LicensedCodecStr
// Define a long name, short name, knob label and codec tag (four character
// code) for the licensed codecs.
// ProRes has unique codec tags (four character codes) for each codec profile
// whereas DNxHD has a single codec tag to alias all codec profiles. This means
// that a separate codec profile knob is needed for DNxHD to select the desired
// target bit rate.
struct LicensedCodecStr
{
  const char* longName; // User readable description of the codec.
  const char* shortName; // Short name for faster internal parsing
  const char* knobLabel; // Knob label in the format "fourcc\tUserLabel"
#if defined(FN_LICENSED_PRORES_DNXHD_CODEC)
  NukeCodecs::CodecFactory::CodecType codecTag; // The ProRes/DNxHD encoder profile (Four Character Code). This is also the FCC used to identify the codec profile in the QuickTime movie.
#else
  int codecTag;
#endif // defined(FN_LICENSED_PRORES_DNXHD_CODEC)
};
typedef struct LicensedCodecStr LicensedCodecStr;

#if defined(FN_LICENSED_PRORES_DNXHD_CODEC)
#define mFnCodecTag(a) a
#else
#define mFnCodecTag(a) 0
#endif

#define MOV64_PRORES_CODEC_NAME "appr\t" APPLE_PRORES_NAME

static const LicensedCodecStr licensedCodecLUT[] = {
  { APPLE_PRORES_4444XQ_NAME,       "fn_ap4x",   "ap4x\t\a",  mFnCodecTag(NukeCodecs::CodecFactory::eProResCodec4444XQ) },
  { APPLE_PRORES_4444_NAME,         "fn_ap4h",   "ap4h\t\a",  mFnCodecTag(NukeCodecs::CodecFactory::eProResCodec4444) },
  { APPLE_PRORES_422HQ_NAME,        "fn_apch",   "apch\t\a",  mFnCodecTag(NukeCodecs::CodecFactory::eProResCodec422HQ) },
  { APPLE_PRORES_422_NAME,          "fn_apcn",   "apcn\t\a",  mFnCodecTag(NukeCodecs::CodecFactory::eProResCodec422) },
  { APPLE_PRORES_422LT_NAME,        "fn_apcs",   "apcs\t\a",  mFnCodecTag(NukeCodecs::CodecFactory::eProResCodec422LT) },
  { APPLE_PRORES_422PROXY_NAME,     "fn_apco",   "apco\t\a",  mFnCodecTag(NukeCodecs::CodecFactory::eProResCodec422Proxy) },
  { APPLE_PRORES_NAME, "fn_prores", MOV64_PRORES_CODEC_NAME,  mFnCodecTag(NukeCodecs::CodecFactory::eProResCodec4444XQ) },
  { AVID_DNXHD_NAME,   "fn_avdn",   "AVdn\t" AVID_DNXHD_NAME, mFnCodecTag(NukeCodecs::CodecFactory::eDNxHDCodecAVdn) },
  { AVID_DNXHR_NAME,   "fn_avdh",   "AVdh\t" AVID_DNXHR_NAME, mFnCodecTag(NukeCodecs::CodecFactory::eDNxHRCodecAVdh) },
  { CODEC_H264_NAME,   "fn_h264",   "h264\t" CODEC_H264_NAME, mFnCodecTag(NukeCodecs::CodecFactory::eH264Codec) },
};

static const NukeCodecs::CodecFactory::_CodecType proresCodecTags[] = {
    mFnCodecTag(NukeCodecs::CodecFactory::eProResCodec4444XQ),
    mFnCodecTag(NukeCodecs::CodecFactory::eProResCodec4444),
    mFnCodecTag(NukeCodecs::CodecFactory::eProResCodec422HQ),
    mFnCodecTag(NukeCodecs::CodecFactory::eProResCodec422),
    mFnCodecTag(NukeCodecs::CodecFactory::eProResCodec422LT),
    mFnCodecTag(NukeCodecs::CodecFactory::eProResCodec422Proxy),
};


static const MovPixelFormatArray kPixelFormatDesc = {
  {"yuvj420p",      "YCbCr 4:2:0 8-bit",      12, "Planar YUV 4:2:0, 12bpp, full range"},
  {"yuvj422p",      "YCbCr 4:2:2 8-bit",      16, "Planar YUV 4:2:2, 16bpp, full range"},
  {"yuvj444p",      "YCbCr 4:4:4 8-bit",      24, "Planar YUV 4:4:4, 24bpp, full range"},
  {"yuv420p",       "YCbCr 4:2:0 8-bit",      12, "Planar YUV 4:2:0, 12bpp"},
  {"yuv422p10le",   "YCbCr 4:2:2 10-bit",     20, "Planar YUV 4:2:2, 20bpp, (1 Cr & Cb sample per 2x1 Y samples), little-endian"},
  {"rgb48be",       "RGB 16-bit",             48, "Packed RGB 16:16:16, 48bpp, 16R, 16G, 16B, The 2-byte value for each R/G/B component is stored as big-endian"},
  {"rgba64be",      "RGBA 16-bit",            64, "Packed RGBA 16:16:16:16, 64bpp, 16R, 16G, 16B, 16A, The 2-byte value for each R/G/B/A component is stored as big-endian"},
  {"rgb24",         "RGB 8-bit",              24, "Packed RGB 8:8:8, 24bpp, RGBRGB"},
  {"rgba",          "RGBA 8-bit",             32, "Packed RGBA 8:8:8:8, 32bpp, RGBARGBA"},
  {"argb",          "RGBA 8-bit",             32, "Packed ARGB 8:8:8:8, 32bpp, ARGBARGB"},

  {"yuv422p",       "YCbCr 4:2:2 8-bit",      16, "Planar YUV 4:2:2, 16bpp"},
  {"rgb555be",      "RGB555",                 16, "Packed RGB 5:5:5, 16bpp, (msb)1X 5R 5G 5B(lsb), big-endian, X=unused"},
  {"gray",          "Gray",                   8,  "Y, 8bpp"},
  {"ya8",           "Gray + Alpha",           16, "8 bits gray (Y), 8 bits alpha"},
  {"gray16be",      "Gray 16-bit",            16, "Y, 16bpp, big-endian"},
  {"ya16be",        "Gray + Alpha 16-bit",    32, "16 bits gray, 16 bits alpha, big-endian"},
  {"monob",         "Monochrome Black",       1,  "Y, 1bpp, 0 is black, 1 is white, \nIn each byte pixels are ordered from the msb to the lsb"}
};


// Helper Functions
namespace {

  std::string PixelFormatDescFrom(const MovPixelFormatName& pixFmtName)
  {
    auto descIt = std::find_if(kPixelFormatDesc.begin(), kPixelFormatDesc.end(), 
      [&pixFmtName](const MovPixelFormat& desc) { return (desc.ffmpegName == pixFmtName); }
    );

    assert(descIt != kPixelFormatDesc.end());

    return descIt->description;
  }


  bool PixelFormatFromFFmpegStr(const MovPixelFormatName& ffmpegPixFmt, MovPixelFormat& pixFmt)
  {
    auto descIt = std::find_if(kPixelFormatDesc.begin(), kPixelFormatDesc.end(), 
      [&ffmpegPixFmt](const MovPixelFormat& desc) { return (desc.ffmpegName == ffmpegPixFmt); }
    );

    if(descIt == kPixelFormatDesc.end()) {
      return false;
    }

    pixFmt = *descIt;

    return true;
  }

  // A whitelist of pixel formats to display
  const std::string kPixelFormatWhitelist[] = {
    "rgb24",    "rgb48be",
    "argb",     "rgba",         "rgba64be",
    "yuv420p",  "yuv422p10le",
    "yuvj444p", "yuvj422p",     "yuvj420p"
  };

  bool hidePixelFormat(AVCodecID codecId, const MovPixelFormatName& name)
  {
    // Uncompressed 10-bit 4:2:2 may only use yuv422p10le
    if (codecId == AV_CODEC_ID_V210)
      return name != "yuv422p10le";

    // Hide if not in the whitelist
    auto it = std::find(std::begin(kPixelFormatWhitelist),
                        std::end(kPixelFormatWhitelist),
                        name);

    return  it == std::end(kPixelFormatWhitelist);
  }


  MovPixelFormatArray PixelFormatsFromAVCodec(const AVCodec* codec)
  {
    MovPixelFormatArray pixelFormats;
    if (codec) {
      const AVPixelFormat* pixfmts = codec->pix_fmts;

      for (; *pixfmts != AV_PIX_FMT_NONE; pixfmts++) {
        MovPixelFormatName pixFmtName = av_get_pix_fmt_name(*pixfmts);
        MovPixelFormat pixFmt;
        if (PixelFormatFromFFmpegStr(pixFmtName, pixFmt)) {
          if (hidePixelFormat(codec->id, pixFmtName)) {
            pixFmt.friendlyName = "\a";
          }
          pixelFormats.push_back(pixFmt);
        }
      }
    }

    return pixelFormats;
  }


  MovPixelFormatName PixelFormatFriendlyNameFrom(const MovPixelFormat& pixFmt)
  {
    return MovPixelFormatName(pixFmt.ffmpegName + "\t" + pixFmt.friendlyName);
  }


  MovPixelFormatNames PixelFormatFriendlyNamesFrom(const MovPixelFormatArray& pixFmts)
  {
    MovPixelFormatNames names;
    names.reserve(pixFmts.size());

    std::transform(
      pixFmts.begin(), pixFmts.end(),
      std::back_inserter(names),
      [](const auto pixFmt) { return PixelFormatFriendlyNameFrom(pixFmt); }
    );

    return names;
  }

  // Knob labels in the format "fourcc\tUserLabel"
  // We want to ignore the fourcc during sorting.
  int compareCodecKnobLabels(const char* a, const char* b)
  {
    // Strip the fourcc codes
    const char* aUserLabel = std::strchr(a, '\t');
    aUserLabel = aUserLabel ? aUserLabel+1 : a;
    const char* bUserLabel = std::strchr(b, '\t');
    bUserLabel = bUserLabel ? bUserLabel+1 : b;

    return strcasecmp(aUserLabel, bUserLabel);
  }

  bool IsLegacyProres(const char* short_name) {
    return strcmp(short_name, "fn_ap4x") == 0 ||
           strcmp(short_name, "fn_ap4h") == 0 ||
           strcmp(short_name, "fn_apch") == 0 ||
           strcmp(short_name, "fn_apcn") == 0 ||
           strcmp(short_name, "fn_apcs") == 0 ||
           strcmp(short_name, "fn_apco") == 0;
  }

  bool IsProres(const char* short_name) {
    return strcmp(short_name, "fn_prores") == 0;
  }

  bool IsDNxHR(const char* short_name) {
    return strcmp(short_name, "fn_avdh") == 0;
  }

  bool IsDNxHD(const char* short_name) {
    return strcmp(short_name, "fn_avdn") == 0;
  }
}




////////////////////////////////////////////////////////////////////////////////
// FnAVFrame
// Wrapper class for the ffmpeg AVFrame structure.
// In order negate the chances of memory leaks when using ffmpeg this class
// provides the initialisation, allocation and release of memory associated
// with an AVFrame structure.
//
// Note that this has been designed to be a drop in replacement for host
// managed memory for colourspace, pixel format and sample format conversions.
// It is not designed to be used with avcodec_decode_video2 as the underlying
// decoder usually manages memory.
//
// Example usage:
//
// Audio
//
//    FnAVFrame avFrame;
//    ret = avFrame.alloc(channels, nb_samples, _avSampleFormat, 1);
//    :
//    ret = swr_convert(_swrContext, avFrame->data, ...);
//    :
//    ret = avcodec_encode_audio2(avCodecContext, &pkt, avFrame, &gotPacket);
//    :
//
// Video
//
//    FnAVFrame avFrame;
//    ret = avFrame.alloc(width(), height(), pixelFormatCodec, 1);
//    if (!ret) {
//      :
//      sws_scale(convertCtx, ..., avFrame->data, avFrame->linesize);
//      :
//    }
//
// IMPORTANT
// This class has been purposefully designed NOT to have parameterised
// constructors or assignment operators. The reason for this is due to the
// complexity of managing the lifetime of the structure and its associated
// memory buffers.
//
FnAVFrame::~FnAVFrame()
{
  // The most important part of this class.
  // Two deallocations may be required, one from the AVFrame structure
  // and one for any image data, AVFrame::data.
  deallocateAVFrameData();
  av_frame_free(&_avFrame);
}

////////////////////////////////////////////////////////////////////////////////
// alloc
// VIDEO SPECIFIC.
// Allocate a buffer or buffers for the AVFrame structure. How many
// buffers depends upon |avPixelFormat| which is a VIDEO format.
//
// @param width Frame width in pixels.
// @param height Frame height in pixels.
// @param avPixelFormat An AVPixelFormat enumeration for the frame pixel format.
// @param align Buffer byte alignment.
//
// @return 0 if successful.
//         <0 otherwise.
//
int FnAVFrame::alloc(int width, int height, enum AVPixelFormat avPixelFormat, int align)
{
  int ret = 0;
  if (!_avFrame)
    _avFrame = av_frame_alloc();
  if (_avFrame) {
    deallocateAVFrameData(); // In case this method is called multiple times on the same object.
    // av_image_alloc will return the size of the buffer in bytes if successful,
    // otherwise it will return <0.
    int bufferSize = av_image_alloc(_avFrame->data, _avFrame->linesize, width, height, avPixelFormat, align);
    if (bufferSize > 0) {
      // Set the frame fields for a video buffer as some
      // encoders rely on them, e.g. Lossless JPEG.
      _avFrame->width = width;
      _avFrame->height = height;
      _avFrame->format = (int)avPixelFormat;
      ret = 0;
    }
    else {
      ret = -1;
    }
  }
  else {
    // Failed to allocate an AVFrame.
    ret = -2;
  }
  return ret;
}

////////////////////////////////////////////////////////////////////////////////
// alloc
// AUDIO SPECIFIC.
// Allocate a buffer or buffers for the AVFrame structure. How many
// buffers depends upon |avSampleFormat| which is an AUDIO format.
//
// @param nbChannels The number of audio channels.
// @param nbSamples The number of audio samples that the buffer will hold.
// @param avSampleFormat An AVSampleFormat enumeration for the audio format.
// @param align Buffer byte alignment.
//
// @return 0 if successful.
//         <0 otherwise.
//
int FnAVFrame::alloc(int nbChannels, int nbSamples, enum AVSampleFormat avSampleFormat, int align)
{
  int ret = 0;
  if (!_avFrame)
    _avFrame = av_frame_alloc();
  if (_avFrame) {
    deallocateAVFrameData(); // In case this method is called multiple times on the same object.
    // av_samples_alloc will return >= if successful, otherwise it will return <0.
    ret = av_samples_alloc(_avFrame->data, _avFrame->linesize, nbChannels, nbSamples, avSampleFormat, align);
    if (ret >= 0) {
      // Set the frame fields for an audio buffer as some
      // encoders rely on them.
      _avFrame->nb_samples = nbSamples;
      _avFrame->format = (int)avSampleFormat;
      ret = 0;
    }
    else {
      ret = -1;
    }
  }
  else {
    // Failed to allocate an AVFrame.
    ret = -2;
  }
  return ret;
}

// Release any memory allocated to the data member variable of
// the AVFrame structure.
void FnAVFrame::deallocateAVFrameData()
{
  if (_avFrame)
    av_freep(_avFrame->data);
}


////////////////////////////////////////////////////////////////////////////////
// AudioReader
// Class that uses the avformat, avcodec, avutil and swresample libraries to
// read, decode and reformat audio data from a file that contains audio.
//
// This class is used if an optional audio file has been specified to be used
// with the mov64Writer. The audio data is interleaved with the video frames
// by the mov64Writer.
//
// The mov64Writer can specify an 'offset' for the audio in order to advance
// or delay the audio with respect to the video. If the audio is delayed then
// the reader will create silence, keep track of the read position and provide
// 'real' audio at the appropriate point. If the audio is advanced then a
// seek operation is performed into the file and reads will begin from that
// position.
//
int AudioReader::open(std::string filename)
{
  int ret = avformat_open_input(&_avFormatContext, filename.c_str(), nullptr, nullptr);
  if (!ret) {
    // Dump file meta data (looking for timecode).
    // Can also do this on each stream.
    //AVDictionaryEntry* tag = 0;
    //while ((tag = av_dict_get(_avFormatContext->metadata, "", tag, AV_DICT_IGNORE_SUFFIX)))
    //  std::cout << tag->key << "=" << tag->value << std::endl;

    ret = avformat_find_stream_info(_avFormatContext, nullptr);
    if (!ret) {
      // Dump information about file onto standard error
      //av_dump_format(_avFormatContext, 0, filename.c_str(), 0);

      ret = av_find_best_stream(_avFormatContext, AVMEDIA_TYPE_AUDIO, -1, -1, nullptr, 0);
      if (ret >= 0) {
        int streamIndex = ret;
        _avStreamAudio.stream = _avFormatContext->streams[streamIndex];
        
        AVCodec* avCodec = avcodec_find_decoder(_avStreamAudio.stream->codecpar->codec_id);
        _avStreamAudio.codecContext = avcodec_alloc_context3(avCodec);
      
        avcodec_parameters_to_context(_avStreamAudio.codecContext, _avStreamAudio.stream->codecpar);

        if (avCodec) {
          _avFrameAudio = av_frame_alloc();
          ret = avcodec_open2(_avStreamAudio.codecContext, avCodec, nullptr);
          if (!ret) {
            // Update the sample format and codec ID that this reader is using
            _avSampleFormat = _avStreamAudio.codecContext->sample_fmt;
            _codecID = _avStreamAudio.codecContext->codec_id; 
          } 
          else  {
            DEBUG_AUDIO_READER("failed to open decoder for codec_id " << _avStreamAudio.codecContext->codec_id);
          }
        }
        else {
          DEBUG_AUDIO_READER("failed to find decoder for codec_id " << _avStreamAudio.codecContext->codec_id);
          ret = -1;
        }
      }
      else {
        DEBUG_AUDIO_READER("unable to find audio stream in file");
      }
    }
  }
  return ret;
}

////////////////////////////////////////////////////////////////////////////////
// read
// Read, decode and resample some audio data from the source file. The format
// of the data will be in the desired target format.
//
// @param avFrame A reference to an FnAVFrame object. This is an input and
//                output parameter. An UNINITIALIZED FnAVFrame is passed to
//                this method which allocs an appropriately sized buffer. The
//                audio is read and converted into this buffer which the caller
//                can then use. When the FnAVFrame object is destroyed it will
//                release the buffer.
//
// @return <0 if unsuccessful, otherwise the number of samples read.
int AudioReader::read(FnAVFrame& avFrame)
{
  int ret = 0;

  if (_avFormatContext && _avStreamAudio.stream) {
    bool readFrame = true;
    while (readFrame) {

      AVCodecContext* avCodecContext = _avStreamAudio.codecContext;
      if (0 > _readPosition) {
        // The audio has been delayed so read from the silence
        // buffer.

        // Convert from stream time to a sample count.
        int nbSamples = (int)(avCodecContext->sample_rate * (0 - _readPosition) * av_q2d(_avStreamAudio.stream->time_base));
        // The QuickTime MOV guidelines from Apple recommend writing
        // audio in 0.5s chunks. So If there is more than 0.5s of
        // silence then then provide 0.5s of audio. Otherwise provide
        // the remaining silence before the start of the audio file.
        if (nbSamples > (avCodecContext->sample_rate >> 1))
          nbSamples = avCodecContext->sample_rate >> 1;
        ret = avFrame.alloc(avCodecContext->channels, nbSamples, _avSampleFormat, 1);
        if (!ret) {
          ret = av_samples_set_silence(avFrame->data, // audio buffer
                                       0, // offset
                                       nbSamples,
                                       avCodecContext->channels,
                                       _avSampleFormat);
          if (!ret) {
            // Update the read position. Convert the sample count into a stream time.
            _readPosition += (int64_t)((double)nbSamples / (avCodecContext->sample_rate * av_q2d(_avStreamAudio.stream->time_base)));
            // Return the number of samples in the buffer of silence.
            ret = nbSamples;
            readFrame = false;
          }
        }
      }
      else {
        AVPacket avPacket = {nullptr};
        av_init_packet(&avPacket);
        ret = av_read_frame(_avFormatContext, &avPacket);
        if (ret == 0) { //no errors
          if (avPacket.stream_index == _avStreamAudio.stream->index) {
            // Decode audio frame.
            int haveDecodedFrame = 0;
            ret = mov64_av_decode(avCodecContext, _avFrameAudio, &haveDecodedFrame, avPacket);
            if (ret >= 0) {
              if (haveDecodedFrame) {
                readFrame = false;

                ret = avFrame.alloc(avCodecContext->channels, _avFrameAudio->nb_samples, _avSampleFormat, 1);
                if (!ret) {
                  if (_swrContext) {
                    // Convert from the source sample format to the target
                    // sample format.
                    ret = swr_convert(_swrContext,
                                      avFrame->data, // out
                                      _avFrameAudio->nb_samples, // out count
                                      const_cast<const uint8_t**>(_avFrameAudio->extended_data), // in
                                      _avFrameAudio->nb_samples); // in count
                  }
                  else {
                    // No format conversion is required so perform a straight copy.
                    ret = av_samples_copy(avFrame->data, // dst
                                          _avFrameAudio->extended_data, // src
                                          0, // dst_offset
                                          0, // src_offset
                                          _avFrameAudio->nb_samples, // nb_samples
                                          avCodecContext->channels, // nb_channels
                                          _avSampleFormat); // sample_fmt
                  }
                  // Update the read position. Convert the sample count into a stream time.
                  _readPosition += (int64_t)(_avFrameAudio->nb_samples * av_q2d(_avStreamAudio.stream->time_base));
                  // Return the number of samples in the buffer.
                  ret = _avFrameAudio->nb_samples;
                }
                else {
                  DEBUG_AUDIO_READER("failed to allocate sample conversion buffer");
                  ret = -3;
                }
              }
            }
            else {
              DEBUG_AUDIO_READER("error decoding audio frame");
              ret = -2;
            }
          }
        } else if ( ret < 0 ) {
          if ( ret == AVERROR_EOF ) {
            mFnAssertMsg( false, "We shouldn't be here as we stop reading after the last packets of the input stream" );
          }
          char szError[1024];
          av_strerror(ret, szError, 1024);
          DEBUG_AUDIO_READER(szError);
        }
        else /*if ( ret > 0 )*/ {
          mFnAssertMsg( false, "Unhandled positive return value from av_read_frame" );
        }
        av_packet_unref(&avPacket);
      }

      if (ret < 0)
        readFrame = false;
    }
  }
  else {
    DEBUG_AUDIO_READER("invalid contexts, call AudioReader::open");
    ret = -1;
  }

  return ret;
}

void AudioReader::close()
{
  _readPosition = 0;

  if (_swrContext) {
    swr_free(&_swrContext);
    _swrContext = nullptr;
  }

  if (_avFrameAudio) {
    av_frame_free(&_avFrameAudio);
    _avFrameAudio = nullptr;
  }

  if (_avStreamAudio.codecContext)
  {
    avcodec_close(_avStreamAudio.codecContext);
    // _avStreamAudio = nullptr;
  }

  if (_avFormatContext)
  {
    avformat_close_input(&_avFormatContext);
    _avFormatContext = nullptr;
  }
}

int AudioReader::setSampleFormat(AVSampleFormat avSampleFormat)
{
  int ret = 0;

  _avSampleFormat = avSampleFormat; // Set the target sample format.

  // If there is a valid audio codec context, reconfigure the
  // resampling context by comparing the source format to
  // the desired target format.
  if (_avStreamAudio.codecContext) {
    AVCodecContext* avCodecContext = _avStreamAudio.codecContext;
    // For an input format that does not match the output format,
    // allocate a resample context to perform the conversion.
    if (avCodecContext->sample_fmt != _avSampleFormat) {
      _swrContext = swr_alloc();
      if (_swrContext) {
        // Configure the resample context to convert from the
        // source sample rate and format to the desired
        // target sample rate and format.
        av_opt_set_int(_swrContext, "in_channel_count", avCodecContext->channels, 0);
        av_opt_set_int(_swrContext, "out_channel_count", avCodecContext->channels, 0);
        av_opt_set_channel_layout(_swrContext, "in_channel_layout", av_get_default_channel_layout(avCodecContext->channels), 0);
        av_opt_set_channel_layout(_swrContext, "out_channel_layout", av_get_default_channel_layout(avCodecContext->channels), 0);
        av_opt_set_int(_swrContext, "in_sample_rate", avCodecContext->sample_rate, 0);
        av_opt_set_int(_swrContext, "out_sample_rate", avCodecContext->sample_rate, 0);
        av_opt_set_sample_fmt(_swrContext, "in_sample_fmt", avCodecContext->sample_fmt, 0);
        av_opt_set_sample_fmt(_swrContext, "out_sample_fmt", _avSampleFormat, 0);
        ret = swr_init(_swrContext);
        if (ret) {
          DEBUG_AUDIO_READER("error initialising context");
          ret = -3;
        }
      }
      else {
        DEBUG_AUDIO_READER("failed to allocate context");
        ret = -2;
      }
    }
    else {
      swr_free(&_swrContext);
      _swrContext = nullptr;
    }
  }

  return ret;
}

////////////////////////////////////////////////////////////////////////////////
// getSampleRate
// Return the sample rate of the audio in the source file.
//
// @return The sample rate if a file has been successfully opened and contains
//         audio. Zero otherwise.
//
int AudioReader::getSampleRate()
{
  int sampleRate = 0;
  if (_avStreamAudio.codecContext) {
    AVCodecContext* avCodecContext = _avStreamAudio.codecContext;
    sampleRate = avCodecContext->sample_rate;
  }
  return sampleRate;
}

////////////////////////////////////////////////////////////////////////////////
// getNumberOfChannels
// Return the channel count of the audio in the source file.
//
// @return The channel count if a file has been successfully opened and contains
//         audio. Zero otherwise.
//
int AudioReader::getNumberOfChannels()
{
  int numberOfChannels = 0;
  if (_avStreamAudio.codecContext) {
    AVCodecContext* avCodecContext = _avStreamAudio.codecContext;
    numberOfChannels = avCodecContext->channels;
  }
  return numberOfChannels;
}

////////////////////////////////////////////////////////////////////////////////
// getDuration
// Return the duration of the audio stream in seconds.
//
// @return The duration of the audio stream in seconds.
//
double AudioReader::getDuration()
{
  double duration = 0.0;
  if (_avStreamAudio.stream)
    duration = _avStreamAudio.stream->duration * av_q2d(_avStreamAudio.stream->time_base);
  return duration;
}

////////////////////////////////////////////////////////////////////////////////
// setStartPosition
// Set the start position of the audio reader. This is used to adjust the audio
// with respect to video in the mov64Writer. This can be used to advance or
// delay the audio. If the audio is advanced the start of the audio will not
// be read. If the audio is delayed then silence is inserted.
//
// @param startTime The time in seconds of the position of the first read.
//                  This value can be zero, positive or negative.
//                  Time zero is the start of the file, all other values are
//                  with respect to this position.
//                  If positive, this will cause the reader to start reading
//                  after of the start of the file. I.e. positive values will
//                  skip the start of the file.
//                  If negative, this will cause the reader to read 'ahead' of
//                  the start of the file. As no data exists, it will provide
//                  silence until the read position arrives at the start of the
//                  file.
//
// @return 0 if successful,
//         <0 otherwise.
//
int AudioReader::setStartPosition(double startTime)
{
  int ret = 0;

  _readPosition = (int64_t)(startTime / av_q2d(_avStreamAudio.stream->time_base));
  int64_t seekTarget = _readPosition;
  if (0 > seekTarget)
    seekTarget = 0; // The |read| method will return silence while |_readPosition| is less than zero.
  ret = av_seek_frame(_avFormatContext, _avStreamAudio.stream->id, seekTarget, 0);
  ret = (ret >= 0) ? 0 : ret;

  return ret;
}



////////////////////////////////////////////////////////////////////////////////
/// mov64Writer
mov64Writer::mov64Writer(Write* iop, bool componentOfMovWriter)
: Writer(iop)
  , formatContext_(nullptr)
  , writeTimecode_(false)
  , fastStart_(false)
  , dnxhdCodecProfile_(0) // Default to DNxHD 4:4:4 10-bit.
  , proresCodecProfile_(2) //Default to HQ
  , h264CodecProfile_(0) //Default to Main
  , dnxhrCodecProfile_(0) //Default to 4:4:4 12-bit
  , encodeVideoRange_(true)
  , _quality(QualityLevel::eQualityLevel_Default)
  , qMin_(kMinimumQuantizeThresholdPerCodec.at("mpeg4").at(QualityLevel::eQualityLevel_Default))
  , qMax_(kMaximumQuantizeThresholdPerCodec.at("mpeg4").at(QualityLevel::eQualityLevel_Default))
  , _pixelFormat(0)
  , isOpen_(false)
  , pts_counter_(0)
  , fps_(DD::Image::root_real_fps())
  , metaEncoder_(0)
  , format_(0)
  , codec_(0) //The actual default is searched from the codec list in the constructor, to deal with custom lib compiles.
  , _audioFile(nullptr)
  , _audioOffset(0.0f)
  , _audioOffsetUnit(0)
  , bitrate_(kDefaultBitrate)  //Redefined as knob default in init.py, to match updated ffmpeg defaults.
  , bitrateTolerance_(kDefaultBitrateTolerance)
  , limitBitrate_(false)
  , gopSize_(12)
  , bFrames_(0)
  , _YCbCr_MatrixType(YCbCr_MatrixType::Auto)
  , mbDecision_(FF_MB_DECISION_SIMPLE)
  , componentOfMovWriter_(componentOfMovWriter)
  , defaultLutUpdated_(false)
  , qualityKnobsUpdated_(false)
  , codecKnob_(nullptr)
  , qualityKnob_(nullptr)
  , quantizerMinKnob_(nullptr)
  , quantizerMaxKnob_(nullptr)
  , bitrateKnob_(nullptr)
  , bitrateToleranceKnob_(nullptr)
#if defined(FN_LICENSED_PRORES_DNXHD_CODEC)
  , licensedCodec_(nullptr)
  , encoder_(nullptr)
  , codecType_(NukeCodecs::CodecFactory::eUnknownCodec)
#endif //defined(FN_LICENSED_PRORES_DNXHD_CODEC)
{
  av_log_set_level(AV_LOG_ERROR);

  streamVideo_ = {0};
  streamAudio_ = {0};

  formatsLongNames_.push_back("default");
  formatsShortNames_.push_back("default");
  void* opaqueMuxerIter = nullptr;
  const AVOutputFormat* fmt = nullptr;
  while ((fmt = av_muxer_iterate(&opaqueMuxerIter))) {
    if (fmt->video_codec != AV_CODEC_ID_NONE) {
      if (Foundry::Nuke::isFormatWhitelistedForWriting( fmt->name ) ) {
        if (fmt->long_name) {
          formatsLongNames_.push_back(std::string(fmt->name) + std::string(" (") + std::string(fmt->long_name) + std::string(")"));
          formatsShortNames_.push_back(fmt->name);
#if FN_MOV64WRITER_PRINT_CODECS
          std::cout << "Format: " << fmt->name << " = " << fmt->long_name << std::endl;
#endif //  FN_MOV64WRITER_PRINT_CODECS
        }
      }
#if FN_MOV64WRITER_PRINT_CODECS
      else {
        std::cout << "Disallowed Format: " << fmt->name << " = " << fmt->long_name << std::endl;
      }
#endif //  FN_MOV64WRITER_PRINT_CODECS
    }
  }
  formatsShortNames_.push_back(nullptr);

  // Apple ProRes and Avid DNxHD support.
  // Add all supported ProRes and DNxHD profiles.
  uint32_t codecIndex = 0;

  while (SIZEOF_ARRAY(licensedCodecLUT) > codecIndex) {
    const LicensedCodecStr* licensedCodec = &licensedCodecLUT[codecIndex];
    codecsShortNames_.push_back(licensedCodec->shortName);
    codecsKnobLabels_.push_back(licensedCodec->knobLabel);
    ++codecIndex;
  }

  void* opaqueCodecIter = nullptr;
  const AVCodec* c = nullptr;
  while ((c = av_codec_iterate(&opaqueCodecIter))) {
    if (c->type == AVMEDIA_TYPE_VIDEO && c->encode2) {
      std::string reason;
      if (Foundry::Nuke::isCodecWhitelistedForWriting( c->name, reason ) &&
          (c->long_name)) {
        //Initialising codec choice here to deal with custom compiles with differing codecs enabled.
        //Also note this is the legacy codec choice - the prores default is set in knobdefaults.py.
        codecsShortNames_.push_back(c->name);
        const char* knobLabel = Foundry::Nuke::getCodecKnobLabel( c->name );
        codecsKnobLabels_.push_back(knobLabel);
#if FN_MOV64WRITER_PRINT_CODECS
        std::cout << "Codec: " << c->name << " = " << c->long_name << std::endl;
#endif //  FN_MOV64WRITER_PRINT_CODECS
      }
      else {
        if (!reason.empty()) {
          codecsShortNames_.push_back(c->name);
          const char* knobLabel = Foundry::Nuke::getCodecKnobLabel( c->name );
          codecsKnobLabels_.push_back(knobLabel);
          disallowedErrorMessages_.insert(StringPair(c->name,reason));
        }
        #if FN_MOV64WRITER_PRINT_CODECS
        if ( !reason.empty() ) {
          std::cout<<reason<<std::endl;
        }
        else {
          std::cout << "Disallowed Codec: " << c->name << " = " << c->long_name << std::endl;
        }

       #endif //  FN_MOV64WRITER_PRINT_CODECS
      }
    }
  }

  // Sort the codec list into knob label alphabetical order
  mFnAssert(codecsShortNames_.size() == codecsKnobLabels_.size());
  // Copy the label/name pairs into a vector
  using StringPair = std::pair<const char*, const char*>;
  std::vector<StringPair> labelNamePairs;
  std::transform(codecsKnobLabels_.begin(), codecsKnobLabels_.end(),
                 codecsShortNames_.begin(), std::back_inserter(labelNamePairs),
                 // Call make_pair with reference template parameters to allow
                 // lvalue arguments.  The arguments are dereferenced when
                 // copied to the StringPair container.
                 std::make_pair<const char*&, const char*&>);
  // Sort the vector
  std::sort(labelNamePairs.begin(), labelNamePairs.end(),
            [](StringPair a, StringPair b) {
              return compareCodecKnobLabels(a.first, b.first) < 0; }
  );
  // Rewrite the string pointers
  std::transform(labelNamePairs.begin(), labelNamePairs.end(),
                 codecsKnobLabels_.begin(), std::mem_fn(&StringPair::first));
  std::transform(labelNamePairs.begin(), labelNamePairs.end(),
                 codecsShortNames_.begin(), std::mem_fn(&StringPair::second));

  codecsShortNames_.push_back(nullptr);
  codecsKnobLabels_.push_back(nullptr);

  // Setting default codec to MPEG4 for age old legacy behaviour
  AVCodec* defaultCompiledInCodec = avcodec_find_encoder(AV_CODEC_ID_MPEG4);

  if (!defaultCompiledInCodec) {
    iop->warning("Mov64Writer: faild to find default encoder: %s", defaultCompiledInCodec->name);
    return;
  }

  currentCodecPixelFormats_  = PixelFormatsFromAVCodec(defaultCompiledInCodec);
  displayedPixelFormatNames_ = PixelFormatFriendlyNamesFrom(currentCodecPixelFormats_);

  auto codecNameItr = std::find_if(codecsShortNames_.begin(), codecsShortNames_.end(),
    [&defaultCompiledInCodec](const char* codecName) -> bool {
        return !strcmp(defaultCompiledInCodec->name, codecName);
      }
  );

  if (codecNameItr != codecsShortNames_.end()) {
    codec_ = static_cast<int>(std::distance(codecsShortNames_.begin(), codecNameItr));
  }
}

mov64Writer::~mov64Writer()
{
}

bool mov64Writer::isRec709Format(const int height) const
{
  // First check for codecs which require special handling:
  //  * JPEG codecs always use Rec 601.
  Knob* codecKnob = iop->knob("mov64_codec");
  // Its possible that replaceable knobs may not exist if file_type knob is blank. This is allowed.
  if (codecKnob) {
    const bool isJpeg = IsJpeg(codecKnob, codec_);
    if (isJpeg) {
      return false;
    }
  }

  // Using method described in step 5 of QuickTimeCodecReader::setPreferredMetadata
  // #cristian: height >= 720 (HD); height < 720 (SD)
  return (height >= 720);
}

bool mov64Writer::isPalFormat(const AVRational& displayRatio) const
{
  const AVRational palRatio = {5, 4};
  const bool isPAL = (displayRatio.num == 576) || (av_cmp_q(palRatio, displayRatio) == 0);

  return isPAL;
}

// Figure out if an FFmpeg codec is definitely YUV based from its underlying storage
// pixel format type.
/*static*/ bool mov64Writer::IsYUV(AVPixelFormat pix_fmt)
{
  // from swscale_internal.h
  const AVPixFmtDescriptor *desc = av_pix_fmt_desc_get(pix_fmt);
  return !(desc->flags & AV_PIX_FMT_FLAG_RGB) && desc->nb_components >= 2;
}

// Figure out if a codec is definitely YUV based from its shortname.
/*static*/ bool mov64Writer::IsYUVFromShortName(const char* shortName, int codecProfile)
{
  return (IsLegacyProres(shortName) ||
          IsProres(shortName) ||
          ((codecProfile > 0) && IsDNxHD(shortName)) ||
          ((codecProfile > 0) && IsDNxHR(shortName)) ||
          !strcmp(shortName, "mjpeg") ||
          !strcmp(shortName, "mpeg1video") ||
          !strcmp(shortName, "mpeg4") ||
          !strcmp(shortName, "v210") ||
          !strcmp(shortName, "mjpegb")  ||
          !strcmp(shortName, "mjpega") ||
          !strcmp(shortName, "fn_h264"));
}

// Figure out if a codec is definitely RGB based from its shortname.
/*static*/ bool mov64Writer::IsRGBFromShortName(const char* shortName, int codecProfile)
{
  return (((codecProfile == 0) && IsDNxHD(shortName)) ||
          ((codecProfile == 0) && IsDNxHR(shortName)) ||
          !strcmp(shortName, "png")  ||
          !strcmp(shortName, "qtrle"));
}

AVColorTransferCharacteristic mov64Writer::getColorTransferCharacteristic() const
{
  Knob* k = iop->knob("colorspace");
  if (k) {
    Enumeration_KnobI* e = k->enumerationKnob();
    if (e) {
      const std::string selection = e->getSelectedItemString();
      if (selection.find("Gamma2.2") != std::string::npos) {
        return AVCOL_TRC_GAMMA22;
      }
    }
  }

  return AVCOL_TRC_UNSPECIFIED;
}

/**
 * called by Write::updateDefaultLUT() from the main thread to figure out
 * what text to put in the "colorspace" widget.
 */
LUT* mov64Writer::defaultLUT() const
{
  AVOutputFormat* fmt = initFormat(/* reportErrors = */ false );
  if (!fmt) {
    return LUT::GetLut(LUT::GAMMA2_2, this);
  }

  int codecTemp = codec_;
  if (iop && iop->knob("mov64_codec"))
    codecTemp = (int)iop->knob("mov64_codec")->get_value();

  const auto shortName = codecsShortNames_[codecTemp];
  // TP434787 & TP395067 Make rec.709 be the default colorspace for exporting these file types
  if (IsDNxHR(shortName) || IsDNxHD(shortName)) {
    return LUT::GetBuiltinLutByName("rec709");
  }
    
  // Hardcode using utility shortname functions, as the pixel format isn't
  // an accurate representation of what the codec should use to cleanly
  // roundtrip with official quicktime and other packages, particularly in
  // the case of quicktime.
  if (IsRGBFromShortName(shortName, dnxhdCodecProfile_)) {
    return LUT::GetLut(LUT::GAMMA1_8, this);
  }
  else if (IsYUVFromShortName(shortName, dnxhdCodecProfile_)) {
    return LUT::GetLut(LUT::GAMMA2_2, this);
  }

  // Fall back to using the current codec's preferred pixel format.

  AVCodecID codecId = AV_CODEC_ID_NONE;
  AVCodec* videoCodec = nullptr;
  const bool success = initCodec(fmt, codecId, videoCodec);
  if (!success) {
    return LUT::GetLut(LUT::GAMMA2_2, this);
  }

  AVPixelFormat targetPixelFormat     = AV_PIX_FMT_YUV420P;
  AVPixelFormat nukeBufferPixelFormat = AV_PIX_FMT_RGB24;
  int outBitDepth = 8;
  getPixelFormats(videoCodec, nukeBufferPixelFormat, targetPixelFormat, outBitDepth);

  // YUV = Gamma2.2, RGB = Gamma1.8
  const bool isYUV = IsYUV(targetPixelFormat);
  return LUT::GetLut(isYUV ? LUT::GAMMA2_2 : LUT::GAMMA1_8, this);
}

bool mov64Writer::isDefaultLUTKnob(DD::Image::Knob* knob) const
{
  return knob->is("mov64_codec") || knob->is("mov64_format");
}

AVOutputFormat* mov64Writer::initFormat(bool reportErrors) const
{
  AVOutputFormat* fmt = nullptr;

  if (!format_) {   
    const auto file = filename();
    const auto fileSpecified = file && strlen(file);
    // If a file name has not yet been set, assume mov format by default
    fmt = av_guess_format(fileSpecified ? nullptr : "mov", file, nullptr);
    if (!fmt && reportErrors) {
      iop->critical("could not deduce output format from file extension");
      return nullptr;
    }
  }
  else {
    fmt = av_guess_format(formatsShortNames_[format_], nullptr, nullptr);
    if (!fmt && reportErrors) {
      iop->critical("could not deduce output format");
      return nullptr;
    }
  }

  return fmt;
}

bool mov64Writer::initCodec(AVOutputFormat* fmt, AVCodecID& outCodecId, AVCodec*& outVideoCodec) const
{
#if defined(FN_LICENSED_PRORES_DNXHD_CODEC)
    licensedCodec_ = nullptr; // Reset.
#endif //defined(FN_LICENSED_PRORES_DNXHD_CODEC)

  outCodecId = fmt->video_codec;
  AVCodec* userCodec = avcodec_find_encoder_by_name(codecsShortNames_[codec_]);
  if (userCodec) {
    outCodecId = userCodec->id;
  }
  else {
    // Check whether the selected codec is Apple ProRes or Avid DNxHD.
    uint32_t codecIndex = 0;
    size_t codecNameLen = strlen(codecsShortNames_[codec_]);
    while (SIZEOF_ARRAY(licensedCodecLUT) > codecIndex) {
      const char* shortName = licensedCodecLUT[codecIndex].shortName;
      size_t shortNameLen = strlen(shortName);
      if ((codecNameLen == shortNameLen) &&
          !strncmp(codecsShortNames_[codec_], shortName, codecNameLen)) {
#if defined(FN_LICENSED_PRORES_DNXHD_CODEC)
        licensedCodec_ = &licensedCodecLUT[codecIndex];
#endif //defined(FN_LICENSED_PRORES_DNXHD_CODEC)
        if (IsDNxHD(shortName) || IsDNxHR(shortName)) {
          outCodecId = AV_CODEC_ID_DNXHD;
        }
        else if (!strncmp(shortName, "fn_h264", 7)) {
          outCodecId = AV_CODEC_ID_H264;
        }
        else {
          outCodecId = AV_CODEC_ID_PRORES;
        }
        break;
      }
      ++codecIndex;
    }
  }

  outVideoCodec = avcodec_find_encoder(outCodecId);
  if (!outVideoCodec) {
    iop->critical("unable to find video codec");
    return false;
  }

  return true;
}

void mov64Writer::getPixelFormats(AVCodec* videoCodec, AVPixelFormat& outNukeBufferPixelFormat, AVPixelFormat& outTargetPixelFormat, int& outBitDepth) const
{
  assert(videoCodec);

  const bool hasAlpha = alphaEnabled();

  // Apple ProRes and Avid DNxHD specific.
  // Provide the pixel format that matches the
  // formats supported by the licensed codec
  // libraries.
  if (AV_CODEC_ID_PRORES == videoCodec->id) {
    outTargetPixelFormat = AV_PIX_FMT_BGRA64;
  }
  else if (AV_CODEC_ID_DNXHD == videoCodec->id) {
    int profile = NukeCodecs::CodecFactory::isCodecTypeDNxHR(licensedCodec_->codecTag) ? dnxhrCodecProfile_ : dnxhdCodecProfile_;
    if (profile < 2) {
      outTargetPixelFormat = AV_PIX_FMT_BGRA64; // 10-bit codec profiles.
    }
    else {
      outTargetPixelFormat = AV_PIX_FMT_BGRA; // 8-bit codec profiles.
    }
  }
  else if (videoCodec->pix_fmts != nullptr) {
    //This is the most frequent path, where we can guess best pix format using ffmpeg.
    //find highest bit depth pix fmt.
    const AVPixelFormat* currPixFormat  = videoCodec->pix_fmts;
    int currPixFormatBitDepth           = outBitDepth;
    while (*currPixFormat != -1) {
      currPixFormatBitDepth             = GetPixelFormatBitDepth(*currPixFormat);
      if (currPixFormatBitDepth  > outBitDepth)
        outBitDepth                     = currPixFormatBitDepth;
      currPixFormat++;
    }

    //figure out nukeBufferPixelFormat from this.
    outNukeBufferPixelFormat = GetPixelFormatFromBitDepth(outBitDepth, hasAlpha);

    //call best_pix_fmt using the full list.
    const int hasAlphaInt = alphaEnabledNoValidation() ? 1 : 0;
    int loss     = 0; //Potentially we should error, or at least report if over a certain value?
    outTargetPixelFormat = avcodec_find_best_pix_fmt_of_list(videoCodec->pix_fmts, outNukeBufferPixelFormat, hasAlphaInt, &loss);

    if (AV_CODEC_ID_QTRLE == videoCodec->id) {
      if (hasAlphaInt &&
         (AV_PIX_FMT_ARGB != outTargetPixelFormat) &&
         (AV_PIX_FMT_RGBA != outTargetPixelFormat) &&
         (AV_PIX_FMT_ABGR != outTargetPixelFormat) &&
         (AV_PIX_FMT_BGRA != outTargetPixelFormat)) {
        // WARNING: Workaround.
        //
        // If the source has alpha, then the alpha channel must be
        // preserved. QT RLE supports an alpha channel, however
        // |avcodec_find_best_pix_fmt_of_list| above seems to default
        // to RGB24 always, despite having support for ARGB and the
        // |hasAlpha| flag being set.
        //
        // So search for a suitable RGB+alpha pixel format that is
        // supported by the QT RLE codec and force the output pixel
        // format.
        //
        currPixFormat = videoCodec->pix_fmts;
        while (*currPixFormat != -1) {
          AVPixelFormat avPixelFormat = *currPixFormat++;
          if ((AV_PIX_FMT_ARGB == avPixelFormat) ||
             (AV_PIX_FMT_RGBA == avPixelFormat) ||
             (AV_PIX_FMT_ABGR == avPixelFormat) ||
             (AV_PIX_FMT_BGRA == avPixelFormat)) {
            outTargetPixelFormat = avPixelFormat;
            break;
          }
        }
      }
    }

    //Unlike the other cases, we're now done figuring out all aspects, so return.
    return;
  }
  else {
    //Lowest common denominator defaults.
    outTargetPixelFormat = AV_PIX_FMT_YUV420P;
  }

  outBitDepth = GetPixelFormatBitDepth(outTargetPixelFormat);
  outNukeBufferPixelFormat = GetPixelFormatFromBitDepth(outBitDepth, hasAlpha);
}

// av_get_bits_per_sample knows about surprisingly few codecs.
// We have to do this manually.
/*static*/ int mov64Writer::GetPixelFormatBitDepth(const AVPixelFormat pixelFormat)
{
  switch (pixelFormat) {
  case AV_PIX_FMT_BGRA64:
  case AV_PIX_FMT_RGBA64:
  case AV_PIX_FMT_RGB48LE:
  case AV_PIX_FMT_RGB48BE:
    return 16;
    break;

  case AV_PIX_FMT_YUV411P:     // Uncompressed 4:1:1 12bit
    return 12;
    break;

  case AV_PIX_FMT_YUV422P10LE: // Uncompressed 4:2:2 10bit - planar
  case AV_PIX_FMT_YUV444P10LE: // Uncompressed 4:4:4 10bit - planar
  case AV_PIX_FMT_YUVA444P:    // Uncompressed packed QT 4:4:4:4
    return 10;
    break;

  case AV_PIX_FMT_YUV420P: // MPEG-1, MPEG-2, MPEG-4 part2 (default)
  case AV_PIX_FMT_YUV444P: // Uncompressed 4:4:4 planar
  default:
    return 8;
    break;
  }
}

/*static*/ AVPixelFormat mov64Writer::GetPixelFormatFromBitDepth(const int bitDepth, const bool hasAlpha)
{
  AVPixelFormat pixelFormat;
  if (hasAlpha) {
    pixelFormat = (bitDepth > 8) ? AV_PIX_FMT_RGBA64 : AV_PIX_FMT_RGB32;
  }
  else {
    pixelFormat = (bitDepth > 8) ? AV_PIX_FMT_RGB48LE : AV_PIX_FMT_RGB24;
  }
  return pixelFormat;
}

/*static*/ void mov64Writer::GetCodecSupportedParams(AVCodec* codec, bool& outLossyParams, bool& outInterGOPParams, bool& outInterBParams)
{
  assert(codec);
  //The flags on the codec can't be trusted to indicate capabilities, so use the props bitmask on the descriptor instead.
  const AVCodecDescriptor* codecDesc = avcodec_descriptor_get(codec->id);

  outLossyParams    = codecDesc ? !!(codecDesc->props & AV_CODEC_PROP_LOSSY) : false;
  outInterGOPParams = codecDesc ? !(codecDesc->props & AV_CODEC_PROP_INTRA_ONLY) : false;
  outInterBParams   = codecDesc ? !(codecDesc->props & AV_CODEC_PROP_INTRA_ONLY) : false;

  //Overrides for specific cases where the codecs don't follow the rules.
  //PNG doesn't observe the params, despite claiming to be lossy.
  if (codecDesc && (codecDesc->id == AV_CODEC_ID_PNG)) {
    outLossyParams = outInterGOPParams = outInterBParams = false;
  }
  //Mpeg4 ms var 3 / AV_CODEC_ID_MSMPEG4V3 doesn't have a descriptor, but needs the params.
  if (codec && (codec->id == AV_CODEC_ID_MSMPEG4V3 || codec->id == AV_CODEC_ID_H264)) {
    outLossyParams = outInterGOPParams = outInterBParams = true;
  }
  //QTRLE supports differing GOPs, but any b frame settings causes unreadable files.
  if (codecDesc && (codecDesc->id == AV_CODEC_ID_QTRLE)) {
    outLossyParams = outInterBParams = false;
    outInterGOPParams = true;
  }
}

std::string mov64Writer::defaultPixelFormatFrom(const AVCodec* codec) const
{
  std::string label;
  const bool haveAlpha = alphaEnabledNoValidation();

  switch (codec->id) {
    case AV_CODEC_ID_MJPEG:
    case AV_CODEC_ID_MJPEGA:
    case AV_CODEC_ID_MJPEGB:
      label = av_get_pix_fmt_name(AV_PIX_FMT_YUVJ444P);
      break;
    case AV_CODEC_ID_MPEG1VIDEO:
    case AV_CODEC_ID_MPEG4:
      label = av_get_pix_fmt_name(AV_PIX_FMT_YUV420P);
      break;
    case AV_CODEC_ID_PNG:
      label = haveAlpha ? av_get_pix_fmt_name(AV_PIX_FMT_RGBA64BE): av_get_pix_fmt_name(AV_PIX_FMT_RGB48BE);
      break;
    case AV_CODEC_ID_QTRLE:
      label = haveAlpha ? av_get_pix_fmt_name(AV_PIX_FMT_ARGB): av_get_pix_fmt_name(AV_PIX_FMT_RGB24);
      break;
    case AV_CODEC_ID_V210:
      label = av_get_pix_fmt_name(AV_PIX_FMT_YUV422P10);
      break;
    default:
      label = av_get_pix_fmt_name(codec->pix_fmts[0]);
      break;
  }

  MovPixelFormat pixFmt;
  bool found = PixelFormatFromFFmpegStr(label, pixFmt);
  assert(found == true);

  return PixelFormatFriendlyNameFrom(pixFmt);
}


namespace {
  bool UsingMov32Encoder(DD::Image::Write& iop)
  {
    if (!iop.knob("meta_encoder"))
      return false;

    auto encoder = static_cast<int>(iop.knob("meta_encoder")->get_value());
    return encoder == 1;
  }
}


void mov64Writer::updatePixelFormatKnobVisibility(AVCodec* codec)
{
  Knob* pixelTypeKnob = iop->knob(kPixelFormatKnob);
  currentCodecPixelFormats_ = PixelFormatsFromAVCodec(codec);
  displayedPixelFormatNames_ = PixelFormatFriendlyNamesFrom(currentCodecPixelFormats_);

  pixelTypeKnob->visible(!displayedPixelFormatNames_.empty());
}


void mov64Writer::updatePixelFormatKnobValues(const AVCodec* codec)
{
  Knob* pixelTypeKnob = iop->knob(kPixelFormatKnob);

  if (!codec) {
    return;
  }

  currentCodecPixelFormats_ = PixelFormatsFromAVCodec(codec);
  displayedPixelFormatNames_ = PixelFormatFriendlyNamesFrom(currentCodecPixelFormats_);

  if (!displayedPixelFormatNames_.empty()) {
    pixelTypeKnob->enable(displayedPixelFormatNames_.size() > 1);
    auto enumerationKnob = pixelTypeKnob->enumerationKnob();
    enumerationKnob->menu(displayedPixelFormatNames_);

    std::string label = defaultPixelFormatFrom(codec);
    SetKnobValue(pixelTypeKnob, displayedPixelFormatNames_, label);
  }
}

void mov64Writer::updateYCbCrMatrixKnob(const char* codecShortName)
{
  const bool isYuv = IsYUVFromShortName(codecShortName, dnxhdCodecProfile_);
  iop->knob("mov64_ycbcr_matrix_type")->visible(isYuv);
  iop->knob("mov64_ycbcr_matrix_type")->enable(isYuv);
}

void mov64Writer::updateLimitBitrateKnob(AVCodec* codec)
{
  iop->knob("mov64_limit_bitrate")->visible(codec ? codec->id == AV_CODEC_ID_MPEG4 : false);
}

////////////////////////////////////////////////////////////////////////////////
// configureAudioStream
// Set audio parameters of the audio stream.
//
// @param avCodec A pointer reference to an AVCodec to receive the AVCodec if
//                the codec can be located and is on the codec white list.
// @param avStream A reference to an AVStream of a audio stream.
//
void mov64Writer::configureAudioStream(AVCodec* avCodec, FnAVStream* fnAVStream)
{
  AVCodecContext* avCodecContext = fnAVStream->codecContext;
  avcodec_get_context_defaults3(avCodecContext, avCodec);
  avCodecContext->sample_fmt = audioReader_->getSampleFormat();
  //avCodecContext->bit_rate    = 64000; // Calculate...
  avCodecContext->sample_rate = audioReader_->getSampleRate();
  avCodecContext->channels = audioReader_->getNumberOfChannels();
}

static std::string Basename( const std::string& in )
{
  size_t pos = in.find_last_of( "/" );
  if (pos == std::string::npos) {
    return in;
  }

  if (pos == in.size()-1) {
    return in;
  }

  return in.substr(pos+1);
}
////////////////////////////////////////////////////////////////////////////////
// configureVideoStream
// Set video parameters of the video stream.
//
// @param avCodec A pointer reference to an AVCodec to receive the AVCodec if
//                the codec can be located and is on the codec white list.
// @param avStream A reference to an AVStream of a audio stream.
//
void mov64Writer::configureVideoStream(AVCodec* avCodec, FnAVStream* fnAVStream)
{
  AVCodecContext* avCodecContext = fnAVStream->codecContext;
  AVStream* avStream = fnAVStream->stream;
  avcodec_get_context_defaults3(avCodecContext, avCodec);

  //Only update the relevant context variables where the user is able to set them.
  //This deals with cases where values are left on an old value when knob disabled.
  bool lossyParams    = false;
  bool interGOPParams = false;
  bool interBParams   = false;
  if (avCodec) GetCodecSupportedParams(avCodec, lossyParams, interGOPParams, interBParams);

  if (lossyParams) {
    avCodecContext->bit_rate = bitrate_;

    if (limitBitrate_) {
      avCodecContext->rc_max_rate =  bitrate_;
      avCodecContext->rc_max_available_vbv_use = 2.0f;
    }

    avCodecContext->bit_rate_tolerance = limitBitrate_ ? 0: bitrateTolerance_;
    avCodecContext->qmin = qMin_;
    avCodecContext->qmax = qMax_;
  }
  avCodecContext->width = width();
  avCodecContext->height = height();

  avCodecContext->color_trc = getColorTransferCharacteristic();

#if !defined(FN_EXAMPLE_PLUGIN)
  #if !defined(FN_PRODUCT_NAME)
  #error "FN_PRODUCT_NAME must be defined!"
  #endif
  
  av_dict_set(&formatContext_->metadata, kMetaKeyApplication, mFnStringize(FN_PRODUCT_NAME), 0);
#endif
  
  DD::Image::Version version = DD::Image::applicationVersion();
  
  av_dict_set(&formatContext_->metadata, kMetaKeyApplicationVersion, version.majorMinorReleaseString().c_str(), 0);
  
  const char* ycbcrmetavalue = isRec709Format(avCodecContext->height) ? "Rec 709" : "Rec 601";
  av_dict_set(&formatContext_->metadata, kMetaKeyYCbCrMatrix, ycbcrmetavalue, 0);

  // Set colorspace metadata key to be the colorspace knob value
  std::string metaKeyColorspace = "";
  Knob* colorspaceKnob = iop->knob("colorspace");
  if (colorspaceKnob) {
    Enumeration_KnobI* enumerationKnob = colorspaceKnob->enumerationKnob();
    if (enumerationKnob) {
      const int idx = enumerationKnob->getSelectedItemIndex();
      metaKeyColorspace =  enumerationKnob->getItemValueString(idx);
      // Handle default
      metaKeyColorspace = stripDefaultLutName(metaKeyColorspace);
    }
  }
  // If we failed to set a value for metaKeyColorspace, then fallback on the current lut name;
  // In OCIO color management this is always LINEAR; see FOUNDRY Bug TP #181134 for more information;
  if (metaKeyColorspace.empty()) {
     metaKeyColorspace = GetLutName(lut());
  }

  if (!metaKeyColorspace.empty()) {
    metaKeyColorspace = Basename(metaKeyColorspace); // strip off cascading menus prefix
    av_dict_set(&formatContext_->metadata, kMetaKeyColorspace, metaKeyColorspace.c_str(), 0);
  }

  av_dict_set(&formatContext_->metadata, kMetaKeyWriter, kMetaValueWriter64, 0);

  //Write the NCLC atom in the case the underlying storage is YUV.
  if(IsYUVFromShortName(codecsShortNames_[codec_], dnxhdCodecProfile_)) {

    const AVRational displayRatio = {avCodecContext->width, avCodecContext->height};

    const bool isHD = isRec709Format(avCodecContext->height);
    const bool isPAL = isPalFormat(displayRatio);

    if (isHD) {
      avCodecContext->color_primaries = AVCOL_PRI_BT709;      // kQTPrimaries_ITU_R709_2
      avCodecContext->color_trc =  AVCOL_TRC_BT709;           // kQTTransferFunction_ITU_R709_2
    }
    else if (isPAL) {
      avCodecContext->color_primaries = AVCOL_PRI_BT470BG;    // kQTPrimaries_EBU_3213
      avCodecContext->color_trc =  AVCOL_TRC_BT709;           // kQTTransferFunction_ITU_R709_2
    }
    else {
      avCodecContext->color_primaries = AVCOL_PRI_SMPTE170M;   // kQTPrimaries_SMPTE_C
      avCodecContext->color_trc =  AVCOL_TRC_BT709;            // kQTTransferFunction_ITU_R709_2
    }

    avCodecContext->colorspace = [this, avCodecContext]() {
        switch (_YCbCr_MatrixType) {
          case YCbCr_MatrixType::Rec601: return AVCOL_SPC_SMPTE170M;
          case YCbCr_MatrixType::Rec709: return AVCOL_SPC_BT709;

          default:
          case YCbCr_MatrixType::Auto: return isRec709Format(avCodecContext->height) ? AVCOL_SPC_BT709 : AVCOL_SPC_SMPTE170M;
        }
      }()
    ;

    av_dict_set(&avStream->metadata, kNCLCPrimariesKey, std::to_string(avCodecContext->color_primaries).c_str(), 0);
    av_dict_set(&avStream->metadata, kNCLCTransferKey, std::to_string(avCodecContext->color_trc).c_str(), 0);
    av_dict_set(&avStream->metadata, kNCLCMatrixKey, std::to_string(avCodecContext->colorspace).c_str(), 0);
  }

  avStream->sample_aspect_ratio.num = 1;
  avStream->sample_aspect_ratio.den = 1;

  // From the Apple QuickTime movie guidelines. Set the
  // appropriate pixel aspect ratio for the movie.
  // Scale by 100 and convert to int for a reliable comparison, e.g.
  // 0.9100000000 != 0.9100002344. This is done as the pixel aspect
  // ratio is provided as a double and the author cannot find a
  // rational representation (num/den) of par.
  int32_t par = (int32_t)(info().format().pixel_aspect() * 100.0);
  if (200 == par) {
    avStream->sample_aspect_ratio.num = 2;
    avStream->sample_aspect_ratio.den = 1;
  }
  else if (150 == par) {
    avStream->sample_aspect_ratio.num = 3;
    avStream->sample_aspect_ratio.den = 2;
  }
  else if (146 == par) {
    // PAL 16:9
    avStream->sample_aspect_ratio.num = 118;
    avStream->sample_aspect_ratio.den = 81;
  }
  else if (133 == par) {
    avStream->sample_aspect_ratio.num = 4;
    avStream->sample_aspect_ratio.den = 3;
  }
  else if (121 == par) {
    // NTSC 16:9
    avStream->sample_aspect_ratio.num = 40;
    avStream->sample_aspect_ratio.den = 33;
  }
  else if (109 == par) {
    // PAL 4:3
    avStream->sample_aspect_ratio.num = 59;
    avStream->sample_aspect_ratio.den = 54;
  }
  else if (100 == par) {
    // Typically HD
    avStream->sample_aspect_ratio.num = 1;
    avStream->sample_aspect_ratio.den = 1;
  }
  else if (91 == par) {
    // NTSC 4:3
    avStream->sample_aspect_ratio.num = 10;
    avStream->sample_aspect_ratio.den = 11;
  }

  avCodecContext->sample_aspect_ratio = avStream->sample_aspect_ratio;

  // timebase: This is the fundamental unit of time (in seconds) in terms
  // of which frame timestamps are represented. For fixed-fps content,
  // timebase should be 1/framerate and timestamp increments should be
  // identical to 1.
  //
  // Bug 23953
  // ffmpeg does a horrible job of converting floats to AVRationals
  // It adds 0.5 randomly and does some other stuff
  // To work around that, we just multiply the fps by what I think is a reasonable number to make it an int
  // and use the reasonable number as the numerator for the timebase.
  // Timebase is not the frame rate; it's the inverse of the framerate
  // So instead of doing 1/fps, we just set the numerator and denominator of the timebase directly.
  // The upshot is that this allows ffmpeg to properly do framerates of 23.78 (or 23.796, which is what the user really wants when they put that in).
  //
  // The code was this:
  //streamVideo_.codec->time_base = av_d2q(1.0 / fps_, 100);
  const float CONVERSION_FACTOR = 1000.0f;
  avCodecContext->time_base.num = (int)CONVERSION_FACTOR;
  avCodecContext->time_base.den = (int)(fps_ * CONVERSION_FACTOR);

  // Apple ProRes and Avid DNxHD specific.
  // The underlying library for writer QuickTime movies has been
  // modified to support customisation of chunking. The QuickTime
  // guidelines from Apple recommend chunk sizes of 4MiB for HD
  // and 2MiB for SD. This is configured by setting the chunksize
  // metadata key with the log2 size of the chunk.
  if ((AV_CODEC_ID_PRORES == avCodecContext->codec_id) ||
      (AV_CODEC_ID_DNXHD == avCodecContext->codec_id) ||
      (AV_CODEC_ID_H264 == avCodecContext->codec_id)) {
    if (avCodecContext->height >= 720) {
      // Assume HD.
      av_dict_set(&avStream->metadata, kChunkSizeKey, "22", 0); // Set 'log2chunksize' to 2^22 = 4MiB.
    }
    else {
      // Assume SD.
      av_dict_set(&avStream->metadata, kChunkSizeKey, "21", 0); // Set 'log2chunksize' to 2^21 = 2MiB.
    }
    // Nuke/NukeStudio only supports progressive.
    // Set this field so that an 'fiel' atom is inserted
    // into the QuickTime 'moov' atom.
    avCodecContext->field_order = AV_FIELD_PROGRESSIVE;
    // If alpha is to be encoded, then the following must
    // be set to 32. This will ensure the correct bit depth
    // information will be embedded in the QuickTime movie.
    // Otherwise the following must be set to 24.
    const bool hasAlpha = alphaEnabled();
    avCodecContext->bits_per_coded_sample = (hasAlpha) ? 32 : 24;
  }

  // Trap fractional frame rates so that they can be specified correctly
  // in a QuickTime movie. The rational number representation of fractional
  // frame rates is 24000/1001, 30000/1001, 60000/1001, etc.
  // WARNING: There are some big assumptions here. The assumption is
  //          that by specifying 23.98, 29.97 on the UI the intended
  //          frame rate is 24/1.001, 30/1.001, etc. so the frame rate
  //          is corrected here.
  int frameRate = (0.0 < fps_) ? (int)fps_ : 0;
  if ((23 == frameRate) || (29 == frameRate) || (59 == frameRate)) {
    avCodecContext->time_base.num = 1001;
    avCodecContext->time_base.den = (frameRate + 1) * 1000;
  }
  else {
    avCodecContext->time_base.num = 100;
    avCodecContext->time_base.den = frameRate * 100;
  }

  avStream->time_base = avCodecContext->time_base;
  avStream->avg_frame_rate = av_inv_q(avCodecContext->time_base);

  if (interGOPParams)
    avCodecContext->gop_size = gopSize_;

  // NOTE: in new ffmpeg, bframes don't seem to work correctly - ffmpeg crashes...
  if (interBParams && bFrames_) {
    avCodecContext->max_b_frames = bFrames_;
    av_opt_set(avCodecContext->priv_data, "b_frame_strategy", "0", 0);
    avCodecContext->b_quant_factor = 2.0f;
  }

  avCodecContext->mb_decision = mbDecision_;

  // Create a timecode stream for QuickTime movies. (There was no
  // requirement at the time of writing for any other file format.
  // Also not all containers support timecode.)
  if (writeTimecode_ && !strcmp(formatContext_->oformat->name, "mov")) {
    // Retrieve the timecode from Nuke/NukeStudio.
    // Adding a 'timecode' metadata item to the video stream
    // metadata will automagically create a timecode track
    // in the QuickTime movie created by the underlying
    // library.
    const MetaData::Bundle& metaData = iop->_fetchMetaData("");
    MetaData::Bundle::PropertyPtr property = metaData.getData("input/timecode");
    if (property) {
      std::string timecode = MetaData::propertyToString(property).c_str();

      // Only write if the timecode is valid
#ifndef FN_EXAMPLE_PLUGIN
      if (Foundry::Base::Timecode::isValidString(timecode)) {

        // Note that the ffmpeg time base is seconds per frame but the foundry
        // time base uses frames per second so we switch num & den
        Foundry::Base::TimeBase timeBase(avCodecContext->time_base.den, avCodecContext->time_base.num);
        Foundry::Base::Timecode::DisplayType displayType = Foundry::Base::Timecode::displayTypeFromString(timecode);
        // Only use time code or drop frame time code
        if (displayType != Foundry::Base::Timecode::kDisplayDropFrameTimecode) {
          displayType = Foundry::Base::Timecode::kDisplayTimecode;
        }
        Foundry::Base::Time time = Foundry::Base::Timecode::stringToTime(timecode, timeBase, displayType);

        if (time != Foundry::Base::kInvalidTime) {
          timecode = Foundry::Base::Timecode::timeToString(time, timeBase, displayType);

          // FFmpeg requires drop frames in the format: [-?]hh:mm:ss;ff not [-?]hh;mm;ss;ff
          size_t semicolon = timecode.rfind(';');
          if (semicolon != std::string::npos && semicolon > 0) {
            // This is a drop frame time code.
            // Replace earlier semicolons.
            semicolon = timecode.rfind(';', semicolon-1);
            while (semicolon != std::string::npos) {
              timecode.replace(semicolon, 1, 1, ':');

              if (semicolon == 0) {
                break;
              }
              semicolon = timecode.rfind(';', semicolon-1);
            }
          }

          av_dict_set(&avStream->metadata, "timecode", timecode.c_str(), 0);
        }
      }
#endif // ndef FN_EXAMPLE_PLUGIN
    }
  }
}

////////////////////////////////////////////////////////////////////////////////
// addStream
// Add a new stream to the AVFormatContext of the file. This will search for
// the codec and if found, validate it against the codec whitelist. If the codec
// can be used, a new stream is created and configured.
//
// @param avFormatContext A reference to an AVFormatContext of the file.
// @param avCodecId An AVCodecID enumeration of the codec to attempt to open.
// @param pavCodec A pointer reference to an AVCodec to receive the AVCodec if
//                 the codec can be located and is on the codec white list.
//                 This can be NULL for non-codec streams, e.g. timecode.
//
// @return A reference to an AVStream if successful.
//         NULL otherwise.
//
void mov64Writer::addStream(AVFormatContext* avFormatContext, enum AVCodecID avCodecId, AVCodec** pavCodec, FnAVStream* fnStreamOut)
{
  AVCodecContext* avCodecContext = nullptr;
  AVStream* avStream = nullptr;
  AVCodec* avCodec = nullptr;
  if (pavCodec) {
    // Find the encoder.
    avCodec = avcodec_find_encoder(avCodecId);
    if (!avCodec) {
      iop->critical("could not find codec");
      return;
    }
    std::string reason;
    const char * name = avCodec->name;
    //we need that ffmpeg can return hw encoder like video toolbox, nvenc, vaapi so on
    if (avCodecId == AV_CODEC_ID_H264) {
      name = "h264";
    }
    if (!Foundry::Nuke::isCodecWhitelistedForWriting(name, reason)) {
      if ( !reason.empty() ) {
        iop->critical(reason.c_str());
        return;
      }
      iop->critical("unsupported codec");
      return;
    }
  }

  avCodecContext = avcodec_alloc_context3(avCodec);
  if (!avCodecContext) {
    iop->critical("could not allocate codec context");
    return;
  }

  avStream = avformat_new_stream(avFormatContext, avCodecContext->codec);
  if (!avStream) {
    iop->critical("could not allocate stream");
    return;
  }

  avStream->id = avFormatContext->nb_streams - 1;


  fnStreamOut->stream       = avStream;
  fnStreamOut->codecContext = avCodecContext;

  switch (avCodec->type) {
  case AVMEDIA_TYPE_AUDIO:
    configureAudioStream(avCodec, fnStreamOut);
    break;

  case AVMEDIA_TYPE_VIDEO:
    configureVideoStream(avCodec, fnStreamOut);
    break;

  default:
    break;
  }

  // Update the caller provided pointer with the codec.
  *pavCodec = avCodec;

  return;
}

////////////////////////////////////////////////////////////////////////////////
// openCodec
// Open a codec.
//
// @param avFormatContext A reference to an AVFormatContext of the file.
// @param avCodec A reference to an AVCodec of the video codec.avcodec_parameters_from_context
// @param avStream A reference to an AVStream of a video stream.avcodec_parameters_from_context
//avcodec_parameters_from_context
// @return 0 if successful,avcodec_parameters_from_context
//         <0 otherwise.
//
int mov64Writer::openCodec(AVFormatContext* avFormatContext, AVCodec* avCodec, FnAVStream* fnAVStream)
{
  AVCodecContext* avCodecContext = fnAVStream->codecContext;
  if (AVMEDIA_TYPE_AUDIO == avCodecContext->codec_type) {
    // Audio codecs.
    if (avcodec_open2(avCodecContext, avCodec, nullptr) < 0) {
      iop->critical("could not open audio codec");
      return -1;
    }
  } 
  else if (AVMEDIA_TYPE_VIDEO == avCodecContext->codec_type) {
    if (!currentCodecPixelFormats_.empty()) {
      std::string name = currentCodecPixelFormats_[_pixelFormat].ffmpegName;
      avCodecContext->pix_fmt = av_get_pix_fmt(name.c_str());
    }
      
    if (AV_CODEC_ID_PRORES == avCodecContext->codec_id ||
        AV_CODEC_ID_DNXHD == avCodecContext->codec_id ||
        AV_CODEC_ID_H264 == avCodecContext->codec_id) {
      // Apple ProRes and Avid DNxHD specific.
      // At this point the underlying writer is configured to
      // write ProRes or DNxHD into a QuickTime mov container.
      // Cache a few parameters that are used purely for
      // configuration and then restore them after the writer
      // has been configured.
      AVPixelFormat avPixelFormatTemp = avCodecContext->pix_fmt;
      const int bits_per_raw_sample = avCodecContext->bits_per_raw_sample;
      const int srcWidth = avCodecContext->width;
      const int srcHeight = avCodecContext->height;
      if (AV_CODEC_ID_DNXHD == avCodecContext->codec_id) {
        if (!NukeCodecs::CodecFactory::isCodecTypeDNxHR(licensedCodec_->codecTag)) {
          // This writer will rescale for any source format
          // that is not natively supported by DNxHD. Check
          // that the source format matches a native DNxHD
          // format. If it's not supported, then force to
          // 1920x1080 (the most flexible format for DNxHD).
          // This will also mean that avcodec_open2 will not
          // fail as the format matches a native DNxHD format.
          // Any other frame dimensions result in error.
          //
          // TODO: 'thin raster' is broken in FFmpeg so the
          //       avcodec_open2 call currently fails. Need
          //       to investigate and fix this problem.
          if ((((1920 == srcWidth) /*|| (1440 == srcWidth)*/) && (1080 == srcHeight)) ||
             (((1280 == srcWidth) /*|| (960 == srcWidth)*/) && (720 == srcHeight))) {
            // No conversion necessary.
          }
          else {
            avCodecContext->width = 1920;
            avCodecContext->height = 1080;
          }
        }
      }
      // #cristian.chera - Workaround - !!FIX ME in FFmpeg!!
      // workaround for odd resolutin in FFmpeg library;
      // for more information see: TP 162306 - [Regression]: Nuke
      // issues errors when trying to render ProRes moves with
      // resolutions that contain odd numbers
      else if (AV_CODEC_ID_PRORES == avCodecContext->codec_id) {
        avCodecContext->width = mFnAlignSize(avCodecContext->width, 2);
      }

      static int dnxhdCodecProfileBitRates[][2] = {
        {440, AV_PIX_FMT_YUV422P10},
        {220, AV_PIX_FMT_YUV422P10},
        {220, AV_PIX_FMT_YUV422P},
        {145, AV_PIX_FMT_YUV422P},
        {36,  AV_PIX_FMT_YUV422P}
      };

      avCodecContext->bit_rate  = dnxhdCodecProfileBitRates[dnxhdCodecProfile_][0] * 1000000;
      avCodecContext->pix_fmt   = (AVPixelFormat)dnxhdCodecProfileBitRates[dnxhdCodecProfile_][1];

      // Restore the desired clip width if it was changed above
      // to workaround the issue in FFmpeg
      if (AV_CODEC_ID_PRORES == avCodecContext->codec_id) {
        avCodecContext->width = srcWidth;
      }
      // #cristian.chera - End Workaround

      // Restore the desired pixel format for the licensed codec
      // library encoder.
      avCodecContext->pix_fmt = avPixelFormatTemp;
      avCodecContext->bits_per_raw_sample = bits_per_raw_sample;

      // Use the correct codec four character code for the
      // ProRes or DNxHD codec profile.
      //
      // !! This must be done AFTER the call to avcodec_open2 !!
      //
#if defined(FN_LICENSED_PRORES_DNXHD_CODEC)
      avCodecContext->codec_tag = SWAP32(licensedCodec_->codecTag); // Convert to a four character code.
#endif // defined(FN_LICENSED_PRORES_DNXHD_CODEC)

      // Avid DNxHD is a little different to Apple ProRes.
      // Apple ProRes simply has bit rate definitions (quality)
      // for each codec profile: 4444, 422HQ, 422, etc. It has
      // no dependency on the dimensions of the frame.
      // Avid DNxHD on the otherhand, has in addition to bit
      // rate definitions (quality), definitions of frame
      // dimensions for each codec profile.
      // The following uses the UI 'quality' setting and the
      // project frame dimensions to select the appropriate
      // compression ID.
      // In the case of ProRes, the codec bit rate is identified
      // by a four character code (the codec tag) which also
      // identifies the codec type. This was done to keep things
      // simple in selecting the appropriate encoder. However
      // for DNxHD there is only a single four character code
      // for all bit rates, so the following determines the
      // codec type from the option selected in the UI.
      if (AV_CODEC_ID_PRORES == avCodecContext->codec_id ||
          AV_CODEC_ID_H264 == avCodecContext->codec_id) {
#if defined(FN_LICENSED_PRORES_DNXHD_CODEC)
        if (IsProres(licensedCodec_->shortName)) {
          codecType_ = proresCodecTags[proresCodecProfile_];
          avCodecContext->codec_tag = SWAP32(proresCodecTags[proresCodecProfile_]);
        }
        else {
            codecType_ = licensedCodec_->codecTag;
        }
#endif // defined(FN_LICENSED_PRORES_DNXHD_CODEC)
      }
#if defined(FN_LICENSED_PRORES_DNXHD_CODEC)
      else {
        if (NukeCodecs::CodecFactory::isCodecTypeDNxHR(licensedCodec_->codecTag)) {
          codecType_ = CodecProfile::dnxhrProfiles[dnxhrCodecProfile_];
        }
        else {
          codecType_ = CodecProfile::dnxhdProfiles[static_cast<int>(1280 == srcWidth)][dnxhdCodecProfile_];
        }
      }

      // Create an instance of the licensed encoder class.
      encoder_.reset(NukeCodecs::CodecFactory::makeEncoder(codecType_));
        
      if (!encoder_.get()) {
        iop->critical("unable to initialise licensed codec");
        return -3;
      }
        
      if (AV_CODEC_ID_H264 == avCodecContext->codec_id) {
        if (configureH264Encoder(avCodecContext) != NukeCodecs::CodecFactory::kError_NoError) {
          encoder_.reset();
          iop->critical("unable to initialise h264 encoder");
          return -3;
        }
      }
#endif //defined(FN_LICENSED_PRORES_DNXHD_CODEC)
    }
    else if (avcodec_open2(avCodecContext, avCodec, nullptr) < 0) {
      iop->critical("unable to open video codec");
      return -4;
    }
      
    if (licensedCodec_ && NukeCodecs::CodecFactory::isCodecTypeDNxHR(licensedCodec_->codecTag)) {
      static const std::array<int, 5> dnxhrFFmpegProfile = {
        FF_PROFILE_DNXHR_444,
        FF_PROFILE_DNXHR_HQX,
        FF_PROFILE_DNXHR_HQ,
        FF_PROFILE_DNXHR_SQ,
        FF_PROFILE_DNXHR_LB
      };
        
      static const std::array<AVPixelFormat, 5> dnxhrPixelFormat = {
        AV_PIX_FMT_RGB48LE,
        AV_PIX_FMT_YUV422P12LE,
        AV_PIX_FMT_YUV422P,
        AV_PIX_FMT_YUV422P,
        AV_PIX_FMT_YUV422P
      };
        
      avCodecContext->profile = dnxhrFFmpegProfile[dnxhrCodecProfile_];
      avCodecContext->pix_fmt = dnxhrPixelFormat[dnxhrCodecProfile_];
    }
  }
  else if (AVMEDIA_TYPE_DATA == avCodecContext->codec_type) {
    // Timecode codecs.
  }

  return 0;
}

////////////////////////////////////////////////////////////////////////////////
// writeAudio
//
// * Read audio from the source file (this will also convert to the desired
//   sample format for the file).
// * Encode
// * Write to file.
//
// NOTE: Writing compressed audio is not yet implemented. E.g. 'flushing' the
//       encoder when the writer has finished writing all the video frames.
//
// @param avFormatContext A reference to an AVFormatContext of the file.
// @param avStream A reference to an AVStream of an audio stream.
// @param flush A boolean value to flag that any remaining frames in the interal
//              queue of the encoder should be written to the file. No new
//              frames will be queued for encoding.
//
// @return 0 if successful,
//         <0 otherwise for any failure to read (and convert) the audio format,
//         encode the audio or write to the file.
//
int mov64Writer::writeAudio(AVFormatContext* avFormatContext, FnAVStream* fnAVStream, bool flush)
{
  int ret = 0;

  FnAVFrame avFrame;
  const int audioReaderResult = audioReader_->read(avFrame);
  if ( audioReaderResult > 0 ) { //read sucessful
    AVPacket pkt = {nullptr}; // data and size must be 0
    av_init_packet(&pkt);
    if (flush) {
      // A slight hack.
      // So that the durations of the video track and audio track will be
      // as close as possible, when flushing (finishing) calculate how many
      // remaining audio samples to write using the difference between the
      // the duration of the video track and the duration of the audio track.
      const double videoTime = (av_stream_get_end_pts(streamVideo_.stream) + av_rescale_q(1, streamVideo_.codecContext->time_base, streamVideo_.stream->time_base)) * av_q2d(streamVideo_.stream->time_base);
      const double audioPts = static_cast<int>(av_stream_get_end_pts(streamAudio_.stream));
      const int samplesToFinish = static_cast<int>((videoTime / av_q2d(streamAudio_.stream->time_base))-audioPts);
      if (0 < samplesToFinish) {
        // Add one sample to the count to guarantee that the audio track
        // will be the same length or very slightly longer than the video
        // track. This will then end the final loop that writes audio up
        // to the duration of the video track.
        if (avFrame->nb_samples > samplesToFinish)
          avFrame->nb_samples = samplesToFinish;
      }
    }

    AVCodecContext* avCodecContext = fnAVStream->codecContext;
    int gotPacket;
    ret = mov64_av_encode(avCodecContext, &pkt, avFrame, &gotPacket);
    if (!ret && gotPacket) {
      pkt.stream_index = fnAVStream->stream->index;
      ret = av_write_frame(avFormatContext, &pkt);
    }

    if (ret < 0) {
      // Report the error.
      char szError[1024];
      av_strerror(ret, szError, 1024);
      iop->error(szError);
    }
  }
  else if ( audioReaderResult < 0 ) {
    //error in read from audio reader
    ret = audioReaderResult;
  }

  return ret;
}

////////////////////////////////////////////////////////////////////////////////
// More of a utility function added since supported DNxHD.
// This codec requires colour space conversions, and more to the point, requires
// VIDEO levels for all formats. Specifically, 4:4:4 requires video levels on
// the input RGB component data!
//
int mov64Writer::colourSpaceConvert(AVFrame* avFrameIn, AVFrame* avFrameOut, AVPixelFormat srcPixelFormat, AVPixelFormat dstPixelFormat, AVCodecContext* avCodecContext, YCbCr_MatrixType conversionMatrix)
{
    
  if (colorSpaceConvertor_) {
    cscParams_.in_planar = reinterpret_cast<const void **>(const_cast<const uint8_t **>(avFrameIn->data));
    cscParams_.in_linesize = avFrameIn->linesize;
      
    cscParams_.out_planar = reinterpret_cast<void **>(avFrameOut->data);
    cscParams_.out_linesize = avFrameOut->linesize;
      
    return colorSpaceConvertor_->Decode(cscParams_);
  }
    
  int ret = 0;

  SwsContext* convertCtx = sws_getContext(width(), height(), srcPixelFormat, // from
                                          avCodecContext->width, avCodecContext->height, dstPixelFormat,// to
                                          SWS_BICUBIC, nullptr, nullptr, nullptr);

  // Only apply colorspace conversions for YUV.
  if (IsYUV(dstPixelFormat)) {
    // Set up the sws (SoftWareScaler) to convert colourspaces correctly, in the sws_scale function below
    const int colorspace = [this, &conversionMatrix]() {
        switch(conversionMatrix) {
          case YCbCr_MatrixType::Rec601: return SWS_CS_ITU601;
          case YCbCr_MatrixType::Rec709: return SWS_CS_ITU709;

          default:
          case YCbCr_MatrixType::Auto: return isRec709Format(height()) ? SWS_CS_ITU709 : SWS_CS_ITU601;
        }  
      }()
    ;
    const int dstRange = (avCodecContext->codec_id == AV_CODEC_ID_MJPEG || avCodecContext->codec_id == AV_CODEC_ID_MJPEGB || avCodecContext->codec_id == AV_CODEC_ID_MJPEGA) ? 1 : 0; // 0 = 16..235, 1 = 0..255

    ret = sws_setColorspaceDetails(convertCtx,
                                   sws_getCoefficients(SWS_CS_DEFAULT), // inv_table
                                   1, // srcRange - 0 = 16..235, 1 = 0..255
                                   sws_getCoefficients(colorspace), // table
                                   dstRange, // dstRange - 0 = 16..235, 1 = 0..255
                                   0, // brightness fixed point, with 0 meaning no change,
                                   1 << 16, // contrast   fixed point, with 1<<16 meaning no change,
                                   1 << 16); // saturation fixed point, with 1<<16 meaning no change);
  }

  sws_scale(convertCtx,
            avFrameIn->data, // src
            avFrameIn->linesize, // src rowbytes
            0,
            height(),
            avFrameOut->data, // dst
            avFrameOut->linesize); // dst rowbytes

  sws_freeContext(convertCtx);

  return ret;
}

bool mov64Writer::alphaEnabled() const
{
  // Note that this is called from the engine function, and so needs to be thread safe.
  ChannelSet writeChannels = iop->channels();
  const ChannelSet validChannels = info().channels();
  writeChannels &= validChannels;
  return writeChannels.contains(Mask_Alpha);
}

bool mov64Writer::alphaEnabledNoValidation() const
{
  ChannelSet writeChannels = iop->channels();
  return writeChannels.contains(Mask_Alpha);
}

int mov64Writer::numberOfDestChannels() const
{
  // 4 channels (RGBA) if alpha is enabled, else 3 (RGB)
  return alphaEnabled() ? 4 : 3;
}


bool mov64Writer::codecIndexIsInRange( size_t codecIndex) const
{
  const bool isInRange = ( codecIndex < codecsShortNames_.size() );
  mFnAssertMsg( isInRange, "codec index is outside range" );
  return isInRange;
}

bool mov64Writer::codecIsDisallowed( const std::string& codecShortName,
                                     std::string& outReason ) const
{
  StringStringMap::const_iterator found =
      disallowedErrorMessages_.find( std::string( codecShortName ) );
  if ( found != disallowedErrorMessages_.end() )
  {
    outReason = found->second;
    return true;
  }
  return false;
}

////////////////////////////////////////////////////////////////////////////////
// * Convert Nuke float RGB values to the pixel format of the encoder.
// * Encode.
// * Write to file.
//
// @param avFormatContext A reference to an AVFormatContext of the file.
// @param avStream A reference to an AVStream of a video stream.
// @param flush A boolean value to flag that any remaining frames in the interal
//              queue of the encoder should be written to the file. No new
//              frames will be queued for encoding.
//
// @return 0 if successful,
//         <0 otherwise for any failure to convert the pixel format, encode the
//         video or write to the file.
//
int mov64Writer::writeVideo(AVFormatContext* avFormatContext, FnAVStream* fnAVStream, bool flush)
{
  if ( isOpen_ ) {
    AVStream* avStream = fnAVStream->stream;
    AVCodecContext* avCodecContext = fnAVStream->codecContext;
    if ( avStream !=nullptr ) {
      // FIXME enum needed for error codes.
      int ret = 0;
      // First convert from Nuke floating point RGB to either 16-bit or 8-bit RGB.
      // Create a buffer to hold either  16-bit or 8-bit RGB.
      // Create another buffer to convert from either 16-bit or 8-bit RGB
      // to the input pixel format required by the encoder.
      AVPixelFormat pixelFormatCodec = avCodecContext->pix_fmt;

      if (!flush) {
        const ChannelSet validChannels = info().channels();
        ChannelSet writeChannels = iop->channels();
        const bool writeAll = writeChannels.all();
        if (writeAll) {
          // Mask to RGBA, because we don't support writing channels outside of this range
          writeChannels &= Mask_RGBA;
        }
        // Only write channels which are valid, as some channels may
        // not be available from the input.
        writeChannels &= validChannels;

        // FIXME This code is currently set up only to be able to write from RGBA channels (despite the
        // fact that the user is able to select channels outside of this range.
        const bool hasNonRGBAChans = (writeChannels.value() & ~Mask_RGBA);
        if (hasNonRGBAChans) {
          ChannelSet invalidChannels(static_cast<ChannelSetInit>(writeChannels.value() & ~Mask_RGBA));
          Channel firstInvalidChannel = invalidChannels.first();
          iop->error("mov64writer ERROR: Cannot write non-RGBA channel - %s", getName(firstInvalidChannel) );
          return -4; // Channel error
        }

        const bool hasAlpha = alphaEnabled();
          
        AVPixelFormat pixelFormatNuke;
        if (hasAlpha) {
          pixelFormatNuke = (avCodecContext->bits_per_raw_sample > 8) ? AV_PIX_FMT_RGBA64 : AV_PIX_FMT_RGBA;
        }
        else {
          pixelFormatNuke = (avCodecContext->bits_per_raw_sample > 8) ? AV_PIX_FMT_RGB48LE : AV_PIX_FMT_RGB24;
        }
          
        int numDestChannels = numberOfDestChannels();
        if (codecType_ == NukeCodecs::CodecFactory::eProResCodec4444XQ ||
            codecType_ == NukeCodecs::CodecFactory::eProResCodec4444) {
          //ProRes 4444 always encoded as b64a, so even if we dont want to save alpha we allocate memory for it
          numDestChannels = 4;
          pixelFormatNuke = AV_PIX_FMT_RGBA64;
        }
          
        if (!inputFrame_ || !outputFrame_) {
          initColorSpaceConvertor(avCodecContext);
        
          //if we succeed to create CSC it means we will convert directly from RGB float to dst PixFmt
          if (colorSpaceConvertor_ || isProRes422() || AV_CODEC_ID_H264 == avCodecContext->codec_id) {
            inputFrame_.reset(av_frame_alloc(), [](AVFrame * frame) { av_freep(&frame->data[0]); av_frame_free(&frame); });
            //little trick: as ffmpeg does not support float point color spaces we alloc it as 16 bit per pixel with x2 width
            av_image_alloc(inputFrame_->data, inputFrame_->linesize, 2 * width(), height(), AV_PIX_FMT_YUV444P16LE, 32);
              
            inputFrame_->width = avCodecContext->width;
            inputFrame_->height = avCodecContext->height;
            inputFrame_->format = avCodecContext->pix_fmt;

              
            if (!isProRes422() && AV_CODEC_ID_H264 != avCodecContext->codec_id) {
              outputFrame_.reset(av_frame_alloc(), [](AVFrame * frame) { av_freep(&frame->data[0]); av_frame_free(&frame); });
              outputFrame_->width = avCodecContext->width;
              outputFrame_->height = avCodecContext->height;
              outputFrame_->format = avCodecContext->pix_fmt;
                
              av_image_alloc(outputFrame_->data, outputFrame_->linesize, outputFrame_->width, outputFrame_->height, pixelFormatCodec, 32);
            }
          }
          else {
            inputFrame_.reset(av_frame_alloc(), [](AVFrame * frame) { av_freep(&frame->data[0]); av_frame_free(&frame); });
            ret = av_image_alloc(inputFrame_->data, inputFrame_->linesize, width(), height(), pixelFormatNuke, 32);
            
            if (ret < 0) {
              inputFrame_.reset();
            }
          }
        }

        if (inputFrame_) {
          ret = 0;
          // Convert floating point values to unsigned values.
          Row row(0, width());
            
          for (int y = 0; y < height(); ++y) {
            get(y, 0, width(), writeChannels, row);
            if (iop->aborted()) {
              return -3; // Input error
            }

            for (Channel z = writeChannels.first(); z != Chan_Black; z = writeChannels.next(z)) {
              const float* FROM = row[z];
              if (colorSpaceConvertor_ || isProRes422() || AV_CODEC_ID_H264 == avCodecContext->codec_id) {
                if (z == Chan_Red || z == Chan_Green || z == Chan_Blue) {
                  float * TO = reinterpret_cast<float *>(inputFrame_->data[z - 1]);
                  int32_t linesize = inputFrame_->linesize[z - 1];
                    
                  TO += (height() - y - 1) * (linesize / sizeof(float));
                  to_float(z - 1, TO, FROM, row[Chan_Alpha], width());
                }
              }
              else if (avCodecContext->bits_per_raw_sample > 8) {
                // void to_short(int z, U16* to, const float* from, const float* alpha, int W, int bits = 16, int delta = 1);
                U16* TO = reinterpret_cast<U16*>(inputFrame_->data[0]);
                // avFrameIn->linesize is in bytes, but stride is U16 (2 bytes), so divide linesize by 2
                TO += (height() - y - 1) * (inputFrame_->linesize[0] / 2) + z - 1;
                to_short(z - 1, TO, FROM, row[Chan_Alpha], width(), 16, numDestChannels);
              } 
              else {
                // void to_byte(int z, uchar* to, const float* from, const float* alpha, int W, int delta = 1);
                uint8_t* TO = inputFrame_->data[0];
                TO += (height() - y - 1) * inputFrame_->linesize[0] + z - 1;
                to_byte(z - 1, TO, FROM, row[Chan_Alpha], width(), numDestChannels);
              }
            }
          }

          // Apple ProRes and Avid DNxHD specific.
          // The underlying codec will handle the RGB -> YCbCr
          // or RGB -> RGB conversion, whatever is most appropriate
          // for the codec. (RGB -> RGB may involve component
          // reordering, scaling, endian swapping, or combinations
          // of these.)
          // The codec handles these conversions as optimally as
          // possible and generally does so better than a third
          // party implementation.
          if (AV_CODEC_ID_PRORES == avCodecContext->codec_id ||
              AV_CODEC_ID_H264 == avCodecContext->codec_id) {
            // For ProRes an intermediate buffer is not required.
            outputFrame_ = inputFrame_;
          }
          else if (AV_CODEC_ID_DNXHD == avCodecContext->codec_id) {

            // Rescale if the source format is not supported by
            // DNxHD.
            int srcWidth = width();
            int srcHeight = height();
            // TODO: 'thin raster' is broken in FFmpeg so the
            //       avcodec_open2 call currently fails. Need
            //       to investigate and fix this problem.
            if (NukeCodecs::CodecFactory::isCodecTypeDNxHR(licensedCodec_->codecTag) ||
                ((((1920 == srcWidth) /*|| (1440 == srcWidth)*/) && (1080 == srcHeight)) ||
                (((1280 == srcWidth) /*|| (960 == srcWidth)*/) && (720 == srcHeight)))) {
              // No conversion necessary.
              outputFrame_ = inputFrame_;
            }
            else {
              // Rescale.
              // Default to resizing to 1920x1080. The reason being
              // that DNxHD supports all bit rates for 1080p. Whereas
              // DNxHD does not support all bitrates for 720p. So 1080p
              // is the most flexible destination format.
              
              if (!outputFrame_) {
                outputFrame_.reset(av_frame_alloc(), [](AVFrame * frame) { av_freep(&frame->data[0]); av_frame_free(&frame); });
                outputFrame_->width = avCodecContext->width;
                outputFrame_->height = avCodecContext->height;
                outputFrame_->format = pixelFormatNuke;
                
                int bufSize = av_image_alloc(outputFrame_->data, outputFrame_->linesize, outputFrame_->width, outputFrame_->height, pixelFormatNuke, 32);
                  
                if (bufSize < 0) {
                  outputFrame_.reset();
                }
              }
              if (outputFrame_) {
                colourSpaceConvert(inputFrame_.get(), outputFrame_.get(), pixelFormatNuke, pixelFormatNuke, avCodecContext, _YCbCr_MatrixType);
              } 
              else {
                // av_image_alloc failed.
                ret = -1;
              }
            }
          }
          else {
            // For any codec other than ProRes or DNxHD an
            // intermediate buffer is allocated for the
            // colour space conversion.
              
            if (!outputFrame_) {
              outputFrame_.reset(av_frame_alloc(), [](AVFrame * frame) { av_freep(&frame->data[0]); av_frame_free(&frame); });

              outputFrame_->width = avCodecContext->width;
              outputFrame_->height = avCodecContext->height;
              outputFrame_->format = avCodecContext->pix_fmt;
              
              int bufferSize = av_image_alloc(outputFrame_->data, outputFrame_->linesize, outputFrame_->width, outputFrame_->height, pixelFormatCodec, 32);
                  
              if (bufferSize < 0) {
                outputFrame_.reset();
              }
            }


            if (outputFrame_) {
              colourSpaceConvert(inputFrame_.get(), outputFrame_.get(), pixelFormatNuke, pixelFormatCodec, avCodecContext, _YCbCr_MatrixType);
            } 
            else {
              // av_image_alloc failed.
              ret = -1;
            }
          }
        }
      }
      else {
        inputFrame_.reset();
        outputFrame_.reset();
        colorSpaceConvertor_.reset();
      }

      if (!ret) {
        bool error = false;

        AVPacket pkt;
        av_init_packet(&pkt);
        // NOTE: If |flush| is true, then avFrameOut will be NULL at this point as
        //       alloc will not have been called.

        const int bytesEncoded = encodeVideo(avCodecContext, outputFrame_.get(), pkt);
        const bool encodeSucceeded = (bytesEncoded > 0);
        if (encodeSucceeded) {
          // Each of these packets should consist of a single frame therefore each one
          // must have a PTS in order to correctly set the frame-rate
          if ((uint64_t)(pkt.pts) != AV_NOPTS_VALUE) {
            pkt.pts = av_rescale_q(pkt.pts, avCodecContext->time_base, avStream->time_base);
          } else {
            pkt.pts = av_rescale_q(pts_counter_, avCodecContext->time_base, avStream->time_base);
            pts_counter_++;
          }

          pkt.dts = AV_NOPTS_VALUE;


          if (!pkt.duration) {
            pkt.duration = av_rescale_q(1, avCodecContext->time_base, avStream->time_base);
          }

          pkt.stream_index = avStream->index;

          const int writeResult = av_write_frame(avFormatContext, &pkt);
          if (pkt.data) {
            av_free(pkt.data);
          }

          const bool writeSucceeded = (writeResult == 0);
          if (!writeSucceeded) {
            // Report the error.
            char szError[1024];
            av_strerror(bytesEncoded, szError, 1024);
            iop->error(szError);
            error = true;
          }
        }
        else {
          if (bytesEncoded == AVERROR(EAGAIN) && !flush) {
            ret = 0;
          }
          else if (bytesEncoded < 0) {
            if (NukeCodecs::CodecFactory::isCodecTypeDNxHD(codecType_)) {
              switch (bytesEncoded) {
                case NukeCodecs::CodecFactory::kError_FailedToCreateEncoder:
                  iop->critical("Failed to encode frame, failed to create DNx encoder");             break;
                case NukeCodecs::CodecFactory::kError_FailedToConfigureEncoder:
                  iop->critical("Failed to encode frame, failed to configure DNx encoder");          break;
                case NukeCodecs::CodecFactory::kError_CodecNotSupported:
                  iop->critical("Failed to encode frame, DNx encoder does not support this codec");  break;
                case NukeCodecs::CodecFactory::kError_IncorrectMemAlignment:
                  iop->critical("Failed to encode frame, the output buffer is not byte aligned");    break;
                case NukeCodecs::CodecFactory::kError_ColorspaceConversion:
                  iop->critical("Failed to encode frame, DNx colour space conversion failed");       break;
                case NukeCodecs::CodecFactory::kError_NoAlphaSupport:
                  iop->critical("Failed to encode frame, DNxHD alpha writing not supported");        break;
                default:
                  iop->critical("Failed to encode frame, unknown error");                            break; 
              }
              ret = -2;
            }
            else {
              // Report the error.
              char szError[1024];
              av_strerror(bytesEncoded, szError, 1024);
              iop->error(szError);
              error = true;
            }
          }
          else if (flush) {
            // Flag that the flush is complete.
            ret = -10;
          }
        }

        if (error) {
          iop->critical("error writing frame to file");
          ret = -2;
        }
      }

      return ret;
    }
    else {
      mFnAssertMsg(isOpen_, "If writer is open there should be an avStream to write to");
      return -6;
    }
  }
  else {
    return -5; //writer is not open!
  }
}

////////////////////////////////////////////////////////////////////////////////
// encodeVideo
// Encode a frame of video.
//
// Note that the uncompressed source frame to be encoded must be in an
// appropriate pixel format for the encoder prior to calling this method as
// this method does NOT perform an pixel format conversion, e.g. through using
// Sws_xxx.
//
// @param avCodecContext A reference to an AVCodecContext of a video stream.
// @param avFrame A constant reference to an AVFrame that contains the source data
//                to be encoded. This must be in an appropriate pixel format for
//                the encoder.
// @param avPacketOut A reference to an AVPacket to receive the encoded frame.
//
// @return <0 for any failure to encode the frame, otherwise the size in byte
//         of the encoded frame.
//
int mov64Writer::encodeVideo(AVCodecContext* avCodecContext, const AVFrame* avFrame, AVPacket& avPacketOut)
{
  int ret, got_packet = 0;

  avPacketOut.data = nullptr;
  avPacketOut.size = 0;

#if defined(FN_LICENSED_PRORES_DNXHD_CODEC)
  if (AV_CODEC_ID_PRORES == avCodecContext->codec_id ||
      AV_CODEC_ID_DNXHD == avCodecContext->codec_id ||
      AV_CODEC_ID_H264 == avCodecContext->codec_id) {
      // Non-I-frame only codecs require 'flushing' of the final
      // frames out of the codec. ProRes and DNxHD are I-frame
      // only codecs and therefore do not need to be 'flushed'.
      // 'Flushing' is indicated through a null pointer argument
      // so trap the null pointer here to avoid passing a null
      // pointer to the ProRes or DNxHD libraries.
      if (avFrame) {
        // Apple ProRes and Avid DNxHD specific.
        // Encode a frame using the licensed encoder.
        NukeCodecs::EncoderParams params;
        if (!isProRes422() && AV_CODEC_ID_H264 != avCodecContext->codec_id) {
          params.in = avFrame->data[0];
          params.inRowBytes = avFrame->linesize[0];
        }
        else {
          params.in_planar = reinterpret_cast<const void **>(const_cast<const uint8_t **>(avFrame->data));
          params.in_linesize = avFrame->linesize;
          params.in_planes = 3;
        }
        params.width = avCodecContext->width;
        params.height = avCodecContext->height;
        params.codecType = codecType_;
        params.dstDataRange = encodeVideoRange_ ? NukeCodecs::EncoderParams::kVideoRange : NukeCodecs::EncoderParams::kFullRange;
        params.numberOfComponents = numberOfDestChannels();
        params.bitsPerComponent = (avCodecContext->bits_per_raw_sample > 8) ? 16 : 8; // Currently on 8 or 16 bits per component.;

        if (AV_CODEC_ID_H264 == avCodecContext->codec_id) {
          NukeCodecs::H264Frame f;
            
          params.out = &f;
          params.outSize = sizeof(f);
          ret = encoder_->Encode(params);
        
          if (ret == NukeCodecs::CodecFactory::kError_NoError) {
            av_packet_from_data(&avPacketOut, static_cast<uint8_t *>(f.data), f.size);
            avPacketOut.pts = AV_NOPTS_VALUE; //todo
            if  (f.key_frame) {
              avPacketOut.flags |= AV_PKT_FLAG_KEY;
            }
            ret = f.size;
          }
          else if (ret == NukeCodecs::CodecFactory::kError_NeedMoreData) {
            ret = AVERROR(EAGAIN);
          }
        }
        else {
          // Alpha DNxHD writing not supported;
          // Make sure the stream isn't a alpha DNxHD Codec,
          // otherwise show an error;
          // Encoder returns an error code for this situation, but
          // error handling isn't good enough yet to catch this exceptions
          const int picSize =  encoder_->GetCompressedFrameSize(params);
          if (picSize > 0) {
            uint8_t* outdata = (uint8_t*)av_malloc(picSize);

            params.out = outdata;
            params.outSize = picSize;
            ret = encoder_->Encode(params);
            if (ret > 0) {
              av_packet_from_data(&avPacketOut, outdata, ret);
              avPacketOut.pts = AV_NOPTS_VALUE;
              avPacketOut.flags |= AV_PKT_FLAG_KEY;
            }
            else {
              ret = -2;
            }
          }
          else {
            ret = picSize;
          }
        }


      }
      else {
        if (AV_CODEC_ID_H264 == avCodecContext->codec_id) {
          encoder_->Flush(false);
        
          NukeCodecs::H264Frame f;
              
          NukeCodecs::EncoderParams params;
          params.out = &f;
          params.outSize = sizeof(f);
          
          ret = encoder_->Encode(params);
      
          if (ret == NukeCodecs::CodecFactory::kError_NoError) {
            av_packet_from_data(&avPacketOut, static_cast<uint8_t *>(f.data), f.size);
            avPacketOut.pts = AV_NOPTS_VALUE; //todo
            if  (f.key_frame) {
              avPacketOut.flags |= AV_PKT_FLAG_KEY;
            }
            ret = f.size;
          }
          else if (ret == NukeCodecs::CodecFactory::kError_EndOfStream) {
            ret = 0;
          }
        }
        else {
        // Signal that there is no further data to encode.
          ret = 0;
        }
      }
  }
  else
#endif // defined(FN_LICENSED_PRORES_DNXHD_CODEC)
  {
    ret = mov64_av_encode(avCodecContext, &avPacketOut, avFrame, &got_packet);
    if (!ret && got_packet) {
      ret = avPacketOut.size;
    }
  }

  return ret;
}

////////////////////////////////////////////////////////////////////////////////
// writeToFile
// Write video and if specifed audio to the movie. Interleave the audio and
// video as specified in the QuickTime movie recommendation. This is to have
// ~0.5s of audio interleaved with the video.
//
// @param avFormatContext A reference to the AVFormatContext of the file to
//                        write.
// @param finalise A flag to indicate that the streams should be flushed as
//                 no further frames are to be encoded.
//
// @return 0 if successful.
//         <0 otherwise.
//
int mov64Writer::writeToFile(AVFormatContext* avFormatContext, bool finalise)
{
  // Write interleaved audio and video if an audio file has
  // been specified. Otherwise just write video.
  //
  // If an audio file has been specified, calculate the
  // target stream time of the audio stream. For a QuickTime
  // movie write the audio in ~0.5s chunks so that there is
  // an approximate 0.5s interleave of audio and video.
  if (streamAudio_.stream && streamVideo_.stream) {
    const double videoTime = av_stream_get_end_pts(streamVideo_.stream) * av_q2d(streamVideo_.stream->time_base);
    // Determine the current audio stream time.
    double audioTime = av_stream_get_end_pts(streamAudio_.stream) * av_q2d(streamAudio_.stream->time_base);
    // Determine the target audio stream time. This is
    // the current video time rounded to the nearest
    // 0.5s boundary.
    const double targetTime = ((int)(videoTime / 0.5)) * 0.5;
    if (audioTime < targetTime) {
      // If audio stream is more than 0.5s behind, write
      // another ~0.5s of audio.
      const double sourceDuration = audioReader_->getDuration();
      int writeResult = 0;
      while ((audioTime < targetTime) && (audioTime < sourceDuration) && (writeResult >= 0)) {
        writeResult = writeAudio( avFormatContext, &streamAudio_, finalise );
        audioTime = av_stream_get_end_pts(streamAudio_.stream) * av_q2d(streamAudio_.stream->time_base);
      }
    }
  }

  return writeVideo(formatContext_, &streamVideo_, finalise);
}

////////////////////////////////////////////////////////////////////////////////
// open
// Internal function to create all the required streams for writing a QuickTime
// movie.
//
// @return true if successful,
//         false otherwise.
//
bool mov64Writer::open()
{
  AVOutputFormat* avOutputFormat = initFormat(/* reportErrors = */ true);
  if (!avOutputFormat) {
    return false;
  }

  if (!formatContext_) {
    avformat_alloc_output_context2(&formatContext_, avOutputFormat, nullptr, filename());
  }

  // Create an audio stream if a file has been provided.
  if (_audioFile && (strlen(_audioFile) > 0)) {
    if (!streamAudio_.stream) {
      // Attempt to create an audio reader.
      audioReader_.reset(new AudioReader());

      // TODO: If the sample format is configurable via a knob, set
      //       the desired format here, e.g.:
      //audioReader_->setSampleFormat(_avSampleFormat);

      if (!audioReader_->open(_audioFile)) {
        AVCodec* audioCodec = nullptr;
        addStream(formatContext_, audioReader_->getCodecID(), &audioCodec, &streamAudio_);
        if (!streamAudio_.stream || !audioCodec) {
          freeFormat();
          return false;
        }

        // Bug 45010 The following flags must be set BEFORE calling
        // openCodec (avcodec_open2). This will ensure that codec
        // specific data is created and initialized. (E.g. for MPEG4
        // the AVCodecContext::extradata will contain Elementary Stream
        // Descriptor which is required for QuickTime to decode the
        // stream.)
        AVCodecContext* avCodecContext = streamAudio_.codecContext;
        if (!strcmp(formatContext_->oformat->name, "mp4") ||
            !strcmp(formatContext_->oformat->name, "mov") ||
            !strcmp(formatContext_->oformat->name, "3gp")) {
          avCodecContext->flags |= AV_CODEC_FLAG_GLOBAL_HEADER;
        }
        // Some formats want stream headers to be separate.
        if (formatContext_->oformat->flags & AVFMT_GLOBALHEADER)
          avCodecContext->flags |= AV_CODEC_FLAG_GLOBAL_HEADER;

        avcodec_parameters_from_context(streamAudio_.stream->codecpar, streamAudio_.codecContext);

        if (openCodec(formatContext_, audioCodec, &streamAudio_) < 0) {
          freeFormat();
          return false;
        }

        // If audio has been specified, set the start position (in seconds).
        // The values are negated to convert them from a video time to an
        // audio time.
        // So for a knob value of -10s, this means that the audio starts at
        // a video time of -10s. So this is converted to +10s audio time. I.e.
        // The video starts +10s into the audio. Vice-versa for a knob value
        // of +10s. The video starts -10s into the audio. In this case the
        // audio reader will provide 10s of silence for the first 10s of
        // video.
        audioReader_->setStartPosition(-((_audioOffsetUnit == 0) ? _audioOffset : (_audioOffset / fps_)));

      }
      else {
        iop->critical("failed to open the audio file\nIt does not contain audio or is an invalid file");
        return false;
      }
    }
  }

  // Create a video stream.
  AVCodecID codecId = AV_CODEC_ID_NONE;
  AVCodec* videoCodec = nullptr;
  if (!initCodec(avOutputFormat, codecId, videoCodec)) {
    freeFormat();
    return false;
  }
    
  // Test if the container recognises the codec type.
  bool isCodecSupportedInContainer = (avformat_query_codec(avOutputFormat, codecId, FF_COMPLIANCE_NORMAL) == 1);
  // mov seems to be able to cope with anything, which the above function doesn't seem to think is the case (even with FF_COMPLIANCE_EXPERIMENTAL)
  // and it doesn't return -1 for in this case, so we'll special-case this situation to allow this
  isCodecSupportedInContainer |= (strcmp(formatContext_->oformat->name, "mov") == 0);
  if (!isCodecSupportedInContainer) {
    iop->critical("the selected codec is not supported in this container.");
    return false;
  }

  // Avid DNxHD specific.
  // Validate the frame dimensions for DNxHD.
  // As its name suggests, DNxHD is HD only!
  if (AV_CODEC_ID_DNXHD == codecId) {
    // Like Apple ProRes, Avid DNxHD has a number of different
    // bit rate selections. However, the highest and  lowest
    // bit rate settings are not available for 720p. The
    // following is used to check ensure that only the valid
    // bit rate options for the frame dimensions are available.
    int w = width();
    int h = height();
    bool supported = true;
    // TODO: 'thin raster' is broken in FFmpeg so the
    //       avcodec_open2 call currently fails. Need
    //       to investigate and fix this problem.
    if (!NukeCodecs::CodecFactory::isCodecTypeDNxHR(licensedCodec_->codecTag) &&
        ((((1280 == w) /*|| (960 == w)*/) && (720 == h)) &&
        ((0 == dnxhdCodecProfile_) || (4 == dnxhdCodecProfile_)))) {
      supported = false;
    }
    if (!supported) {
      std::ostringstream oss;
      oss << "DNxHD does not support the selected bit rate for " << w << "x" << h;
      oss << std::endl << "Supported bit rates are:";
      oss << std::endl << CodecName::AVID_DNXHD_422_220X_NAME;
      oss << std::endl << CodecName::AVID_DNXHD_422_220_NAME;
      oss << std::endl << CodecName::AVID_DNXHD_422_145_NAME;
      iop->critical(oss.str().c_str());
      return false;
    }
  }

  AVPixelFormat targetPixelFormat     = AV_PIX_FMT_YUV420P;
  AVPixelFormat nukeBufferPixelFormat = AV_PIX_FMT_RGB24;
  int outBitDepth                     = 8;
  getPixelFormats(videoCodec, nukeBufferPixelFormat, targetPixelFormat, outBitDepth);

  if (!streamVideo_.stream) {
    addStream(formatContext_, codecId, &videoCodec, &streamVideo_);
    if (!streamVideo_.stream|| !videoCodec) {
      freeFormat();
      return false;
    }

    AVCodecContext* avCodecContext = streamVideo_.codecContext;
    avCodecContext->pix_fmt = targetPixelFormat;
    avCodecContext->bits_per_raw_sample = outBitDepth;

    // Now that the stream has been created, and the pixel format
    // is known, for DNxHD, set the YUV range.
    if (AV_CODEC_ID_DNXHD == avCodecContext->codec_id) {
      // Set the metadata for the YUV range. This modifies the appropriate
      // field in the 'ACLR' atom in the video sample description.
      // Set 'legal range' = 1 or 'full range' = 2.
      avCodecContext->color_range = encodeVideoRange_ ? AVColorRange::AVCOL_RANGE_MPEG : AVColorRange::AVCOL_RANGE_JPEG;
    }

    // Bug 45010 The following flags must be set BEFORE calling
    // openCodec (avcodec_open2). This will ensure that codec
    // specific data is created and initialized. (E.g. for MPEG4
    // the AVCodecContext::extradata will contain Elementary Stream
    // Descriptor which is required for QuickTime to decode the
    // stream.)
    if (!strcmp(formatContext_->oformat->name, "mp4") ||
        !strcmp(formatContext_->oformat->name, "mov") ||
        !strcmp(formatContext_->oformat->name, "3gp")) {
      avCodecContext->flags |= AV_CODEC_FLAG_GLOBAL_HEADER;
    }
    // Some formats want stream headers to be separate.
      if (formatContext_->oformat->flags & AVFMT_GLOBALHEADER)
        avCodecContext->flags |= AV_CODEC_FLAG_GLOBAL_HEADER;

    if (openCodec(formatContext_, videoCodec, &streamVideo_) < 0) {
      freeFormat();
      return false;
    }

    avcodec_parameters_from_context(streamVideo_.stream->codecpar, avCodecContext);

    if (!(avOutputFormat->flags & AVFMT_NOFILE)) {
      if (avio_open(&formatContext_->pb, filename(), AVIO_FLAG_WRITE) < 0) {
        iop->critical("unable to open file");
        freeFormat();
        return false;
      }
    }

    std::string movflags = "write_colr";
    if (fastStart_) {
      movflags += "+faststart";
    }

    AVDictionary* header_params = nullptr;
    av_dict_set(&header_params, "movflags", movflags.c_str(), 0);
    const bool writeHeaderResult = (avformat_write_header(formatContext_, &header_params) == 0);
    if (!writeHeaderResult) {
      iop->critical("An error has occurred during file writing. Please see the log for more details.");
      freeFormat();
      return false;
    }
  }

  // Special behaviour.
  //
  // R.e. libavformat/movenc.c::mov_write_udta_tag
  // R.e. libavformat/movenc.c::mov_write_string_metadata
  //
  // Prevent the @swr libavformat reference from appearing
  // in the 'udta' atom by setting the 'encoder' key in the
  // metadata to null. From movenc.c a zero length value
  // will not be written to the 'udta' atom.
  //
  AVDictionaryEntry* tag = av_dict_get(formatContext_->metadata, "encoder", nullptr, AV_DICT_IGNORE_SUFFIX);
  if (tag)
  {
#if defined(FN_LICENSED_PRORES_DNXHD_CODEC)
    if (licensedCodec_) {
      av_dict_set(&formatContext_->metadata, "encoder", licensedCodec_->longName, 0);
    }
    else
#endif
    {
      av_dict_set(&formatContext_->metadata, "encoder", videoCodec->name, 0);
    }
  }

#if FN_DEBUG
  const bool hasVideoStream = (streamVideo_.stream !=NULL);
  mFnAssertMsg( hasVideoStream, "Writer reports to be successfully open but video "
             "stream is not initialized" );
#endif //FN_DEBUG
  return true;
}

////////////////////////////////////////////////////////////////////////////////
// execute
// This is called once for each frame of video that is to be encoded and
// written to the file.
//
// For the first frame, configure the writer for encoding video and if an audio
// file has been specified, configure for encoding audio.
//
// Audio is written in ~0.5s chunks (r.e. ProRes QuickTime Guidelines.rtf in
// Thirdparty/ProResCodec/<version>/src/docs). The audio/video interleave in
// the file is achieved by monitoring the duration of the video stream as each
// video frame is encoded and written to the file. At every 0.5s interval of
// the video stream, a 0.5s chunk of audio is written to the file.
//
// When writing is completed there is a final call to |finish()|. The video
// stream is likely to be ahead of the audio stream at this point. Audio is
// written to the file so that the duration of the audio stream will match the
// duration of the video stream.
//
void mov64Writer::execute()
{
  size_t index = static_cast<size_t>(codec_);
  if (iop && iop->knob("mov64_codec"))
    index = (int)iop->knob("mov64_codec")->get_value();

  const bool codecIsInRange  = codecIndexIsInRange( index );
  if ( codecIsInRange ) {
    const std::string shortName( codecsShortNames_[index] );
    std::string errorMessage;
    if ( codecIsDisallowed( shortName, errorMessage ) )
    {
      iop->critical(errorMessage.c_str());
      return;
    }
  }

  if ( !isOpen_ ) {
    //One could think that since we do it just for the first frame, this open
    //could go in a symetric call to finish (start) that could be called from
    //beginExecuting. The problem is that at the beginExecuting stage, the optree
    //hasn't been validated, therefore the output format has not been set to the
    //correct one, resulting in 640x480 avformat.
    isOpen_ = open();
  }

  if (isOpen_) {
    writeToFile(formatContext_, false);
  }
}

////////////////////////////////////////////////////////////////////////////////
// finish
// Complete the encoding, finalise and close the file.
// Some video codecs may have data remaining in their encoding buffers. This
// must be 'flushed' to the file.
// This method flags the 'flush' to get the remaining encoded data out of the
// encoder(s) and into the file.
// Audio is written after video so it may require more writes from the audio
// stream in order to ensure that the duration of the audio and video tracks
// are the same.
//
void mov64Writer::finish()
{
  if ( isOpen_) {
    bool flushFrames = true;
    while (flushFrames) {
      // Continue to write the audio/video interleave while there are still
      // frames in the video and/or audio encoder queues, without queuing any
      // new data to encode.
      flushFrames = !writeToFile(formatContext_, true) ? true : false;
    }
    // The audio is written in ~0.5s chunks only when the video stream position
    // passes a multiple of 0.5s interval. Flushing the video encoder may result
    // in less than 0.5s of video being written to the file so at this point the
    // video duration may be longer than the audio duration. This final stage
    // writes enough audio samples to the file so that the duration of both
    // audio and video are equal.
    if (streamAudio_.stream && streamVideo_.stream) {
      //add frame duration to end pts so we dont discard last audio frame
      const double videoTime = (av_stream_get_end_pts(streamVideo_.stream) + av_rescale_q(1, streamVideo_.codecContext->time_base, streamVideo_.stream->time_base)) * av_q2d(streamVideo_.stream->time_base);
      // Determine the current audio stream time.
      double audioTime = av_stream_get_end_pts(streamAudio_.stream) * av_q2d(streamAudio_.stream->time_base);
      if (audioTime < videoTime) {
        int ret = 0;
        const double sourceDuration = audioReader_->getDuration();
        while ((audioTime < videoTime) && (audioTime < sourceDuration) && !ret) {
          ret = writeAudio(formatContext_, &streamAudio_, true);
          audioTime = av_stream_get_end_pts(streamAudio_.stream) * av_q2d(streamAudio_.stream->time_base);
        }
      }
    }

    const char* oldCodecName = nullptr;
    AVCodecContext* avCodecContext = streamVideo_.codecContext;
    if (AV_CODEC_ID_PRORES == avCodecContext->codec_id ||
        AV_CODEC_ID_DNXHD == avCodecContext->codec_id ||
        AV_CODEC_ID_H264 == avCodecContext->codec_id) {
      // Apple ProRes and Avid DNxHD specific.
      // At this point the underlying writer is configured to
      // write ProRes or DNxHD into a QuickTime mov container.
      if (streamVideo_.stream) {
        // The assumption is that the codec name stored in the AVCodecContext is
        // not important for FFmpeg encoding and can be overridden to support the
        // official codec names in the video tag.

        // Special behaviour.
        //
        // R.e. libavformat/movenc.c::mov_write_video_tag
        // R.e. libavformat/movenc.c::find_compressor
        //
        // Fix the codec name field. Replace the generic 'prores'
        // or 'dnxhd' tag used by the avcodec codec with the correct
        // codec profile name (the same string that appears if
        // encoding using QuickTime).
        // The codec name field is statically allocated so there is
        // no need to cache the original pointer. This results in
        // the correct codec name appearing in the QuickTime movie.
        AVCodec* avCodec = const_cast<AVCodec*>(avCodecContext->codec); // Cast away const to modify
        oldCodecName = avCodec->name; // Cache so that this can be restored
        if (AV_CODEC_ID_DNXHD == avCodecContext->codec_id) {
          avCodec->name = NukeCodecs::CodecFactory::isCodecTypeDNxHR(licensedCodec_->codecTag) ? "DNxHR" :  "DNxHD"; // Replace with the correct codec profile name.
        }
#if defined(FN_LICENSED_PRORES_DNXHD_CODEC)
        else {
          avCodec->name = licensedCodec_->longName; // Replace with the correct codec profile name.
        }
#endif // defined(FN_LICENSED_PRORES_DNXHD_CODEC)
      }
    }

    // Finalise the movie.
    av_write_trailer(formatContext_);

    if (streamVideo_.stream) {
      if (oldCodecName) {
        // Restore the original codec library name for the licensed codec (see above)
        AVCodec* avCodec = const_cast<AVCodec*>(avCodecContext->codec); // Cast away const to modify.
        avCodec->name = oldCodecName;
        oldCodecName = nullptr;
      }
    }

    freeFormat();
  }

  pts_counter_ = 0;
}

void mov64Writer::knobs(Knob_Callback f)
{
  // We no longer support mov32... All these knobs would error the script
  Obsolete_knob(f, "mov32_new_qt_behavior", nullptr);
  Obsolete_knob(f, "mov32_ycbcr_matrix", nullptr);
  Obsolete_knob(f, "mov32_codec", nullptr);
  Obsolete_knob(f, "mov32_codecoptions", nullptr);
  Obsolete_knob(f, "mov32_pixel_format", nullptr);
  Obsolete_knob(f, "mov32_fps", nullptr);
  Obsolete_knob(f, "mov32_quality", nullptr);
  Obsolete_knob(f, "mov32_keyframerate", nullptr);
  Obsolete_knob(f, "mov32_audiofile", nullptr);
  Obsolete_knob(f, "mov32_audio_offset", nullptr);
  Obsolete_knob(f, "mov32_units", nullptr);
  Obsolete_knob(f, "mov32_settings", nullptr);
  Obsolete_knob(f, "mov32_write_timecode", nullptr);
  Obsolete_knob(f, "mov32_advanced", nullptr);
  Obsolete_knob(f, "mov32_flatten", nullptr);
  Obsolete_knob(f, "mov32_use_format_aspect", nullptr);
  Obsolete_knob(f, "mov32_ycbcr_matrix_type", nullptr);
  Obsolete_knob(f, "mov32_write_nclc", nullptr);
  Obsolete_knob(f, "mov32_write_gamma", nullptr);
  Obsolete_knob(f, "mov32_write_prores", nullptr);
  Obsolete_knob(f, "meta_codec", nullptr);

  static const char* kEncoderLabels[] = { "mov64", "mov32", nullptr };
  Enumeration_knob(f, &metaEncoder_, kEncoderLabels, "meta_encoder", "encoder");
  SetFlags(f, DD::Image::Knob::INVISIBLE);
  Tooltip(f, "Select library used to decode");
  if (!f.makeKnobs()) {
    if (metaEncoder_ != 0) {
      iop->internalError("mov32 is no longer supported!");
      return;
    }
  }


  static std::vector<const char*> formatsAliases;

  formatsAliases.resize(formatsLongNames_.size());
  for (int i = 0; i < static_cast<int>(formatsLongNames_.size()); ++i)
    formatsAliases[i] = formatsLongNames_[i].c_str();
  formatsAliases.push_back(nullptr);

  // The 'format' knob is no longer supported - we will ALWAYS set mov64_format to default (even for mov)
  Obsolete_knob(f, "format", "if {$value ne {MOV format (mov)}} { alert {ffmpeg format is no longer supported} }");
  Enumeration_knob(f, &format_, &formatsAliases[0], "mov64_format");
  SetFlags(f, Knob::HIDDEN);

  if (!componentOfMovWriter_)
    Obsolete_knob(f, "codec", Foundry::Nuke::getMov64CodecTranslationScript());
  // Store the codec knob for use by the movWriter via getCodecKnob
  codecKnob_ = Enumeration_knob(f, &codec_, &codecsKnobLabels_[0], "mov64_codec", "Codec");
  SetFlags(f, Knob::KNOB_CHANGED_ALWAYS | Knob::ALWAYS_SAVE | Knob::EARLY_STORE);
  Tooltip(f, "Selects the codec to be used during encoding. This selection will affect available knobs and option settings.");

  // Obsolete knobs could stomp on shared mov32Writer ones, when used in the combined movWriter.
  // We need to handle them in Tcl.
  Obsolete_knob(f, "fps", componentOfMovWriter_ ? "knob mov64_fps $value\nknob mov32_fps $value"
                : "knob mov64_fps $value");
    
  Float_knob(f, &fps_, IRange(0.0, 120.0f), "mov64_fps", "FPS");
  Tooltip(f, "Sets the playback rate in frames per second for the output file.");

  Obsolete_knob(f, "dnxhd_codec_profile", "knob mov64_dnxhd_codec_profile $value");
  Enumeration_knob(f, &dnxhdCodecProfile_, CodecProfile::dnxhdCodecProfileLabels.data(), "mov64_dnxhd_codec_profile", "Codec Profile" );
  SetFlags(f, Knob::HIDDEN | Knob::DISABLED); // Disabled by default, enabled by selected a DNxHD codec profile.
  Tooltip(f, "Only for the Avid DNxHD codec, select the target bit rate for the encoded movie.");
    
  Enumeration_knob(f, &dnxhrCodecProfile_, CodecProfile::dnxhrProfileNames.data(), "mov64_dnxhr_codec_profile", "Codec Profile" );
  SetFlags(f, Knob::HIDDEN | Knob::DISABLED); // Disabled by default, enabled by selected a DNxHR codec profile.
  Tooltip(f, "Select DNxHR codec profile.");
    
  Enumeration_knob(f, &proresCodecProfile_, CodecProfile::proresProfiles.data(), "mov_prores_codec_profile", "Codec Profile" );
  //no need to disable as ProRes is default codec
  //SetFlags(f, Knob::HIDDEN | Knob::DISABLED);
  Tooltip(f, "Selects the quality and bit rate for the selected Codec.");
  Enumeration_knob(f, &h264CodecProfile_, CodecProfile::h264Profiles.data(), "mov_h264_codec_profile", "Codec Profile" );
  SetFlags(f, Knob::HIDDEN | Knob::DISABLED);
  Tooltip(f, "Select H264 codec profile.");

  Enumeration_knob(f, &_pixelFormat, nullptr, kPixelFormatKnob, "Pixel Format");
  SetFlags(f, Knob::SAVE_MENU | Knob::ALWAYS_SAVE | Knob::NO_ANIMATION | Knob::DISABLED | Knob::EXPAND_TO_CONTENTS | Knob::HIDDEN); // Save the entries for load/save and copy/paste
  Tooltip(f,
    "Lists pixel formats supported by the current codec.\n"
    "<b>Pixel colorspace</b> - either RGB(A) or YCbCr(A).\n"
    "<b>Pixel layout</b> - 4:2:2, 4:4:4, 4:4:4:4, etc. This defines how the chroma channels in the buffer are arranged."
    " 4:4:4 buffers have lower spatial chroma sampling than 4:2:2, so they are generally preferred when available.\n" 
    "<b>Pixel bit depth</b> - 8-bit or 16-bit. This sets the encoding depth used when decompressing the frames."
    " A large bit depth gives higher accuracy at the cost of speed and memory usage.\n"
    "This setting defaults to the best format accepted by the codec."
  );

  if (!f.makeKnobs()) {
    auto enumerationKnob = iop->knob(kPixelFormatKnob)->enumerationKnob();
    enumerationKnob->menu(displayedPixelFormatNames_);
  }

  qualityKnob_ = Enumeration_knob(f, &_quality, &kEncodeVideoQualityLabels[0], "mov64_quality", "Quality");
  SetFlags(f, Knob::HIDDEN | Knob::ALWAYS_SAVE | Knob::KNOB_CHANGED_ALWAYS);
  Tooltip(f, "The quality control is a preset for four of the parameters in the Advanced section. "
             "Best produces a large file size with high quality, whilst Least produces a much "
             "smaller size of lower quality.");

  static const char* kYCbCr_MatrixType_Labels[] = {
    "Auto",
    "Rec 601",
    "Rec 709",
    nullptr
  };
  Enumeration_knob(f, reinterpret_cast<int*>(&_YCbCr_MatrixType), kYCbCr_MatrixType_Labels, "mov64_ycbcr_matrix_type", "YCbCr Matrix");
  SetFlags(f, Knob::NO_ANIMATION);
  Tooltip(f,
    "Sets the way RGB is converted to YCbCr. <b>Rec 601</b> and <b>Rec 709</b> follow the ITU.BC specifications. "
    "This setting is only available when you’re working with a Y’CbCr-based pixel type.\n"
    "<b>Auto</b> uses a set of heuristics to determine the best colour matrix to use.\n"
    "<b>Format-based</b> sets the color matrix to <b>Rec 601</b> for formats with a width below 840 pixels and "
    "<b>Rec 709</b> for formats with a width of 840 pixels or above."
  );
    
  Obsolete_knob(f, "audiofile", componentOfMovWriter_ ? "knob mov64_audiofile $value\nknob mov32_audiofile $value"
                : "knob mov64_audiofile $value");
  File_knob(f, &_audioFile, "mov64_audiofile", "Audio File");
  Tooltip(f, "Allows user to specify a separate audio file to include in the output. Either enter the " 
          "filepath manually or click the browse button to locate the audio file.");

  Obsolete_knob(f, "audio_offset", componentOfMovWriter_ ? "knob mov64_audio_offset $value\nknob mov32_audio_offset $value"
                : "knob mov64_audio_offset $value");
  Float_knob(f, &_audioOffset, IRange(-1000.0, 1000.0), "mov64_audio_offset", "Audio Offset");
  ClearFlags(f, Knob::SLIDER);
  SetFlags(f, Knob::NO_ANIMATION);
  Tooltip(f, "Sets the start time of any audio file specified in the audio file control. The unit of measure is specified using the units control. "
          "Negative values cause the audio to start before the video and vice versa.");

  Obsolete_knob(f, "units", componentOfMovWriter_ ? "knob mov64_units $value\nknob mov32_units $value"
                : "knob mov64_units $value");
  static const char* offset_units[] = {
    "Seconds", "Frames", nullptr
  };
  Enumeration_knob(f, &_audioOffsetUnit, offset_units, "mov64_units", "Units");
  ClearFlags(f, Knob::STARTLINE);
  Tooltip(f, "Sets the unit of measure for the <b>Audio offset</b> control:  <b>Seconds</b> or <b>Frames</b>");

  Obsolete_knob(f, "Advanced", "knob mov64_advanced $value");
  BeginClosedGroup(f, "mov64_advanced", "Advanced");
    
  Obsolete_knob(f, "dnxhd_encode_video_range", "knob mov64_dnxhd_encode_video_range $value");
  Obsolete_knob(f, "mov64_dnxhd_encode_video_range", "knob dataRange $value");
  static const char* encodeVideoLevelsLabels[] = {
    "Full Range", "Video Range", nullptr
  };
  Enumeration_knob(f, &encodeVideoRange_, encodeVideoLevelsLabels, "dataRange", "Data Range" );
  SetFlags(f, Knob::DISABLED); // Disabled by default, enabled by selected a DNxHD codec profile.
  Tooltip(f, "Sets the minimum and maximum color range values for encoding the output file:\n"
          "<b>Full Range</b> - encodes with the full data range (0-255 for 8-bit or 0-1023 for 10-bit).\n"
          "<b>Video Range</b> - encodes with a reduced, or video/legal, data range (16-240 for 8-bit or 64-960 for 10-bit).");

  Bool_knob(f, &fastStart_, "mov64_fast_start", "Fast Start");
  Tooltip( f, "Write decoding critical metadata (moov atom) at beginning of the file to allow playback when streaming.");
  SetFlags(f, Knob::STARTLINE | Knob::NO_ANIMATION);
    
  Obsolete_knob(f, "writeTimeCode", componentOfMovWriter_ ? "knob mov64_write_timecode $value\nknob mov32_write_timecode $value"
                  : "knob mov64_write_timecode $value");
  Bool_knob(f, &writeTimecode_, "mov64_write_timecode", "Write Timecode" );

  Tooltip(f, "When enabled, Nuke adds a timecode track to the generated file.\n" 
          "<b>Note:</b>  The timecode is read from the input/timecode metadata key pair. If this field is blank, the timecode is not written"
          " into the file. The timecode is automatically read from any file containing a timecode track and propagated through the tree.");
    
  ClearFlags(f, Knob::STARTLINE);
  //Knobs related to inter frame compression settings.
  Obsolete_knob(f, "gop_size", "knob mov64_gop_size $value");
  Int_knob(f, &gopSize_, IRange(0, 30), "mov64_gop_size", "GOP Size");
  SetFlags(f, Knob::ALWAYS_SAVE | Knob::NO_ANIMATION | Knob::HIDDEN);
  Tooltip (f, "Specifies how many frames may be grouped together by the codec to form a compression GOP. Exercise caution "
             "with this control as it may impact whether the resultant file can be opened in other packages. Only supported by "
             "certain codecs.");

  Obsolete_knob(f, "b_frames", "knob mov64_b_frames $value");
  Int_knob(f, &bFrames_, IRange(0, FF_MAX_B_FRAMES), "mov64_b_frames", "B Frames");
  SetFlags(f, Knob::FORCE_RANGE | Knob::ALWAYS_SAVE | Knob::NO_ANIMATION | Knob::HIDDEN);
  Tooltip (f, "Controls the maximum number of B frames found consecutively in the resultant stream, where zero means no limit "
             "imposed. Only supported by certain codecs.");
  Named_Text_knob(f, "mov64_bitrate_divider", nullptr);

  //ffmpeg offers a number of controls over quality, bitrate, compression, profile/level etc.
  //Most don't do anything for the codecs we offer, so we stick with those known to have an impact,
  //enabling/disabling in knob_changed for the codecs they affect. The bitrate/tolerance and quality
  //cnotrols all balance against each other, so tweaking them requires paying attention to all 3.
  //Potentially they could be another layer on top of predefined settings, but no other ffmpeg gui
  //has got this working cleanly, presumably due to the interop between controls.
  //Codec values, initialised from defaults in ffmpeg, are set to always save to ensure compat
  //between different versions of ffmpeg.
  //Knobs related to lossy compression settings.

  Bool_knob(f, &limitBitrate_, "mov64_limit_bitrate", "Limit Bitrate");
  SetFlags(f,  Knob::STARTLINE | Knob::HIDDEN);
  Tooltip (f, "Limits the output bit rate to a certain range will clamp the peak bit-rate within a specified window.\n"
           "Only supported by certain codecs.");

  Obsolete_knob(f, "bitrate", "knob mov64_bitrate $value");
  bitrateKnob_ = Int_knob(f, &bitrate_, IRange(0.0, kDefaultBitrate), "mov64_bitrate", "Bitrate");
  SetFlags(f, Knob::SLIDER | Knob::LOG_SLIDER | Knob::ALWAYS_SAVE | Knob::KNOB_CHANGED_ALWAYS | Knob::HIDDEN);
  Tooltip (f, "Sets the target bitrate the codec will attempt to reach, within the confines of the bitrate tolerance and Quality Min/Max settings.\n"
           "Values are initially set by the Quality preset.\n"
           "Only supported by certain codecs.");

  Obsolete_knob(f, "bitrate_tolerance", "knob mov64_bitrate_tolerance $value");
  bitrateToleranceKnob_ = Int_knob(f, &bitrateTolerance_, IRange(833, kDefaultBitrateTolerance), "mov64_bitrate_tolerance", "Bitrate Tolerance");
  SetFlags(f, Knob::SLIDER | Knob::LOG_SLIDER | Knob::ALWAYS_SAVE | Knob::KNOB_CHANGED_ALWAYS | Knob::HIDDEN);
  Tooltip (f, "Sets the amount that the bitrate can vary from the bitrate setting. Setting this tolerance too low will result in renders failing.\n"
           "Values are initially set by the Quality preset.\n"
           "Only supported by certain codecs.");

  //Knobs related to intra frame compression settings.
  static const char* qualityRangeTooltip = "The quality range the codec is allowed to vary the image data quantiser "
                                           "between to attempt to hit the desired bitrate. Higher values mean increased "
                                           "image degradation is possible, but with the upside of lower bit rates. "
                                           "Only supported by certain codecs.";
  
  Obsolete_knob(f, "quality_min", "knob mov64_quality_min $value");
  quantizerMinKnob_ = Int_knob(f, &qMin_, IRange(1, 31), "mov64_quality_min", "Quantizer Min");
  SetFlags(f, Knob::NO_ANIMATION | Knob::ALWAYS_SAVE | Knob::KNOB_CHANGED_ALWAYS | Knob::HIDDEN);
  Tooltip (f, qualityRangeTooltip);

  Obsolete_knob(f, "quality_max", "knob mov64_quality_max $value");
  quantizerMaxKnob_ = Int_knob(f, &qMax_,  IRange(1, 31), "mov64_quality_max", "Quantizer Max");
  SetFlags(f, Knob::NO_ANIMATION | Knob::ALWAYS_SAVE | Knob::KNOB_CHANGED_ALWAYS | Knob::HIDDEN);
  Tooltip (f, qualityRangeTooltip);
  ClearFlags(f, Knob::STARTLINE);

  Obsolete_knob(f, "mov64_write_nclc", nullptr);

  //Badly formatted names obsoleted.
  Obsolete_knob(f, "bitrateTol", "knob mov64_bitrate_tolerance $value");
  Obsolete_knob(f, "gopSize", "knob mov64_gop_size $value");
  Obsolete_knob(f, "bFrames", "knob mov64_b_frames $value");

  //Removed from panel - should never have been exposed as very low level control.
  //Maintained for compatibility with old scripts and unforseen circumstances.
  Obsolete_knob(f, "mbDecision", "knob mov64_mbDecision $value");
  static const char* mbDecisionTypes[] = {
    "FF_MB_DECISION_SIMPLE", "FF_MB_DECISION_BITS", "FF_MB_DECISION_RD", nullptr
  };
  Enumeration_knob(f, &mbDecision_, mbDecisionTypes, "mov64_mbDecision", "macro block decision mode");
  SetFlags(f, Knob::INVISIBLE);

  // Update the default LUT at launch
  if (!f.makeKnobs() && !defaultLutUpdated_) {
    iop->updateDefaultLUT();
    defaultLutUpdated_ = true;
  }

  EndGroup(f);

  Divider(f);

  static const auto DNxHRSDKVersion = DNxHR_SDK_Version();
  Named_Text_knob(f, "DNxHR_SDK_Version", "SDK Build:", DNxHRSDKVersion.c_str());
  SetFlags(f, Knob::ALWAYS_SAVE | Knob::INVISIBLE);

#ifdef PLUGINS_UPDATE
  Newline(f);
  Named_Text_knob(f, "Nuke_FileIO_Version", "Plugins Build:", PluginsInstallerVersion);
  SetFlags(f, Knob::ALWAYS_SAVE);
  Tooltip(f, "Plugins package version that this file was originally loaded with.");
#endif

  // In the case of a knob store, we still need to initialise the current pixel formats properly
  // as this otherwise would only get initialised via a knob_changed call will which may not
  // happen if this op is getting regenrated.
  if (!f.makeKnobs() && !f.queryingKnobs()) {
    Knob* codecKnob = iop->knob("mov64_codec");

    if (codecKnob) {
      const size_t index = static_cast<size_t>(codecKnob->get_value());
      AVCodec* codec = avcodec_find_encoder_by_name(codecsShortNames_[index]);

      if (codec) {
        currentCodecPixelFormats_ = PixelFormatsFromAVCodec(codec);
        displayedPixelFormatNames_ = PixelFormatFriendlyNamesFrom(currentCodecPixelFormats_);
      }
    }
  }
}

void mov64Writer::initDynamicKnobs()
{
  if (!qualityKnobsUpdated_) {
    updateQalityKnobFromAdvancedSettings();
    qualityKnobsUpdated_ = true;
  }
    
  handleLegacyProRes();
}

int mov64Writer::knob_changed(DD::Image::Knob* knob)
{
  if ( knob->is("file" ) ) {
    // Switch the 'format' knob based on the new filename suffix
    std::string suffix;
    FileOp::StripSuffix(filename(), &suffix);
    if (!suffix.empty()) {
      // Compare found suffix to known formats
      const size_t lengthWithoutNullTerminator = formatsShortNames_.size() - 1;
      for (size_t i = 0; i < lengthWithoutNullTerminator; ++i) {
        const char* formatName = formatsShortNames_[i];
        if (suffix.compare(formatName) == 0) {
          iop->knob("mov64_format")->set_value(static_cast<double>(i));
          break;
        }
      }
    }
    return 1;
  }

  if (knob->is("mov64_quality")) {
    updateAdvancedQualityKnobs();
    return 1;
  }

  if (knob->is("mov64_quality_min") ||
      knob->is("mov64_quality_max") ||
      knob->is("mov64_bitrate") ||
      knob->is("mov64_bitrate_tolerance")) {
    updateQalityKnobFromAdvancedSettings();
    return 1;
  }

  const bool updateAdvancedParams = (knob == &Knob::showPanel || knob->is("file_type") || knob->is("mov64_codec") || knob->is("mov64_advanced")) && (!UsingMov32Encoder(*iop));
  if (updateAdvancedParams) {
    //The advanced params are enabled/disabled based on the codec chosen and its capabilities.
    //We also investigated setting the defaults based on the codec defaults, however all current
    //codecs defaulted the same, and as a user experience it was pretty counter intuitive.
    //Check knob exists, to deal with cases where Nuke might not have updated underlying writer (#44774)
    //(we still want to use showPanel to update when loading from script and the like).
    if(iop->knob("mov64_codec")) {
      const size_t index = static_cast<size_t>(iop->knob("mov64_codec")->get_value());
      const bool codecIsInRange = codecIndexIsInRange( index );
      if ( codecIsInRange ) {
        const std::string shortName( codecsShortNames_[index] );
        std::string errorMessage;
        if ( codecIsDisallowed( shortName, errorMessage ) )
        {
          iop->critical(errorMessage.c_str());
          return 1;
        }
      }

      AVCodec* codec = avcodec_find_encoder_by_name(codecsShortNames_[index]);

      updatePixelFormatKnobVisibility(codec);

      if (knob->is("mov64_codec") || knob->is("file_type")) {
        updatePixelFormatKnobValues(codec);
        updateQalityKnobFromAdvancedSettings();
      }

      updateLimitBitrateKnob(codec);
      updateYCbCrMatrixKnob(codecsShortNames_[index]);

      bool lossyParams = false;
      bool interGOPParams = false;
      bool interBParams = false;

      if (codec) {
        GetCodecSupportedParams(codec, lossyParams, interGOPParams, interBParams);
      }
      else if (strcmp(codecsShortNames_[index], "fn_h264") == 0) {
        lossyParams = interGOPParams = interBParams = true;
      }

      if (knob->is("mov64_codec")) {
        if (lossyParams) {
          qualityKnob_->set_value(static_cast<int>(eQualityLevel_High));
          updateAdvancedQualityKnobs();
        }
      }

      bitrateKnob_->visible(lossyParams);
      bitrateToleranceKnob_->visible(lossyParams);
      if (iop->knob("mov64_bitrate_divider")) {
        iop->knob("mov64_bitrate_divider")->visible(lossyParams);
      }
        
      qualityKnob_->visible(lossyParams);
      quantizerMinKnob_->visible(lossyParams);
      quantizerMaxKnob_->visible(lossyParams);

      iop->knob("mov64_gop_size")->visible(interGOPParams);
      iop->knob("mov64_b_frames")->visible(interBParams);

      //We use the bitrate to set the min range for bitrate tolerance.
      updateBitrateToleranceRange();

      // Only enable the DNxHD codec profile knob if the Avid
      // DNxHD codec has been selected.
      // Only enable the video range knob if the Avid DNxHD codec
      // has been selected.
      const bool is_dnxhd = IsDNxHD(codecsShortNames_[index]);
      const bool is_dnxhr = IsDNxHR(codecsShortNames_[index]);
      iop->knob("dataRange")->enable(is_dnxhd || is_dnxhr);
      iop->knob("mov64_dnxhd_codec_profile")->visible(is_dnxhd);
      iop->knob("mov64_dnxhd_codec_profile")->enable(is_dnxhd);
      iop->knob("mov64_dnxhr_codec_profile")->visible(is_dnxhr);
      iop->knob("mov64_dnxhr_codec_profile")->enable(is_dnxhr);
      
      const bool prores_codec = strcmp(codecsShortNames_[index], "fn_prores") == 0;
      iop->knob("mov_prores_codec_profile")->enable(prores_codec);
      iop->knob("mov_prores_codec_profile")->visible(prores_codec);
        
      const bool h264_codec = strcmp(codecsShortNames_[index], "fn_h264") == 0;
      iop->knob("mov_h264_codec_profile")->enable(h264_codec);
      iop->knob("mov_h264_codec_profile")->visible(h264_codec);

      auto dnxhrSdkKnob = iop->knob("DNxHR_SDK_Version");
      if (dnxhrSdkKnob) {
        dnxhrSdkKnob->visible(is_dnxhd || is_dnxhr);
      }
    }

    return 1;
  }

  if (knob->is("mov64_fps") || knob->is("mov64_bitrate")) {
    updateBitrateToleranceRange();
    return 1;
  }

  return Writer::knob_changed(knob);
}

void mov64Writer::updateAdvancedQualityKnobs()
{
  const size_t codecIndex = static_cast<size_t>(codecKnob_->get_value());
  const bool codecIsInRange = codecIndexIsInRange( codecIndex );
  if ( codecIsInRange ) {
    const std::string shortName( codecsShortNames_[codecIndex] );
    const int qualityIndex = static_cast<int>(qualityKnob_->get_value());
    const QualityLevel qualityLevel = static_cast<QualityLevel>(qualityIndex);

    if (qualityLevel != eQualityLevel_Custom) {
      const int minQ = kMinimumQuantizeThresholdPerCodec.at(shortName).at(qualityLevel);
      const int maxQ = kMaximumQuantizeThresholdPerCodec.at(shortName).at(qualityLevel);

      quantizerMinKnob_->set_value(minQ);
      quantizerMaxKnob_->set_value(maxQ);
      
      if (shortName == "fn_h264") {
        double bitrate = static_cast<double>(kH264FullHDBitrate.at(qualityLevel));
        bitrateKnob_->set_value(bitrate);
        bitrateToleranceKnob_->set_value(0);
      }
      else {
        bitrateKnob_->set_value(kDefaultBitrate);
        bitrateToleranceKnob_->set_value(kDefaultBitrateTolerance);
      }
      
    }
      
      
  }
}

void mov64Writer::updateQalityKnobFromAdvancedSettings()
{
  const size_t codecIndex = static_cast<size_t>(codecKnob_->get_value());
  const bool codecIsInRange = codecIndexIsInRange( codecIndex );
  if ( codecIsInRange ) {
    const std::string shortName( codecsShortNames_[codecIndex] );

    if (kMinimumQuantizeThresholdPerCodec.find(shortName) == kMinimumQuantizeThresholdPerCodec.end()) {
      // Quality knob is not used for this codec
      return;
    }

    const int minQ = static_cast<int>(quantizerMinKnob_->get_value());
    const int maxQ = static_cast<int>(quantizerMaxKnob_->get_value());
    const int bitRate = static_cast<int>(bitrateKnob_->get_value());
    const int bitRateTolerance = static_cast<int>(bitrateToleranceKnob_->get_value());

    if (bitRate == kDefaultBitrate && bitRateTolerance == kDefaultBitrateTolerance) {
      const auto minimumQuantizeThreshold = kMinimumQuantizeThresholdPerCodec.at(shortName);
      const auto maximumQuantizeThreshold = kMaximumQuantizeThresholdPerCodec.at(shortName);

      for (int qualityIndex = eQualityLevel_Best; qualityIndex <= eQualityLevel_Least; ++qualityIndex) {
        const QualityLevel qualityLevel = static_cast<QualityLevel>(qualityIndex);
        if (minimumQuantizeThreshold.at(qualityLevel) == minQ &&
            maximumQuantizeThreshold.at(qualityLevel) == maxQ) {
          qualityKnob_->set_value(qualityLevel);
          return;
        }
      }
    }
  }

  qualityKnob_->set_value(eQualityLevel_Custom);
}

////////////////////////////////////////////////////////////////////////////////
// updateBitrateToleranceRange
// Utility to update tolerance knob's slider range.
// Employed in place of chained knob_changed calls.
// Only valid for use from knob_changed.
//
void mov64Writer::updateBitrateToleranceRange()
{
  //Bitrate tolerance should in theory be allowed down to target bitrate/target framerate.
  //We're not force limiting the range since the upper range is not bounded.
  double minRange = bitrateKnob_->get_value() / iop->knob("mov64_fps")->get_value();
  bitrateToleranceKnob_->set_range(minRange, kDefaultBitrateTolerance, false);
}

////////////////////////////////////////////////////////////////////////////////
// freeFormat
// Release resources.
//
void mov64Writer::freeFormat()
{
  if (streamVideo_.stream) {
    avcodec_free_context(&streamVideo_.codecContext);
    streamVideo_.codecContext = nullptr;
    streamVideo_.stream = nullptr;
    // free(streamVideo_);
    // streamVideo_ = nullptr;
  }
  if (streamAudio_.stream) {
    avcodec_free_context(&streamAudio_.codecContext);
    streamAudio_.codecContext = nullptr;
    streamAudio_.stream = nullptr;
    // free(streamAudio_);
    // streamAudio_ = nullptr;
  }
  if (!(formatContext_->oformat->flags & AVFMT_NOFILE))
    avio_close(formatContext_->pb);
  avformat_free_context(formatContext_);
  formatContext_ = nullptr;
  isOpen_ = false;
}

void mov64Writer::initColorSpaceConvertor(AVCodecContext* ctx) {
  cscParams_.out = nullptr;
  cscParams_.in_planes = 3;
  cscParams_.width = ctx->width;
  cscParams_.height = ctx->height;
    
  cscParams_.srcDataRange = NukeCodecs::DecoderParams::kVideoRange;
    
  if (AV_CODEC_ID_PRORES == ctx->codec_id || AV_CODEC_ID_DNXHD == ctx->codec_id || AV_CODEC_ID_H264 == ctx->codec_id) {
    return;
  }
    
  colorSpaceConvertor_.reset(NukeCodecs::CodecFactory::makeDecoder(NukeCodecs::CodecFactory::eColorSpaceConvertor));
   
  switch (_YCbCr_MatrixType) {
    case YCbCr_MatrixType::Auto: cscParams_.colorMatrix = isRec709Format(ctx->height) ? NukeCodecs::DecoderParams::eColorMatrix::kRec709 : NukeCodecs::DecoderParams::eColorMatrix::kRec601;
        break;
    case YCbCr_MatrixType::Rec601: cscParams_.colorMatrix = NukeCodecs::DecoderParams::eColorMatrix::kRec601; break;
    case YCbCr_MatrixType::Rec709: cscParams_.colorMatrix = NukeCodecs::DecoderParams::eColorMatrix::kRec709; break;
  }
    
  switch (ctx->pix_fmt) {
    case AV_PIX_FMT_YUVJ420P:
      cscParams_.srcDataRange = NukeCodecs::DecoderParams::kFullRange;
    case AV_PIX_FMT_YUV420P:
      cscParams_.srcSubsampling = NukeCodecs::DecoderParams::k420;
      cscParams_.bitDepth = 8;
      cscParams_.out_planes = 3;
      break;
    case AV_PIX_FMT_YUVJ422P:
      cscParams_.srcDataRange = NukeCodecs::DecoderParams::kFullRange;
    case AV_PIX_FMT_YUV422P:
      cscParams_.srcSubsampling = NukeCodecs::DecoderParams::k422;
      cscParams_.bitDepth = 8;
      cscParams_.out_planes = 3;
      break;
    case AV_PIX_FMT_YUVJ444P:
      cscParams_.srcDataRange = NukeCodecs::DecoderParams::kFullRange;
    case AV_PIX_FMT_YUV444P:
      cscParams_.srcSubsampling = NukeCodecs::DecoderParams::k444;
      cscParams_.bitDepth = 8;
      cscParams_.out_planes = 3;
      break;
    case AV_PIX_FMT_YUV420P16LE:
      cscParams_.srcSubsampling = NukeCodecs::DecoderParams::k420;
      cscParams_.bitDepth = 16;
      cscParams_.out_planes = 3;
      break;
    case AV_PIX_FMT_YUV422P16LE:
      cscParams_.srcSubsampling = NukeCodecs::DecoderParams::k422;
      cscParams_.bitDepth = 16;
      cscParams_.out_planes = 3;
      break;
    case AV_PIX_FMT_YUV444P16LE:
      cscParams_.srcSubsampling = NukeCodecs::DecoderParams::k444;
      cscParams_.bitDepth = 16;
      cscParams_.out_planes = 3;
      break;
    case AV_PIX_FMT_YUV420P12LE:
      cscParams_.srcSubsampling = NukeCodecs::DecoderParams::k420;
      cscParams_.bitDepth = 12;
      cscParams_.out_planes = 3;
      break;
    case AV_PIX_FMT_YUV422P12LE:
      cscParams_.srcSubsampling = NukeCodecs::DecoderParams::k422;
      cscParams_.bitDepth = 12;
      cscParams_.out_planes = 3;
      break;
    case AV_PIX_FMT_YUV444P12LE:
      cscParams_.srcSubsampling = NukeCodecs::DecoderParams::k444;
      cscParams_.bitDepth = 12;
      cscParams_.out_planes = 3;
      break;
    case AV_PIX_FMT_YUV420P10LE:
      cscParams_.srcSubsampling = NukeCodecs::DecoderParams::k420;
      cscParams_.bitDepth = 10;
      cscParams_.out_planes = 3;
      break;
    case AV_PIX_FMT_YUV422P10LE:
      cscParams_.srcSubsampling = NukeCodecs::DecoderParams::k422;
      cscParams_.bitDepth = 10;
      cscParams_.out_planes = 3;
      break;
    case AV_PIX_FMT_YUV444P10LE:
      cscParams_.srcSubsampling = NukeCodecs::DecoderParams::k444;
      cscParams_.bitDepth = 10;
      cscParams_.out_planes = 3;
      break;
    default:
      colorSpaceConvertor_.reset();
      break;
  }
}

bool mov64Writer::isProRes422() const
{
  return codecType_ == NukeCodecs::CodecFactory::eProResCodec422 ||
         codecType_ == NukeCodecs::CodecFactory::eProResCodec422LT ||
         codecType_ == NukeCodecs::CodecFactory::eProResCodec422Proxy ||
         codecType_ == NukeCodecs::CodecFactory::eProResCodec422HQ;
}

void mov64Writer::handleLegacyProRes() {
  auto codecKnob = iop->knob("mov64_codec");
  if(codecKnob) {
    const size_t index = static_cast<size_t>(codecKnob->get_value());
    if (IsLegacyProres(codecsShortNames_[index])) {
      auto it = std::find(codecsShortNames_.begin(), codecsShortNames_.end(), "fn_prores");
      if (it != codecsShortNames_.end()){
        codecKnob->set_value(static_cast<double>(it - codecsShortNames_.begin()));

        auto metaKnob = iop->knob("meta_codec");

        if (metaKnob) {
          Enumeration_KnobI* enumerationKnob = metaKnob->enumerationKnob();
          if (enumerationKnob) {
            SetKnobValue(metaKnob, enumerationKnob->menu(), MOV64_PRORES_CODEC_NAME);
          }
        }

        auto proresProfileKnob = iop->knob("mov_prores_codec_profile");

        if (proresProfileKnob) {
          for (size_t i = 0; i < sizeof(licensedCodecLUT); ++i) {
            if (strcmp(licensedCodecLUT[i].shortName, codecsShortNames_[index]) == 0) {
                
              for (size_t j = 0; j < sizeof(proresCodecTags); ++j) {
                if (licensedCodecLUT[i].codecTag == proresCodecTags[j]) {
                
                  proresProfileKnob->set_value(static_cast<double>(j));
                  break;
                }
              }
                
              break;
            }
          }
        }
      }
    }
  }
}

int32_t mov64Writer::configureH264Encoder(AVCodecContext * ctx) {

  updateAdvancedQualityKnobs();

  NukeCodecs::H264Params params;
  params.width = width();
  params.height = height();
  params.fps = fps_;

  //bitrate limits set for Full HD, so here we rescale
  double bitrate = bitrateKnob_->get_value();
  bitrate *= width() * height();
  bitrate /= (1920 * 1080);

  params.alloc_func = av_malloc;
  params.free_func = av_free;
  params.bitrate = static_cast<uint32_t>(ceil(bitrate)) * 1000;
  params.tolerance = static_cast<uint32_t>(bitrateToleranceKnob_->get_value()) * 1000;
  params.gop_size = gopSize_;
  params.profile = static_cast<int32_t>(iop->knob("mov_h264_codec_profile")->get_value());

  NukeCodecs::H264Frame frame;
  params.avcc = &frame;

  int32_t res = encoder_->ConfigureEncoder(&params);

  if (res == NukeCodecs::CodecFactory::kError_NoError) {
    ctx->extradata_size = frame.size;
    ctx->extradata = static_cast<uint8_t *>(frame.data);
  }

  return res;
}

static Writer* build(Write* iop)
{
  return new mov64Writer(iop);
}

const Writer::Description mov64Writer::d("mov\0avi\0mp4\0m4v\0mov64\0ffmpeg\0", "mov", build);