// Copyright (c) 2014 The Foundry Visionmongers Ltd.  All Rights Reserved.

#ifndef NUKE_MOV64WRITER_H
#define NUKE_MOV64WRITER_H

#include "DDImage/Writer.h"

#ifndef INT64_C
# define INT64_C(c) (c ## LL)
#endif
#ifndef UINT64_C
# define UINT64_C(c) (c ## ULL)
#endif

#include "mov64Common.h"
#define NUKE_MOV64WRITER_CPP
#include "movCommon.h"

extern "C" {
  #ifdef _WIN32
    #pragma warning(push, 0) // We dont care about warnings from the libav headers
  #endif
  #include <errno.h>
  #include <libavformat/avformat.h>
  #include <libavcodec/avcodec.h>
  #include <libavutil/imgutils.h>
  #include <libavformat/avio.h>
  #include <libavutil/opt.h>
  #include <libswscale/swscale.h>
  #include <libavutil/avutil.h>
  #include <libavutil/error.h>
  #include <libswresample/swresample.h>
  #include <libavutil/intreadwrite.h>
  #ifdef _WIN32
    #pragma warning(pop)
  #endif
}
#include <map>
#include <string>

#ifndef FN_EXAMPLE_PLUGIN
////////////////////////////////////////////////////////////////////////////////
// FN_LICENSED_PRORES_DNXHD_CODEC
// Comment out this preprocessor definition if building this file WITHOUT a
// valid Apple ProRes license or valid Avid DNxHD license.
# define FN_LICENSED_PRORES_DNXHD_CODEC
#endif // ifndef FN_EXAMPLE_PLUGIN

////////////////////////////////////////////////////////////////////////////////
#if defined(FN_LICENSED_PRORES_DNXHD_CODEC)
# include <NukeCodecs/fnNukeCodecs.h> // propriatary Foundry codec management
#endif
struct LicensedCodecStr; // Forward declaration.
////////////////////////////////////////////////////////////////////////////////

#define DEBUG_AUDIO_READER(x) std::cout << x << std::endl
//#define DEBUG_AUDIO_READER(x) ;


enum class YCbCr_MatrixType : int {
  Auto = 0,
  Rec601 = 1,
  Rec709 = 2,
};

enum QualityLevel {
  eQualityLevel_Best = 0,
  eQualityLevel_High,
  eQualityLevel_Medium,
  eQualityLevel_Low,
  eQualityLevel_Least,
  eQualityLevel_Custom,

  eQualityLevel_Default = eQualityLevel_High
};

////////////////////////////////////////////////////////////////////////////////
// FnAVFrame
// Wrapper class for the FFmpeg AVFrame structure.
// In order negate the chances of memory leaks when using FFmpeg this class
// provides the initialisation, allocation and release of memory associated
// with an AVFrame structure.
//
// Note that this has been designed to be a drop in replacement for host
// managed memory for colourspace, pixel format and sample format conversions.
// It is not designed to be used with avcodec_decode_video2 as the underlying
// decoder usually manages memory.
//
// Example usage:
//
// Audio
//
//    FnAVFrame avFrame;
//    ret = avFrame.alloc(channels, nb_samples, _avSampleFormat, 1);
//    :
//    ret = swr_convert(_swrContext, avFrame->data, ...);
//    :
//    ret = avcodec_encode_audio2(avCodecContext, &pkt, avFrame, &gotPacket);
//    :
//
// Video
//
//    FnAVFrame avFrame;
//    ret = avFrame.alloc(width(), height(), pixelFormatCodec, 1);
//    if (!ret) {
//      :
//      sws_scale(convertCtx, ..., avFrame->data, avFrame->linesize);
//      :
//    }
//
// IMPORTANT
// This class has been purposefully designed NOT to have parameterised
// constructors or assignment operators. The reason for this is due to the
// complexity of managing the lifetime of the structure and its associated
// memory buffers.
//
class FnAVFrame
{
public:
  ////////////////////////////////////////////////////////////////////////////////
  // Ctor. Initialise the AVFrame structure.
  FnAVFrame()
    : _avFrame(nullptr)
  {
  }

  ////////////////////////////////////////////////////////////////////////////////
  // Dtor. Release all resources. Release the AVFrame buffers and the
  // AVFrame structure.
  ~FnAVFrame();

  ////////////////////////////////////////////////////////////////////////////////
  // alloc
  // VIDEO SPECIFIC.
  // Allocate a buffer or buffers for the AVFrame structure. How many
  // buffers depends upon |avPixelFormat| which is a VIDEO format.
  //
  // @param width Frame width in pixels.
  // @param height Frame height in pixels.
  // @param avPixelFormat An AVPixelFormat enumeration for the frame pixel format.
  // @param align Buffer byte alignment.
  //
  // @return 0 if successful.
  //         <0 otherwise.
  //
  int alloc(int width, int height, enum AVPixelFormat avPixelFormat, int align);

  ////////////////////////////////////////////////////////////////////////////////
  // alloc
  // AUDIO SPECIFIC.
  // Allocate a buffer or buffers for the AVFrame structure. How many
  // buffers depends upon |avSampleFormat| which is an AUDIO format.
  //
  // @param nbChannels The number of audio channels.
  // @param nbSamples The number of audio samples that the buffer will hold.
  // @param avSampleFormat An AVSampleFormat enumeration for the audio format.
  // @param align Buffer byte alignment.
  //
  // @return 0 if successful.
  //         <0 otherwise.
  //
  int alloc(int nbChannels, int nbSamples, enum AVSampleFormat avSampleFormat, int align);

  // operator dereference overload.
  AVFrame* operator->() const { return _avFrame; }
  // operator type cast overload.
  operator AVFrame*() { return _avFrame; }

private:
  AVFrame* _avFrame;

  // Release any memory allocated to the data member variable of
  // the AVFrame structure.
  void deallocateAVFrameData();

  // Probably do not need the following or the headaches
  // of trying to support the functionality.
  // Hide the copy constuctor. Who would manage memory?
  FnAVFrame(FnAVFrame& avFrame) { }
  // Hide the assignment operator. Who would manage memory?
  FnAVFrame& operator=(const FnAVFrame& rhs) { return *this; }
};

typedef struct FnAVStream {
  AVStream* stream;
  AVCodecContext* codecContext;
} FnAVStream;

////////////////////////////////////////////////////////////////////////////////
// AudioReader
// Class that uses the avformat, avcodec, avutil and swresample libraries to
// read, decode and reformat audio data from a file that contains audio.
//
// This class is used if an optional audio file has been specified to be used
// with the mov64Writer. The audio data is interleaved with the video frames
// by the mov64Writer.
//
// The mov64Writer can specify an 'offset' for the audio in order to advance
// or delay the audio with respect to the video. If the audio is delayed then
// the reader will create silence, keep track of the read position and provide
// 'real' audio at the appropriate point. If the audio is advanced then a
// seek operation is performed into the file and reads will begin from that
// position.
//
class AudioReader
{
public:
  AudioReader()
    : _avFormatContext(nullptr)
    , _avFrameAudio(nullptr)
    , _swrContext(nullptr)
    , _avSampleFormat(AV_SAMPLE_FMT_S16)
    , _codecID(AV_CODEC_ID_PCM_S16LE)
    , _readPosition(0)
  {
    _avStreamAudio = {0};
  }

  ~AudioReader()
  {
    close();
  }

  ////////////////////////////////////////////////////////////////////////////////
  // open
  // Attempt to open the file and determine whether the file contains any audio
  // and whether that audio can be decoded.
  //
  // @param filename The name of the file containing audio.
  //
  // @return 0 if successful, non-zero otherwise.
  //
  int open(std::string filename);

  ////////////////////////////////////////////////////////////////////////////////
  // read
  // Read, decode and resample some audio data from the source file. The format
  // of the data will be in the desired target format.
  //
  // @param avFrame A reference to an FnAVFrame object. This is an input and
  //                output parameter. An UNINITIALIZED FnAVFrame is passed to
  //                this method which allocs an appropriately sized buffer. The
  //                audio is read and converted into this buffer which the caller
  //                can then use. When the FnAVFrame object is destroyed it will
  //                release the buffer.
  //
  // @return <0 if unsuccessful, otherwise the number of samples read.
  //
  int read(FnAVFrame& avFrame);

  ////////////////////////////////////////////////////////////////////////////////
  // close
  // Release all resources.
  //
  void close();

  /////////////////////////////////////////////////////////////////////////////////
  // setSampleFormat
  // Specify the desired output sample format of the reader.
  // This will configure an appropriate resampling context that is used in the
  // Read method to convert from a source format to the desired target format.
  //
  // @param avSampleFormat The desired target sample format.
  //
  // @return 0 if successful, non-zero otherwise.
  //
  int setSampleFormat(AVSampleFormat avSampleFormat);

  /////////////////////////////////////////////////////////////////////////////////
  // getSampleFormat
  // Return the target format of the reader.
  //
  // @return An AVSampleFormat enumeration of the target sample format.
  //
  AVSampleFormat getSampleFormat()
  {
    return _avSampleFormat;
  }

  ////////////////////////////////////////////////////////////////////////////////
  // getSampleRate
  // Return the sample rate of the audio in the source file.
  //
  // @return The sample rate if a file has been successfully opened and contains
  //         audio. Zero otherwise.
  //
  int getSampleRate();

  ////////////////////////////////////////////////////////////////////////////////
  // getNumberOfChannels
  // Return the channel count of the audio in the source file.
  //
  // @return The channel count if a file has been successfully opened and contains
  //         audio. Zero otherwise.
  //
  int getNumberOfChannels();

  ////////////////////////////////////////////////////////////////////////////////
  // getDuration
  // Return the duration of the audio stream in seconds.
  //
  // @return The duration of the audio stream in seconds.
  //
  double getDuration();

  ////////////////////////////////////////////////////////////////////////////////
  // setStartPosition
  // Set the start position of the audio reader. This is used to adjust the audio
  // with respect to video in the mov64Writer. This can be used to advance or
  // delay the audio. If the audio is advanced the start of the audio will not
  // be read. If the audio is delayed then silence is inserted.
  //
  // @param startTime The time in seconds of the position of the first read.
  //                  This value can be zero, positive or negative.
  //                  Time zero is the start of the file, all other values are
  //                  with respect to this position.
  //                  If positive, this will cause the reader to start reading
  //                  after of the start of the file. I.e. positive values will
  //                  skip the start of the file.
  //                  If negative, this will cause the reader to read 'ahead' of
  //                  the start of the file. As no data exists, it will provide
  //                  silence until the read position arrives at the start of the
  //                  file.
  //
  // @return 0 if successful,
  //         <0 otherwise.
  //
  int setStartPosition(double startTime);

  ////////////////////////////////////////////////////////////////////////////////
  // getCodecId
  // Return the codec ID of the audio stream.
  //
  // @return The codec ID of the audio stream.
  //
  AVCodecID getCodecID() const { return _codecID; }

private:
  AVFormatContext* _avFormatContext; // A format context is required for the file handling.
  FnAVStream _avStreamAudio; // This provided all manipulation of the stream and codec context.
  AVFrame* _avFrameAudio; // This is provide so that it is initialized once and used many times. The decoded manages the read buffer.
  SwrContext* _swrContext;

  AVSampleFormat _avSampleFormat; // Desired target format.
  AVCodecID _codecID; // Codec ID of the audio stream

  int64_t _readPosition;
};

////////////////////////////////////////////////////////////////////////////////
// mov64Writer
//
using MovPixelFormatName  = std::string;
using MovPixelFormatNames = std::vector<MovPixelFormatName>;

struct MovPixelFormat {
  MovPixelFormatName ffmpegName;
  MovPixelFormatName friendlyName;
  uint32_t    bitsPerPixel;
  std::string description;

  // The ffmpeg format string is unique
  bool operator == (const MovPixelFormat &pf) {
    return (ffmpegName == pf.ffmpegName);
  }
};

using MovPixelFormatArray = std::vector<MovPixelFormat>;

class mov64Writer : public DD::Image::Writer
{
private:
  typedef std::map< std::string, std::string > StringStringMap;
  typedef std::pair< std::string, std::string > StringPair;



public:
  explicit mov64Writer(DD::Image::Write* iop, bool componentOfMovWriter = false);
  ~mov64Writer() override;

  DD::Image::Knob* getCodecKnob() { return codecKnob_; }

  bool movie() const override { return true; }

  DD::Image::LUT* defaultLUT() const override;
  bool open();
  void execute() override;
  void finish() override;
  bool isDefaultLUTKnob(DD::Image::Knob* knob) const override;
  void knobs(DD::Image::Knob_Callback f) override;
  void initDynamicKnobs() override;
  int knob_changed(DD::Image::Knob* knob) override;

  static const Writer::Description d;

  const char* const* getCodecNames() const
  {
    return &codecsKnobLabels_.front();
  }

private:
  void freeFormat();
  AVColorTransferCharacteristic getColorTransferCharacteristic() const;
  AVPixelFormat                 getPixelFormat(AVCodec* videoCodec) const;
  AVOutputFormat*               initFormat(bool reportErrors) const;
  bool                          initCodec(AVOutputFormat* fmt, AVCodecID& outCodecId, AVCodec*& outCodec) const;

  void getPixelFormats(AVCodec*          videoCodec,
                       AVPixelFormat&    outNukeBufferPixelFormat,
                       AVPixelFormat&    outTargetPixelFormat,
                       int&              outBitDepth) const;

  int encodeVideo(AVCodecContext* avCodecContext, const AVFrame* avFrame, AVPacket& avPacketOut);
  void updateBitrateToleranceRange();
  void updateAdvancedQualityKnobs();
  void updateQalityKnobFromAdvancedSettings();
  bool isRec709Format(const int height) const;
  bool isPalFormat(const AVRational& displayRatio) const;
  std::string defaultPixelFormatFrom(const AVCodec* codec) const;
  void updatePixelFormatKnobVisibility(AVCodec* codec);
  void updatePixelFormatKnobValues(const AVCodec* codec);
  void updateYCbCrMatrixKnob(const char* codecShortName);
  void updateLimitBitrateKnob(AVCodec* codec);
  static bool IsYUV(AVPixelFormat pixelFormat);
  static bool IsYUVFromShortName(const char* shortName, int codecProfile);
  static bool IsRGBFromShortName(const char* shortName, int codecProfile);


  static int            GetPixelFormatBitDepth(const AVPixelFormat pixelFormat);
  static AVPixelFormat  GetPixelFormatFromBitDepth(const int bitDepth, const bool hasAlpha);
  static void           GetCodecSupportedParams(AVCodec* codec, bool& outLossyParams,
                                                bool& outInterGOPParams, bool& outInterBParams);


private:
  AVFormatContext*  formatContext_;
  FnAVStream streamVideo_;
  FnAVStream streamAudio_;

  AVStream* streamTimecode_;

  MovPixelFormatArray currentCodecPixelFormats_;
  MovPixelFormatNames displayedPixelFormatNames_;
  std::vector<std::string> formatsLongNames_;
  std::vector<const char*> formatsShortNames_;
  std::vector<const char*> codecsShortNames_;
  std::vector<const char*> codecsKnobLabels_;
  StringStringMap disallowedErrorMessages_;
  bool writeTimecode_; // Flag to add a timecode track to the movie.
  bool fastStart_;
  int dnxhdCodecProfile_;
  int proresCodecProfile_;
  int h264CodecProfile_;
  int dnxhrCodecProfile_;
  int encodeVideoRange_; // DNxHD specific, flag for full scale or video/legal range encoding.
  int _quality;
  int _pixelFormat;
  bool isOpen_; // Flag for the configuration state of the FFmpeg components.
  uint64_t pts_counter_;
  // knobs variables
  float fps_;
  int metaEncoder_;
  int format_;
  int codec_;
  const char* _audioFile;
  float _audioOffset;
  int _audioOffsetUnit;
  int bitrate_;
  int bitrateTolerance_;
  bool limitBitrate_;
  int gopSize_;
  int bFrames_;
  YCbCr_MatrixType _YCbCr_MatrixType;
  int mbDecision_;
  int qMin_;
  int qMax_;
  bool componentOfMovWriter_;
  bool defaultLutUpdated_;
  bool qualityKnobsUpdated_;

  DD::Image::Knob* codecKnob_;
  DD::Image::Knob* qualityKnob_;
  DD::Image::Knob* quantizerMinKnob_;
  DD::Image::Knob* quantizerMaxKnob_;
  DD::Image::Knob* bitrateKnob_;
  DD::Image::Knob* bitrateToleranceKnob_;

#if defined(FN_LICENSED_PRORES_DNXHD_CODEC)
  // Apple ProRes and Avid DNxHD codec specific.
  mutable const struct LicensedCodecStr* licensedCodec_; // Pointer into licensedCodecLUT for the currently selected ProRes and DNxHD codecs. Mutable is used as this is simply a pointer into an existing table of const data. The data is never modified. This pointer is used to reference the selected ProRes or DNxHD codec and unfortunately this happens inside a const function, so mutable is used to move the pointer to the appropriate table entry during configuration.
  std::unique_ptr<NukeCodecs::IEncoder> encoder_;
  std::unique_ptr<NukeCodecs::IDecoder> colorSpaceConvertor_;
  NukeCodecs::DecoderParams cscParams_;
  //we use shared_ptr here as both frames can point to the same AVFrame (ProRes, DNxHD case)
  std::shared_ptr<AVFrame> inputFrame_;
  std::shared_ptr<AVFrame> outputFrame_;
  NukeCodecs::CodecFactory::CodecType codecType_;
#endif

  std::unique_ptr<AudioReader> audioReader_;

  void configureAudioStream(AVCodec* avCodec, FnAVStream* fnAVStream);
  void configureVideoStream(AVCodec* avCodec, FnAVStream* fnAVStream);
  void addStream(AVFormatContext* avFormatContext, enum AVCodecID avCodecId, AVCodec** pavCodec, FnAVStream* fnStreamOut);
  int openCodec(AVFormatContext* avFormatContext, AVCodec* avCodec, FnAVStream* fnAVStream);
  int writeAudio(AVFormatContext* avFormatContext, FnAVStream* fnAVStream, bool flush);
  int writeVideo(AVFormatContext* avFormatContext, FnAVStream* fnAVStream, bool flush);
  int writeToFile(AVFormatContext* avFormatContext, bool finalise);

  int colourSpaceConvert(AVFrame* avFrameIn, AVFrame* avFrameOut, AVPixelFormat srcPixelFormat, AVPixelFormat dstPixelFormat, AVCodecContext* avCodecContext, YCbCr_MatrixType matrix);

  // Returns true if the selected channels contain alpha and that the channel is valid
  bool alphaEnabled() const;

  bool alphaEnabledNoValidation() const;

  // Returns the nmber of destination channels this will write into.
  int numberOfDestChannels() const;

  bool codecIndexIsInRange( size_t codecIndex) const;
  bool codecIsDisallowed( const std::string& codecShortName, std::string& reason ) const;
    
  void initColorSpaceConvertor(AVCodecContext* ctx);
  bool isProRes422() const;
  void handleLegacyProRes();
  int32_t configureH264Encoder(AVCodecContext * ctx);
};


#endif // NUKE_MOV64WRITER_H

